/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	Date.h                                          *
 *                                                          *
 * Content: Implementation of Date class based on code from *
 *          Chris Laurel <claurel@shatters.net>             *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
// Copyright (C) 2001 Chris Laurel <claurel@shatters.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <cmath>
#include <iomanip>
#include "time.h"
#include "Date.h"

char JWDate::weekday[7][10] = { "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday" };

JWDate::JWDate(int Y, int M, int D)
{
    year = Y;
    month = M;
    day = D;
    hour = 0;
    minute = 0;
    seconds = 0.0;
}

JWDate::JWDate(double jd)
{
  setJuliandate(jd);
}

JWDate::setJuliandate(double jd)
{
    int a = (int) (jd + 0.5);
    double c;
    if (a < 2299161)
    {
        c = a + 1524;
    }
    else
    {
        double b = (int) ((a - 1867216.25) / 36524.25);
        c = a + b - (int) (b / 4) + 1525;
    }

    int d = (int) ((c - 122.1) / 365.25);
    int e = (int) (365.25 * d);
    int f = (int) ((c - e) / 30.6001);

    double dday = c - e - (int) (30.6001 * f) + ((jd + 0.5) - (int) (jd + 0.5));

    /* This following used to be 14.0, but gcc was computing it incorrectly, so
       it was changed to 14 */
    month = f - 1 - 12 * (int) (f / 14);
    year = d - 4715 - (int) ((7.0 + month) / 10.0);
    day = (int) dday;
    
    double dhour = (dday - day) * 24;
    hour = (int) dhour;

    double dminute = (dhour - hour) * 60;
    minute = (int) dminute;

    seconds = (dminute - minute) * 60;
}

JWDate::JWDate(QDomElement node)
{
	char* stopstring;	// temporary variable for type conversion
	std::string value;	// temporary container for attribute values
	
	// First set date to the "real" time and date.
	// We do this because maybe (node == NULL) or some attributes are not given.
	struct tm *newtime;
	long ltime;
	
	time( &ltime );
	
	/* Obtain coordinated universal time: */
	newtime = gmtime( &ltime );
	
	year = newtime->tm_year + 1900; // tm_year is years since 1900!
	month = newtime->tm_mon + 1;	// tm_mon starts with 0!
	day = newtime->tm_mday;
	hour = newtime->tm_hour;
	minute = newtime->tm_min;
	seconds = newtime->tm_sec;
	
	if (node.isNull())
		return;

	// get the julian start time of simulation if given
	double julian;
	if (node.hasAttribute("JULIANDATE"))
	{
		julian = (node.attribute("JULIANDATE")).toDouble();
    setJuliandate(julian);
		return;
	}

	// no julian date so get yyyy/mm/dd hh:min:sec
	if (node.hasAttribute("YEAR"))
	{
		value = node.attribute("YEAR");
		year = (int)strtol( value.c_str(), &stopstring, 10);
	}
	if (node.hasAttribute("MONTH"))
	{
		value = node.attribute("MONTH");
		month = (int)strtol( value.c_str(), &stopstring, 10);
	}
	if (node.hasAttribute("DAY"))
	{
		value = node.attribute("DAY");
		day = (int)strtol( value.c_str(), &stopstring, 10);
	}
	if (node.hasAttribute("HOUR"))
	{
		value = node.attribute("HOUR");
		hour = (int)strtol( value.c_str(), &stopstring, 10);
	}
	if (node.hasAttribute("MINUTE"))
	{
		value = node.attribute("MINUTE");
		minute = (int)strtol( value.c_str(), &stopstring, 10);
	}
	if (node.hasAttribute("SECOND"))
	{
		value = node.attribute("SECOND");
		seconds = strtod( value.c_str(), &stopstring);
	}
}

// Convert a calendar date to a Julian date
double JWDate::doubleValue()
{
    int y = year, m = month;
    if (month <= 2)
    {
        y = year - 1;
        m = month + 12;
    }

    // Correct for the lost days in Oct 1582 when the Gregorian calendar
    // replaced the Julian calendar.
    int B = -2;
    if (year > 1582 || (year == 1582 && (month > 10 || (month == 10 && day >= 15))))
    {
        B = y / 400 - y / 100;
    }

    return (floor(365.25 * y) +
            floor(30.6001 * (m + 1)) + B + 1720996.5 +
            day + hour / 24.0 + minute / 1440.0 + seconds / 86400.0);
}

char *JWDate::print()
{
	char *buf = new char[80];
	sprintf(buf, "%s %04d/%02d/%02d %02d:%02d:%02d",JWDate::weekday[((int)(this->doubleValue()+0.5))%7],year,month,day,hour,minute,(int)seconds);
	return buf;
}


