/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	GLFont.h                                        *
 *                                                          *
 * Content: Implementation of texture font                  *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#include "GLFont.h"
#include <qapplication.h>
#include <qfile.h> 

GLFont::GLFont()
{
  fWidth = 16;
  fHeight = 16;
  fSpacing = 10; 
  fxCount = 16;
  fyCount = 16;
  fdWidth = 640;
  fdHeight = 480;
  fStartPos = 0;
  fScale = 16;
}

GLFont::~GLFont()
{
 KillFont();
}

GLvoid GLFont::KillFont(GLvoid)							// Delete The Font From Memory
{
	glDeleteLists(fBase,fxCount * fyCount);				// Delete Allocated Display Lists
}


GLvoid GLFont::Print(GLint x, GLint y, char *string)	// Where The Printing Happens
{
	// init text output
	glDisable(GL_LIGHTING);								// Turn off lighting
	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, fTexture.texID);		// Select Our Font Texture
	glDisable(GL_DEPTH_TEST);							// Disables Depth Testing
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glPushMatrix();										// Store The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix
	glOrtho(0,fdWidth,0,fdHeight,-1,1);					// Set Up An Ortho Screen
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glPushMatrix();										// Store The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix
	glScalef(fdWidth/800.0f, fdHeight/600.0f, 1);
	glPushAttrib(GL_BLEND);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_ALPHA_TEST);
	glTranslated(x,600 - y,0);						// Position The Text (0,0 - Bottom Left)
	glScalef(fScale/fHeight, fScale/fHeight, 1);
	glListBase(fBase - fStartPos);						// Choose The Font Set (0 or 1)
	glCallLists(strlen(string),GL_BYTE,string);			// Write The Text To The Screen
	glPopAttrib();	//GL_BLEND
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glPopMatrix();										// Restore The Old Projection Matrix
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glPopMatrix();										// Restore The Old Projection Matrix
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
}

GLvoid GLFont::PrintTest(GLint x, GLint y)	// Where The Printing Happens
{
	// init text output
	glDisable(GL_LIGHTING);								// Turn off lighting
	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, fTexture.texID);		// Select Our Font Texture
	glDisable(GL_DEPTH_TEST);							// Disables Depth Testing
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glPushMatrix();										// Store The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix
	glOrtho(0,fdWidth,0,fdHeight,-1,1);					// Set Up An Ortho Screen
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glPushMatrix();										// Store The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix
	glScalef(fdWidth/800.0f, fdHeight/600.0f, 1);
	glPushAttrib(GL_BLEND);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_ALPHA_TEST);
	for (int Ly = 0; Ly < 16; Ly++)
	{
		for (int Lx = 0; Lx < 16; Lx++)
		{
			glLoadIdentity();									// Reset The Modelview Matrix
			glTranslated(x + 16*Lx,600 - y - 16*Ly,0);		// Position The Text (0,0 - Bottom Left)
			glScalef(fScale/fHeight, fScale/fHeight, 1);
			glCallList(fBase + Lx + 16 * Ly);					// Write The Text To The Screen
		}
	}
	glPopAttrib();	//GL_BLEND
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glPopMatrix();										// Restore The Old Projection Matrix
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glPopMatrix();										// Restore The Old Projection Matrix
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
}

void GLFont::SetBase(int Base)
{
 fStartPos = Base;
}

void GLFont::SetTexture(char *Tex, int xCount, int yCount)
{
	fTexture.LoadImage(Tex, true);
	fTexture.activate(TEXTURE::T_NORMAL);

	fxCount     = xCount;
	fyCount     = yCount;
}

void GLFont::SetFontProperties(int Width, int Height, int Spacing)
{
	fWidth      = Width;
	fHeight     = Height;
	fSpacing    = Spacing;
}

void GLFont::SetDisplayMode(int dWidth, int dHeight)
{
 fdWidth  = dWidth;
 fdHeight = dHeight;
}

void GLFont::SetFontSize(int s)
{
	fScale = s;
}


GLvoid GLFont::BuildFont()								// Build Our Font Display List
{
	int     loop;
	float	cx;											// Holds Our X Character Coord
	float	cy;											// Holds Our Y Character Coord
    float   cwx;                                        // CharWidth in texture units
	float   cwy;                                        // CharHeight in texture units

	cwx   = (1.0f / (float)fTexture.width) * fWidth;
	cwy   = (1.0f / (float)fTexture.height) * fHeight;
	fBase=glGenLists(fxCount * fyCount);				// Creating Display Lists
	glBindTexture(GL_TEXTURE_2D, fTexture.texID);		// Select Our Font Texture
	for (loop=0; loop<(fxCount * fyCount); loop++)		// Loop Through All Lists
	{
		cx=float(loop%fxCount) * cwx;					// X Position Of Current Character
		cy=float(loop/fyCount) * cwy;					// Y Position Of Current Character

		glNewList(fBase + loop,GL_COMPILE);				// Start Building A List
			glBegin(GL_QUADS);							// Use A Quad For Each Character
				glTexCoord2f(cx,1-cy-cwy);				// Texture Coord (Bottom Left)
				glVertex2i(0,0);						// Vertex Coord (Bottom Left)
				glTexCoord2f(cx+cwx,1-cy-cwy);			// Texture Coord (Bottom Right)
				glVertex2i(fWidth - 1,0);				// Vertex Coord (Bottom Right)
				glTexCoord2f(cx+cwx,1-cy);				// Texture Coord (Top Right)
				glVertex2i(fWidth - 1,fHeight -1);		// Vertex Coord (Top Right)
				glTexCoord2f(cx,1-cy);					// Texture Coord (Top Left)
				glVertex2i(0,fHeight -1);				// Vertex Coord (Top Left)
			glEnd();									// Done Building Our Quad (Character)
			glTranslated(fSpacing,0,0);					// Move To The Right Of The Character
		glEndList();									// Done Building The Display List
	}													// Loop Until All Are Built
}

void GLFont::init(QDomElement fontNode)
{
  if (!fontNode.isNull())
  {
	QString val;

	// initialize font
	char filename[80] = {0};
	sprintf(filename, "%s", (const char *)(fontNode.attribute("FILENAME")));
	int numx = (fontNode.attribute("NUMX")).toInt();
	int numy = (fontNode.attribute("NUMY")).toInt();
	int cwidth = (fontNode.attribute("CWIDTH")).toInt();
	int cheight = (fontNode.attribute("CHEIGHT")).toInt();
	int cspace = (fontNode.attribute("CSPACE")).toInt();

	// Set Font Texture, How Many Characters per row and how many columns
	SetTexture(filename,numx, numy);
	// Set Font Width and Font Height in Texture, and Character Spacing
	SetFontProperties(cwidth,cheight,cspace);

	BuildFont();    // Build The Font
	SetBase(32);	// Adjust Base Start Pos (First Set of Fonts)
  }
  else
  {
	  qWarning("Could not initialize GLFont!");
  }
}


