/************************************************************
 * Infinite Universe Engine:                                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File: JWGameEngine.cpp                                   *
 *                                                          *
 * Content: Implementation of the main graphics engine      *
 *          that controls, updates and renders all objects  *
 *          in our virtual universe                         *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#include "JWGameEngine.h"
#include "GUI/JWInfiniteUniverseEngine.h"
#include <qapplication.h>
#include <qgl.h>
#include "glext.h"
#include "Date.h"
#include "Model.h"
#include "JWPlanet.h"
#include "JWRing.h"
#include "JWSolarSystem.h"
#include "Starfield.h"
#include <qmessagebox.h> 
#include "qtxml.h"
#include "JWAstro.h"
#include "profile.h"
#include <qframe.h>

/**
 * The following section contains support for specific GL extensions we use
 */

// JWGLEXT

// some gobal rendering variables we need here
bool multitextureSupported = true;    // Flag Indicating Whether Multitexturing Is Supported
bool useMultitexture = true;          // Use It If It Is Supported?
GLint maxTexelUnits = 1;              // Number Of Texel-Pipelines. This Is At Least 1.

#if defined(_WIN32)
PFNGLMULTITEXCOORD1FARBPROC         glMultiTexCoord1fARB = NULL;
PFNGLMULTITEXCOORD2FARBPROC         glMultiTexCoord2fARB = NULL;
PFNGLMULTITEXCOORD3FARBPROC         glMultiTexCoord3fARB = NULL;
PFNGLMULTITEXCOORD4FARBPROC         glMultiTexCoord4fARB = NULL;
PFNGLACTIVETEXTUREARBPROC             glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;
#endif

/*
 Here we have some global needed extension checking code!
 */
#define __ARB_ENABLE true                  // Used To Disable ARB Extensions Entirely
//#define EXT_INFO                         // Do You Want To See Your Extensions At Start-Up?
#define MAX_EXTENSION_SPACE 10240          // Characters for Extension-Strings
#define MAX_EXTENSION_LENGTH 256           // Maximum Of Characters In One Extension-String

// Always Check For Extension-Availability During Run-Time!
// Here We Go!
bool isInString(char *string, const char *search) 
{
	int pos=0;
	int maxpos=strlen(search)-1;
	int len=strlen(string);	
	char *other;
	for (int i=0; i<len; i++) 
  {
		if ((i==0) || ((i>1) && string[i-1]=='\n')) 
    {				// New Extension Begins Here!
			other=&string[i];			
			pos=0;													// Begin New Search
			while (string[i]!='\n') 
      {								// Search Whole Extension-String
				if (string[i]==search[pos]) pos++;					// Next Position
				if ((pos>maxpos) && string[i+1]=='\n') return true; // We Have A Winner!
				i++;
			}			
		}
	}	
	return false;													// Sorry, Not Found!
}

// isMultitextureSupported() Checks At Run-Time If Multitexturing Is Supported
bool initMultitexture(void) 
{
	char *extensions;	
	extensions=strdup((char *) glGetString(GL_EXTENSIONS));			// Fetch Extension String
	int len=strlen(extensions);
	for (int i=0; i<len; i++)										// Separate It By Newline Instead Of Blank
		if (extensions[i]==' ') extensions[i]='\n';

#ifdef EXT_INFO
		QMessageBox::information(0, "supported GL extensions", extensions);
#endif

	if (isInString(extensions,"GL_ARB_multitexture")				// Is Multitexturing Supported?
		&& __ARB_ENABLE)												// Override-Flag
	{	
		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&maxTexelUnits);
#if defined(_WIN32)
		glMultiTexCoord1fARB  = (PFNGLMULTITEXCOORD1FARBPROC)   wglGetProcAddress("glMultiTexCoord1fARB");
		glMultiTexCoord2fARB  = (PFNGLMULTITEXCOORD2FARBPROC)   wglGetProcAddress("glMultiTexCoord2fARB");
		glMultiTexCoord3fARB  = (PFNGLMULTITEXCOORD3FARBPROC)   wglGetProcAddress("glMultiTexCoord3fARB");
		glMultiTexCoord4fARB  = (PFNGLMULTITEXCOORD4FARBPROC)   wglGetProcAddress("glMultiTexCoord4fARB");
		glActiveTextureARB    = (PFNGLACTIVETEXTUREARBPROC)     wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREARBPROC) wglGetProcAddress("glClientActiveTextureARB");		
#endif
#ifdef EXT_INFO
	QMessageBox::information(0, "Infinite Universe", "The GL_ARB_multitexture extension will be used.");
#endif
	  return true;
	}
	return false;
}
// end JWGLEXT



namespace jwengine {

	JWGameEngine *jwgame;
	JWInfiniteUniverseEngine *mainwindow;

	// text field to display initialization messages
	std::string message;

	void setMessage(std::string msg, TEXTURE *tex)
	{
		jwengine::message = msg;
		int fWidth = jwengine::jwgame->width();
		int fHeight = jwengine::jwgame->height();
		
		glClearColor(0,0,0,0);
		glClear( GL_COLOR_BUFFER_BIT );
		glColor3f(1,1,1);
		
		if (tex)
		{
			// init screen output
			glDisable(GL_LIGHTING);								// Turn off lighting
			glEnable(GL_TEXTURE_2D);
			
			glBindTexture(GL_TEXTURE_2D, tex->texID);		// Select Our Font Texture
			glDisable(GL_DEPTH_TEST);							// Disables Depth Testing
			glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
			glPushMatrix();										// Store The Projection Matrix
			glLoadIdentity();									// Reset The Projection Matrix
			glOrtho(0,fWidth,0,fHeight,-1,1);					// Set Up An Ortho Screen
			glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
			glPushMatrix();										// Store The Modelview Matrix
			glLoadIdentity();									// Reset The Modelview Matrix
			glPushAttrib(GL_BLEND);
			glDisable(GL_BLEND);
			
			// now draw a textured quad
			glBegin(GL_QUADS);							// Use A Quad For Each Character
				glTexCoord2f(0,0);						// Texture Coord (Bottom Left)
				glVertex2i(0,0);						// Vertex Coord (Bottom Left)
				glTexCoord2f(1,0);						// Texture Coord (Bottom Right)
				glVertex2i(fWidth,0);					// Vertex Coord (Bottom Right)
				glTexCoord2f(1,1);						// Texture Coord (Top Right)
				glVertex2i(fWidth,fHeight);				// Vertex Coord (Top Right)
				glTexCoord2f(0,1);						// Texture Coord (Top Left)
				glVertex2i(0,fHeight);					// Vertex Coord (Top Left)
			glEnd();									// Done Building Our Quad (Character)
			
			glPopAttrib();	//GL_BLEND
			glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
			glPopMatrix();										// Restore The Old Projection Matrix
			glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
			glPopMatrix();										// Restore The Old Projection Matrix
		}
		
		jwengine::jwgame->m_fFont.SetFontSize(32);
		jwengine::jwgame->m_fFont.Print(10, 300, (char *)jwengine::getMessage());
		qWarning(jwengine::getMessage());
		jwengine::jwgame->swapBuffers();				
	}
	
	const char *getMessage()
	{
		const char *cstr;
		cstr = jwengine::message.c_str();
		return cstr;
	}


}

// constructor with no real functionality (separate method)
JWGameEngine::JWGameEngine( QWidget* parent, const char* name ) : QGLWidget( parent, name )
{
	jwengine::jwgame = this;

	progress = 0;

	isReady = false;
	paused = false;
  stereomode = false;
  screenDistance = 4.0; // this is for the IBDS powerwall
  eyeSeparation = 0.1;  // simple eye distance

	// process some standard initializations
	wireframe = false;	// show wireframe terrain
	sunmode = false;	// show sun as massive model
	crashDetect = true;	// enables collision detection
	fog_mode = true;	// enables fog on planets
	ground_fog_depth = 40000; // this fog depth MUST be minimum 1000m!
	light_mode = false; // enables spacecraft light
	tex_mode = true;	// texturing on/off
	orbit = true;		// camera belongs to nearest planet?
	BFculling = true;	// enable backface culling
	diamondStep = true;	// tesselation regular or diamond structure?
	update_mode = true;	// toggles update of LOD tesselation for all planets (performance measurement)
	showProfiler = false;		// draw the profiler data (on/off), only for debug purposes (performance measurement)
  showInfos = true;
	camMode = false;	// move camera instead of player model (free camera)
	show_model = false; // the player model should not be displayed
	clipping = true;	// we enable view frustum clipping by default
	lineLOD = 0;		// special debug option to show the bounding radius of the patches
	QualityScale = 0;	// infinite different qualities (0=low, 1=medium, 2=high) 3,4,.. 
  userQuality = 0;  // minimum quality the user wishes
	timeSpeedup = 100.0;// as default 1 sec is 100 sec in virtual space (goes from 1 to 10000)
	screenshots = 0;	// current screenshot number

	setFocusPolicy(QWidget::StrongFocus);
	qtime.start();
}


// destructor: delete all pointer objects from memory
JWGameEngine::~JWGameEngine()
{
	// planets are destroyed in sol automatically!
	if (camera != 0)	{	delete camera; camera = 0; }
	if (playerModel != 0)	{	delete playerModel; playerModel = 0; }
}

//
// Handles mouse press events for the connect widget.
//

void JWGameEngine::mousePressEvent( QMouseEvent *e )
{
	if (e->button() & QWidget::LeftButton)
		o_mb = 1;
	if (e->button() & QWidget::RightButton)
		o_mb = 2;

	o_mx = e->x();
	o_my = e->y();
}


//
// Handles mouse release events for the connect widget.
//

void JWGameEngine::mouseReleaseEvent( QMouseEvent *e )
{
	o_mb = 0;
}


//
// Handles mouse move events for the connect widget.
//
void JWGameEngine::mouseMoveEvent( QMouseEvent *e )
{
	int mouse_x, mouse_y;
	mouse_x = e->x();
	mouse_y = e->y();
	QMouseEvent *me = new QMouseEvent(QEvent::MouseMove, QPoint(o_mx-mouse_x, o_my-mouse_y), 0, e->state());
	o_mx = mouse_x;	
	o_my = mouse_y;
	
	if (!paused)
	{
		// check which object should be moved because we can vary our focus of interest
		C3DBase *activeObj = 0;
		if (camMode)	// we want to move the camera
			activeObj = camera;
		else if (camera->GetParent())  // we directly control the player model if present
			activeObj = camera->GetParent();
		
		if (activeObj)
			activeObj->reactOnInput(me);
	}
}


/**
* Process input of keyboard
* @param k		a QT key event
*/
void JWGameEngine::keyPressEvent( QKeyEvent *k )
{

	// decide what to do in the engine for all cases
  switch ( k->key() ) 
	{
		
	case QEvent::Key_Escape:
    QApplication::exit();
		return;
	
	case QEvent::Key_H:
	case QEvent::Key_F1:
		paused = !paused;
		return;

	case QEvent::Key_B:	BFculling = !BFculling; return;
	case QEvent::Key_C:	camMode = !camMode; return;

  case QEvent::Key_D:
		if (k->state() & QEvent::ShiftButton)
			screenDistance -= 0.1;  // 10 cm closer
		else
			screenDistance += 0.1;  // 10 cm more
		if (screenDistance > 10)
			screenDistance = 10;
		if (screenDistance < 0.1)
			screenDistance = 0.1;    
    return;

  case QEvent::Key_E:
		if (k->state() & QEvent::ShiftButton)
			eyeSeparation -= 0.01;  // 1 cm closer
		else
			eyeSeparation += 0.01;  // 1 cm more
		if (eyeSeparation > 1)
			eyeSeparation = 1;
		if (eyeSeparation < 0.02)
			eyeSeparation = 0.02;    
    return;

  case QEvent::Key_PageUp:
    fovy = Clamp(10.0, 90.0, fovy - 5.0);
    return;

  case QEvent::Key_PageDown:
    fovy = Clamp(10.0, 90.0, fovy + 5.0);
    return;

  case QEvent::Key_F:
		if (!fog_mode)
		{
			ground_fog_depth = 625.0f;
			fog_mode = true;
		}
		else if (ground_fog_depth > 80000.0f)
			fog_mode = false;
		else
			ground_fog_depth *= 2;
		return;
	case QEvent::Key_G:	lineLOD = (lineLOD+1) % 24; return;
	case QEvent::Key_F2:	this->SaveScreenshot(); return;
	case QEvent::Key_I:	showInfos = !showInfos; return;
	case QEvent::Key_L: light_mode = !light_mode; return;
	case QEvent::Key_M:	show_model = !show_model; return;
	case QEvent::Key_O:	orbit = !orbit; return;
	case QEvent::Key_P:	showProfiler = !showProfiler; return;
	case QEvent::Key_Q:
		if (k->state() & QEvent::ShiftButton)
      userQuality--;
    else
      userQuality++;
    userQuality = CLAMP(0,5,userQuality);
    return;
	case QEvent::Key_R:	
		camera->ResetRotation();
		camera->SetPosition(CDoubleVector(0.0, 0.0, 2 * playerModel->boundingRadius));
		camera->speed = 0;
		return;
		
	case QEvent::Key_S:	stereomode = !stereomode; return;
	case QEvent::Key_T:	tex_mode = !tex_mode; return;
	case QEvent::Key_U:	update_mode = !update_mode; return;
	case QEvent::Key_V:	clipping = !clipping; return;
	case QEvent::Key_W:
		wireframe = !wireframe;
		if(wireframe == true) 
		{
			glPolygonMode(GL_FRONT, GL_LINE);
			glPolygonMode(GL_BACK, GL_LINE);
		}
		else
		{
			glPolygonMode(GL_FRONT, GL_FILL);
			glPolygonMode(GL_BACK, GL_FILL);
		}
		return;
		
	case QEvent::Key_Z:
		if (k->state() & QEvent::ShiftButton)
			timeSpeedup /= 10;
		else
			timeSpeedup *= 10;
		if (timeSpeedup > 1000000)
			timeSpeedup = 1000000;
		if (timeSpeedup < 0.01)
			timeSpeedup = 0.01;
		return;
	}

	// here we look what to do only when game is active
	if (!paused)
	{
	// check which object should be moved because we can vary our focus of interest
	C3DBase *activeObj = 0;
	if (camMode)	// we want to move the camera
		activeObj = camera;
	else if (camera->GetParent())  // we directly control the player model if present
		activeObj = camera->GetParent();

	if (activeObj)
		activeObj->reactOnInput(k);
	}

}

// do some timer triggered animation!
void JWGameEngine::animate()
{
    updateGL();
}

/**
  Paint the box. The actual openGL commands for drawing the box are
  performed here.
*/

void JWGameEngine::paintGL()
{
	if (isReady) // all objects in universe created
	{
		int msecs = qtime.restart();
		if (!paused)
		{
#if defined(_WIN32)
#ifdef _DEBUG
      // increment frame counter
      CProfileManager::Increment_Frame_Counter();

      // reset statistics every 100 frames
      if (CProfileManager::Get_Frame_Count_Since_Reset() > 100)
        CProfileManager::Reset();
#endif
#endif
      // call and measure LOD-update for the engine
      {
        PROFILE("LOD Update");
			  updateEngine(msecs);
      }
      // call and measure the rendering time
      {
        PROFILE("Render");
			  draw(msecs);	
      }
		}
		else
			drawMenu();
	}
	else
	{
		setAutoBufferSwap(false);
		this->run();
		setAutoBufferSwap(true);
	}
}

/**
  Set up the OpenGL rendering state, and define display list
*/
void JWGameEngine::initializeGL()
{
	// we need special extensions (see "glext.h")
	useMultitexture = initMultitexture();

	fovy = 45.0;	// fovy is vertical angle of viewport (45 by default, 30 better for first-person feeling)

	// GL stuff
	glPolygonMode(GL_BACK, GL_FILL);
	glPolygonMode(GL_FRONT, GL_FILL);
	glFrontFace(GL_CCW);
	glCullFace( GL_BACK );
	if (BFculling)
		glEnable( GL_CULL_FACE );
	
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations

	glShadeModel(GL_SMOOTH);

	glEnable(GL_BLEND);									// Turn on blending
	glEnable(GL_LIGHTING);								// Turn on lighting
	glEnable(GL_COLOR_MATERIAL);						// Allow color
	glEnable(GL_NORMALIZE);								// i normalize almost everything but this IS still necessary!!!

	// global lighting parameters
	GLfloat lmodel_ambient[] = { 0.2f, 0.2f, 0.2f, 1.0f };	// still show 20% when there is no light at all!
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);// expensive specular lighting!	

	// GL_LIGHT1 is a spotlight at camera position
	GLfloat light1_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light1_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light1_specular[] = {0.5, 0.5, 0.5, 1.0 };
	glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.5f);
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.000001f);
	glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 30.0f);
	glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 2.0f);
	GLfloat spot_direction[] = { 0, -0.2f, -1.0f };
	glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, spot_direction);

	/*
	QTXML	qtXml;

	// fetch parameters from XML files
	qtXml.readFile("Data/xml/config.xml");

	// create a system independant texture font
	QDomElement e = (qtXml.readUniqueNode("PARAMS:FONT")).toElement();
	m_fFont.init(e);
	*/

	// now start a thread to load the user defined data!

	//QThread::start(); // this will execute initEngine()!
	//this->run();
}

/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void JWGameEngine::resizeGL( int w, int h )
{
  glViewport( 0, 0, (GLint)w, (GLint)h );
	// Set up the viewing parameters.
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void JWGameEngine::run()
{
	initEngine();
	isReady = true;
}

/**
 * Initializes the engine, sets rendering options and creates all objects (planets, starfield, player model)
 */
void JWGameEngine::initEngine()
{
	QTXML	qtXml;
	QDomElement e;

	// fetch parameters from XML files
	qtXml.readFile("Data/xml/config.xml");

	// create a system independant texture font
	e = (qtXml.readUniqueNode("PARAMS:FONT")).toElement();
	m_fFont.init(e);

	e = (qtXml.readUniqueNode("PARAMS:PLAYER")).toElement();
	std::string playerdir = (const char *)(e.attribute("FILENAME", "none"));	// the player model (a jet)	
	std::string playerparent = (const char *)(e.attribute("PARENT", "none"));	// starting player position (Earth)

	e = (qtXml.readUniqueNode("PARAMS:PLANETLIST")).toElement();
	std::string planetdir = (const char *)(e.attribute("FILENAME", "none"));	// planet definition file (planets.xml)

	// set the date and time in universe!
	e = (qtXml.readUniqueNode("PARAMS:DATE")).toElement();
	JWDate today = JWDate(e);

	virtualTime = today.doubleValue();

	// check if we have a player model
	if (playerdir != "none")
	{
		qtXml.readFile(playerdir.c_str()); // Read in the config file

		e = (qtXml.readUniqueNode("MODEL")).toElement();
		std::string modeldir = (const char *)(e.attribute("DIR", "Data/Models/"));
		std::string plmodel = (const char *)(e.attribute("FILE", "jet.ms3d"));
		double plsize = (e.attribute("SIZE", "20.0")).toDouble();
		//deprecated: double plrot = (e.attribute("ROTATE", "0.0")).toDouble();
		
		// init player
		playerModel = new MilkshapeModel();	// Memory To Hold The Model
		if ( ((MilkshapeModel *)playerModel)->loadModelData( modeldir.c_str(), plmodel.c_str() ) == false )		// Loads The Model And Checks For Errors
			//qWarning("Couldn't load the player model");
			MessageBox(0,"No valid ms3d model!!!","Warning", MB_OK);

		((MilkshapeModel *)playerModel)->SetSize((float)plsize);
		//deprecated: ((MilkshapeModel *)playerModel)->baserotation = plrot;
	}
	else
	{
		// dummy "NULL" object (simulate 1st person view)
		playerModel = new C3DObject();
		playerModel->boundingRadius = 1.0;
	}

	// create a starfield background sphere with 10000 randomly distributed stars
	sf = new Starfield();

	// load the planet definition file into parser (main object is solarsytem)
	qtXml.readFile(planetdir.c_str());
	// initialize our solar system object
	e = (qtXml.readUniqueNode("SOLARSYSTEM")).toElement();
	sol.init(e);
	jwengine::mainwindow->initPlanetchoice();

	// bind player to parent planet if given
	if (playerparent != "none")
		playerModel->SetParent(sol.getObjectByName(playerparent.c_str()));
	if (playerModel->GetParent() != 0)
		playerModel->SetPosition(CDoubleVector(0.0, 0.0, playerModel->GetParent()->boundingRadius + 10000.0));	// enough to be outside every planet/star
	else
		playerModel->SetPosition(CDoubleVector(0.0, 0.0, 1.0e9));	// enough to be outside every planet/star
	playerModel->speed = 0;
	playerModel->accel = 1000;	// accelerate with 1000m/s^2	
	// place camera around planet
	camera = new C3DCamera();
	camera->SetParent(playerModel);
	camera->SetPosition(CDoubleVector(0.0, 0.0, 2 * playerModel->boundingRadius));
	camera->boundingRadius = 1;
	camera->speed = 0;
	camera->accel = 1;
}


/**
 * This method moves all active objects in the scene and calls the update functions of these objects.
 * The most important update is to recompute the tesselation of the planets
 * through calling "updateGeometry"
 * The other lines move the planets on their way around the sun
 * and of course the player object and the cameras (for now there only exists one).
 * @param millisec		time spent since last update in milliseconds
 */
void JWGameEngine::updateEngine(int millisec)
{
	virtualTime += (((millisec * timeSpeedup) * 0.001) / 86400.0);	// virtualTime is in julian date, so 1 second is 1/86400 day

	playerModel->move(C3DObject::FORWARD, playerModel->speed * millisec * 0.001);
	if (camMode)
		camera->move(C3DObject::FORWARD, camera->speed * millisec * 0.001);

	// now move all planets in this system and update their geometry
	sol.updateGeometry(camera);
	
	// we have to place the playerModel + camera at a different node in the scenegraph
	// if it belongs to the nearest planet (orbit around means constant relative position to that planet)
	C3DBase *np = sol.getNearestObject(camera);
	float pdist = (float)Abs(np->getHeight_above_Surface(camera));

	// we are closer than 1000km so we should be in orbit! (measurement will be changed if this is not sensible)
	if (pdist < Max(1000000.0, 5*np->boundingRadius))
	{
		// check if we are in orbit mode of np
		// if not we must change into
		if (np != playerModel->GetParent())
		{
			// we want to get into orbit, so first leave former system and then enter orbit
			if (playerModel->GetParent() != 0)
			{
				CDoubleVector d = playerModel->GetFullPosition();
				CDoubleQuaternion cview = playerModel->GetViewQuaternion(0);
				
				playerModel->ResetRotation();
				playerModel->Rotate(cview);
				playerModel->SetParent(0);
				playerModel->SetPosition(d);
			}
			// player and np belong to different coordinate systems, so use GetFullPosition!
			// Get the rotation of parent object
			CDoubleMatrix rotMat = np->GetViewMatrix(0);
			rotMat.Transpose();
			CDoubleVector d = rotMat * (playerModel->GetFullPosition() - np->GetFullPosition());
			CDoubleQuaternion cview = (np->GetViewQuaternion(0)).Inverse();
			playerModel->Rotate(cview);
			playerModel->SetParent(np);
			playerModel->SetPosition(d);
			playerModel->speed = 0.0;	// full stop since we got into the orbit!
		}
	}
	else
	{
		// no orbit required so change to full position
		if (playerModel->GetParent() != 0)
		{
			CDoubleVector d = playerModel->GetFullPosition();
			CDoubleQuaternion cview = playerModel->GetViewQuaternion(0);

			playerModel->ResetRotation();
			playerModel->Rotate(cview);
			playerModel->SetParent(0);
			playerModel->SetPosition(d);
		}
	}


	// collision detection with planetary surface!
	JWPlanet *nplan = dynamic_cast <JWPlanet *> (np);
	if ((crashDetect) && (nplan != 0))
	{
		// get Players Position in relation to nplan's coordinates
		C3DBase *cp = C3DBase::GetCommonParent(nplan, playerModel);
		CDoubleVector diff;
		if (playerModel->GetParent() != 0)
		  diff = playerModel->GetParent()->GetCameraMatrix(cp) * (playerModel->GetFullPosition(cp) - nplan->GetFullPosition(cp));
		else
		  diff = (playerModel->GetFullPosition() - nplan->GetFullPosition());

		double h = nplan->getHeight_above_Surface(playerModel);
		if (h < playerModel->boundingRadius)
		{
			CDoubleVector t;
			t = diff;
			t.Normalize();
			h = playerModel->boundingRadius - h;
			playerModel->SetPosition(playerModel->GetPosition() + t * h);
		}

		// now re-check with camera too
		// get camera Position in relation to nplan's coordinates
		cp = C3DBase::GetCommonParent(nplan, camera);
		if (camera->GetParent() != 0)
		  diff = camera->GetParent()->GetCameraMatrix(cp) * (camera->GetFullPosition(cp) - nplan->GetFullPosition(cp));
		else
		  diff = (camera->GetFullPosition() - nplan->GetFullPosition());

		h = nplan->getHeight_above_Surface(camera);
		if (h < camera->boundingRadius)
		{
			CDoubleVector t;
			t = diff;
			t.Normalize();
			h = camera->boundingRadius - h;
			camera->SetPosition(camera->GetPosition() + t * h);
		}
	}
}

void JWGameEngine::initCamera()
{
  camera->initFrustum(fovy, (double)viewWidth/(double)viewHeight * fovy, this->clipping);
}

extern int LOD_quality;
/**
 * This method draws our whole universe (for now some planets)
 * It also prints status information and a help screen
 * @param millisec		time since last rendered frame
 */
void JWGameEngine::draw(int millisec)
{
	// varibles for status information
	static char qualityName[6][20] = { "low", "medium", "high", "very high", "super high", "perfect" };
	static char szFrameCount[80] = {0};
	static char szSpeed[80] = {0};
	static char szDate[80] = {0};
	static char szAltitude[160] = {0};
	static char szTriag[80] = {0};
	static int nTime = 0;
	static int nFrames = 0;

  glViewport( 0, 0, width(), height() );
  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  QWidget *d = QApplication::desktop();
  int sw=d->width();                   // returns screen width
  int sh=d->height();                  // returns screen height
  int xpo = jwengine::mainwindow->geometry().x() + jwengine::mainwindow->glframe->geometry().x();
  int ypo = jwengine::mainwindow->geometry().y() + jwengine::mainwindow->glframe->geometry().y();

  // if we are in stereomode we must adjust the viewport parameters
  if (stereomode)
  {
    int leftBorder = xpo;
    int rightBorder = sw - width() - leftBorder;
    this->viewWidth = sw/2 - leftBorder - rightBorder;
    this->viewHeight = height();
  }
  else
  {
    viewWidth = width();
    viewHeight = height();
  }

  // Adjust Fonts 2D Drawing Rectangle Size 
	m_fFont.SetDisplayMode(viewWidth,viewHeight);
  
  int renderLoops = 1;
  if (stereomode)
    renderLoops = 2;

  int view;
  for (view=0; view < renderLoops; view++)
  {
    if (stereomode)
    {
      if (view == 0)
        activeView = CAM_LEFT;
      else
        activeView = CAM_RIGHT;
    }
    else
      activeView = CAM_CENTER;

    activeRange = V_FAR;

    glViewport( view*(sw/2), 0, viewWidth, viewHeight);
    // here we start drawing all objects in the universe
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    initViewFrustum();
    // make sure we manipulate the right thing
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if (activeView == CAM_RIGHT)
      glDrawBuffer(GL_BACK_RIGHT);
    else
      glDrawBuffer(GL_BACK_LEFT);
    glClear(GL_DEPTH_BUFFER_BIT);
    
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);
    
    // for now spotlight belongs to camera which defines our viewport -> fixed position
    // to do: integrate dynamic lights at any position
    GLfloat light1_position[] = { 0, 0, 0, 1.0 };
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    
    glDisable(GL_FOG);		// first render objects without fog
    glDisable(GL_BLEND);
    glDisable(GL_ALPHA_TEST);
    
    glEnable(GL_LIGHT0);
    // is our camera spotlight active?
    if (light_mode)
      glEnable(GL_LIGHT1);
    else
      glDisable(GL_LIGHT1);
    // backface culling could be turned off (on by default)
    if (BFculling)
      glEnable( GL_CULL_FACE );
    
    starAlpha = 1.0;
    
    // now compute status variables
    // Determine the FPS; recompute it every second
    nTime += millisec;
    if(nTime >= 1000)
    {
      float fFPS = (float)(nFrames * 1000) / (float)nTime;
      if (fFPS > 25.0f)
        QualityScale++;
      if (fFPS < 12.0f)
        QualityScale--;
      QualityScale = CLAMP(userQuality, 5, QualityScale);
      // when measuring performance we force constant quality no dynamic switch allowed!
      if (showProfiler)
        QualityScale = userQuality;
      sprintf(szFrameCount, "%2.2f FPS, %s Quality (minimum is %s Quality), %s fog view", fFPS, qualityName[MIN(5,QualityScale)], qualityName[MIN(5,userQuality)], distStr(ground_fog_depth));
      sprintf(szSpeed, "Speed %s/s (%2.2f * c)", distStr(playerModel->speed), playerModel->speed/3.0e8); 
      JWDate day(virtualTime);
      sprintf(szDate, "Date: %s  (timeSpeedup  = %6.2f )", day.print(), timeSpeedup); 
      nTime = nFrames = 0;
    }
    nFrames++;
    
    
    // get the nearest planet to the camera in this solar system
    C3DBase *nearestObject = sol.getNearestObject(camera);
    JWPlanet *nearestPlanet = dynamic_cast <JWPlanet *> (nearestObject);
    
    // compute height above terrain and altitude (they are not the same!)
    double alt, hNN;
    double alpha, beta, h;
    
    alt = (nearestObject->getHeight_above_Surface(camera));
    
    if (nearestPlanet)
    {
      hNN = ((nearestPlanet->GetFullPosition()-camera->GetFullPosition()).Magnitude() - nearestPlanet->radius);
      CIntVector campos = nearestPlanet->lastpos;
      CoordinatesToPolar(campos.x, campos.y, campos.z, &alpha, &beta, &h);
      alpha = Clamp(-180, 180, 180.0/PI * alpha);
      beta = Clamp(-90, 90, 90 - 180.0/PI * beta);
      sprintf(szAltitude, "nearest planet is %s: %s to ground, %s NN, NS[%3.2f] x WE[%3.2f]", nearestPlanet->getName(), distStr(alt), distStr(hNN), beta, alpha); 
      
      if (nearestPlanet->hasSky)
      {
        // my own approximation of atmosphere density (linear is not physically correct but looks nicer!)
        // cscale is factor to blend the atmosphere coloring into the general background color
        float cscale = (float)Clamp(0.0,1.0, (1-hNN/nearestPlanet->ATM_RANGE));
        // compute sun vector relative to nearest planet, needed for realistic lighting
        CDoubleVector d1 = - nearestPlanet->GetFullPosition(); // sun is always at (0,0,0)!
        CDoubleVector d2 = camera->GetFullPosition() - nearestPlanet->GetFullPosition();
        d1.Normalize();
        d2.Normalize();
        if (nearestPlanet->style == JWPlanet::SUN)
          d1 = d2;
        // cscale now is a product of atmospheric lighting and sun lighting
        // examples: cscale goes towards zero on the night side of the planet or if we are too high and leave the atmosphere (hNN > ATM_RANGE) 
        //           cscale goes towards one  if we are on the day side and deep enough in the atmosphere layer (near ground)
        cscale = (float)Clamp(0.0,1.0, (cscale * ((d1 | d2) + 1.0)));
        // above 90% density we do not see stars
        starAlpha =1 - 1.2f*cscale;
      }
    }
    else
    {
      hNN = alt; alpha = 0; beta = 0;
      sprintf(szAltitude, "nearest object is %s: %s to surface", nearestObject->getName(), distStr(alt)); 
    }
        
    // now adjust stereo parameters
    double eyeOffset;
    if (activeView == CAM_CENTER)
      eyeOffset = 0.0;
    else if (activeView == CAM_LEFT)
      eyeOffset = -eyeSeparation/2;
    else if (activeView == CAM_RIGHT)
      eyeOffset = eyeSeparation/2;
    CDoubleVector camshift = camera->GetRightAxis();
    camshift.Normalize();
    camshift *= eyeOffset;

    camera->SetPosition(camera->GetPosition() + camshift);

    sf->draw(camera);
    
    sol.draw(camera);
    
    if (playerModel && show_model)
      playerModel->draw(camera);
    
    camera->SetPosition(camera->GetPosition() - camshift);

    // compute number of rendered trianlges
    int numTris = sol.numTriangles;
    sprintf(szTriag,"%10d Triangles, %2d LODmin, %2d lineLOD", numTris, ((LODObject *)nearestObject)->computeLOD_min(camera), lineLOD); 
    
    // print put the statistics
    QTime ct = QTime::currentTime();
    glColor3f(1.0f, (float)(0.5*sin(PI*ct.second())+0.5), (float)(0.5*cos(PI*ct.second())+0.5));
    
    // display text messages
    //m_fFont.PrintTest(100, 200);
    m_fFont.SetFontSize(16);

    if (showInfos)
    {
      m_fFont.Print(10, 20, szDate);			// date
      m_fFont.Print(10, 40, szFrameCount);	// FPS counter
      m_fFont.Print(10, 60, szAltitude);		// altitude, hNN
      
      CDoubleVector v = - camera->GetFullPosition();	// sun is always at (0,0,0)!
      alt = v.Magnitude();
      if (playerModel->GetParent())
        sprintf(szAltitude, "%s to Sun, Camera in orbit of %s", distStr(alt), playerModel->GetParent()->getName()); 
      else
        sprintf(szAltitude, "%s to Sun, Camera in free space", distStr(alt)); 
      
      m_fFont.Print(10, 80, szAltitude);		// distance to sun
      
      m_fFont.Print(10, 100, szSpeed);			// speed of player
      m_fFont.Print(10, 120, szTriag);			// number of triangles / culled triangles
      m_fFont.Print(10, 140, "Press 'H' for help.");	// help screen
      
      sprintf(szAltitude, "Screensize %d x %d, Position of glframe (%d, %d) with size %d x %d", sw,sh, xpo, ypo, width(), height()); 
      m_fFont.Print(10, 160, szAltitude);
      
      if (stereomode)
      {
        sprintf(szAltitude, "Eye separation: %fm, distance to screen: %fm", eyeSeparation, screenDistance); 
        m_fFont.Print(10, 180, szAltitude);
      }
    }
    
    // we only can use the profiler with Windows for now
#if defined(_WIN32)
#ifdef _DEBUG
    if (showProfiler)
    {
      m_fFont.Print(10, 200, "Performance monitor");	// help screen
      CProfileIterator *perfIT = CProfileManager::Get_Iterator();
      CProfileNode *actualNode;
      // print out statistics
      int line = 220;
      for (perfIT->First(); !perfIT->Is_Done(); perfIT->Next())
      {
        sprintf(szAltitude, "Name:%s, Time/Frame:%f ms, Calls/Frame:%d",
          perfIT->Get_Current_Name(),            
          (perfIT->Get_Current_Total_Time() / (float)CProfileManager::Get_Frame_Count_Since_Reset()),
          (int)(perfIT->Get_Current_Total_Calls() / (float)CProfileManager::Get_Frame_Count_Since_Reset()) );
        
        m_fFont.Print(10, line, szAltitude);		// profiler data
        line += 20;
        // print children
        actualNode = perfIT->CurrentParent;
        perfIT->CurrentParent = perfIT->CurrentChild;
        
        for (perfIT->First(); !perfIT->Is_Done(); perfIT->Next())
        {
          sprintf(szAltitude, "Name:%s, Time/Frame:%f ms (%02d%% of %s), Calls/Frame:%d",
            perfIT->Get_Current_Name(),            
            (perfIT->Get_Current_Total_Time() / (float)CProfileManager::Get_Frame_Count_Since_Reset()),
            (int)(100*perfIT->Get_Current_Total_Time() / perfIT->Get_Current_Parent_Total_Time()),
            perfIT->Get_Current_Parent_Name(),            
            (int)(perfIT->Get_Current_Total_Calls() / (float)CProfileManager::Get_Frame_Count_Since_Reset()) );
          
          m_fFont.Print(30, line, szAltitude);		// profiler data
          line += 20;
        }
        perfIT->CurrentChild = perfIT->CurrentParent;
        perfIT->CurrentParent = actualNode;
      }
      CProfileManager::Release_Iterator(perfIT);
    }
#endif
#endif
  }
}

void JWGameEngine::drawMenu()
{
	static char tempstr[80];

	glClearColor(0.7f,0.9f,1,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(0.0f, 0.0f, 0.0f);

	m_fFont.SetFontSize(24);
	m_fFont.Print(10, 20, "Diese Hilfeseite mit 'H' oder 'F1' verlassen.");
	m_fFont.Print(10, 60, "Die Anwendung mit 'ESC' verlassen.");

	m_fFont.Print(10, 100, "Navigation:");

	m_fFont.SetFontSize(16);
	m_fFont.Print(40, 120, "Mouse mit linker Taste gepresst bewegen, um die Blickrichtung zu aendern. (Free View)");
	m_fFont.Print(40, 140, "Mouse mit rechter Taste gepresst bewegen fuer Rollbewegung um die Sichtachse. (Roll)");
	m_fFont.Print(40, 160, "'SPACE' druecken fuer sofortigen Stopp.");
	m_fFont.Print(40, 180, "'1' bis '0' setzen Beschleunigung (1 = 0.1m/(ss), 2 = 1m/(ss), ... 0 = Lichtgeschwindigkeit/s)");
	sprintf(tempstr, "Momentante Geschwindigkeit betraegt %s/s (%2.2f * c)", distStr(playerModel->speed), playerModel->speed/3.0e8);
	m_fFont.Print(40, 200, tempstr);
	m_fFont.Print(40, 220, "'NACH OBEN' druecken, um mit dem gesetzen Faktor zu beschleunigen.");
	m_fFont.Print(40, 240, "'NACH UNTEN' druecken, um zu bremsen.");
	m_fFont.Print(40, 260, "Taste '+' zoomt an Planeten heran, '-' zoomt heraus.");
	m_fFont.Print(40, 280, "Mit 'C' kann Kamera unabhaengig zum Flugzeugmodell bewegt werden (Look around)");
	if (camMode)
		m_fFont.Print(40, 300, "Kamera wird momentan unabhaengig bewegt. Wieder 'C' druecken, um normal zu navigieren.");
	else
		m_fFont.Print(40, 300, "Kamera wird momentan relativ zum Flugzeugmodell bewegt.");
	m_fFont.Print(40, 320, "'R' setzt Kamera zurueck, falls sie mit 'C' bewegt wurde.");

	m_fFont.SetFontSize(24);
	m_fFont.Print(10, 360, "Grafikoptionen:");

	m_fFont.SetFontSize(16);
	if (BFculling)
	m_fFont.Print(40, 380, "'B' (de)aktiviert backface culling in OpenGL (Im Moment AN).");
	else
	m_fFont.Print(40, 380, "'B' (de)aktiviert backface culling in OpenGL (Im Moment AUS).");

	m_fFont.Print(40, 400, "'F' (de)aktiviert Nebel fuer atmosphaerische Effekte");
	if (fog_mode)
		sprintf(tempstr, "    Atmosphaerische Sichtweite = %s", distStr(ground_fog_depth));
	else
		sprintf(tempstr, "Nebel ist nicht aktiv; Sichtweite unendlich!");
	m_fFont.Print(40, 420, tempstr);

	sprintf(tempstr, "'G' markiere alle Patches mit LOD %d", this->lineLOD);
	m_fFont.Print(40, 440, tempstr);
	if (this->light_mode)
	m_fFont.Print(40, 460, "'L' (de)aktiviert Scheinwerfer der Kamera (Spotlight ist AN).");
	else
	m_fFont.Print(40, 460, "'L' (de)aktiviert Scheinwerfer der Kamera (Spotlight ist AUS).");

	m_fFont.Print(40, 480, "'M' (de)aktiviert die Anzeige des Flugzeugs im Vordergrund.");
	m_fFont.Print(40, 500, "'Q'(+'SHIFT') schaltet durch die 5 Qualitaetsstufen \"low\"-\"super high\"");
	m_fFont.Print(40, 520, "'W' (de)aktiviert den Wireframe-Modus.");
	m_fFont.Print(40, 540, "'Z'(+'SHIFT') bestimmt den timeSpeedup-Faktor (0.01 - 10000), welcher die Tageslaenge definert.");
	m_fFont.Print(40, 560, "    (1 = Echtzeit, d.h. Tag dauert 24h | 10000 = Tag dauert 8,6 sec)");
}


// SaveScreenshot()
// desc: takes a snapshot of the current window contents by retrieving
//       the screen data with glReadPixels() and writing the data to a file
void JWGameEngine::SaveScreenshot()
{
  screenshots++;
  if (screenshots <= 999)
  {	  
		char filename[20] = {0};
    if (stereomode)
		  sprintf(filename, "frame%03d.jps", screenshots);
    else
		  sprintf(filename, "frame%03d.jpg", screenshots);


		TEXTURE screentex;
		// round size so that we have aligned data
		int w = (width()/8)*8;
		int h = (height()/8)*8;

		screentex.init(	w, h, GL_RGB);	// width and height of the OpenGL window
		// read the image data from the window
		glReadPixels(0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, screentex.imageData);
		// write the image data to a file
		screentex.WriteJPEG(filename);
  }
}


void JWGameEngine::initViewFrustum()
{
  double nearRange, farRange;

  if (activeRange == V_FAR)
  {
    nearRange = 1000.0;
    farRange = 160000000.0;
  }
  else
  {
		int gldepth;
		glGetIntegerv(GL_DEPTH_BITS, &gldepth);
    // near clipping at 0.2m, far clipping at infinity is theoretically possible! (see opengl.org)
    // we limit far depth so that 24 bit z-buffer is sufficient (16 bit is too inaccurate so set near up to 10m!)
		if (gldepth >= 24)
    {
      nearRange = 0.2;
      farRange = 160000.0;
    }
		else
    {
      nearRange = 10.0;
      farRange = 160000.0;
    }
  }


  // create a really nice frustum computation (so far everything is symmetric)
  double fovy_HALF = (double)(PI/180*(fovy)/2);
	double fovx_HALF = (double)(PI/180*(fovy)/2) * (double)viewWidth/(double)viewHeight;
  double right = screenDistance * tan(fovx_HALF);
  double left = -right;
  double top = screenDistance * tan(fovy_HALF);
  double bottom = -top;

  // now adjust stereo parameters
  double eyeOffset;
  if (activeView == CAM_CENTER)
    eyeOffset = 0.0;
  else if (activeView == CAM_LEFT)
    eyeOffset = -eyeSeparation/2;
  else if (activeView == CAM_RIGHT)
    eyeOffset = eyeSeparation/2;

  right -= eyeOffset;
  left -= eyeOffset;

  // project frustum extends to near plane
  double planeProjector = (nearRange / screenDistance);
  right *= planeProjector;
  left *= planeProjector;
  top *= planeProjector;
  bottom *= planeProjector;
  
  // finish
  glFrustum(left, right, bottom, top, nearRange, farRange);
  // instead of
  // gluPerspective(fovy, (double)viewWidth/(double)viewHeight, nearRange, farRange);
}



