/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWGameEngine.h                                  *
 *                                                          *
 * Content: Header file of the main graphics engine         *
 *			that controls, updates and renders all objects  *
 *			in our virtual universe                         *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#ifndef JWGAMEENGINE_H
#define JWGAMEENGINE_H

// QT defs
#include <qgl>
//#include <qthread.h> 
#include <qdatetime> 

#include <stdio>
#include "GLFont.h"
#include "JWSolarSystem.h"

class Model;
class JWPlanet;
class Starfield;

/**
 * This class manages the whole engine needed to display an interactive solar system.
 */
class JWGameEngine : public QGLWidget //, public QThread
{
  Q_OBJECT

public:
  typedef enum RENDER_MODE { CAM_CENTER=0, CAM_LEFT=1, CAM_RIGHT=2 }; 
  typedef enum VIEW_RANGE { V_NEAR=0, V_FAR=1 };

	// Engine options.
	bool wireframe;						// show wireframe terrain
	bool sunmode;							// show sun as massive model or only sprite
	bool crashDetect;					// enables collision detection between player and planet
	bool fog_mode;						// enables fog on planets
	float ground_fog_depth;		// this fog depth MUST be minimum 1000m!
	bool light_mode;					// enables spacecraft light
	bool tex_mode;						// texturing on/off
	bool orbit;								// player belongs to nearest planet?
	bool BFculling;						// enable backface culling
	bool diamondStep;					// tesselation regular or diamond structure?
	bool update_mode;					// toggles update of LOD tesselation for all planets (performance measurement)
	bool showProfiler;				// draw the profiler data (on/off), only for debug purposes (performance measurement)
  bool showInfos;           // draw status information (on/off)
	bool camMode;							// move camera instead of player model (free camera)
	bool show_model;					// should the player object be shown (on/off)
	bool clipping;
	bool paused;							// if we are in pause mode (menu)
  bool stereomode;          // do we render to stereo screen?
  double eyeSeparation;     // distance between left and right eye
  double screenDistance;    // distance to the screen (focus of eye lens)
  RENDER_MODE activeView;   // to distinguish between stereo and normal view (frustum setup)
  VIEW_RANGE activeRange;   // do we draw V_NEAR (0.1 - 10000) of V_FAR (> 10000) geometry at the moment?
	float starAlpha;					// how opaque do we draw our star background
	int lineLOD;
	int QualityScale;					// performance vs. quality factor for quadtree update
  int userQuality;          // minimum quality the user wishes
	int	screenshots;					// current screenshot number

	double timeSpeedup;				// how fast the time goes on
	double virtualTime;				// the time in our universe as julian date!
	QTime qtime;							// a QT FPS counter!
	
	// Window/Fullscreen variables.
  int viewWidth, viewHeight;  // width and height of the active viewport
	double fovy; // field of view angle in degrees (45.0) as default

	// Mouse input variables.
	int o_mx, o_my, o_mb; 

	// Display font
	GLFont m_fFont;		// Font Class  

	// Thread control variables	
	bool isReady;	// all data for the game engine is present
	float progress; // if we can compute how far the progress is we will use that variable in range [0,1]

	// The solar system object holds a vector of planets (maybe additional objects in later versions)
	JWSolarSystem sol;
	Starfield *sf;
	// Viewer variables.
	C3DCamera *camera;		// Camera defines what we see. It can be fixed at player model or move around independently.
	C3DObject *playerModel;	// playerModel is what we move in our universe (for now it's only a jet but in later versions
							// this should become an avatar that interacts with everything e.g. pilot of the jet!)

	// constructor
	JWGameEngine( QWidget* parent, const char* name );
	// destructor: delete all pointer objects from memory
	~JWGameEngine();

	// this thread function calls initEngine!
	virtual void run();

	/**
	* Initializes the engine, sets rendering options and creates all objects (planets, starfield, player model)
	*/
	void initEngine();
	
	/**
	* This method moves all active objects in the scene and calls the update functions of these objects.
	* The most important update is to recompute the tesselation of the planets
	* through calling "updateGeometry"
	* The other lines move the planets on their way around the sun
	* and of course the player object and the cameras (for now there only exists one).
	* @param millisec		time spent since last update in milliseconds
	*/
	void updateEngine(int millisec);
  void initCamera();
  void initViewFrustum();

	void draw(int millisec);
	void drawMenu();
	
	/**
	* This method is important for detail rendering because we
	* only want to compute and display details like clouds, fog, trees, buildings
	* for the planet we are in orbit.
	* The ohter planets are too far away to show details!
	* @param cam	A C3DCamera object that represents our active camera view.  
	* @return		The planet or other very large object that is the nearest one to the C3DCamera cam.
	*/
	LODObject* getNearestPlanet(C3DCamera* cam);

	// SaveScreenshot()
	// desc: takes a snapshot of the current window contents by retrieving
	//       the screen data with glReadPixels() and writing the data to a file
	void SaveScreenshot();

public slots:

	void animate();

protected:

    void initializeGL();
    void paintGL();
    void resizeGL( int w, int h );
		void keyPressEvent( QKeyEvent *k ); // reimplement keyboard functions
		void mousePressEvent( QMouseEvent *e );
		void mouseReleaseEvent( QMouseEvent *e );
		void mouseMoveEvent( QMouseEvent *e );

};

class JWInfiniteUniverseEngine;

// here we position global help routines
namespace jwengine {

	// here i define my "static" variables

	// a global shortcut to the engine (to get/set its options!)
	extern JWGameEngine *jwgame;
	// a global shortcur to the main window (to get/set its options!)
	extern JWInfiniteUniverseEngine *mainwindow;
	// text field to display initialization messages in the engine!
	extern std::string message;

	void setMessage(std::string msg, TEXTURE *tex=0);

	const char *getMessage();
};

#endif


