// JWNoise.cpp
// The white noise generator implementation
#include "JWNoise.h"
#include "Noise.h"
#include <limits.h>

// this initializes the noise field and permutates it
void JWNoise::Init(int nDimensions, unsigned int nSeed)
{
	m_nDimensions = MIN(nDimensions, MAX_DIMENSIONS);
  CRandom rn(nSeed);
  
	int i, j, k;
  int temp;
  for (i=0; i<NOISESIZE; i++)
    noiseMap[i] = i;
  
  for (i=0; i < 10000; i++)
  {
      j = rn.RandomI(0,NOISEMASK);
      k = rn.RandomI(0,NOISEMASK);
      temp = noiseMap[j];
      noiseMap[j] = noiseMap[k];
      noiseMap[k] = temp;
  }
}
		
// simple noise, repeating after NOISESIZE!
unsigned int JWNoise::simpleNoise(unsigned int x) { return noiseMap[(x & NOISEMASK)]; }
	
// 4-dimensional multilevel noise
// scrambles all bits of a 4d-int-vector to white noise
double JWNoise::Lattice(unsigned int x, unsigned int y, unsigned int z, unsigned int t)
{ 
	unsigned int ergebnis = noiseMap[0];
	while (x || y || z || t)
	{
	  ergebnis += simpleNoise(x + simpleNoise(y + simpleNoise(z + simpleNoise(t))));
	  x >>= 8;
	  y >>= 8;
	  z >>= 8;
	  t >>= 8;
  }
	ergebnis &= NOISEMASK;
	return (((ergebnis << 1)/(double)NOISESIZE) - 1.0);
}
	
// return noise in [-1..1] from interpolated 4d-lattice for whole integer space!
double JWNoise::Noise(double *f)
{
  double MAXUINT = UINT_MAX + 1.0;

	unsigned int n[MAX_DIMENSIONS];			// Indexes to pass to lattice function
	double r[MAX_DIMENSIONS];		        // Remainders to pass to lattice function
	double w[MAX_DIMENSIONS];       	  // Cubic values to pass to interpolation function

	for(int i=0; i<m_nDimensions; i++)
	{
    f[i] = Mod(f[i], MAXUINT);
		n[i] = (unsigned int)f[i];
		r[i] = f[i] - n[i];
		w[i] = Cubic(r[i]);
	}

	double fValue;
	switch(m_nDimensions)
	{
		case 1:
			fValue = Lerp(Lattice(n[0]), Lattice(n[0]+1), w[0]);
			break;
		case 2:
			fValue = Lerp(Lerp(Lattice(n[0], n[1]), Lattice(n[0]+1, n[1]), w[0]),
						  Lerp(Lattice(n[0], n[1]+1), Lattice(n[0]+1, n[1]+1), w[0]),  w[1]);
			break;
		case 3:
			fValue = Lerp(Lerp(Lerp(Lattice(n[0], n[1], n[2]), Lattice(n[0]+1, n[1], n[2]),	w[0]),
							   Lerp(Lattice(n[0], n[1]+1, n[2]), Lattice(n[0]+1, n[1]+1, n[2]),	w[0]),
							   w[1]),
						  Lerp(Lerp(Lattice(n[0], n[1], n[2]+1), Lattice(n[0]+1, n[1], n[2]+1),	w[0]),
							   Lerp(Lattice(n[0], n[1]+1, n[2]+1), Lattice(n[0]+1, n[1]+1, n[2]+1),	w[0]),
							   w[1]),
						  w[2]);
			break;
		case 4:
			fValue = Lerp(Lerp(Lerp(Lerp(Lattice(n[0], n[1], n[2], n[3]),
										 Lattice(n[0]+1, n[1], n[2], n[3]),
										 w[0]),
									Lerp(Lattice(n[0], n[1]+1, n[2], n[3]),
										 Lattice(n[0]+1, n[1]+1, n[2], n[3]),
										 w[0]),
									w[1]),
									Lerp(Lerp(Lattice(n[0], n[1], n[2]+1, n[3]),
										 Lattice(n[0]+1, n[1], n[2]+1, n[3]),
										 w[0]),
									Lerp(Lattice(n[0], n[1]+1, n[2]+1, n[3]),
										 Lattice(n[0]+1, n[1]+1, n[2]+1, n[3]),
										 w[0]),
									w[1]),
							   w[2]),
						  Lerp(Lerp(Lerp(Lattice(n[0], n[1], n[2], n[3]+1),
										 Lattice(n[0]+1, n[1], n[2], n[3]+1),
										 w[0]),
									Lerp(Lattice(n[0], n[1]+1, n[2], n[3]+1),
										 Lattice(n[0]+1, n[1]+1, n[2], n[3]+1),
										 w[0]),
									w[1]),
									Lerp(Lerp(Lattice(n[0], n[1], n[2]+1, n[3]+1),
										 Lattice(n[0]+1, n[1], n[2]+1, n[3]+1),
										 w[0]),
									Lerp(Lattice(n[0], n[1]+1, n[2]+1, n[3]+1),
										 Lattice(n[0]+1, n[1]+1, n[2]+1, n[3]+1),
										 w[0]),
									w[1]),
							   w[2]),
						  w[3]);
			break;
	}
	return CLAMP(-0.99999f, 0.99999f, fValue);

}
	
