/************************************************************
* Infinite Universe Engine:				                *
*                                                          *
* Interactive simulation of a solar system and             *
* real-time rendering of planetary bodies with OpenGL      *
* copyright  2002 Jochen Winzen                           *
* winzen@online.de                                         *
* -------------------------------------------------------- *
* File:	JWObject.cpp                                    *
*                                                          *
* Content: Implementation of some basic 3D object classes  *
*                                                          *
* Author:  Jochen Winzen                                   *
* Version: 2002/12/08                                      *
************************************************************
*/
//#if defined(_CC_MSVC_)
//#pragma warning(disable:4786)
//#endif

#include "JWObject.h"
#include "JWStaticObject.h"
#include "JWNoise.h"

void C3DBase::setName(const char *pName)
{ name = pName; }

const char *C3DBase::getName()
{ return name.c_str(); }

// local position routines
void C3DBase::SetPosition(const CDoubleVector &v)
{ m_vPosition = v; }

CDoubleVector C3DBase::GetPosition()
{ return m_vPosition; }

// scenegraph management
void C3DBase::SetParent(C3DBase *p)
{ m_pParent = p; }

C3DBase *C3DBase::GetParent()
{ return m_pParent; }

C3DBase *C3DBase::GetCommonParent(C3DBase *obj1, C3DBase *obj2)			
{
	// this method returns the first parent that both objects share
	// we need that parent as a common reference for position calculations
	if ((obj1 == 0) || (obj2 == 0))
		return 0;
	// create a vector of parent objects for each obj
	std::vector<C3DBase *> parvec1, parvec2;
	C3DBase *parobj, *cmpobj;
	
	parobj = obj1;
	while(parobj != 0)
	{
		parvec1.push_back(parobj);
		parobj = parobj->GetParent();
	}
	parobj = obj2;
	while(parobj != 0)
	{
		parvec2.push_back(parobj);
		parobj = parobj->GetParent();
	}
	
	// now return the "nearest" common parent if there is one
	
	C3DBase *commonParent = 0;
	
	while (!parvec1.empty() && !parvec2.empty())
	{
		parobj = parvec1.back();// as we go from back to front we walk down to the children
		cmpobj = parvec2.back();// to get the "nearest" common object on the path to obj1 and obj2
		parvec1.pop_back();		// shrink list
		parvec2.pop_back();		// ""
		
		if (parobj != cmpobj)
			return commonParent;
		else
			commonParent = parobj;	// parobj and cmpobj are equal so a possible common parent!
	}
	// at least one list is empty (shorter path)
	return commonParent;
};

CDoubleVector C3DBase::GetFullPosition(C3DBase *commonParent)
{
  return GetRelativePosition(commonParent, CDoubleVector(0,0,0));
}

CDoubleVector C3DBase::GetFullPosition()
{
  return GetRelativePosition(0, CDoubleVector(0,0,0));
}

// offset is a arbitrary vector relative to "this"
// now we translate this local position into an relative position to the commonParent
CDoubleVector C3DBase::GetRelativePosition(C3DBase *commonParent, CDoubleVector offset)	
{
		// this method returns absolute coordinates for objects that belong to local systems
		if (this == commonParent)
      return offset;
    else
    {
      // Get the rotation of this object
      CDoubleMatrix rotMat = *this;
      // and rotate offset about this, add this->m_vPosition to get coordinates relative to parent
      if (m_pParent != 0)
        return m_pParent->GetRelativePosition(commonParent, rotMat * offset + m_vPosition);
      else
        return (rotMat * offset + m_vPosition);
    }
}

CDoubleMatrix C3DBase::GetViewMatrix(C3DBase *commonParent)
{ return ((CDoubleMatrix)GetViewQuaternion(commonParent));}

CDoubleMatrix C3DBase::GetCameraMatrix(C3DBase *commonParent)
{
		CDoubleMatrix rotMat;
    rotMat = GetViewQuaternion(commonParent);
    rotMat.Transpose();
    return rotMat;
}

CDoubleQuaternion C3DBase::GetViewQuaternion(C3DBase *commonParent)
{
		// we do not check 0 since commonParent IS a parent of this!
		if (this == commonParent)
      return CDoubleQuaternion(0,0,0,1);
    else if (m_pParent == commonParent)
      return *this;
    else
      return (m_pParent->GetViewQuaternion(commonParent))* (*this);
}

CDoubleMatrix C3DBase::GetModelMatrix(C3DBase *pCamera)
{
		// This method delivers the ModelMatrix for the object we want to look at with pCamera.
		// The matrix is computed against the common parent to enhance numerical stability!
		// With this hierarchic method we can render infinitly large scenens (a universe or more)!
		// We do not need to rotate the position difference vector since m will be multiplied by
		// pCamera->GetCameraMatrix(cp) which shifts all positions relative to the camera!
		C3DBase *commonParent = C3DBase::GetCommonParent(this, pCamera);
    CDoubleMatrix m;
    m.ModelMatrix( this->GetViewQuaternion(commonParent),
      (GetFullPosition(commonParent) - pCamera->GetFullPosition(commonParent)) );
    return m;
}

CDoubleMatrix C3DBase::GetScaledModelMatrix(C3DBase *pCamera)
{
	// This code scales the planet's size and distance to the camera down when it's too far away.
	
	// This solves a problem with many video card drivers where objects too far away aren't rendering properly
	// It also alleviates the Z-buffer precision problem caused by having your near and far clipping planes too far apart.
	static double MAX_DISTANCE=(double)16000000.0;		// Distance to desired far clipping plane
	double MAX_DISCERNABLE;
	static double HALF_MAX=(double)(MAX_DISTANCE*0.5);	// Everything between HALF_MAX and MAX_DISCERNABLE is scaled exponentially between HALF_MAX and MAX_DISTANCE
	CDoubleMatrix m = GetModelMatrix(pCamera);
	CDoubleVector v;
	C3DBase *commonParent = C3DBase::GetCommonParent(this, pCamera);
	v = (GetFullPosition(commonParent) - pCamera->GetFullPosition(commonParent));
	
	double dDistance = v.Magnitude();
	if(dDistance > HALF_MAX)
	{
		// Beyond this distance, everything is rendered at MAX_DISTANCE
		MAX_DISCERNABLE = Max((double)MAX_DISTANCE, (double)(1000*boundingRadius));
		v /= dDistance;
		double dFactor = MAX_DISTANCE;
		if(dDistance < MAX_DISCERNABLE)
			dFactor = (double)(HALF_MAX + HALF_MAX * (1.0 - exp((HALF_MAX - dDistance) / MAX_DISCERNABLE)));
		else
			dDistance = MAX_DISCERNABLE;
		
		v *= dFactor;
		m.f2[3][0] = v.x;
		m.f2[3][1] = v.y;
		m.f2[3][2] = v.z;
		dFactor /= dDistance;
		m.Scale(dFactor, dFactor, dFactor);
	}
	return m;
}


/**
* This method is important for detail rendering because we
* only want to compute and display details like clouds, fog, trees, buildings
* for the planet we are in orbit.
* The ohter planets are too far away to show details!
* @param cam	A C3DBase object that represents our active camera view.  
* @return		The planet that is the nearest one to the C3DBase cam.
*/
C3DBase *C3DBase::getNearestObject(C3DBase* cam)
{
	// Iterator is used to loop through the vector sol.planets.
	std::vector<C3DBase *>::iterator theIterator;
	double maxdist = 1.0e80; // really far away ;-)
	C3DBase *retobj = 0;
	// Lookup contents of the vector sol.planets.
	for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
	{
		// compute distance between planet and camera
		(*theIterator)->camdist = (*theIterator)->getHeight_above_Surface(cam);
		if ((*theIterator)->camdist < maxdist)
		{
			retobj = (*theIterator);
			maxdist = (*theIterator)->camdist;
		}
	}
	
	return retobj;
}

void C3DBase::warpToPosition(C3DBase *obj, C3DBase *dest)
{
		CDoubleVector newpos = CDoubleVector(1,0,0)*(obj->boundingRadius + dest->boundingRadius);
    obj->SetPosition(newpos);
    obj->SetParent(dest);
}

C3DBase *C3DBase::getObjectByName(const char *pname)
{
		// Iterator is used to loop through the components
		std::vector<C3DBase *>::iterator theIterator;
    // Lookup contents of the vector sol.planets.
    for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
    {
      // compare name field
      if (strcmp((*theIterator)->getName(), pname) == 0)
        return (*theIterator);
    }
    return 0;
}

// now we can compare 2 components and sort the vector according!
bool C3DBase::distanceSort( C3DBase *x, C3DBase *y) { 
		return x->camdist < y->camdist;
};

// now we can compare 2 components alphabetically and sort the vector according!
bool C3DBase::alphaSort( C3DBase *x, C3DBase *y) { 
		return (strcmp(x->getName(), y->getName()) < 0);
};

// Test if this object belongs to the hierarchy of possibleParent
bool C3DBase::belongsToObject(C3DBase *possibleParent)
{
		if (m_pParent)
    {
      if (m_pParent == possibleParent)
        return true;
      else
        return m_pParent->belongsToObject(possibleParent);
    }
    // this does not belong to the hierarchy of possibleParent
    return false;
};

/**
* Computes the height of the object over the surface of this geometry.
* We search the affected quadtree and delegate the query to it.
* @param obj		the object
* @return			height above surface in meters
*/
double C3DBase::getHeight_above_Surface(C3DBase *obj)
{
		CDoubleVector diff = obj->GetFullPosition() - this->GetFullPosition();
    // since we have no surface simply return distance to this object
    double dist = diff.Magnitude() - obj->boundingRadius - this->boundingRadius;
    return dist;
};


/**
 * A simple movement function for the object with 6 DOF
 * @param type		Defines how to move in space e.g. translate, rotate etc.
 * @param amount	How far should we move
 */
void C3DObject::move(MOVEMENT type, double amount)
{
	CDoubleVector t;
	switch(type)
	{
	case FORWARD:
		this->SetPosition(this->GetPosition() + this->GetViewAxis() * amount);
		break;
	case ROTATE:
		this->Rotate(this->GetUpAxis(), amount );
		break;
	case LOOKUP:
		this->Rotate(this->GetRightAxis(), amount );
		break;
	case ZOOM:
			t = this->GetPosition();
			this->SetPosition(t * amount);
		break;
	case ROLL:
		this->Rotate(this->GetViewAxis(), amount );
		break;
	}
	this->Normalize();
}

void C3DObject::reactOnInput(QEvent *e)
{
	QKeyEvent *ke=0;
	QMouseEvent *me=0;
	switch (e->type())
	{
	case QEvent::KeyPress:
		ke = (QKeyEvent *)e;
		
		switch ( ke->key() ) 
		{
			
		case QEvent::Key_Up:		
			speed += accel;
			break;
			
		case QEvent::Key_Down:
			if (speed > 0)
				speed /= 2;
			else
				speed -= accel;
			break;
			
		case QEvent::Key_Space:
			speed = 0;
			break;
			
		case QEvent::Key_Plus:
			move(C3DObject::ZOOM, 0.99);
			break;
			
		case QEvent::Key_Minus:
			move(C3DObject::ZOOM, 1/0.99);
			break;
			
		case QEvent::Key_1:	accel = 0.1; return;
		case QEvent::Key_2:	accel = 1; return;
		case QEvent::Key_3:	accel = 10; return;
		case QEvent::Key_4:	accel = 100; return;
		case QEvent::Key_5:	accel = 1000; return;
		case QEvent::Key_6:	accel = 10000; return;
		case QEvent::Key_7:	accel = 100000; return;
		case QEvent::Key_8:	accel = 1000000; return;
		case QEvent::Key_9:	accel = 10000000; return;
		case QEvent::Key_0:	accel = 300000000; return;
		}
		
		break;
		
	case QEvent::MouseMove:
		me = (QMouseEvent *)e;

		if (me->state() & QEvent::LeftButton) {
			move(C3DObject::ROTATE, (double)(me->x()/100.0));						
			move(C3DObject::LOOKUP, (double)(me->y()/100.0));
		}
		else if (me->state() & QEvent::RightButton) {
			move(C3DObject::ROLL, (double)(-me->x()/100.0));						
		}		
		
		break;
	}
}


// initialize static variables here (stupid concept!)
int LODObject::instancecount = 0;
double LODObject::screen_error = 0.1;
int LODObject::grid_size = 0;
int LODObject::gs = 0;		// initial grid size of a quadtree patch e.g 16x16 (gs = grid_size+1)
int *LODObject::RegularIndexList[4] = {0,0,0,0};
int *LODObject::indexList[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int LODObject::indexLenght[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int *LODObject::fanindexList[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int LODObject::fanindexLenght[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int *LODObject::stripindexList[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int LODObject::stripindexLenght[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
float *LODObject::colorBlend[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
LODObject::TRIANGULATION LODObject::tmode = LODObject::TRIANGLE_FAN; // this is the fastest mode!

LODObject::LODObject()
{
	numTriangles = 0;
	m_pParent = NULL;
	boundingRadius = 0;
	lastpos = CIntVector(0,0,0);
	
	LODObject::instancecount++;
	// create triangulation lists with first planet
	if (LODObject::instancecount == 1)
		// for now let this be constant, creates 8x8 quadtree patches for ALL planets!
		LODObject::initTriangulation(8);
	
	INVALID = new JWStaticObject();	// we need a dummy static object as geom. occluder since 0 is only used for free space
	INVALID->isSolid = true;

	// default clear fields
  axis_angle = 0;
	orbit = 0;
	year_length = 0;
	day_length = 0;

	char tempc[80];
	sprintf(tempc, "unknown%d", LODObject::instancecount);	// each object has a name attribute, can be unknown ;-)
	this->setName(tempc);
	sprintf(center, "none");	// each object has a center attribute, which can be none at all		
};

LODObject::~LODObject()
{
	int i;
	
	LODObject::instancecount--;
	
	if (LODObject::instancecount == 0)
	{
		// delete the colorBlend-arrays
		for( i=0; i<16; i++)
		{
			if (colorBlend[i] != 0) { delete[] colorBlend[i]; colorBlend[i] = 0; }
		}
		// delete precomputed triangulation lists when the last planet is deleted!
		indexList[3]=0;
		indexList[5]=0;
		indexList[8+7]=0;
		indexList[8+5]=0;
		indexList[2*8+1]=0;
		indexList[2*8+3]=0;
		indexList[3*8+1]=0;
		indexList[3*8+7]=0;
		
		for( i=0; i<32; i++)
			if (indexList[i] != 0)	{ delete[] indexList[i]; indexList[i]=0; }

		fanindexList[3]=0;
		fanindexList[5]=0;
		fanindexList[8+7]=0;
		fanindexList[8+5]=0;
		fanindexList[2*8+1]=0;
		fanindexList[2*8+3]=0;
		fanindexList[3*8+1]=0;
		fanindexList[3*8+7]=0;
		
		for( i=0; i<32; i++)
			if (fanindexList[i] != 0)	{ delete[] fanindexList[i]; fanindexList[i]=0; }

		stripindexList[3]=0;
		stripindexList[5]=0;
		stripindexList[8+7]=0;
		stripindexList[8+5]=0;
		stripindexList[2*8+1]=0;
		stripindexList[2*8+3]=0;
		stripindexList[3*8+1]=0;
		stripindexList[3*8+7]=0;
		
		for( i=0; i<32; i++)
			if (stripindexList[i] != 0)	{ delete[] stripindexList[i]; stripindexList[i]=0; }

		for( i=0; i<4; i++)
			if (RegularIndexList[i] != 0)	{ delete[] RegularIndexList[i]; RegularIndexList[i]=0; }
	}
	
	if (INVALID != 0) { delete INVALID; INVALID = 0; }
	objects.erase(objects.begin(), objects.end());
	
};

// all LODobjects have the computeLOD_min function
/**
* We compute the smallest LOD we actually need if we are looking on terrain from height altitude
* @param altitude	height above terrain in meters
* @result			result is stored most often in "LOD_min" which defines how many LODs must be computed
*/
int LODObject::computeLOD_min(C3DCamera *camera)
{
	// in Hhe "altitude" passen (2*tan(fovy/2)*altitude/element_size) Elemente auf den Bildschirm
	// im feinsten LOD werden diese in GRID_SIZE Elemente gepackt
	double altitude = Abs(getHeight_above_Surface(camera));
	// how much base units do we see if we look straight on the object (equals 2 for 90 fovy)
  double fovView = 2*tan(camera->fovy/2*PI/180.0)  * (altitude / element_size);
	// scaling with grid_size so that LODs are not too rough!
	return log2((unsigned int)Clamp(LODObject::grid_size, cube_size, fovView * LODObject::screen_error)); 
};

/* create a specific triangulation scheme for all quadtree patches of LODObjects!
* @param pGrid_size	constant that defines the basic grid size of a quadtree patch i.e. 8x8, 16x16 etc. (8 seems to be best choice!)
*/
void LODObject::initTriangulation(int pGrid_size)
{
	int i,l, base, gh, gf;
	grid_size = pGrid_size;
	gs = grid_size+1;
	
	gh = grid_size/2;
	gf = grid_size/4;

	// now compute 16 color blend arrays for all cases
	for (l = 0; l < 16; l++)
	{
		char filename[80];
		sprintf(filename, "Data/blend/blend%02d.bmp", l);

		TEXTURE *Texi = new TEXTURE();
		// load blending image
		if (!Texi->LoadImage(filename, false))
			  qWarning("Could not open texture file %s", filename);
		int byp = Texi->bpp/8;
		// init array
		colorBlend[l] = new float[4*gs*gs];

		for (i=0; i < 4*gs*gs; i++)
			colorBlend[l][i] = 1.0f;
		int x,y;
		for (y=0; y < gs; y++)
		{
			for (x=0; x < gs; x++)
			{
				// now fill in the correct alpha value
				if (Texi->imageData)
					colorBlend[l][4*(y*gs + x) + 3] = CLAMP(0,1, ((Texi->imageData[byp*(y*gs+x) + 0] & 0xFF) / 256.0f));
				else
					colorBlend[l][4*(y*gs + x) + 3] = 1;
			}
		}
		delete Texi;
	}

	// precompute triangulation lists for all 4 child-regions of a quad
	// create diamond triangulation structure
	// use with GL_TRIANGLES
	for (i=0; i<32; i++)
	{
		indexList[i] = 0;
	}
	
	for( i=0; i<4; i++)
	{
		base = (i & 1)*gh + (i & 2)*gh/2*gs; // uneven i means +gh in this row (x value), (i&2) means +gh rows!
		// make 8 lists for each subquad
		// divide space so:
		//
		//  ______________
		// |\            /|     The even numbers consist of 2*gf*gf triangles  -------
		// |  \ 4 or 5 /  |                                (6*gf*gf vertices)  \|/|\|/
		// |    \    /    |                                                      ---
		// | 6 or \/  2 or|                                                      \|/
		// |  7   /\   3  |                                                       
		// |    /    \    |   The odd numbers consist of 2*gf*gf-gf triangles  -------                                                
		// |  / 0 or 1 \  |   and realize the transitions between LODs         \ /|\ /
		// |/____________\|                                                      ---
		//                                                                       \|/
		//
		for (l = 0; l < 8; l++)
		{
			indexList[8*i + l] = new int[6*gf*gf -3*(l & 1)*gf]; // #vertices see above
			int *indexXY = new int[12*gf*gf];	// every vertex is stored here as (x,y), so 2*#vertices
			int indexCount = 0;
			int x,y;
			// create Lists for one section (here 0,1 adaption to other sectors is done by rotation of (x,y) coords!)
			for (y = -gf; y < 0; y++)
			{
				for (x = y+1; x < -y; x += 2)
				{
					if (y == -gf)
					{
						if  ((l & 1)==1)
						{
							// add 1 big triangle
							indexXY[indexCount + 0] = x-1;
							indexXY[indexCount + 1] = y;
							indexXY[indexCount + 2] = x+1;
							indexXY[indexCount + 3] = y;
							indexXY[indexCount + 4] = x;
							indexXY[indexCount + 5] = y+1;
							indexCount += 6;
						}
						else
						{
							// add 2 normal triangles
							// left triangle
							indexXY[indexCount + 0] = x-1;
							indexXY[indexCount + 1] = y;
							indexXY[indexCount + 2] = x;
							indexXY[indexCount + 3] = y;
							indexXY[indexCount + 4] = x;
							indexXY[indexCount + 5] = y+1;
							indexCount += 6;
							// right triangle
							indexXY[indexCount + 0] = x;
							indexXY[indexCount + 1] = y;
							indexXY[indexCount + 2] = x+1;
							indexXY[indexCount + 3] = y;
							indexXY[indexCount + 4] = x;
							indexXY[indexCount + 5] = y+1;
							indexCount += 6;
						}
					}
					else
					{
						// add 4 normal triangles
						// upper left triangle
						indexXY[indexCount + 0] = x-1;
						indexXY[indexCount + 1] = y;
						indexXY[indexCount + 2] = x;
						indexXY[indexCount + 3] = y;
						indexXY[indexCount + 4] = x;
						indexXY[indexCount + 5] = y+1;
						indexCount += 6;
						// upper right triangle
						indexXY[indexCount + 0] = x;
						indexXY[indexCount + 1] = y;
						indexXY[indexCount + 2] = x+1;
						indexXY[indexCount + 3] = y;
						indexXY[indexCount + 4] = x;
						indexXY[indexCount + 5] = y+1;
						indexCount += 6;
						// lower left triangle
						indexXY[indexCount + 0] = x-1;
						indexXY[indexCount + 1] = y;
						indexXY[indexCount + 2] = x;
						indexXY[indexCount + 3] = y-1;
						indexXY[indexCount + 4] = x;
						indexXY[indexCount + 5] = y;
						indexCount += 6;
						// lower right triangle
						indexXY[indexCount + 0] = x;
						indexXY[indexCount + 1] = y;
						indexXY[indexCount + 2] = x;
						indexXY[indexCount + 3] = y-1;
						indexXY[indexCount + 4] = x+1;
						indexXY[indexCount + 5] = y;
						indexCount += 6;
					}
				}
			}
			
			int rotationCounter = l/2;
			int idx, r, tempXY;
			
			for (idx = 0; idx < indexCount; idx += 2)
			{
				// rotate (x,y) about r*90 degrees!
				for (r = 0; r < rotationCounter; r++)
				{
					tempXY = indexXY[idx + 0];
					indexXY[idx + 0] = - indexXY[idx + 1];
					indexXY[idx + 1] = tempXY;
				}		
				// compute real index for indexList now
				indexList[8*i + l][idx/2] = base + (indexXY[idx + 1] + gf)*gs + (indexXY[idx + 0] + gf); 	
			}
			indexLenght[8*i + l] = indexCount/2;
			
			if (indexXY != 0) { delete[] indexXY; indexXY = 0; }
		}
		// remove the wrong indexLists (inside the quad is no bigger LOD possible!)
		switch(i)
		{
		case 0:
			if (indexList[8*i + 3] != 0) { delete[] indexList[8*i + 3]; indexList[8*i + 3] = 0; }
			indexList[8*i + 3] = indexList[8*i + 2];
			indexLenght[8*i + 3] = indexLenght[8*i + 2];
			if (indexList[8*i + 5] != 0) { delete[] indexList[8*i + 5]; indexList[8*i + 5] = 0; }
			indexList[8*i + 5] = indexList[8*i + 4];
			indexLenght[8*i + 5] = indexLenght[8*i + 4];
			break;
		case 1:
			if (indexList[8*i + 7] != 0) { delete[] indexList[8*i + 7]; indexList[8*i + 7] = 0; }
			indexList[8*i + 7] = indexList[8*i + 6];
			indexLenght[8*i + 7] = indexLenght[8*i + 6];
			if (indexList[8*i + 5] != 0) { delete[] indexList[8*i + 5]; indexList[8*i + 5] = 0; }
			indexList[8*i + 5] = indexList[8*i + 4];
			indexLenght[8*i + 5] = indexLenght[8*i + 4];
			break;
		case 2:
			if (indexList[8*i + 1] != 0) { delete[] indexList[8*i + 1]; indexList[8*i + 1] = 0; }
			indexList[8*i + 1] = indexList[8*i + 0];
			indexLenght[8*i + 1] = indexLenght[8*i + 0];
			if (indexList[8*i + 3] != 0) { delete[] indexList[8*i + 3]; indexList[8*i + 3] = 0; }
			indexList[8*i + 3] = indexList[8*i + 2];
			indexLenght[8*i + 3] = indexLenght[8*i + 2];
			break;
		case 3:
			if (indexList[8*i + 1] != 0) { delete[] indexList[8*i + 1]; indexList[8*i + 1] = 0; }
			indexList[8*i + 1] = indexList[8*i + 0];
			indexLenght[8*i + 1] = indexLenght[8*i + 0];
			if (indexList[8*i + 7] != 0) { delete[] indexList[8*i + 7]; indexList[8*i + 7] = 0; }
			indexList[8*i + 7] = indexList[8*i + 6];
			indexLenght[8*i + 7] = indexLenght[8*i + 6];
			break;
		}
	}
	
	// precompute triangulation lists for all 4 child-regions of a quad
	// create diamond triangulation structure
	// use with GL_TRIANGLE_FAN
	for (i=0; i<32; i++)
	{
		fanindexList[i] = 0;
	}
	
	for( i=0; i<4; i++)
	{
		base = (i & 1)*gh + (i & 2)*gh/2*gs; // uneven i means +gh in this row (x value), (i&2) means +gh rows!
		// make 8 lists for each subquad
		// divide space so:
		//
		//  ______________
		// |\            /|     The even numbers consist of 2*gf*gf triangles  -------
		// |  \ 4 or 5 /  |                       (in a Fan:3*gf*gf vertices)  \|/|\|/
		// |    \    /    |                                                      ---
		// | 6 or \/  2 or|                                                      \|/
		// |  7   /\   3  |                                                       
		// |    /    \    |   The odd numbers consist of 2*gf*gf triangles     -------                                                
		// |  / 0 or 1 \  |   and realize the transitions between LODs         \ /|\ /
		// |/____________\|                                                      ---
		//                                                                       \|/
		//
		for (l = 0; l < 8; l++)
		{
			fanindexList[8*i + l] = new int[3*gf*gf]; // #vertices see above
			int *indexXY = new int[6*gf*gf];	// every vertex is stored here as (x,y), so 2*#vertices
			int indexCount = 0;
			int x,y;

			if  ((l & 1)==1)
			{
				// now create a list for 2 triangle fans with one big side!
				// left triangle fan
				x = -1; y = -1;
				indexXY[indexCount + 0] = x;
				indexXY[indexCount + 1] = y;
				indexXY[indexCount + 2] = x-1;
				indexXY[indexCount + 3] = y-1;
				indexXY[indexCount + 4] = x-1;
				indexXY[indexCount + 5] = y-1;
				indexXY[indexCount + 6] = x+1;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x+1;
				indexXY[indexCount + 9] = y;
				indexXY[indexCount + 10] = x+1;
				indexXY[indexCount + 11] = y+1;
				indexCount += 12;
				// right triangle fan
				x = 1; y = -1;
				indexXY[indexCount + 0] = x;
				indexXY[indexCount + 1] = y;
				indexXY[indexCount + 2] = x-1;
				indexXY[indexCount + 3] = y+1;
				indexXY[indexCount + 4] = x-1;
				indexXY[indexCount + 5] = y;
				indexXY[indexCount + 6] = x-1;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x-1;
				indexXY[indexCount + 9] = y-1;
				indexXY[indexCount + 10] = x+1;
				indexXY[indexCount + 11] = y-1;
				indexCount += 12;
			}
			else
			{
				// now create a list for 2 triangle fans with all sides refined!
				// left triangle fan
				x = -1; y = -1;
				indexXY[indexCount + 0] = x;
				indexXY[indexCount + 1] = y;
				indexXY[indexCount + 2] = x-1;
				indexXY[indexCount + 3] = y-1;
				indexXY[indexCount + 4] = x;
				indexXY[indexCount + 5] = y-1;
				indexXY[indexCount + 6] = x+1;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x+1;
				indexXY[indexCount + 9] = y;
				indexXY[indexCount + 10] = x+1;
				indexXY[indexCount + 11] = y+1;
				indexCount += 12;
				// right triangle fan
				x = 1; y = -1;
				indexXY[indexCount + 0] = x;
				indexXY[indexCount + 1] = y;
				indexXY[indexCount + 2] = x-1;
				indexXY[indexCount + 3] = y+1;
				indexXY[indexCount + 4] = x-1;
				indexXY[indexCount + 5] = y;
				indexXY[indexCount + 6] = x-1;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x;
				indexXY[indexCount + 9] = y-1;
				indexXY[indexCount + 10] = x+1;
				indexXY[indexCount + 11] = y-1;
				indexCount += 12;
			}
			int rotationCounter = l/2;
			int idx, r, tempXY;
			
			for (idx = 0; idx < indexCount; idx += 2)
			{
				// rotate (x,y) about r*90 degrees!
				for (r = 0; r < rotationCounter; r++)
				{
					tempXY = indexXY[idx + 0];
					indexXY[idx + 0] = - indexXY[idx + 1];
					indexXY[idx + 1] = tempXY;
				}		
				// compute real index for indexList now
				fanindexList[8*i + l][idx/2] = base + (indexXY[idx + 1] + gf)*gs + (indexXY[idx + 0] + gf); 	
			}
			fanindexLenght[8*i + l] = indexCount/2;
			
			if (indexXY != 0) { delete[] indexXY; indexXY = 0; }
		}
		// remove the wrong indexLists (inside the quad is no bigger LOD possible!)
		switch(i)
		{
		case 0:
			if (fanindexList[8*i + 3] != 0) { delete[] fanindexList[8*i + 3]; fanindexList[8*i + 3] = 0; }
			fanindexList[8*i + 3] = fanindexList[8*i + 2];
			fanindexLenght[8*i + 3] = fanindexLenght[8*i + 2];
			if (fanindexList[8*i + 5] != 0) { delete[] fanindexList[8*i + 5]; fanindexList[8*i + 5] = 0; }
			fanindexList[8*i + 5] = fanindexList[8*i + 4];
			fanindexLenght[8*i + 5] = fanindexLenght[8*i + 4];
			break;
		case 1:
			if (fanindexList[8*i + 7] != 0) { delete[] fanindexList[8*i + 7]; fanindexList[8*i + 7] = 0; }
			fanindexList[8*i + 7] = fanindexList[8*i + 6];
			fanindexLenght[8*i + 7] = fanindexLenght[8*i + 6];
			if (fanindexList[8*i + 5] != 0) { delete[] fanindexList[8*i + 5]; fanindexList[8*i + 5] = 0; }
			fanindexList[8*i + 5] = fanindexList[8*i + 4];
			fanindexLenght[8*i + 5] = fanindexLenght[8*i + 4];
			break;
		case 2:
			if (fanindexList[8*i + 1] != 0) { delete[] fanindexList[8*i + 1]; fanindexList[8*i + 1] = 0; }
			fanindexList[8*i + 1] = fanindexList[8*i + 0];
			fanindexLenght[8*i + 1] = fanindexLenght[8*i + 0];
			if (fanindexList[8*i + 3] != 0) { delete[] fanindexList[8*i + 3]; fanindexList[8*i + 3] = 0; }
			fanindexList[8*i + 3] = fanindexList[8*i + 2];
			fanindexLenght[8*i + 3] = fanindexLenght[8*i + 2];
			break;
		case 3:
			if (fanindexList[8*i + 1] != 0) { delete[] fanindexList[8*i + 1]; fanindexList[8*i + 1] = 0; }
			fanindexList[8*i + 1] = fanindexList[8*i + 0];
			fanindexLenght[8*i + 1] = fanindexLenght[8*i + 0];
			if (fanindexList[8*i + 7] != 0) { delete[] fanindexList[8*i + 7]; fanindexList[8*i + 7] = 0; }
			fanindexList[8*i + 7] = fanindexList[8*i + 6];
			fanindexLenght[8*i + 7] = fanindexLenght[8*i + 6];
			break;
		}
	}
	
	// precompute triangulation lists for all 4 child-regions of a quad
	// create diamond triangulation structure
	// use with GL_TRIANGLE_STRIP
	for (i=0; i<32; i++)
	{
		stripindexList[i] = 0;
	}
	
	for( i=0; i<4; i++)
	{

		base = (i & 1)*gh + (i & 2)*gh/2*gs; // uneven i means +gh in this row (x value), (i&2) means +gh rows!
		// make 8 lists for each subquad
		// divide space so:
		//
		//  ______________
		// |\            /|     The even numbers consist of 2*gf*gf triangles  -------
		// |  \ 4 or 5 /  |                                (6*gf*gf vertices)  \|/|\|/
		// |    \    /    |                                                      ---
		// | 6 or \/  2 or|                                                      \|/
		// |  7   /\   3  |                                                       
		// |    /    \    |   The odd numbers consist of 2*gf*gf-gf triangles  -------                                                
		// |  / 0 or 1 \  |   and realize the transitions between LODs         \ /|\ /
		// |/____________\|                                                      ---
		//                                                                       \|/
		//
		for (l = 0; l < 8; l++)
		{
			stripindexList[8*i + l] = new int[13]; // #vertices see above
			int *indexXY = new int[26];	// every vertex is stored here as (x,y), so 2*#vertices
			int indexCount = 0;
			int x,y;

			if  ((l & 1)==0)
			{
				// now create a list for 1 triangle strip with all sides refined!
				x = 0; y = 0;
				indexXY[indexCount + 0] = x+2;
				indexXY[indexCount + 1] = y-2;
				indexXY[indexCount + 2] = x+1;
				indexXY[indexCount + 3] = y-1;
				indexXY[indexCount + 4] = x+1;
				indexXY[indexCount + 5] = y-2;
				indexXY[indexCount + 6] = x;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x;
				indexXY[indexCount + 9] = y-2;
				indexXY[indexCount + 10] = x-1;
				indexXY[indexCount + 11] = y-1;
				indexXY[indexCount + 12] = x-1;
				indexXY[indexCount + 13] = y-2;
				indexXY[indexCount + 14] = x-2;
				indexXY[indexCount + 15] = y-2;
				indexXY[indexCount + 16] = x-1;
				indexXY[indexCount + 17] = y-1;
				indexXY[indexCount + 18] = x-1;
				indexXY[indexCount + 19] = y-1;
				indexXY[indexCount + 20] = x;
				indexXY[indexCount + 21] = y;
				indexXY[indexCount + 22] = x;
				indexXY[indexCount + 23] = y-1;
				indexXY[indexCount + 24] = x+1;
				indexXY[indexCount + 25] = y-1;
				indexCount += 26;
			}
			else
			{
				// now create a list for 1 triangle strip with one big side!
				x = 0; y = 0;
				indexXY[indexCount + 0] = x+2;
				indexXY[indexCount + 1] = y-2;
				indexXY[indexCount + 2] = x+1;
				indexXY[indexCount + 3] = y-1;
				indexXY[indexCount + 4] = x;
				indexXY[indexCount + 5] = y-2;
				indexXY[indexCount + 6] = x;
				indexXY[indexCount + 7] = y-1;
				indexXY[indexCount + 8] = x-1;
				indexXY[indexCount + 9] = y-1;
				indexXY[indexCount + 10] = x;
				indexXY[indexCount + 11] = y-2;
				indexXY[indexCount + 12] = x-2;
				indexXY[indexCount + 13] = y-2;
				indexXY[indexCount + 14] = x-1;
				indexXY[indexCount + 15] = y-1;
				indexXY[indexCount + 16] = x-1;
				indexXY[indexCount + 17] = y-1;
				indexXY[indexCount + 18] = x;
				indexXY[indexCount + 19] = y-1;
				indexXY[indexCount + 20] = x;
				indexXY[indexCount + 21] = y;
				indexXY[indexCount + 22] = x+1;
				indexXY[indexCount + 23] = y-1;
				indexCount += 24;
			}
			int rotationCounter = l/2;
			int idx, r, tempXY;
			
			for (idx = 0; idx < indexCount; idx += 2)
			{
				// rotate (x,y) about r*90 degrees!
				for (r = 0; r < rotationCounter; r++)
				{
					tempXY = indexXY[idx + 0];
					indexXY[idx + 0] = - indexXY[idx + 1];
					indexXY[idx + 1] = tempXY;
				}		
				// compute real index for indexList now
				stripindexList[8*i + l][idx/2] = base + (indexXY[idx + 1] + gf)*gs + (indexXY[idx + 0] + gf); 	
			}
			stripindexLenght[8*i + l] = indexCount/2;
			
			if (indexXY != 0) { delete[] indexXY; indexXY = 0; }
		}
		// remove the wrong indexLists (inside the quad is no bigger LOD possible!)
		switch(i)
		{
		case 0:
			if (stripindexList[8*i + 3] != 0) { delete[] stripindexList[8*i + 3]; stripindexList[8*i + 3] = 0; }
			stripindexList[8*i + 3] = stripindexList[8*i + 2];
			stripindexLenght[8*i + 3] = stripindexLenght[8*i + 2];
			if (stripindexList[8*i + 5] != 0) { delete[] stripindexList[8*i + 5]; stripindexList[8*i + 5] = 0; }
			stripindexList[8*i + 5] = stripindexList[8*i + 4];
			stripindexLenght[8*i + 5] = stripindexLenght[8*i + 4];
			break;
		case 1:
			if (stripindexList[8*i + 7] != 0) { delete[] stripindexList[8*i + 7]; stripindexList[8*i + 7] = 0; }
			stripindexList[8*i + 7] = stripindexList[8*i + 6];
			stripindexLenght[8*i + 7] = stripindexLenght[8*i + 6];
			if (stripindexList[8*i + 5] != 0) { delete[] stripindexList[8*i + 5]; stripindexList[8*i + 5] = 0; }
			stripindexList[8*i + 5] = stripindexList[8*i + 4];
			stripindexLenght[8*i + 5] = stripindexLenght[8*i + 4];
			break;
		case 2:
			if (stripindexList[8*i + 1] != 0) { delete[] stripindexList[8*i + 1]; stripindexList[8*i + 1] = 0; }
			stripindexList[8*i + 1] = stripindexList[8*i + 0];
			stripindexLenght[8*i + 1] = stripindexLenght[8*i + 0];
			if (stripindexList[8*i + 3] != 0) { delete[] stripindexList[8*i + 3]; stripindexList[8*i + 3] = 0; }
			stripindexList[8*i + 3] = stripindexList[8*i + 2];
			stripindexLenght[8*i + 3] = stripindexLenght[8*i + 2];
			break;
		case 3:
			if (stripindexList[8*i + 1] != 0) { delete[] stripindexList[8*i + 1]; stripindexList[8*i + 1] = 0; }
			stripindexList[8*i + 1] = stripindexList[8*i + 0];
			stripindexLenght[8*i + 1] = stripindexLenght[8*i + 0];
			if (stripindexList[8*i + 7] != 0) { delete[] stripindexList[8*i + 7]; stripindexList[8*i + 7] = 0; }
			stripindexList[8*i + 7] = stripindexList[8*i + 6];
			stripindexLenght[8*i + 7] = stripindexLenght[8*i + 6];
			break;
		}
	}

	// precompute triangulation lists for all 4 child-regions of a quad
	// use a regular grid with triangle_strip instead of "diamonds"
	//
	//  -----------
	// |    /|    /
	// |  /  |  /   ....
	// |/    |/
	//  -----------
	// to do: implement crack fixing for this structure too
	for (i=0; i<4; i++)
	{
		RegularIndexList[i] = 0;
	}
	
	int j,k;
	for( i=0; i<4; i++)
	{
		RegularIndexList[i] = new int[2*gh*(gh+1)]; // 2 Triangles for each cell, gh*gh cells
		base = (i & 1)*gh + (i & 2)*gh/2*gs; // uneven i means +gh in this row (x value), (i&2) means +gh rows!
		
		for (j=0; j < gh; j++)
		{
			for (k=0; k <= gh; k++)
			{
				RegularIndexList[i][2*(j*(gh+1)+k)+0] = base + (j+1)*gs + k;
				RegularIndexList[i][2*(j*(gh+1)+k)+1] = base + j*gs + k;
			}
		}
	}
};	



/** This method checks if a ray given by (o + t*d) intersects the triangle (v0v1v2) or not
* @param u	the barycentric coordinate for v1
* @param v	the barycentric coordinate for v2  
* @param t  parameter on ray
* @return   intersects (true/false) and correct values u,v,t
*/
bool RayTriangleIntersection(CDoubleVector o, CDoubleVector d, CDoubleVector v0, CDoubleVector v1, CDoubleVector v2, double *u, double *v, double *t)
{
  // code follows Real-Time Rendering, p.305
  CDoubleVector e1 = v1-v0;
  CDoubleVector e2 = v2-v0;
  CDoubleVector p = d ^ e2; // cross product

  double a = e1 | p;		// determinant for solutions of projection into u,v space
  if ((a > -1.0e-5) && (a < 1.0e-5))
	  return false;

  double f = 1/a;
  CDoubleVector s = o-v0;
  *u = f*(s | p);
  if ((*u < 0.0) || (*u > 1.0))
	  return false;
  CDoubleVector q = s ^ e1;
  *v = f*(d | q);
  if ((*v < 0.0) || ((*u+*v) > 1.0))
	  return false;
  *t = f*(e2 | q);
  return true;
}





