/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWObject.h                                      *
 *                                                          *
 * Content: Header file for some basic 3D object classes    *
 *          e.g movable objects and cameras                 *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#ifndef JWOBJECT_H
#define JWOBJECT_H

#include "Matrix.h"		// algebra functions and classes (Vector, Matrix etc.)
#include "Texture.h"
#include "qtxml.h"
#include "profile.h"
#include "JWNoise.h"

class C3DBase;
class C3DObject;
class C3DCamera;
class LODObject;
class JWStaticObject;

/****************************************************************************
* Class: C3DBase
*****************************************************************************
* This class represents the most basic 3D object in the scene. It has a 3D
* position, an orientation, and a parent which provides its frame of reference
* in the scene. For instance, this could be a gun turret on a ship. It's
* really part of the ship, but can also be treated as a separate object. This
* is really just a base class meant to be derived from to provide some
* standardized functionality.
* Note: This class is derived from CDoubleQuaternion so it will inherit useful
*       functions like Rotate(), GetViewAxis(), GetUpAxis(), and GetRightAxis().
****************************************************************************/
class C3DBase : public CDoubleQuaternion
{
protected:
	C3DBase *m_pParent;			// this object belongs to m_pParent and rotates+translates with it!
	CDoubleVector m_vPosition;	// this is the local position relative to m_pParent
	double camdist;				// distance to camera (will be set internally and used to make a back to front sorting for drawing!)
  std::string name;				// reference name of this object (must be at leat local unique)

public:
	double boundingRadius;	// this variable defines the radius of a bounding sphere to that object!
	int numTriangles;		// total amount of triangles/frame
	// here we store all objects that belong to this one
	// NOTE: These objects must be part of this (like tires on a car) and
	// MUST NEVER leave the boundingRadius since we want to optimize our scenegraph!
	std::vector<C3DBase *> components;

	C3DBase() : CDoubleQuaternion(0.0, 0.0, 0.0, 1.0), m_vPosition(0.0), m_pParent(0), boundingRadius(1), numTriangles(0)	{	};
	virtual ~C3DBase() { components.erase(components.begin(), components.end()); };

	// all objects can have a predefined behaviour on events (key_events to steer a vehicle etc.)
	virtual void reactOnInput(QEvent *e) { };
	// all objects inherit an updateGeometry function
	virtual void updateGeometry(C3DCamera *pCamera) { };
	// all objects inherit an empty draw function
	virtual void draw(C3DCamera *) { };

  void setName(const char *pName);
	const char *getName();
	// local position routines
	void SetPosition(const CDoubleVector &v);
	CDoubleVector GetPosition();
	// scenegraph management
	void SetParent(C3DBase *p);
	C3DBase *GetParent();

	static C3DBase *GetCommonParent(C3DBase *obj1, C3DBase *obj2);

	CDoubleVector GetFullPosition(C3DBase *commonParent);
	CDoubleVector GetFullPosition();

	// offset is a arbitrary vector relative to "this"
	// now we translate this local position into an relative position to the commonParent
	CDoubleVector GetRelativePosition(C3DBase *commonParent, CDoubleVector offset);

	CDoubleMatrix GetViewMatrix(C3DBase *commonParent);

	CDoubleMatrix GetCameraMatrix(C3DBase *commonParent);

	CDoubleQuaternion GetViewQuaternion(C3DBase *commonParent);

	CDoubleMatrix GetModelMatrix(C3DBase *pCamera);

	CDoubleMatrix GetScaledModelMatrix(C3DBase *pCamera);

	static void warpToPosition(C3DBase *obj, C3DBase *dest);

	C3DBase *getObjectByName(const char *pname);

	C3DBase *getNearestObject(C3DBase* cam);

	// now we can compare 2 components and sort the vector according!
	static bool distanceSort( C3DBase *x, C3DBase *y);

	// now we can compare 2 components alphabetically and sort the vector according!
	static bool alphaSort( C3DBase *x, C3DBase *y);

	// Test if this object belongs to the hierarchy of possibleParent
	bool belongsToObject(C3DBase *possibleParent);

	/**
	* Computes the height of the object over the surface of this geometry.
	* We search the affected quadtree and delegate the query to it.
	* @param obj		the object
	* @return			height above surface in meters
	*/
	virtual double getHeight_above_Surface(C3DBase *obj);
};

// end C3DBase



/****************************************************************************
* Class: C3DObject
*****************************************************************************
* Adding to C3DBase, this class represents a movable object in the scene. It
* has a velocity vector and certain methods to make using it easier.
****************************************************************************/
class C3DObject : public C3DBase
{
public:
	typedef enum MOVEMENT{ FORWARD=1, ROTATE=2, LOOKUP=3, ZOOM=4, ROLL=5 };		// specific cases of movement (6 DOF)

	CDoubleVector m_vVelocity;
	double speed; // somewhat redundant but easier for my purposes than using a real velocity-vector
	double accel;

	C3DObject() : C3DBase(), m_vVelocity(0.0)	{ speed = 0; accel = 0; }
	void SetVelocity(CDoubleVector &v)			{ m_vVelocity = v; }
	CDoubleVector GetVelocity()					{ return m_vVelocity; }
	void AddVelocity(CDoubleVector &v)			{ m_vVelocity += v; }
	void UpdatePosition(double fSeconds)		{ m_vPosition += m_vVelocity * fSeconds; }
	void UpdatePosition(int millisec)			{ UpdatePosition((double)(millisec * 0.001)); }

	virtual void reactOnInput(QEvent *e);

	/**
	* A simple movement function for the object with 6 DOF
	* @param type		Defines how to move in space e.g. translate, rotate etc.
	* @param amount	How far should we move
	*/
	void move(MOVEMENT type, double amount);
};
// end C3DObject


/****************************************************************************
* Class: C3DCamera
*****************************************************************************
* every camera has its own view frustum!
* 2002/07/11 IMPORTANT NOTE: I tried some different versions of the frustum and isVisible methods
* BUT especially isVisible is tested VERY often and has to be fast, faster, fastest!!!
* So the only way was to use the FullPosition values instead of relative positions!
* An early test showed that this DOUBLES the program speed (20 FPS instead of 10 FPS)!!!!!!
*****************************************************************************/
class C3DCamera : public C3DObject
{
protected:
	bool clipping;
	CDoubleVector base;
	CDoubleVector view;
	CDoubleVector up;
	CDoubleVector right;
	CDoubleVector clip[5];

public:
	double fovy,fovx;

  inline bool isVisible(C3DBase *obj)
  {
    PROFILE("isVisible");
    
    if (!clipping)
      return true;
    
    CDoubleVector diff = obj->GetFullPosition() - base;
    // wenn Objekt hinter Kamera (mit Sicherheitsabstand boundingRadius)
    double dist;
    int i;
    // perform clipping
    for (i=0; i<5; i++) 
    {
      dist = diff | clip[i];
      if (dist < -obj->boundingRadius)
        return false;
    }
    return true;
  };
		
	bool isVisible(C3DBase *planet, CDoubleVector newcenter, double newboundingRadius)
	{
    PROFILE("isVisible");

    if (!clipping)
			return true;

		CDoubleVector diff = planet->GetFullPosition() - base;
		diff += (planet->GetViewMatrix(0) * newcenter);
		// wenn Objekt hinter Kamera (mit Sicherheitsabstand boundingRadius)
		double dist;
		int i;
		// perform clipping
		for (i=0; i<5; i++) 
		{
			dist = diff | clip[i];
			if (dist < -newboundingRadius)
				return false;
		}
		return true;
	};

	void initFrustum(double pfovy, double pfovx, bool pClipping)
	{
		fovy = pfovy;
		fovx = pfovx;
		clipping = pClipping;
		base = this->GetFullPosition();
		CDoubleQuaternion q = this->GetViewQuaternion(0);
		view = q.GetViewAxis();
		view.Normalize();
		up = q.GetUpAxis();
		up.Normalize();
		right = q.GetRightAxis();
		right.Normalize();
		// create normal clipping planes (+1 against numeric errors!)
		double fovy_HALF = (double)(PI/180*(fovy+1)/2);
		double fovx_HALF = (double)(PI/180*(fovx+1)/2);
		clip[0] = view;
		clip[1] = view + right/(double)tan(fovx_HALF);
		clip[1].Normalize();
		clip[2] = view - right/(double)tan(fovx_HALF);
		clip[2].Normalize();
		clip[3] = view - up/(double)tan(fovy_HALF);
		clip[3].Normalize();
		clip[4] = view + up/(double)tan(fovy_HALF);
		clip[4].Normalize();
	};
};
// end C3DCamera




/****************************************************************************
* Class: LODObject
*****************************************************************************
* This complex class contains all quadtree- and LOD-management algorithms
*****************************************************************************/
class LODObject : public C3DBase
{
private:
	static int instancecount;	// we count the number of LODObjects created!
	static double screen_error; // a measurement for maximal possible screen pixel error (default: 1/10 of screenwidth)

public:
	typedef enum TRIANGULATION { TRIANGLES=0, TRIANGLE_FAN=1, TRIANGLE_STRIP=2 };

	static TRIANGULATION tmode;

	static int grid_size , gs;		// initial grid size of a quadtree patch e.g 16x16 (gs = grid_size+1)
	static int *RegularIndexList[4];// indexListen fr die 4 Teilbereiche des Quad zur Darstellung mit regulrem Gitter
	static int *indexList[32];		// indexListen fr die 4 Teilbereiche des Quad zur Darstellung als "Diamond"-Triangles
	static int indexLenght[32];		// counts how many indexes are used!
	static int *fanindexList[32];		// indexListen fr die 4 Teilbereiche des Quad zur Darstellung als "Diamond"-Triangles
	static int fanindexLenght[32];		// counts how many indexes are used!
	static int *stripindexList[32];		// indexListen fr die 4 Teilbereiche des Quad zur Darstellung als "Diamond"-Triangles
	static int stripindexLenght[32];		// counts how many indexes are used!
	static float *colorBlend[16];	// 16 color arrays used to smoothly blend the global texture to pure vertex colors! 

	JWNoise jwfrac;			// noise generator (1 for each object!)

	char center[80];		// name of center object of this object (in most cases 0, but for moons, rings etc. the parent planet!)
	CIntVector lastpos;		// surface position of camera in last rendered frame (for update semantics)

	int HF_width, HF_height;// real size of a loaded map for heightfield HF, if artificial map equal to HF_size
	int HF_size;			// HF_size is a power of 2 (e.g 1024) and <= HF_height
	int HF_LOD;				// LOD of textured size

	bool fineLOD;			// this is a switch for the depth buffer is so far that we use a finer precision
							// for the fine LOD patches (<= LOD_quality) of the nearest planet!
	int LODalt;				// minimal LOD to display
	int LOD_diff;			// LOD difference between 2 following frames

	std::vector<JWStaticObject *> objects;	// here we store all nice planetary objects!
	JWStaticObject	*INVALID;	// a default NOP object!
	
	double H_RANGE;		// maximum height value of terrain in m
	int H_LOG;			// log2(H_RANGE) to save extra computations

	double axis_angle;	// angle against (1,0,0) in radians
	double orbit;		// distance of planet to its center (sun or other planet)
	// NOTE: if you want clockwise rotation year_length/day_length must be negative!
	double year_length; // time length of one full round about its center in standard days (i.e. earth=365, moon=29, ...)
	double day_length;  // rotation time in standard hours (i.e. earth=24, moon=0, ...)

	int cube_size;			// a power of 2 that defines the LOD cube of the planet
	double element_size;	// how much m is a LOD_min-elements width (sth. like smallest voxel ;-) 

	LODObject();
	// destructor
	virtual ~LODObject();
		
	void setCenter(const char *pName) { sprintf(center, "%s", pName); };
	char *getCenter() { return center; };
	
	int computeLOD_min(C3DCamera *camera);

	virtual void init(QDomElement node) = 0;

	/**

	* This method converts an arbitrary (x,y,z)-vector into a virtual cube position (CIntVector)
	* We need the int^3-vector for LOD management and random functions!
	* @param x		x position
	* @param y		y position
	* @param z		z position
	* @param c_s	index of the responsible cube side (0-5) for the position (x,y,z)
	* @param dist   distance from base geometry surface in meters
	* @return		Virtual cube position in range [-CUBE_HALF, CUBE_HALF]^3 and c_s
	*/
	virtual CIntVector CoordinatesToSurfaceVector(const double x, const double y, const double z, int *c_s, double *dist) = 0;

	virtual CDoubleVector SurfaceVectorToCoordinates(CIntVector pos, int c_s, double h) = 0;
	virtual CDoubleVector SurfaceVectorNormal(CIntVector pos, int c_s) = 0;


	virtual double get_HeightField_And_RandomOffset(CIntVector point, int LOD, double h_old) = 0;

	/**
	* This method creates a color value for each terrain vertex according to the planet type (see style def.) 
	* For now only height of terrain is regarded to create a color.
	* @param hf		the height value of a vertex for that we compute the terrain color
	* @param pos	vertex position
	* @return		quadrupel (r,g,b,a) as color
	*/
	virtual CFloatVector4 computeColor(const double hf, CIntVector pos = CIntVector(0,0,0)) = 0;

  /* create a specific triangulation scheme for all quadtree patches of LODObjects!
	* @param pGrid_size	constant that defines the basic grid size of a quadtree patch i.e. 8x8, 16x16 etc. (8 seems to be best choice!)
	*/
	static void initTriangulation(int pGrid_size = 8);

};

/** This method checks if a ray given by (o + t*d) intersects the triangle (v0v1v2) or not
* @param u	the barycentric coordinate for v1
* @param v	the barycentric coordinate for v2  
* @param t  parameter on ray
* @return   intersects (true/false) and correct values u,v,t
*/
extern bool RayTriangleIntersection(CDoubleVector o, CDoubleVector d, CDoubleVector v0, CDoubleVector v1, CDoubleVector v2, double *u, double *v, double *t);

#endif // JWOBJECT_H


