/************************************************************
 * Infinite Universe Engine:				                        *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2003 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWPlanet.cpp                                      *
 *                                                          *
 * Content: Implementation of a fractal planet with all     *
 *          necessary attributes                            *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2003/04/01                                      *
 ************************************************************
 */

#include <qgl.h>
#include "glext.h"
#include <cmath>
#include "JWPlanet.h"
#include "JWQuadtree.h"
#include "JWGameEngine.h"
#include "dem/ddgegrid.h"
#include "JWStaticObject.h"
#include "JWNoise.h"

// initialize static variables here (stupid concept!)
int JWPlanet::planetcount = 0;										// count the number of planets
TEXTURE *JWPlanet::staticTexDetail = 0;						// detail texture for terrain
TEXTURE *JWPlanet::staticTexFineDetail = 0;				// fine detail texture for terrain
TEXTURE *JWPlanet::staticTexSuperFineDetail = 0;	// super fine detail texture for terrain
GLuint JWPlanet::waterId[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

TEXTURE *JWStar::texCorAll = 0;	// a static corona texture for all stars that do not have their own!
int JWStar::starcount = 0;			// number of stars


// begin JWPlanet

/**
 * This method converts an arbitrary (x,y,z)-vector into a virtual cube position (CIntVector) for that planet.
 * We need the int^3-vector for LOD management and random functions!
 * @param x		x position
 * @param y		y position
 * @param z		z position
 * @param c_s	index of the responsible cube side (0-5) for the position (x,y,z)
 * @param dist  height above base geometry in meters
 * @return		Virtual cube position in range [-CUBE_HALF, CUBE_HALF]^3 and c_s
 */
CIntVector JWPlanet::CoordinatesToSurfaceVector(const double x, const double y, const double z, int *c_s, double *dist)
{
	int CUBE_HALF = cube_size/2;
	int xt,yt,zt;

	*dist = sqrt(x*x + y*y + z*z) - radius;

	double ax = Abs(x);
	double ay = Abs(y);
	double az = Abs(z);
	// determine which component is most significant
	double maxx = Max(ax, Max(ay,az));
		
	// scale maxx to CUBE_HALF and shift coordinates to lie within [0, CUBE_SIZE]
	double scale = CUBE_HALF / maxx;
	xt = (int)(scale * x);
	yt = (int)(scale * y);
	zt = (int)(scale * z);
	
	if (az == maxx)
	{
		if (zt < 0)	{ *c_s = 2; zt = -CUBE_HALF; }
		else		{ *c_s = 0; zt = CUBE_HALF;  }
	}
	else if (ax == maxx)
	{
		if (xt < 0)	{ *c_s = 3; xt = -CUBE_HALF;  }
		else		{ *c_s = 1; xt = CUBE_HALF;  }
	}
	else {
		if (yt < 0)	{ *c_s = 5; yt = -CUBE_HALF; }
		else		{ *c_s = 4; yt = CUBE_HALF;  }
	}

	return CIntVector(xt,yt,zt);
}

CDoubleVector JWPlanet::SurfaceVectorToCoordinates(CIntVector pos, int c_s, double h)
{
	CDoubleVector vec = CDoubleVector(pos.x, pos.y, pos.z);
	vec.Normalize();
	return (vec * (radius + h));
}

CDoubleVector JWPlanet::SurfaceVectorNormal(CIntVector pos, int c_s)
{
	CDoubleVector vec = CDoubleVector(pos.x, pos.y, pos.z);
	vec.Normalize();
	return vec;
}

void JWPlanet::init(QDomElement node)
{
	std::string value;	// temporary container for attribute values

	// set the name of planet
	if (node.hasAttribute("NAME"))	// really needed attribute!
	{
		value = ( node.attribute("NAME") );
		this->setName(value.c_str());
	}

	// read the style parameter
	PLANET_TYPE pstyle = style;
	if (node.hasAttribute("TYPE"))
	{
		value = (node.attribute("TYPE") );
		if (value == "SUN"  ) pstyle = JWPlanet::SUN;
		if (value == "EARTH") pstyle = JWPlanet::EARTH;
		if (value == "MOON" ) pstyle = JWPlanet::MOON;
		if (value == "MARS" ) pstyle = JWPlanet::MARS;
		if (value == "GAS"  ) pstyle = JWPlanet::GAS;
		if (value == "CLOUDS"  ) pstyle = JWPlanet::CLOUDS;
	}
	
	// get the radius of planet
	double pradius = radius;
	if (node.hasAttribute("RADIUS"))
		pradius = ( node.attribute("RADIUS") ).toDouble();

	// get the heightrange
	double pH_R;
	if (node.hasAttribute("HEIGHT"))
		pH_R = ( node.attribute("HEIGHT") ).toDouble();
	else
		pH_R = pradius * 0.001;  // a 1000km planet has 10km H_RANGE

  double pangle;
	if (node.hasAttribute("AXIS_ANGLE"))
		pangle = ( node.attribute("AXIS_ANGLE") ).toDouble();
	else
		pangle = 0;

	// now we can init the geometry!
	this->initPlanet(pstyle, pradius, pH_R, pangle);

	// set the orbit of planet (direct write)
	if (node.hasAttribute("ORBIT"))
		orbit = ( node.attribute("ORBIT") ).toDouble();
	else
		orbit = (rand() % 200 + 100) * 1000000.0e3;  // if we are too lazy to define a orbit

	// set the year duration of planet (direct write)
	if (node.hasAttribute("YEAR_DURATION"))
		year_length = ( node.attribute("YEAR_DURATION") ).toDouble();
	else
		year_length = (rand() % 1000 + 100); // if we are too lazy to define a year duration in days

	// set the day duration of planet (direct write)
	if (node.hasAttribute("DAY_DURATION"))
		day_length = ( node.attribute("DAY_DURATION") ).toDouble();
	else
		day_length = (rand() % 20 + 10); // if we are too lazy to define a day duration in hours
	
	// get the detail texture parameters
	char *finetex1 = 0;
	char *finetex2 = 0;
	char *finetex3 = 0;
	if (node.hasAttribute("FINETEX1"))	// really needed attribute!
	{
		finetex1 = new char[80];
		sprintf(finetex1, "%s", (const char *)( node.attribute("FINETEX1") ));
	}
	if (node.hasAttribute("FINETEX2"))	// really needed attribute!
	{
		finetex2 = new char[80];
		sprintf(finetex2, "%s", (const char *) ( node.attribute("FINETEX2") ));
	}
	if (node.hasAttribute("FINETEX3"))	// really needed attribute!
	{
		finetex3 = new char[80];
		sprintf(finetex3, "%s", (const char *) ( node.attribute("FINETEX3") ));
	}

	// now init the detail textures
	this->initTextures(finetex1, finetex2, finetex3);

	// set the center of planet if it has one i.e. the center of moon is Earth! (call by "name")
	if (node.hasAttribute("CENTER"))	// really needed attribute!
	{
		value = node.attribute("CENTER");
		this->setCenter(value.c_str());
	}

	// before we complete generation we loop through the subnodes
	QDomNode n;
	JWStaticObject *obj;

	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have a plant object
			if (n.nodeName() == "PLANT")
			{
				obj = new JWPlant();
				obj->init(n.toElement());
				this->objects.push_back(obj);
			}
			// we have a house object
			else if (n.nodeName() == "HOUSE")
			{
				obj = new JWHouse();
				obj->init(n.toElement());
				this->objects.push_back(obj);
			}
			// we have a rock object
			else if (n.nodeName() == "ROCK")
			{
				obj = new JWAsteroid();
				obj->init(n.toElement());
				this->objects.push_back(obj);
			}
			// we have a water definition
			else if (n.nodeName() == "WATER")
			{
				this->initWater(n.toElement());
			}
			// we have a sky definition
			else if (n.nodeName() == "SKY")
			{
				this->initSky(n.toElement());
			}
			// we have a sky definition
			else if (n.nodeName() == "CLOUDS")
			{
				cloudSphere = new JWPlanet();
				cloudSphere->init(n.toElement());
        cloudSphere->SetPosition(CDoubleVector(0,0,0));
        cloudSphere->SetParent(this);
        cloudSphere->day_length = -10 * day_length;
        hasClouds = true;
			}
			// we have a digital evaluation map (DEM)
			else if (n.nodeName() == "DEM")
			{
				this->loadDEM(n.toElement());
			}
		}
	}

	this->SetPosition(CDoubleVector(this->orbit, 0, 0));	// default position to prevent chaos

	char *texfile = 0;
	if (node.hasAttribute("TEXTURE"))	// really needed attribute!
	{
		texfile = new char[80];
		sprintf(texfile, "%s", (const char *)(node.attribute("TEXTURE")) );
	}

	if (texfile != 0)
	{
		colorMap = new TEXTURE();
		// check loading process for errors
		if(!colorMap->LoadImage(texfile, false))
		{
			delete colorMap;
			colorMap = 0;
		}
		// do activate, but we still need the raw image data!
		colorMap->keepRawData = true;
		colorMap->activate(TEXTURE::T_NORMAL);
	}

	// now we can create a nice status screen!
	char message[80] = {0};
	sprintf(message, "Please wait! Initializing planet: %s ...", this->getName());
	jwengine::setMessage(message, colorMap);


	char *heightfile = 0;
	if (node.hasAttribute("HEIGHTFIELD"))	// really needed attribute!
	{
		heightfile = new char[80];
		sprintf(heightfile, "%s", (const char *)( node.attribute("HEIGHTFIELD")) );
	}

	bool HFisComplete = true;	// when we load a planetary heightmap very often only the land mass is given, so compute random oceans!
	if (node.hasAttribute("HFISCOMPLETE"))
	{
		value = ( node.attribute("HFISCOMPLETE") );
		if (value == "true" ) HFisComplete = true;
		if (value == "false") HFisComplete = false;
	}

	this->initHeightField(heightfile, HFisComplete);

  if (hasWater)
  {
    initWaterSphere();
  }
}

void JWPlanet::initWater(QDomElement node)
{
	QDomNode n;
	double red=0.5, green=0.75, blue=1;
	
	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have a color definition
			if (n.nodeName() == "COLOR")
			{
				QTXML::getColor(n.toElement(), red,green,blue);
			}
		}
	}
	this->initWater((float)red, (float)green, (float)blue);
}

void JWPlanet::initSky(QDomElement node)
{
	QDomNode n;
	double rh=0.5, gh=0.8, bh=1;
	double rc=1, gc=0.3, bc=0;
	double atm_range;
	
	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have a color definition
			if (n.nodeName() == "COLOR_DN")
			{
				QTXML::getColor(n.toElement(), rh,gh,bh);
			}
			// we have a color definition
			if (n.nodeName() == "COLOR_ME")
			{
				QTXML::getColor(n.toElement(), rc,gc,bc);
			}
		}
	}
	// get the atm_range value
	if (node.hasAttribute("ATM_RANGE"))
		atm_range = (node.attribute("ATM_RANGE") ).toDouble();
	else
		atm_range = 10000.0;

	this->initSky( rh,  gh,  bh, rc, gc, bc, atm_range);
}

/**
 * Constructor of our planet. We initialize basic geometry and generate index lists for triangulation.
 * @param pStyle		Still the main influence factor for appearance. It's a constant like MOON, EARTH,... and is used very often
 *						to switch between the different cases. TO DO: find a more generalized approach
 * @param rad			radius of the planet in meters! (It is most intuitive to use a real world system)
 * @param H_R			height range of terrain in meters. That means how high are mountains (< H_R), how deep are oceans/valleys (> -H_R)
 */
void JWPlanet::initPlanet(PLANET_TYPE pStyle, double rad, double H_R, double paxis_angle )
{
	int LODmax;

	style = pStyle;
	if (style == JWPlanet::EARTH)
		this->cityDistance = 10;

	axis_angle = PI*paxis_angle/180;
  radius = rad;								// planet with radius rad
	H_RANGE = H_R;								// mountains can be H_RANGE meters high
	H_LOG = log2((unsigned int)H_RANGE)+1+log2((unsigned int)LODObject::grid_size);	// every H_RANGE meters can have a height difference of H_RANGE (mountanious terrain!)
	boundingRadius = radius + 2*H_RANGE;

	LODmax = log2((unsigned int)Clamp(1,0x04000000,2.0*PI*radius/4));
	cube_size = (1 << LODmax);	// resolution of each cube side
	element_size = (double)((2.0*PI*radius)/(4.0*cube_size)); // smallest voxel-size of planet (if possible ca. 1m)

	// simple pic of cube:
	//       *---------*  
	//      /  | [4]  /|  
	//     /         / |   ^ y                  [number] is front-face
	//    *---------*  |   |                     number  is a backface
	//    |    |   2|[1]   |                     for cube-side 0-3: dy is simple y-up from -CUBE_SIZE/2 to CUBE_SIZE/2
	//    | 3       |  |   |                                        dx goes conter-clockwise around y-axis
	//    |    [0]- |- *   *-----------> x                  side 4:   dx is z-axix, dy is x-axis
	//    |   /     | /   /                                 side 5:   dx is -z-axis, dy is x-axis 
	//    | /     5 |/   /
	//    *---------*   z
}

// constructor
JWPlanet::JWPlanet() : LODObject()
{
	year_length = 365;
	day_length = 24;

	HF = 0;
  colorMap = 0;
	for (int i=0; i<6; i++)
	{
		jwq[i] = 0;
		Tex[i] = 0;
	}

	numTriangles=0;	// total amount of triangles/frame
	hasWater = false;	// defines if we have oceans on planet for height < 0 (not variable for now)
  waterSphere = 0;
	hasSky = false;		// defines if planet has an atmosphere
	hasClouds = false;		// defines if planet has clouds
  cloudSphere = 0;
	fineLOD = false;	// the draw routines automatically set this property if needed!
	
	cityDistance = 0;

	radius = (rand() % 100 + 1) * 1000000.0;  // if we are too lazy to define a radius

	// init the random generator!
	jwfrac.Init(3, 1234);	// 3 dimensional noise with fixed seed 1234 (to do: make it more variable)

	style = JWPlanet::MOON;	// moons are the default

	JWPlanet::planetcount++;

	// init static detail textures
	if (JWPlanet::planetcount == 1)
	{
		staticTexDetail = new TEXTURE();
		staticTexDetail->createFractal(256,256, GL_LUMINANCE);
		staticTexDetail->activate(TEXTURE::T_TERRAIN);
		
		staticTexFineDetail = new TEXTURE();
		staticTexFineDetail->createFractal(256,256, GL_LUMINANCE);
		staticTexFineDetail->activate(TEXTURE::T_DETAIL);
		
		staticTexSuperFineDetail = new TEXTURE();
		staticTexSuperFineDetail->createFractal(256,256, GL_LUMINANCE);
		staticTexSuperFineDetail->activate(TEXTURE::T_DETAIL);

    // load static water textures
    int width, height;
    int i;
    GLubyte *imageData = NULL;
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    /* Load the caustic ripple textures. */
    printf("loading caustics:");
    for (i=0; i<32; i ++) {
      char filename[80];
      
      sprintf(filename, "Data/Water/caust%02d.bw", i);
      printf(" %d", i);
      fflush(stdout);
      imageData = TexclassRGB::read_alpha_texture(filename, &width, &height);
      if (imageData == NULL) {
        fprintf(stderr, "\n%s: could not load image file\n", filename);
        exit(1);
      }
      glGenTextures(1, &waterId[i]);
      glBindTexture(GL_TEXTURE_2D, waterId[i]);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      TEXTURE::build2DMipmaps(GL_ALPHA, width, height, imageData, TEXTURE::MIPMAP_WHITE);
      //gluBuild2DMipmaps(GL_TEXTURE_2D, GL_LUMINANCE, width, height, GL_LUMINANCE, GL_UNSIGNED_BYTE, imageData);
      free(imageData);
    }
    printf(".\n");	
	}
}

// destructor
JWPlanet::~JWPlanet()
{
	int i;

	for(i=0; i<6; i++)
	{
		if (jwq[i] != 0) { delete jwq[i]; jwq[i]=0; }
    if (style != JWPlanet::WATER)
	  	if (Tex[i] != 0) { delete Tex[i]; Tex[i]=0; }
	}
  // waterSpheres do not have their own textures they belong to a master sphere
  if (style != JWPlanet::WATER)
  {
    if (cloudSphere) { delete cloudSphere; cloudSphere = 0; }
    if (waterSphere) { delete waterSphere; waterSphere = 0; }
    if (HF) { delete[] HF; HF = 0; }
    if (colorMap) { delete colorMap; colorMap = 0; }
	  if (TexDetail && (TexDetail != staticTexDetail)) { delete TexDetail; TexDetail = 0; }
	  if (TexFineDetail && (TexFineDetail != staticTexFineDetail)) { delete TexFineDetail; TexFineDetail = 0; }
	  if (TexSuperFineDetail && (TexSuperFineDetail != staticTexSuperFineDetail)) { delete TexSuperFineDetail; TexSuperFineDetail = 0; }
  }
	// delete all reference objects!
	dems.erase(dems.begin(), dems.end());

	JWPlanet::planetcount--;
	if (JWPlanet::planetcount == 0)
	{
		delete staticTexDetail; staticTexDetail = 0;
		delete staticTexFineDetail; staticTexFineDetail = 0;
		delete staticTexSuperFineDetail; staticTexSuperFineDetail = 0;
	}
}	

/**
* This method inits our planetary textures which can also be fractal generated.
* We use 3 multiplicative layers of multitexturing for the planet's surface that vary the vertex colors
* @param tex			1st texture layer (high LODs > 15)
* @param finetex        2nd texture layer (medium LODs > 7)
* @param superfinetex   3rd texture layer ("detail" texture for smallest LODs <= 7)
*/
void JWPlanet::initTextures(char *tex, char *finetex, char *superfinetex)
{
	TexDetail = new TEXTURE();
	// if texture loading fails create fractal!
	if (!TexDetail->LoadImage(tex, false))
	{
		delete TexDetail;
		TexDetail = staticTexDetail;
	}
	else
		TexDetail->activate(TEXTURE::T_DETAIL);

	TexFineDetail = new TEXTURE();
	// if texture loading fails create fractal!
	if(!TexFineDetail->LoadImage(finetex, false))
	{
		delete TexFineDetail;
		TexFineDetail = staticTexFineDetail;
	}
	else
		TexFineDetail->activate(TEXTURE::T_DETAIL);

	TexSuperFineDetail = new TEXTURE();
	// if texture loading fails create fractal!
	if (!TexSuperFineDetail->LoadImage(superfinetex, false))
	{
		delete TexSuperFineDetail;
		TexSuperFineDetail = staticTexSuperFineDetail;
	}
  else
	  TexSuperFineDetail->activate(TEXTURE::T_DETAIL);
}

void JWPlanet::loadDEM(QDomElement node)
{
	int i,j;
	// init a new dem grid
	ElevationGrid *eg = new ElevationGrid();

	// now load the file
	char *ftmp = 0;
	if (node.hasAttribute("FILENAME"))
	{
		ftmp = new char[80];
		sprintf(ftmp, "%s", (const char *)(node.attribute("FILENAME")) );
		qWarning("Now loading file %s", ftmp);
	}	

	std::string fname = ftmp;
	// use ending for specific loading methods
	std::string format = fname.substr(fname.length()-3, fname.length());
	if (strcmp(format.c_str(),"dem") == 0)
	{
		if (!eg->LoadFromDEM(fname.c_str()))
			qWarning("Could not open DEM file %s", fname.c_str());
	}
	else if (strcmp(format.c_str(), ".bt") == 0)
	{
		if (!eg->LoadFromBT(fname.c_str()))
			qWarning("Could not open BT file %s", fname.c_str());
	}
	else
	{
		// assume that an heightmap image is given!
		// we need another eg constructor!
		delete eg;
		// get the additional values (lat/lon, min_h, max_h)
		DRECT area;
		area.left = (node.attribute("LEFT")).toDouble();
		area.right = (node.attribute("RIGHT")).toDouble();
		area.bottom = (node.attribute("BOTTOM")).toDouble();
		area.top = (node.attribute("TOP")).toDouble();
		double min_h = (node.attribute("MIN_H")).toDouble();
		double max_h = (node.attribute("MAX_H")).toDouble();

		// now read image and convert to ElevationGrid
		TEXTURE *ttmp = new TEXTURE();
		ttmp->LoadImage(ftmp);

		int m_iColumns, m_iRows;
		m_iColumns = ttmp->width;
		m_iRows = ttmp->height;

		// initialize Grid with float values
		eg = new ElevationGrid(area, m_iColumns, m_iRows, true);
		eg->m_bUTM = false;	// no UTM zone!
		eg->ComputeCornersFromExtents();

		float fvalue;
		float vrange = (float)(max_h - min_h); 
		int byp = ttmp->bpp / 8;

		for (i = 0; i < m_iColumns; i++)
		{
			for (j = 0; j < m_iRows; j++)
			{
				fvalue = vrange*((ttmp->imageData[byp*(j*m_iColumns+i) + 0] & 0xFF) / 256.0f) + (float)min_h;
				eg->SetFValue(i, j, fvalue);
			}
		}

		// delete texture
		delete ttmp;
	}
	// we need sensible polar coordiantes!
	if (eg->m_bUTM)
		eg->ConvertCornersToLatlon();

	// now compute a sensible HF_LOD for this grid (we must know how fine/coarse the data is)
	int hw, hh;
	eg->GetDimensions(hw, hh);
	int demHF_size = (int)((90/(eg->m_Corners_lat[2].x - eg->m_Corners_lat[0].x))*hw); // how large would this grid be if it covered the whole cube side?
	eg->HF_LOD = log2((unsigned int)this->cube_size)-log2((unsigned int)demHF_size);

	// make water deeper to avoid z-fighting!
	float hfval;
	for (i = 0; i < hw; i++) {
		for (j = 0; j < hh; j++)
		{
			hfval = eg->GetFValue(i, j);
			if ((hfval <= 0) && (hfval > -50))
				hfval -= (float)Min(50, (1 << eg->HF_LOD)*0.5);
			eg->SetFValue(i,j,hfval);
		}
	}

  // do not add noise if DEM is still fine!
  if (eg->HF_LOD < 7)
    eg->HF_LOD = 0;
  else
    eg->HF_LOD += log2((unsigned int)LODObject::grid_size); // a HF_LOD should not be less than 3 (log2(grid_size))

  // check height values!
	eg->FindHeightExtents();

	// load texture if present
	char *textmp = 0;
	if (node.hasAttribute("TEXTURE"))
	{
		textmp = new char[80];
		sprintf(textmp, "%s", (const char *)(node.attribute("TEXTURE")) );
		eg->tex = new TEXTURE();
		eg->tex->LoadImage(textmp);
	}	

	// set name of DEM area
	eg->SetName((const char *)(node.attribute("NAME")));

	// create a navpoint that is in center of this DEM
	double alpha = (eg->m_Corners_lat[0].x + eg->m_Corners_lat[2].x) * 0.5;
	double beta = (eg->m_Corners_lat[0].y + eg->m_Corners_lat[2].y) * 0.5;
	alpha = alpha * PI/180;
	beta = (90 - beta) * PI/180;
	double h = (radius + H_RANGE); // start 10km above surface
	// compute real coordiantes from the polar coordinates in DEM format
	double x,y,z;
	PolarToCoordinates(&x,&y,&z,alpha, beta, h);

	C3DBase *navpoint = new C3DBase();
	navpoint->SetPosition(CDoubleVector(x,y,z));
	navpoint->SetParent(this);
	navpoint->setName(eg->GetName());
	this->components.push_back(navpoint);

	// we want to know somthing about these things!
	qWarning("Area %s: NS[%3.2f, %3.2f] x WE[%3.2f, %3.2f], min_h: %f, max_h: %f", eg->GetName(), (float)eg->m_Corners_lat[0].y, (float)eg->m_Corners_lat[2].y, (float)eg->m_Corners_lat[0].x, (float)eg->m_Corners_lat[2].x, eg->m_fMinHeight, eg->m_fMaxHeight );

	// add this DEM to our vector
	dems.push_back(eg);
}

/**
* Here we init a global height field (HF) for the whole planet.
* We can either use a bitmap file source (filename in tex) or create a random HF of size pHF_size^2
* @param tex		filename of bitmap that is used as height field (if 0 we create a fractal map)
* @param pHF_size	size of the wanted HF, a bigger map will look nicer but current systems are limited! (TNT2: 256, GeForce3: 4096)
*/
void JWPlanet::initHeightField(char *tex, bool HFisComplete, int pHF_size)
{
  // if HF is not 0 we externally defined it and also give correct parameters!
	if (HF != 0)
  {
      HF_size = pHF_size;
      // all LODs from LODmax downto HF_LOD have valid heightfield values
      HF_LOD = log2((unsigned int)cube_size)-log2((unsigned int)HF_size)+log2((unsigned int)LODObject::grid_size);
      // create a fractal heightfield
      HF_width = HF_size;
      HF_height = HF_size;
  }
  // only create HF if it does not exist yet!
  else
  {
    // load heightfield
    TEXTURE *Texi = new TEXTURE();
    // test if texture loading was successful
    if (!Texi->LoadImage(tex, false))
    {
      // texture load not possible: tex == 0 or filename not correct!
      HF_size = pHF_size;
      // all LODs from LODmax downto HF_LOD have valid heightfield values
      HF_LOD = log2((unsigned int)cube_size)-log2((unsigned int)HF_size)+log2((unsigned int)LODObject::grid_size);
      // create a fractal heightfield
      HF_width = HF_size;
      HF_height = HF_size;
      HF = create_fractal(HF_size); 
    }
    else
    {
      // loading texture file success
      // NOTE: HF_size is changing into user defined size from loaded file! (ignore pHF_size here)
      HF_width = Texi->width;
      HF_height = Texi->height;
      HF_size = 1 << log2((unsigned int)HF_height);
      // all LODs from LODmax downto HF_LOD have valid heightfield values
      HF_LOD = log2((unsigned int)cube_size)-log2((unsigned int)HF_size)+log2((unsigned int)LODObject::grid_size);
      HF = new float[HF_width*HF_height];
      
      float *HFtemp = 0;
      if (!HFisComplete)
        HFtemp = create_fractal(HF_width);
      
      int k,j;
      double h;
      int byp = Texi->bpp/8;
      
      for (k=0; k<HF_height; k++)
        for (j=0; j<HF_width; j++)
        {
          if (HFisComplete)
            h = H_RANGE*((Texi->imageData[byp*(k*HF_width+j) + 0] & 0xFF) / 128.0 - 1.0);
          else
          {
            if ((Texi->imageData[byp*(k*HF_width+j) + 0] & 0xFF) == 0)
              h = -10.0 - ABS(HFtemp[k*HF_width + j]);
            else
              h = H_RANGE*((Texi->imageData[byp*(k*HF_width+j) + 0] & 0xFF) / 256.0);
          }
          HF[k*HF_width + j] = (float)h;
        }
        
        
        if (HFtemp != 0) { delete[] HFtemp; HFtemp = 0; }
    }
    if (Texi != 0) { delete Texi; Texi=0; }
  }
  
	int	LODmax = log2((unsigned int)cube_size);
	int i;
	// create 6 starting quadtree
	int ch = cube_size/2;
	// init quadtree[0]
	CIntVector start = CIntVector(0, 0, ch);
	dx_s[0] = CIntVector(cube_size, 0, 0);
	dy_s[0] = CIntVector(0, cube_size, 0);
	if (jwq[0] != 0)
		{ delete jwq[0]; jwq[0] = 0; }
	jwq[0] = new JWQuadtree(0, cube_size/2, cube_size/2, LODmax, this);
	jwq[0]->initAsCubeSide(start, dx_s[0], dy_s[0]);
	createCubeTexture(0, start, dx_s[0], dy_s[0]);

	// init quadtree[1]
	start = CIntVector(ch, 0, 0);
	dx_s[1] = CIntVector(0, 0, -cube_size);
	dy_s[1] = CIntVector(0, cube_size, 0);
	if (jwq[1] != 0)
		{ delete jwq[1]; jwq[1] = 0; }
	jwq[1] = new JWQuadtree(1, cube_size/2, cube_size/2, LODmax, this);
	jwq[1]->initAsCubeSide(start, dx_s[1], dy_s[1]);
	createCubeTexture(1, start, dx_s[1], dy_s[1]);

	// init quadtree[2]
	start = CIntVector(0, 0, -ch);
	dx_s[2] = CIntVector(-cube_size, 0, 0);
	dy_s[2] = CIntVector(0, cube_size, 0);
	if (jwq[2] != 0)
		{ delete jwq[2]; jwq[2] = 0; }
	jwq[2] = new JWQuadtree(2, cube_size/2, cube_size/2, LODmax, this);
	jwq[2]->initAsCubeSide(start, dx_s[2], dy_s[2]);
	createCubeTexture(2, start, dx_s[2], dy_s[2]);
	
	// init quadtree[3]
	start = CIntVector(-ch, 0, 0);
	dx_s[3] = CIntVector(0, 0, cube_size);
	dy_s[3] = CIntVector(0, cube_size, 0);
	if (jwq[3] != 0)
		{ delete jwq[3]; jwq[3] = 0; }
	jwq[3] = new JWQuadtree(3, cube_size/2, cube_size/2, LODmax, this);
	jwq[3]->initAsCubeSide(start, dx_s[3], dy_s[3]);
	createCubeTexture(3, start, dx_s[3], dy_s[3]);

	// init quadtree[4]
	start = CIntVector(0, ch, 0);
	dx_s[4] = CIntVector(0, 0, cube_size);
	dy_s[4] = CIntVector(cube_size, 0, 0);
	if (jwq[4] != 0)
		{ delete jwq[4]; jwq[4] = 0; }
	jwq[4] = new JWQuadtree(4, cube_size/2, cube_size/2, LODmax, this);
	jwq[4]->initAsCubeSide(start, dx_s[4], dy_s[4]);
	createCubeTexture(4, start, dx_s[4], dy_s[4]);

	// init quadtree[5]
	start = CIntVector(0, -ch, 0);
	dx_s[5] = CIntVector(0, 0, -cube_size);
	dy_s[5] = CIntVector(cube_size, 0, 0);
	if (jwq[5] != 0)
		{ delete jwq[5]; jwq[5] = 0; }
	jwq[5] = new JWQuadtree(5, cube_size/2, cube_size/2, LODmax, this);
	jwq[5]->initAsCubeSide(start, dx_s[5], dy_s[5]);
	createCubeTexture(5, start, dx_s[5], dy_s[5]);

	for (i=0; i<6; i++)
	{
		dx_s[i] /= cube_size;
		dy_s[i] /= cube_size;
	}
}

void JWPlanet::createCubeTexture(const int id, const CIntVector base, const CIntVector dx, const CIntVector dy)
{
	int u,v; // etwas anders als sonst: u fr Zeilen, v fr Spalten
	CDoubleVector temp;
	CIntVector surfacePoint;
	CFloatVector4 color;
	double h;
	
	char filename[80] = {0};
	sprintf(filename, "Data/cubemaps/%scube%02d.jpg", this->getName(), id);
	
	int LOD = log2((unsigned int)cube_size);
	
	int hs = MIN(HF_size, 1024); // we do not allow more than 1k*1k textures per side!

  // if we externally set a texture do not create!
	if (Tex[id] != 0)
    return;

	Tex[id] = new TEXTURE();
		
	// now create cube side texture if we have no precomputed image!
  if (style != JWPlanet::CLOUDS)
  {
    if (!Tex[id]->LoadImage(filename))
    {
      Tex[id]->init(hs, hs, GL_RGB);
      
      for(u=0; u <hs; u++)
      {
        for(v=0; v <hs; v++)
        {
          surfacePoint = base + (dx/hs) * (v-hs/2) + (dy/hs) * (u-hs/2);
          h = get_HeightField_And_RandomOffset(surfacePoint, LOD, 0);
          color = computeColor(h, surfacePoint);
          Tex[id]->imageData[3*(u*hs+v) + 0] = (unsigned char)((color.r)*255);
          Tex[id]->imageData[3*(u*hs+v) + 1] = (unsigned char)((color.g)*255);
          Tex[id]->imageData[3*(u*hs+v) + 2] = (unsigned char)((color.b)*255);
        }
      }
      // now save this cube texture so that we won't have to recompute it everytime!
      Tex[id]->WriteJPEG(filename);		
    }
  }
  else
  {
		Tex[id]->init(hs, hs, GL_RGBA);
		
		for(u=0; u <hs; u++)
		{
			for(v=0; v <hs; v++)
			{
				surfacePoint = base + (dx/hs) * (v-hs/2) + (dy/hs) * (u-hs/2);
				h = get_HeightField_And_RandomOffset(surfacePoint, LOD, 0);
				color = computeColor(h, surfacePoint);
				Tex[id]->imageData[4*(u*hs+v) + 0] = (unsigned char)((color.r)*255);
				Tex[id]->imageData[4*(u*hs+v) + 1] = (unsigned char)((color.g)*255);
				Tex[id]->imageData[4*(u*hs+v) + 2] = (unsigned char)((color.b)*255);
				Tex[id]->imageData[4*(u*hs+v) + 3] = (unsigned char)((color.a)*255);
			}
		}
  }

	Tex[id]->activate(TEXTURE::T_CUBEMAP);
}


/** Uses fractal midpoint displacement to calculate a new landscape.
* @param MAP	size of HF we want to create
* @return		the created HF
*/
float* JWPlanet::create_fractal(int MAP)
{
	// init Array
	float *y = new float[MAP*MAP];
	int bsize, csize;
	int x, z, i;
	int r = MAP; // NEW: relative to height field size!
	float maxH;
	// init with zero height
	for(x=0;x<MAP*MAP;x++)
		y[x] = 0.0f;
		
	bsize = MAP;
	
	for(i=0; i < log2((unsigned int)MAP); i++) {
		// Generate random height field points (approx. between -2.5*MAP and +2.5*MAP)
		for(x=0;x<MAP;x+=bsize) {
			for(z=0;z<MAP;z+=bsize) {
				y[index(x,z,MAP,MAP)] += (rand()%(r+1)-r/2);
			}
		}
		
		// Decrease altitude increment step
		if(i>4)
			r=r/2;
		
		// Diamond step
		csize = bsize/2;
		if(csize>0) {
			for(x=0; x<MAP; x+=bsize) {
				for(z=0; z<MAP; z+=bsize) {
					if( x < MAP)
						y[index(x+csize,z,MAP,MAP)] = (y[index(x+bsize,z,MAP,MAP)]+y[index(x,z,MAP,MAP)])/2;
					if( z < MAP)
						y[index(x,z+csize,MAP,MAP)] = (y[index(x,z+bsize,MAP,MAP)]+y[index(x,z,MAP,MAP)])/2;
					if ( x < MAP && z < MAP )
						y[index(x+csize,z+csize,MAP,MAP)] = ( y[index(x,z,MAP,MAP)] + y[index(x+bsize,z,MAP,MAP)] + y[index(x,z+bsize,MAP,MAP)] + y[index(x+bsize,z+bsize,MAP,MAP)]) / 4;
				}
			}
		}
		bsize = csize;
	}
	maxH = 0.0f;
	// normalize field
	for(x=0;x<MAP*MAP;x++)
		maxH = (float)Max(Abs(y[x]),maxH);

	for(x=0;x<MAP*MAP;x++)
	{
		y[x] = y[x]/maxH;	// bring it into range -1,1
		y[x] = y[x]*(float)H_RANGE;// rescale to H_RANGE!
	}

	return y;
}

/**
* Computes the height of the player object over terrain of this planet.
* We search the affected quadtree and delegate the query to it.
* @param player		the player object
* @return			height above terrain in meters
*/
double JWPlanet::getHeight_above_Surface(C3DBase *obj)
{
	// get Players Position in relation to Planet's coordinates
	C3DBase *cp = C3DBase::GetCommonParent(this, obj);
	CDoubleVector diff = this->GetCameraMatrix(cp) * (obj->GetFullPosition(cp) - this->GetFullPosition(cp));
	int c_s;
	double dist;
	CIntVector pos = CoordinatesToSurfaceVector(diff.x, diff.y, diff.z, &c_s, &dist);
	// returns real height of player above the terrain
	double h = jwq[c_s]->getHeight_of_Terrain(pos);
	return (dist - h);
}

extern int LOD_quality; // LOD at which we switch to midpoint displacement instead of shpere interpolation + redefine Vertex positons!

/**
* The "heart" of my algorithm. We need to recompute the adaptive LOD structure every frame. (TO DO: can we dynamically adjust this?)
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::updateGeometry(C3DCamera *pCamera)
{	
	// get Players Position in relation to Planet's coordinates
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	CDoubleVector t = this->GetCameraMatrix(cp) * (pCamera->GetFullPosition(cp) - this->GetFullPosition(cp));

	CIntVector newpos;
	int c_s;
	double diff;
	// We compute the new cube position that the camera has on the planets surface.
	newpos = CoordinatesToSurfaceVector(t.x, t.y, t.z, &c_s, &diff);
	// Get the difference between these 2 vectors to speed up LOD updating!
	// Reason: when moving very fast we have to create too much new quadtrees!
	int dist = ABS(newpos.x - lastpos.x) + ABS(newpos.y - lastpos.y) + ABS(newpos.z - lastpos.z);
	// we just compared the movement for LOD purposes, now use the new position
	lastpos = newpos;
	LODalt = (int)computeLOD_min(pCamera);
	// now compute how much we'll have to update and adapt the quality!
	LOD_diff = log2((unsigned int)dist) - LODalt;	// how many scales of LOD does the distance vary?
	if (LOD_diff > 0)
		LODalt = CLAMP(LODalt, log2((unsigned int)cube_size), LODalt + LOD_diff);
  if (style == JWPlanet::CLOUDS)
    LODalt = MAX(LOD_quality, LODalt);

	// update quadtrees
	int i;
	for (i=0; i<6; i++)
		jwq[i]->update(pCamera, lastpos, LODalt);

  // NEW: 2003/02/14 if planet has a waterSphere we directly update its quadtree!!!
  if (hasWater)
  {
   	waterSphere->LODalt = (int)waterSphere->computeLOD_min(pCamera);
  	waterSphere->LOD_diff = log2((unsigned int)dist) - waterSphere->LODalt;	// how many scales of LOD does the distance vary?
	  if (waterSphere->LOD_diff > 0)
		waterSphere->LODalt = CLAMP(waterSphere->LODalt, log2((unsigned int)cube_size), waterSphere->LODalt + waterSphere->LOD_diff);
  
    for (i=0; i<6; i++)
		  waterSphere->jwq[i]->update(pCamera, lastpos, waterSphere->LODalt);
  }

  if (hasClouds)
  {

		double day_cycle = 2*PI*(Mod(jwengine::jwgame->virtualTime*24, Abs(cloudSphere->day_length))) / cloudSphere->day_length;
    cloudSphere->ResetRotation();
    cloudSphere->Rotate(cloudSphere->GetUpAxis(), day_cycle);
    cloudSphere->updateGeometry(pCamera);
  }
}


/**
* Implementation of the virtual draw method of all C3DBase objects. We draw a whole planet by calling special submethods!
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::draw(C3DCamera *camera)
{
	double dist = (GetFullPosition()-camera->GetFullPosition()).Magnitude();
	double hNN = dist - radius;

	numTriangles = 0;

	// do not draw if we are too far away!
	if (dist > 10000 * this->boundingRadius)
		return;

	//check visibility
	if (!camera->isVisible(this))
		return;

	// draw very simple sphere if we are still far away!
	if (dist > 400 * this->boundingRadius)
	{
		// Multiply the current view matrix by the model matrix and draw the planet
		CDoubleMatrix m = GetScaledModelMatrix(camera);
		C3DBase *cp = C3DBase::GetCommonParent(this, camera);
		glPushMatrix();
		// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
		glLoadMatrixd(camera->GetCameraMatrix(cp)*m);

		glColor4fv((this->computeColor(0)));
		GLUquadricObj *qobj = gluNewQuadric();
		gluQuadricDrawStyle(qobj, GLU_FILL); /* smooth shaded */
		gluSphere(qobj, radius, 8, 4); // make big enough to avoid near clip plane
		gluDeleteQuadric(qobj);
		numTriangles = 64;

		glPopMatrix();		
		return;
	}	

	
	glDisable(GL_FOG);
	if (hasSky)
	{
		if (hNN <= 0)
		{
			if (!fineLOD && hasSky && !jwengine::jwgame->wireframe)
				drawSky(camera);
			updateFog(camera);
		}
		else
		{
			updateFog(camera);
			if (!fineLOD && hasSky && !jwengine::jwgame->wireframe)
				drawSky(camera);			
		}
	}


	// high LODs (incl. water) should have higlights!
	GLfloat no_mat[] = { 0, 0, 0, 1.0 };
	GLfloat diff_mat[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat no_shininess[] = { 0.0 };
	glMaterialfv(GL_FRONT, GL_SPECULAR, no_mat);
	glMaterialfv(GL_FRONT, GL_EMISSION, no_mat);
	glMaterialfv(GL_FRONT, GL_SHININESS, no_shininess);
	// reset the ambient and diffuse properties (they could be corrupted!)
	glMaterialfv(GL_FRONT, GL_AMBIENT, diff_mat);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_mat);
	
	drawTerrain(camera);	// first draw the high LODs

  // draw them first so we see them from underwater!
  if (hasClouds)
  {
    cloudSphere->numTriangles = 0;

    cloudSphere->fineLOD = fineLOD;
    // if we are below the clouds clip them inverse
    if (hNN <= (cloudSphere->radius - this->radius))
    {
    	glFrontFace(GL_CW);
      cloudSphere->drawTerrain(camera);
    	glFrontFace(GL_CCW);
    }
  }

  if (fineLOD) // this only happens for the nearest planet!!!
	{
		// draw trees in a 2nd method AFTER terrain to increase modularity and minimize context switches!
		// we only need to draw objects if there are some
		glMaterialfv(GL_FRONT, GL_SPECULAR, no_mat);

    if (!objects.empty())
			drawStaticObjects(camera);

    // draw water in a 3rd method AFTER terrain to avoid blending errors!
		if (hasWater)
			waterSphere->drawWater(camera);
  }

  if (hasClouds)
  {
    // if we are above the clouds draw them last!
    if (hNN > (cloudSphere->radius - this->radius))
    {
      glDisable(GL_FOG);
      glDisable(GL_DEPTH_TEST);
      // if we are above the clouds draw them last!
      cloudSphere->drawTerrain(camera);
      glEnable(GL_DEPTH_TEST);
    }

    numTriangles += cloudSphere->numTriangles;
  }

  // only draw fog if we are in atmosphere of this planet, this is checked by updateFog!
	if ( glIsEnabled(GL_FOG) && (fineLOD || (LODalt > LOD_quality)))
		drawFog(camera);
}

#if defined(_WIN32)
extern PFNGLACTIVETEXTUREARBPROC       glActiveTextureARB;
extern PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;
#endif
	
/**
* Here we draw the terrain of the planet
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::drawTerrain(C3DCamera *pCamera)
{
	// Multiply the current view matrix by the model matrix and draw the planet
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glPushMatrix();
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	if (style == SUN)
	  glDisable(GL_LIGHTING);	// Turn off lighting
	else
	  glEnable(GL_LIGHTING);	// Turn on lighting
  
  if (style == JWPlanet::CLOUDS)
  {   
		glAlphaFunc(GL_GREATER,0.75f); // Set Alpha Testing  
    glDisable(GL_ALPHA_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }
 
  GLuint t1, t2, t3, t4;    // texture saving stack

	//check visibility
	if (pCamera->isVisible(this))
	{
		int i;
		for (i=0; i<6; i++)
		{
			if (jwengine::jwgame->tex_mode)
			{
        // save the texID parameters
        t1 = Tex[i]->texID; 
        t2 = TexDetail->texID;
        t3 = TexFineDetail->texID;
        t4 = TexSuperFineDetail->texID;

				if ((style == JWPlanet::SUN) || (style == JWPlanet::GAS))
				{
  				if ((style == JWPlanet::SUN) && (!jwengine::jwgame->sunmode))
            Tex[i]->texID = 0;

          TexDetail->texID = 0;
          TexFineDetail->texID = 0;
          TexSuperFineDetail->texID = 0;
				}

        if (style == JWPlanet::CLOUDS)
        {
          TexFineDetail->texID = 0;
          TexSuperFineDetail->texID = 0;
        }

        glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, Tex[i]->texID);	// select Texture for this cube side			
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, TexDetail->texID);	// select detail Texture
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, TexFineDetail->texID);	// select detail Texture
			}

			if ((jwq[i] != 0) && (jwq[i]->isActive)) jwq[i]->draw(pCamera);

			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);

        // restore the texID values!
        Tex[i]->texID = t1; 
        TexDetail->texID = t2;
        TexFineDetail->texID = t3;
        TexSuperFineDetail->texID = t4;
			}
		}
	}

	if (jwengine::jwgame->BFculling)
		glEnable( GL_CULL_FACE );
	glEnable(GL_LIGHTING);	// Turn on lighting

  if (style == JWPlanet::CLOUDS)
    glDisable(GL_BLEND);

	glPopMatrix();
}

/**
* Initialize water for the planet with color (r,g,b). We load 32 animated water textures and set hasWater = true
* @param r		color attribute
* @param g		color attribute
* @param b		color attribute
*/
void JWPlanet::initWater(float r, float g, float b)
{
  hasWater = true;
  rgbWater[0] = r;
  rgbWater[1] = g;
  rgbWater[2] = b;
}

void JWPlanet::initWaterSphere()
{
  // a planet with water needs a waterSphere object!
  if (style != JWPlanet::WATER)
  {
    waterSphere = new JWPlanet();
	  // now we can init the geometry!
  	waterSphere->initPlanet(JWPlanet::WATER, this->radius, this->H_RANGE);
    waterSphere->hasWater = true;
    waterSphere->rgbWater[0] = rgbWater[0];
    waterSphere->rgbWater[1] = rgbWater[1];
    waterSphere->rgbWater[2] = rgbWater[2];

    // instead of initTextures:
    waterSphere->TexDetail = TexDetail;
    waterSphere->TexFineDetail = TexFineDetail;
    waterSphere->TexSuperFineDetail = TexSuperFineDetail;
    waterSphere->SetPosition(CDoubleVector(0,0,0));
    waterSphere->SetParent(this);
    waterSphere->HF = HF;
    for (int i=0; i<6; i++)
      waterSphere->Tex[i] = Tex[i];
   	waterSphere->initHeightField(0, true, this->HF_size);
  }
}



/**
* Here we draw a water surface for planets with hasWater == true
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::drawWater(C3DCamera *pCamera)
{
	numTriangles = 0;

  // Multiply the current view matrix by the model matrix and draw the planet
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glPushMatrix();
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	glDisable( GL_CULL_FACE );	// Near planet I don't use backface culling to have more flexibility (caves etc!)
	glEnable(GL_DEPTH_TEST);
	
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_ALPHA_TEST);
  glEnable(GL_BLEND);

	waterNum = (waterNum +1) % 32;

	//checkVisibility();
	if (pCamera->isVisible(this))
	{
		int i;
		for (i=0; i<6; i++)
		{
			if (jwengine::jwgame->tex_mode)
			{
				// TEXTURE-UNIT #0		
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, TexDetail->texID);	// select detail Texture
				// TEXTURE-UNIT #1:
				glActiveTextureARB(GL_TEXTURE1_ARB);					// select fine detail
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, TexFineDetail->texID);		
				// TEXTURE-UNIT #2:
				glActiveTextureARB(GL_TEXTURE2_ARB);					// select fine detail
				glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, JWPlanet::waterId[waterNum]);		
			}

			if ((jwq[i] != 0) && (jwq[i]->isActive)) jwq[i]->drawWater(pCamera);

			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			}
		}
	}

	glEnable(GL_LIGHTING);								// Turn on lighting
	if (jwengine::jwgame->BFculling)
		glEnable( GL_CULL_FACE );

	glPopMatrix();

  this->GetParent()->numTriangles += numTriangles;
}

/**
* Here we draw the static objects that are placed on the terrain.
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::drawStaticObjects(C3DCamera *pCamera)
{
	// Multiply the current view matrix by the model matrix and draw the planet
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glPushMatrix();
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	glDisable( GL_CULL_FACE );	// Most objects don't use backface culling to have more flexibility (caves etc!)
	//checkVisibility;
	if (pCamera->isVisible(this))
	{
		int i;
		for (i=0; i<6; i++)
		{
			if (jwengine::jwgame->tex_mode)
			{
				// TEXTURE-UNIT #2:
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
				// TEXTURE-UNIT #1:
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
				// TEXTURE-UNIT #0		
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
			}

			if ((jwq[i] != 0) && (jwq[i]->isActive)) jwq[i]->drawStaticObjects(pCamera);

			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			}
		}
	}

	if (jwengine::jwgame->BFculling)
		glEnable( GL_CULL_FACE );
	glDisable(GL_ALPHA_TEST);   // Disable Alpha Testing to draw blended geometry!
	glDisable(GL_BLEND);

	glPopMatrix();
}


/**
 * This is one of the core routines. Here we get a height value for every position on our planet
 * We first check the loaded DEMs (1), then use the global HF for high LODs (2) or 3D noise for fine LODs (3)
 * @param point		position of vertex on virtual cube [0,cube_size]^3
 * @param LOD		the level of detail we need for this vertex
 * @param h_old		interpolated height value from higher LODs (good for multifractal functions and coast line sharpness)
 * @return			the current height value at this position and this LOD
 */
double JWPlanet::get_HeightField_And_RandomOffset(CIntVector point, int LOD, double h_old)
{
  // water itself has no sensible values
  if (style == JWPlanet::WATER)
    return ((JWPlanet *)GetParent())->get_HeightField_And_RandomOffset(point, LOD, h_old);

  if (style == JWPlanet::CLOUDS)
    return 0;

	double h, hRand;
	double alpha, beta;

	int LODcoastline = LOD_quality;
	
	// -- (1) --
	// check the DEM data of this planet
	if (!dems.empty())
	{
		// compute polar coordinates in DEM format
		CoordinatesToPolar(point.x, point.y, point.z, &alpha, &beta, &h);
		alpha = Clamp(-180, 180, 180.0/PI * alpha);
		beta = Clamp(-90, 90, 90 - 180.0/PI * beta);
		// 1st attempt: cycle through the dems list and return height value of the first suitable candidate!
		std::vector<ElevationGrid *>::iterator theIterator;
		ElevationGrid *eg;
		// lookup contents of the Vector objects
		for (theIterator = dems.begin(); theIterator != dems.end(); theIterator++)
		{
			eg = (*theIterator);
			// if point is in DEM covered area
			if ( (eg->m_Corners_lat[0].x <= alpha) && (alpha <= eg->m_Corners_lat[2].x) &&
				(eg->m_Corners_lat[0].y <= beta) && (beta <= eg->m_Corners_lat[2].y) )
			{
				// remap position into DEM patch
				int hw, hh;
				int hx,hy;
				double rx, ry;
				eg->GetDimensions(hw, hh);
				alpha = (hw-1) * (alpha - eg->m_Corners_lat[0].x) / (eg->m_Corners_lat[2].x - eg->m_Corners_lat[0].x);
				beta = (hh-1) * (beta - eg->m_Corners_lat[0].y) / (eg->m_Corners_lat[2].y - eg->m_Corners_lat[0].y);
				// bicubic interpolation between DEM points
				hx = (int)alpha;
				hy = (int)beta;
				rx = Cubic(alpha-hx);
				ry = Cubic(beta-hy);
				h = (1-rx)*(1-ry)*eg->GetFValue((hx % hw), (hy % hh));
				h += rx*(1-ry)*eg->GetFValue(((hx+1) % hw), (hy % hh));
				h += (1-rx)*ry*eg->GetFValue((hx % hw), ((hy+1) % hh));
				h += rx*ry*eg->GetFValue(((hx+1) % hw), ((hy+1) % hh));
				
				// return height if LOD not too fine
				if (LOD >= (eg->HF_LOD-2))
					return h;
				else
					LODcoastline = 0; // force smooth approximation
			}
		}
	}
	
	// -- (2) --
	// no DEM, height of big LODs are taken from global height map (HF)
	if (LOD >= MIN(H_LOG, HF_LOD))
	{
		// bicubic interpolation between the HF map points
		int hw = HF_width;
		int hh = HF_height;
		int hx,hy;
		double rx, ry;
		CoordinatesToPolar(point.x, point.y, point.z, &alpha, &beta, &h);
		alpha = hw*(alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, hw]
		beta = hh*(1-beta/PI);		  // convert from [0,PI] to [0,hh]
		hx = (int)alpha;
		hy = (int)beta;
		rx = Cubic(alpha-hx);
		ry = Cubic(beta-hy);
		h = (1-rx)*(1-ry)*HF[(hy % hh)*hw + (hx % hw)];
		h += rx*(1-ry)*HF[(hy % hh)*hw + ((hx+1) % hw)];
		h += (1-rx)*ry*HF[((hy+1) % hh)*hw + (hx % hw)];
		h += rx*ry*HF[((hy+1) % hh)*hw + ((hx+1) % hw)];

		return h;
	}
	// -- (3) --
	// small LODs return random values
	else //if (LOD < MIN(jwp->H_LOG, jwp->HF_LOD))
	{
		// real 3D noise for small LODs (mulitfractal midpoint displacement)
    double f[4] = {point.x, point.y, point.z, 0};
		hRand = (Abs(h_old)/2+H_RANGE/4)*jwfrac.Noise(f);
		hRand = hRand * pow(0.5, (H_LOG - LOD));

		h = h_old + hRand;
		
		if((hasWater) && (LOD >= LODcoastline))
			if (h*h_old < 0)
			{
				hRand = -hRand;
				h += 2*hRand;
			}
		return h;
	}
}


/**
* This method creates a color value for each terrain vertex according to the planet type (see style def.) 
* For now only height of terrain is regarded to create a color.
* @param hf		the height value of a vertex for that we compute the terrain color
* @param pos	vertex position
* @return		quadrupel (r,g,b,a) as color
*/
CFloatVector4 JWPlanet::computeColor(const double hf, CIntVector pos)
{
	float h = (float)(hf/(2*H_RANGE)); // height in [-0.5,0.5]
	double alpha, beta, htmp;
	int hw, hh;
	TEXTURE *coltex = 0; // this points to the color data (global map or DEM texture!)
	CFloatVector4 color;

  if (((h <= 0) && (hasWater)) || (style == JWPlanet::WATER))
	{
		return CFloatVector4((h + 1.0f)*rgbWater[0], (h + 1.0f)*rgbWater[1], (h + 1.0f)*rgbWater[2], 1.0);
	}

	// check the DEM data of this planet
	if (!dems.empty())
	{
		// compute polar coordinates in DEM format
		CoordinatesToPolar(pos.x, pos.y, pos.z, &alpha, &beta, &htmp);
		alpha = Clamp(-180, 180, 180.0/PI * alpha);
		beta = Clamp(-90, 90, 90 - 180.0/PI * beta);
		// 1st attempt: cycle through the dems list and return texture of the first suitable candidate!
		std::vector<ElevationGrid *>::iterator theIterator;
		ElevationGrid *eg;
		bool search = true;
		// lookup contents of the Vector objects
		for (theIterator = dems.begin(); (theIterator != dems.end()) && search; theIterator++)
		{
			eg = (*theIterator);
			// if a texture is given
			if (eg->tex)
			{
				// if point is in DEM covered area
				if ( (eg->m_Corners_lat[0].x <= alpha) && (alpha <= eg->m_Corners_lat[2].x) &&
					(eg->m_Corners_lat[0].y <= beta) && (beta <= eg->m_Corners_lat[2].y) )
				{
					// we ue this texture as colormap!
					coltex = eg->tex;
					hw = coltex->width;
					hh = coltex->height;
					// remap position into DEM patch
					alpha = (hw-1) * (alpha - eg->m_Corners_lat[0].x) / (eg->m_Corners_lat[2].x - eg->m_Corners_lat[0].x);
					beta = (hh-1) * (beta - eg->m_Corners_lat[0].y) / (eg->m_Corners_lat[2].y - eg->m_Corners_lat[0].y);
					search = false;
				}
			}
		}
	}
	// no suitable DEM texture so use colorMap if present
	if ((coltex == 0) && (this->colorMap))
	{
		coltex = this->colorMap;
		hw = coltex->width;
		hh = coltex->height;
		CoordinatesToPolar(pos.x, pos.y, pos.z, &alpha, &beta, &htmp);
		alpha = hw*(alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, hw]
		beta = hh*(1-beta/PI);		  // convert from [0,PI] to [0,hh]
	}

	if (coltex)
	{
		// bilinear interpolation between the colorMap points
		float col[4];
		
		int bypp = coltex->bpp/8;

		int hx,hy;
		double rx, ry;
		hx = (int)alpha;
		hy = (int)beta;
		rx = Cubic(alpha-hx);
		ry = Cubic(beta-hy);
		for (int i=0; i < bypp; i++)
		{
			h = (1-rx)*(1-ry)*coltex->imageData[bypp*((hy % hh)*hw + (hx % hw)) + i];
			h += rx*(1-ry)*coltex->imageData[bypp*((hy % hh)*hw + ((hx+1) % hw)) + i];
			h += (1-rx)*ry*coltex->imageData[bypp*(((hy+1) % hh)*hw + (hx % hw)) + i];
			h += rx*ry*coltex->imageData[bypp*(((hy+1) % hh)*hw + ((hx+1) % hw)) + i];
			col[i] = h / 255.0f;
		}
    if (bypp < 4)
		  col[3] = 1.0f;
		color = col;	
	}
	else switch(style)
	{
	case EARTH:
		// hohe Stellen (> 6km) werden hellgrau-wei
		if(h > 0.3) {
			color = CFloatVector4(h + 0.5f, h + 0.5f, h + 0.5f, 1);
		}	else if(h > 0.2){
			// 4-6km haben beigebraune Farbe
			color = CFloatVector4( h + 0.75f, (h + 0.75f)*0.8f, (h + 0.75f)*0.5f, 1);
		}	else if(h > 0){
			// zw. 0m und 4km kommt grn
			color = CFloatVector4(h + 0.6f, (h + 0.6f)*0.6f, 0.0f, 1);
		}	else {
			// unterwasser Sandfarbe
			color = CFloatVector4(h + 1.0f, (h + 1.0f)*0.8f, (h + 1.0f)*0.6f, 1);
		}
		break;
		
	case SUN:
	case GAS:
		color = CFloatVector4( 1.0f, 0.8f + 0.5f*h, 0.3f, 1);
		break;

	case MOON:	
		color = CFloatVector4( 0.5f + 0.5f*h, 0.5f + 0.5f*h, 0.5f + 0.5f*h, 1);
		break;

	case MARS:	
		color = CFloatVector4( 0.75f + 0.5f*h, 0.4f + 0.4f*h, 0.2f + 0.4f*h, 1);
		break;
	}

  if (style == CLOUDS)
  {
    // interprete as pure alpha map
    if (coltex && (coltex->bpp < 32)) // a 32 bit image can have colored clouds!
    {
      color.a = color.r;
      color.r = 1;
      color.g = 1;
      color.b = 1;
    }
  }

	float cRand = (((int)(hf) % 16) -8)/256.0f; // add a little bit low level noise!

	color.r = (float)Clamp(0.0f, 0.9999f, color.r+cRand);
	color.g = (float)Clamp(0.0f, 0.9999f, color.g+cRand);
	color.b = (float)Clamp(0.0f, 0.9999f, color.b+cRand);
	return color;
}


/**
* Init the sky of a planet with the colors (rdn,gdn,bdn) and (rme,gme,bme) for day/night and morning/evening atmosphere color
* @param rdn	color attribute
* @param gdn	color attribute
* @param bdn	color attribute
* @param rme	color attribute
* @param gme	color attribute
* @param bme	color attribute
* @param atmh	sets the ATM_RANGE for the planet that means the thickness of the atmosphere in meters
*/
void JWPlanet::initSky(float rdn, float gdn, float bdn, float rme, float gme, float bme, float atmh)
{
	hasSky = true;
	// "normal" rgb color of sky
	rgbSkyDayNight[0] = rdn;
	rgbSkyDayNight[1] = gdn;
	rgbSkyDayNight[2] = bdn;
	// atmospheric glaze in the morning or evening ("Abendrot")
	rgbSkyMorningEvening[0] = rme;
	rgbSkyMorningEvening[1] = gme;
	rgbSkyMorningEvening[2] = bme;
	this->ATM_RANGE = atmh;
	boundingRadius += atmh;
}

/**
* Here we draw the sky effect of a planet. We draw a circle of quads around the camera position
* that clips with the horizon of the planet. This models an orthogonal cut through the atmosphere layer
* and gives the nice effect of air around the planet in space and nice color variations on ground
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::drawSky(C3DCamera* pCamera)
{
	// compute the skycolor
	float hNN = (float)((GetFullPosition()-pCamera->GetFullPosition()).Magnitude() - radius);
	// my own approximation of atmosphere density (linear is not physically correct but looks nicer!)
	// cscale is factor to blend the atmosphere coloring into the general background color
	float cscale = (float)Clamp(0.0,1.0, (1-hNN/ATM_RANGE));
	// compute sun vector relative to nearest planet, needed for realistic lighting
	CDoubleVector d1 = - GetFullPosition(); // sun is always at (0,0,0)!
	CDoubleVector d2 = pCamera->GetFullPosition() - GetFullPosition();
	d1.Normalize();
	d2.Normalize();
	if (this->style == JWPlanet::SUN)
		d1 = d2;
	// cscale now is a product of atmospheric lighting and sun lighting
	// examples: cscale goes towards zero on the night side of the planet or if we are too high and leave the atmosphere (hNN > ATM_RANGE) 
	//           cscale goes towards one  if we are on the day side and deep enough in the atmosphere layer (near ground)
	cscale = (float)Clamp(0.0,1.0, (cscale * ((d1 | d2) + 1.0)));

	// translate Sky to camera position so that only rotations affect the model matrix (simulates infinite distance!)
	CDoubleVector diff, tmp;
	if (this == pCamera->GetParent())
		diff = pCamera->GetPosition();
	else
	{
		diff = pCamera->GetFullPosition() - this->GetFullPosition();
		CDoubleMatrix rotMat = *this;
		// Get the INVERSE rotation of planet
		rotMat.Transpose();
		// and rotate diff with it to get diff as local vector!
		diff = rotMat * diff;		
	}
	// return if we are too far away from this planet to see it's atmosphere
	if (diff.Magnitude() > (50.0*this->radius))
		return;
	// now compute 2 normal vectors to diff!
	CDoubleVector up = diff;
	up.Normalize();
	if (Abs(up | this->GetUpAxis()) < 0.7)
		tmp = this->GetUpAxis();
	else
		tmp = this->GetRightAxis();
	CDoubleVector u1 = tmp ^ up;
	CDoubleVector u2 = u1 ^ up;

	// now we have the vector where we hit the planet sphere and 2 good normal vectors
	double difflen = diff.Magnitude();
	double rbottom = (radius - 2*H_RANGE);
	double rtop = (radius + ATM_RANGE);
	double rmid = radius + Min(H_RANGE, ATM_RANGE/2);
	double alpha = asin(Clamp(0.0001,  0.9999, (radius / difflen))); // surpress singularity effects
	double l = sin(alpha) * cos(alpha) * difflen;
	CDoubleVector z = up * (sin(alpha) * radius);
	CDoubleVector skyPoint;

	// transform finished, now draw
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE );
	glDisable(GL_DEPTH_TEST);

	//sky.draw(pCamera, up, u1, u2, alpha);		
	glDisable(GL_ALPHA_TEST);
 	glBlendFunc(GL_SRC_ALPHA, GL_ONE);//GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	glPushMatrix();
	// now draw the skycolor background!
	glLoadIdentity();
	glDepthMask(false);
	glColor4f((float)(rgbSkyDayNight[0]*cscale), (float)(rgbSkyDayNight[1]*cscale), (float)(rgbSkyDayNight[2]*cscale), cscale);
	GLUquadricObj *qobj = gluNewQuadric();
	gluQuadricDrawStyle(qobj, GLU_FILL); /* smooth shaded */
	gluSphere(qobj, 10000.0, 20, 16); // make big enough to avoid near clip plane, at cloud distance!!!
	gluDeleteQuadric(qobj);
	glDepthMask(true);
	
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	// now draw the atmosphere slices
	CDoubleMatrix rotMat = *this;
	// Get the INVERSE rotation of planet
	rotMat.Transpose();
	// and rotate position with it to get diff as local vector!
	CDoubleVector sunvec = rotMat * (- GetFullPosition()); // vector to sun which is the center of this system
	sunvec.Normalize();
	if (this->style == JWPlanet::SUN)
		sunvec = up;
	float sunscale;	// ranges from 1(direct sun) to 0(night side)
	float r,g,b;
	
	int i;
	for (i = 0; i<128; i++)
	{
		glBegin(GL_QUADS);
		skyPoint = z + u1 * (sin(PI * i/64.0)*l) + u2 * (cos(PI * i/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rmid;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * (i+1)/64.0)*l) + u2 * (cos(PI * (i+1)/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rmid;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * (i+1)/64.0)*l) + u2 * (cos(PI * (i+1)/64.0)*l);
		skyPoint.Normalize();
		skyPoint *= rtop;
		glColor4f(rgbSkyDayNight[0]*cscale, rgbSkyDayNight[1]*cscale, rgbSkyDayNight[2]*cscale, 0);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * i/64.0)*l) + u2 * (cos(PI * i/64.0)*l);
		skyPoint.Normalize();
		skyPoint *= rtop;
		glColor4f(rgbSkyDayNight[0]*cscale, rgbSkyDayNight[1]*cscale, rgbSkyDayNight[2]*cscale, 0);
		glVertex3dv(skyPoint);
		// take 2 quads to achieve stronger density!
		skyPoint = z + u1 * (sin(PI * i/64.0)*l) + u2 * (cos(PI * i/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rmid;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * (i+1)/64.0)*l) + u2 * (cos(PI * (i+1)/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rmid;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * (i+1)/64.0)*l) + u2 * (cos(PI * (i+1)/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rbottom;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		skyPoint = z + u1 * (sin(PI * i/64.0)*l) + u2 * (cos(PI * i/64.0)*l);
		skyPoint.Normalize();
		sunscale = (float)Abs(sunvec | skyPoint);
		r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];
		sunscale = (float)Clamp(0,0.8,1+(sunvec | skyPoint));
		skyPoint *= rbottom;
		glColor4f(r, g, b, sunscale);
		glVertex3dv(skyPoint);
		glEnd();
	}

	numTriangles += 256;

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);			// Turn on lighting
	glDisable(GL_BLEND);
	if (jwengine::jwgame->BFculling)
	  glEnable( GL_CULL_FACE );
	glPopMatrix();		
}

/**
* We compute if we are inside the atmosphere (if there is one) and how dense fogging should be.
* The visible distance can vary from from short view (1000m) up to no fog at all.
* We also have to check the day/night cycle to adjust the fog color.
* @param camera		camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::updateFog(C3DCamera* camera)
{
	float alt = (float)getHeight_above_Surface(camera);
	float hNN = (float)((GetFullPosition()-camera->GetFullPosition()).Magnitude() - radius);
	
	// we are too high to see fog, so return immediately
	if ((hNN >= ATM_RANGE) || (!jwengine::jwgame->fog_mode))
	{
		glDisable(GL_FOG);								// Disables GL_FOG
		return;
	}

	float cscale = (float)Clamp(0.0,1.0, (1-hNN/ATM_RANGE)); // my own approximation of atmosphere density
	CDoubleVector d1 = - GetFullPosition();
	CDoubleVector d2 = camera->GetFullPosition() - GetFullPosition();
	d1.Normalize();
	d2.Normalize();
	cscale = (float)Clamp(0.0,1.0, (cscale * ((d1 | d2) + 1.0)));

	// compute nice underwater fog + atmospheric fog!
	CFloatVector4 color;
	float ground_fog_depth = jwengine::jwgame->ground_fog_depth; // this fog depth MUST be minimum 1000m!
	color = computeColor(0);
	// dark_sun is the loss of light in the deep sea (at -3km there is almost no light left)
	float dark_sun = (float)Clamp(0.05,1.0,1.0 + 3*hNN/(H_RANGE));
	// the lamp of playerModel enlightens the scnene
	if (jwengine::jwgame->light_mode)
		dark_sun = (float)Clamp(0.1,1.0,dark_sun+0.2);
	glFogi(GL_FOG_MODE, GL_LINEAR);				        // Fog Mode
	glHint(GL_FOG_HINT, GL_NICEST);					    // Fog Hint Value
	if ((hNN < 0) && (hasWater))
	{
		color.r = (float)min(0.5*color.r,(double)(rgbWater[0]*cscale))*dark_sun;
		color.g = (float)min(0.5*color.g,(double)(rgbWater[1]*cscale))*dark_sun;
		color.b = (float)min(0.5*color.b,(double)(rgbWater[2]*cscale))*dark_sun;
		fog_depth = 5000*dark_sun;
		glFogf(GL_FOG_END, fog_depth);				// Fog End Depth
		glFogf(GL_FOG_START, 0.0f);					// Fog Start Depth
	}
	else
	{
		// do make fog more red on evening!
		float sunscale = (float)Abs(d1 | d2);
		color.r = sunscale * rgbSkyDayNight[0] + (1 - sunscale) * rgbSkyMorningEvening[0];
		color.g = sunscale * rgbSkyDayNight[1] + (1 - sunscale) * rgbSkyMorningEvening[1];
		color.b = sunscale * rgbSkyDayNight[2] + (1 - sunscale) * rgbSkyMorningEvening[2];

		alt = (float)Min(alt, hNN);
		//if (alt < 100) // near ground is a dense fog
		//	fog_depth = ground_fog_depth;
		//else
		fog_depth = (float)Clamp(ground_fog_depth, 4000000, 10*alt+((alt-100)/40)*(hNN/40) );
		glFogf(GL_FOG_END, fog_depth);				// Fog End Depth
		glFogf(GL_FOG_START, 0.0f);					// Fog Start Depth
		

		if (this->style == JWPlanet::EARTH)
		{
			// now mix up color with a density parameter that fades to white if near fog!
			// adjust the near fog color with cscale too!
			float fogdensity = cscale * (float)Clamp(0,1, (log(1000.0) / log(fog_depth)) - 0.25);
			color *= (1-fogdensity);
			color += (CFloatVector(1) * fogdensity);
		}
		
		// day/night darkening
		color *= cscale;

	}
	if (this->style == JWPlanet::SUN)
		color = computeColor(H_RANGE);	// a sun has a lightening fog! (we dive through the corona!)
	GLfloat	fogColor[4] = {color.r,color.g,color.b, 0.5f};		// Fog Color
	glFogfv(GL_FOG_COLOR, fogColor);					// Set Fog Color
	
	if (hasSky)	// a planet with sky has atmosphereic fog
		glEnable(GL_FOG);								// Enables GL_FOG
	else if ((hNN < 0) && (hasWater))	// no sky but water
		glEnable(GL_FOG);								// Enables GL_FOG
	else
		glDisable(GL_FOG);								// Disables GL_FOG
}

/**
* We draw a fog sphere around the camera. In sky fogged geometry only gets fog color but underwater we can clip at far distance!
* @param camera		camera that defines our viewport in the virtual world ;-)
*/
void JWPlanet::drawFog(C3DCamera* pCamera)
{
	glPushMatrix();
	glLoadIdentity();
	// transform finished, now draw
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable( GL_CULL_FACE );

	// above Sea Level we do not fog objects in the sky! so only create fog
	CDoubleVector diff = pCamera->GetFullPosition() - this->GetFullPosition();
	if ( ((diff.Magnitude() - radius) >= 0) || !hasWater)
		glColorMask(0,0,0,1);

	glColor3f(0,0,0);

	GLUquadricObj *qobj = gluNewQuadric();
	gluQuadricDrawStyle(qobj, GLU_FILL); /* smooth shaded */
	gluQuadricNormals(qobj, GLU_SMOOTH);
	gluSphere(qobj, fog_depth, 20, 16);
	gluDeleteQuadric(qobj);

	glColorMask(1,1,1,1);
	// underwater we want the skybox to be a clipping backgrond!
	glEnable(GL_LIGHTING);			// Turn on lighting
	if (jwengine::jwgame->BFculling)
	  glEnable( GL_CULL_FACE );
	glPopMatrix();		
}

// end JWPlanet

//*************************************************************************************************************************

// begin JWStar

JWStar::JWStar()
{
	JWStar::starcount++;
	// create corona texture with first star
	if (JWStar::starcount == 1)
	{
		// create a corona texture
		texCorAll = new TEXTURE();
		texCorAll->init(256,256,GL_ALPHA);
		int x,y;
		unsigned char intensity;
		for (x=0; x < 256; x++)
		{
			for (y=0; y < 256; y++)
			{
				//intensity = (unsigned char)CLAMP(0,255, 512 - 4*sqrt((127.5-x)*(127.5-x) + (127.5-y)*(127.5-y)) );
				intensity = (unsigned char)CLAMP(0,255, 340 - (8.0/3.0)*sqrt((127.5-x)*(127.5-x) + (127.5-y)*(127.5-y)) );
				texCorAll->imageData[y*256 + x] = intensity;
			}
		}
		texCorAll->activate(TEXTURE::T_NORMAL);			
	}
		
	lightnum = -1;
	style = JWPlanet::SUN;
	TexStar = 0;
};

JWStar::~JWStar()
{
	// delete the stars own texture
	if (TexStar && (TexStar != texCorAll))
	{	delete TexStar; TexStar = 0; }
		
	// delete all stars corona texture
	JWStar::starcount--;
	if (JWStar::starcount == 0)
	{	delete texCorAll; texCorAll = 0; }
}


void JWStar::init(QDomElement node)
{
	QDomDocument newdef("dummy star");
	QDomElement dplanet = newdef.createElement("STAR");
	dplanet.setAttribute("NAME","Sun X1");
	dplanet.setAttribute("TYPE", "SUN");
	dplanet.setAttribute("RADIUS", 500000000.0); // 500000 km
	dplanet.setAttribute("HEIGHT", 1.0e5); // 100 km
	dplanet.setAttribute("ORBIT", 0.0);
	dplanet.setAttribute("YEAR_DURATION", 0.0);
	dplanet.setAttribute("DAY_DURATION", 0.0);
	
	QDomElement dlight = newdef.createElement("LIGHT");
	dlight.setAttribute("NUMBER", 0);
	QDomElement dcolor  = newdef.createElement("COLOR");
	dcolor.setAttribute("RED", 1.0);
	dcolor.setAttribute("GREEN", 1.0);
	dcolor.setAttribute("BLUE", 1.0);
	dlight.appendChild(dcolor);
	dplanet.appendChild(dlight);
	QDomElement dsky  = newdef.createElement("SKY");
	dsky.setAttribute("ATM_RANGE", 100000000.0); // 100000 km
	dplanet.appendChild(dsky);
	QDomElement dcolor1  = newdef.createElement("COLOR_DN");
	dcolor1.setAttribute("RED", 1.0);
	dcolor1.setAttribute("GREEN", 1.0);
	dcolor1.setAttribute("BLUE", 0.3);
	dsky.appendChild(dcolor1);
	QDomElement dcolor2  = newdef.createElement("COLOR_ME");
	dcolor2.setAttribute("RED", 1.0);
	dcolor2.setAttribute("GREEN", 0.5);
	dcolor2.setAttribute("BLUE", 0.3);
	dsky.appendChild(dcolor2);
	
	if (node.isNull())
		node = dplanet;
	// we can use these methods with and without specifing parameters!
	this->initCorona((node.namedItem("CORONA")).toElement());
	this->initLight((node.namedItem("LIGHT")).toElement());
	// all other features can be handled in JWPlanet
	JWPlanet::init(node);
}


/**
* This method creates a color value for each terrain vertex according to the planet type (see style def.) 
* For now only height of terrain is regarded to create a color.
* @param hf		the height value of a vertex for that we compute the terrain color
* @param pos	vertex position
* @return		quadrupel (r,g,b,a) as color
*/
CFloatVector4 JWStar::computeColor(const double hf, CIntVector pos)
{
	float r,g,b;
	float h = (float)(hf/(2*H_RANGE) + 0.5f); // height in [0,1]
	
	r = h * rgbSunHot[0] + (1-h) * rgbSunCold[0];
	g = h * rgbSunHot[1] + (1-h) * rgbSunCold[1];
	b = h * rgbSunHot[2] + (1-h) * rgbSunCold[2];

	float cRand = (((int)(hf) % 16) -8)/256.0f; // add a little bit low level noise!

	r = (float)Clamp(0.0f, 0.9999f, r+cRand);
	g = (float)Clamp(0.0f, 0.9999f, g+cRand);
	b = (float)Clamp(0.0f, 0.9999f, b+cRand);	

	return CFloatVector4(r,g,b,1);
}

/** general draw method
* @param camera		camera that defines our viewport in the virtual world ;-)
*/
void JWStar::draw(C3DCamera *camera)
{
	double dist = (GetFullPosition()-camera->GetFullPosition()).Magnitude();

	numTriangles = 0;

  glDisable(GL_FOG);
	if (!fineLOD && hasSky && !jwengine::jwgame->wireframe)
		drawSky(camera);			
	// ALWAYS draw corona since it sets the GL_LIGHT!!!
	if (!fineLOD)
		drawCorona(camera);	

	// do not draw details if we are too far away!
	if (dist <= 400 * this->boundingRadius)
	{
		//check visibility
		if (!camera->isVisible(this))
			return;
		
    glDisable(GL_CULL_FACE);
		drawTerrain(camera);
	}
}

void JWStar::initLight(QDomElement node)
{
	char* stopstring;	// temporary variable for type conversion
	std::string value;	// temporary container for attribute values
	
	QDomNode n;
	double red=1, green=1, blue=1;
	int lId = -1;
	
	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have a color definition
			if (n.nodeName() == "COLOR")
				QTXML::getColor(n.toElement(), red,green,blue);
		}
	}
	// get the light number value
	if (node.hasAttribute("NUMBER"))
	{
		long lvar;
		value = (node.attribute("NUMBER") );
		lvar = strtol( value.c_str(), &stopstring, 10);
		switch(lvar)
		{
		case 0: lId = GL_LIGHT0; break;
		case 1: lId = GL_LIGHT1; break;
		case 2: lId = GL_LIGHT2; break;
		case 3: lId = GL_LIGHT3; break;
		case 4: lId = GL_LIGHT4; break;
		case 5: lId = GL_LIGHT5; break;
		case 6: lId = GL_LIGHT6; break;
		case 7: lId = GL_LIGHT7; break;
		}
	}	

	if (lId != -1)
		this->activateLight(lId, red,green,blue);
}

/**
* Generate a point light source of color (r,g,b) and number lId at this star position
* @param lId	the light number in OpenGL for that light
* @param r		color attribute
* @param g		color attribute
* @param b		color attribute
*/
void JWStar::activateLight(int lId, float r, float g, float b)
{
	lightnum = lId;
	lr = r;
	lg = g;
	lb = b;
	GLfloat light_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	GLfloat light_diffuse[] = { r, g, b, 1.0f };
	GLfloat light_specular[] = { r, g, b, 1.0f };
	GLfloat light_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	glLightfv(lightnum, GL_AMBIENT, light_ambient);
	glLightfv(lightnum, GL_DIFFUSE, light_diffuse);
	glLightfv(lightnum, GL_SPECULAR, light_specular);
	glLightfv(lightnum, GL_POSITION, light_position);
	
	glEnable(lId);								// Turn on a light with defaults set
}

void JWStar::initCorona(QDomElement node)
{
	QDomNode n;
	double rh=1, gh=1, bh=1;
	double rc=1, gc=1, bc=0.3;
	char *texfile=0;
	
	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have a color definition
			if (n.nodeName() == "COLOR_HOT")
			{
				QTXML::getColor(n.toElement(), rh,gh,bh);
			}
			// we have a color definition
			if (n.nodeName() == "COLOR_COLD")
			{
				QTXML::getColor(n.toElement(), rc,gc,bc);
			}
		}
	}
	// get the texture file value
	if (node.hasAttribute("TEXTURE"))
	{
		texfile = new char[80];
		sprintf(texfile, "%s", (const char *)(node.attribute("TEXTURE")) );
	}	
	this->initCorona( rh,  gh,  bh, rc, gc, bc, texfile);
}


/**
* Init the corona of a star with the colors (rh,gh,bh) and (rc,gc,bc) for hot and cold plasma zones
* @param rh		color attribute
* @param gh		color attribute
* @param bh		color attribute
* @param rc		color attribute
* @param gc		color attribute
* @param bc		color attribute
* @param texfile	texture for corona, to do: compute this automatically, no bmp needed!
*/
void JWStar::initCorona(float rh, float gh, float bh,float rc,float gc,float bc, char *texfile)
{
	rgbSunHot[0] = rh;
	rgbSunHot[1] = gh;
	rgbSunHot[2] = bh;
	rgbSunCold[0] = rc;
	rgbSunCold[1] = gc;
	rgbSunCold[2] = bc;

	if (texfile)
	{
		TexStar = new TEXTURE();
		if (TexStar->LoadImage(texfile, false))
		{
			TexStar->activate(TEXTURE::T_NORMAL);
			return;
		}
		else
			delete TexStar;
	}

	// still here so there was no corona image given
	TexStar = texCorAll;
}

/** special draw method, shows a colored texture decal sprite
* @param camera		camera that defines our viewport in the virtual world ;-)
*/
void JWStar::drawCorona(C3DCamera *pCamera)
{
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	CDoubleMatrix m = pCamera->GetCameraMatrix(cp) * GetScaledModelMatrix(pCamera);
	
	glPushMatrix();
	glLoadMatrixd(m);
	
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glColor4f(1.0f,1.0f,1.0f,1.0f);
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, TexStar->texID);	// select Texture for this star		
	
	CDoubleVector d, tmp;
	d = pCamera->GetFullPosition(cp) - this->GetFullPosition(cp);
	// Get the INVERSE rotation of planet
	CDoubleMatrix rotMat = this->GetCameraMatrix(cp);
	// and rotate diff with it to get diff as local vector!
	d = rotMat * d;		
	// now compute 2 normal vectors to diff!
	CDoubleVector up = d;
	up.Normalize();
	if (Abs(up | this->GetUpAxis()) < 0.7)
		tmp = this->GetUpAxis();
	else
		tmp = this->GetRightAxis();
	CDoubleVector v1 = tmp ^ up;
	CDoubleVector v2 = v1 ^ up;
	v1 *= 4*radius;
	v2 *= 4*radius;
				
	float fadeout = Clamp(0, 1, 0.1*(d.Magnitude()-radius-ATM_RANGE)/ATM_RANGE); // if we come close to the star the corona is messy
	// place light at right position!
	if (lightnum != -1)
	{
		// This is a star and the corona has color of light
		GLfloat light_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		glLightfv(lightnum, GL_POSITION, light_position);
		glColor4f(lr, lg, lb, fadeout);
	}
	else
		glColor4f(rgbSkyDayNight[0], rgbSkyDayNight[1], rgbSkyDayNight[2], fadeout);	// use atmosphere color for corona
	
	CDoubleVector v;
	glBegin(GL_QUADS);
	v = v1 + v2;
	glTexCoord2f(1.0, 1.0);
	glVertex3dv(v);
	v = v1 - v2;
	glTexCoord2f(0.0, 1.0);
	glVertex3dv(v);
	v = -v1 - v2;
	glTexCoord2f(0.0, 0.0);
	glVertex3dv(v);
	v = -v1 + v2;
	glTexCoord2f(1.0, 0.0);
	glVertex3dv(v);
	glEnd();

	numTriangles += 2;
	
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
}

// end JWStar




