/************************************************************
 * Infinite Universe Engine:                                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2003 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:  JWPlanet.h                                        *
 *                                                          *
 * Content: Header file of a fractal planet with all        *
 *          necessary attributes                            *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2003/04/01                                      *
 ************************************************************
 */

#ifndef JWPLANET_H
#define JWPLANET_H

#include "Matrix.h"
#include "JWObject.h"
#include "Texture.h"
#include "texload.h"
#include "qtxml.h"

class JWQuadtree;
class ElevationGrid;

class JWPlanet : public LODObject {
private:
  // reference counter for the number of planets
  static int planetcount;
  // default detail textures for all planets (used LOD for earth)
  static TEXTURE *staticTexDetail;          // LOD > 15
  static TEXTURE *staticTexFineDetail;      // LOD > 7
  static TEXTURE *staticTexSuperFineDetail; // LOD <= 7
  static GLuint waterId[32];                // 32 animated water textures

public:
  typedef enum PLANET_TYPE { SUN=1, EARTH=2, MOON=3,
                             MARS=4, GAS=5, WATER=6, CLOUDS=7 };
  // the style is important for the color scheme of the planet
  PLANET_TYPE style;

  // 6 quadtrees store the geometry of the planet (one per cube side)
  JWQuadtree *jwq[6];
  // vectors needed for orientation on each surface
  CIntVector dx_s[6];
  CIntVector dy_s[6];
  
  float *HF;                  // heightfield for planet
  TEXTURE *Tex[6];            // texture for each cube side
  TEXTURE *TexDetail;         // detail texture for terrain
  TEXTURE *TexFineDetail;     // fine detail texture for terrain
  TEXTURE *TexSuperFineDetail;// super fine detail texture
  TEXTURE *colorMap;          // use a planetary texture as color map

  std::vector<ElevationGrid *> dems; // here we store digital elevation maps

  int waterNum;               // current used water texture
  bool hasWater;              // planet has oceans
  float rgbWater[3];          // color of water
  JWPlanet *waterSphere;      // help object for water

  bool hasSky;                   // planet has atmosphere
  float rgbSkyDayNight[3];       // 1st color of atmosphere
  float rgbSkyMorningEvening[3]; // 2nd color of atmosphere
  bool hasClouds;             // planet has clouds
  JWPlanet *cloudSphere;      // help object for clouds
  float ATM_RANGE;            // height of atmospheric hull
  float fog_depth;            // depth of view in atmosphere
  
  // factor between 1 and 100 (0 to disable), 
  // defines how dense cities are placed on planet
  int cityDistance;       

  double radius;              // radius of this planet
  
  // ******************** function declarations ****************************
  // *********************** class interface *******************************

  JWPlanet();           // constructor
  virtual ~JWPlanet();  // destructor

  /**
  * This method initializes the whole planet and fills in all data.
  * @param node    the <PLANET>...</PLANET> node in a XML document
  */
  virtual void init(QDomElement node);

  /**
  * The "heart" of my algorithm. We recompute the LOD structure every frame.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  virtual void updateGeometry(C3DCamera *pCamera);

  /**
  * We draw the whole planet by calling special subroutines.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  virtual void draw(C3DCamera *camera);

  // ******************** function declarations ****************************
  // ********************** internal routines ******************************

  // *********** conversions between integer and double space **************

  /**
  * This method converts a (x,y,z)-vector into a cube position (CIntVector).
  * We need the int^3-vector for LOD management and random noise functions.
  * @param x    x position
  * @param y    y position
  * @param z    z position
  * @param c_s  index of the cube side (0-5) for the position (x,y,z)
  * @param dist distance from geometry surface (in meters above sea level)
  * @return     cube position in range [-CUBE_HALF, CUBE_HALF]^3, c_s and dist
  */
  virtual CIntVector CoordinatesToSurfaceVector(const double x,
                                                const double y,
                                                const double z,
                                                int *c_s, double *dist);
  /**
  * This method computes the vertex in double space for a given cube position.
  * @param pos   position in integer space
  * @param c_s   cube side of integer position
  * @param h     height of terrain at pos (in meters above sea level)
  * @return      vertex position
  */
  virtual CDoubleVector SurfaceVectorToCoordinates(CIntVector pos,
                                                   int c_s,
                                                   double h);

  /**
  * This method computes the vertex normal for a given cube position.
  * @param pos   position in integer space
  * @param c_s   cube side of integer position
  * @return      vertex normal
  */
  virtual CDoubleVector SurfaceVectorNormal(CIntVector pos, int c_s);

  /**
  * This is one of the core routines. Here we get a height value for every 
  * position on our planet. We first check the loaded DEMs (1), then use the 
  * global HF for high LODs (2) or 3D noise for fine LODs (3).
  * @param point    cube position in range [-CUBE_HALF, CUBE_HALF]^3
  * @param LOD      the level of detail we need for this vertex
  * @param h_old    interpolated height value from higher LODs 
                    (good for multifractal functions and sharp coast lines)
  * @return         the current height value at this position and this LOD
  */
  double get_HeightField_And_RandomOffset(CIntVector point,
                                          int LOD,
                                          double h_old);

  /**
  * This method creates a color value for each terrain vertex according to the 
  * planet type. Only height and position on terrain are used to create a color.
  * @param hf   height value of vertex where the terrain color is computed
  * @param pos  integer position of vertex
  * @return     quadrupel (r,g,b,a) as color
  */
  CFloatVector4 computeColor(const double hf,
                             CIntVector pos = CIntVector(0,0,0));
  
  /**
  * Computes the height of the given object over terrain of this planet.
  * We search the affected quadtree and delegate the query to it.
  * @param obj    object e.g. player model
  * @return       height above terrain in meters
  */
  virtual double getHeight_above_Surface(C3DBase *obj);

  // ********************* internal init subroutines ***********************

  /**
  * Geometry construction of our planet. We initialize basic geometry and 
  * generate index lists for triangulation.
  * @param pStyle   Still the main influence factor for appearance. 
                    It's a constant like MOON, EARTH,... and is used very often
  *                 to switch between the different cases.
  * @param rad      radius of the planet in meters
  * @param H_R      height range of terrain in meters.
  *                 That means how high are mountains (< H_R), 
  *                 how deep are oceans/valleys (> -H_R)
  * @param axis_angle   angle agaist up-axis
  */
  void initPlanet(PLANET_TYPE pStyle, double rad, double H_R,
                  double paxis_angle = 0);

  /**
  * This method inits our planetary textures which can also be fractal generated.
  * We use 3 multiplicative layers of multitexturing for the planet's surface.
  * @param tex            1st texture layer (high LODs > 15)
  * @param finetex        2nd texture layer (medium LODs > 7)
  * @param superfinetex   3rd texture layer ("detail" texture for LODs <= 7)
  */
  void initTextures(char *tex=0, char *finetex=0, char *superfinetex=0);

  /**
  * Here we compute the 6 cube side textures for a planet.
  * @param id     cube side number
  * @param base   center position of the quadtree for this cube side
  * @param dx,dy  vectors for 2-dimensional surface orientation
  */
  void createCubeTexture(const int id, const CIntVector base, 
                         const CIntVector dx, const CIntVector dy);
  
  /**
  * Here we init a global height field (HF) for the whole planet.
  * We can either use a bitmap file source (filename in tex) 
  * or create a random HF of size pHF_size^2 with create_fractal.
  * @param tex      filename of bitmap that is used as height field
                    (if 0 we create a fractal map)
  * @param HFisComplete If false, underwater terrain uses fractal map.
  * @param pHF_size size of the HF, a bigger map will look nicer but
                    current systems are limited. (TNT2: 256, GeForce3: 4096)
  */
  void initHeightField(char *tex, bool HFisComplete, int pHF_size=256);

  /**
  * Uses fractal midpoint displacement to calculate a new landscape.
  * @param MAP  size of HF we want to create
  * @return     the created HF
  */
  float* create_fractal(int MAP);

  /**
  * This function loads digital elevation maps.
  * @param node    A <DEM ... /> node in the XML document
  */
  void loadDEM(QDomElement node);

  /**
  * This function inits the water surface for a planet.
  * @param node    A <WATER ... /> node in the XML document
  */
  void initWater(QDomElement node);

  /**
  * Initialize water for the planet with color (r,g,b). 
  * We load 32 animated water textures and set hasWater = true
  * @param r    color attribute
  * @param g    color attribute
  * @param b    color attribute
  */
  void initWater(float r, float g, float b);

  /**
  * Creates a subobject for water surface rendering.
  */
  void initWaterSphere();

  /**
  * This function inits the atmosphere for a planet.
  * @param node    A <SKY ... /> node in the XML document
  */
  void initSky(QDomElement node);

  /**
  * Init the sky of a planet with the colors (rdn,gdn,bdn) and (rme,gme,bme) 
  * for day/night and morning/evening atmosphere colors.
  * @param rdn  color attribute
  * @param gdn  color attribute
  * @param bdn  color attribute
  * @param rme  color attribute
  * @param gme  color attribute
  * @param bme  color attribute
  * @param atmh sets the ATM_RANGE for the planet that means 
                the thickness of the atmosphere in meters
  */
  void initSky(float rdn=0, float gdn=0, float bdn=0,
               float rme=0, float gme=0, float bme=0, float atmh=10000.0f);

  // *********************** special draw routines *************************

  /**
  * Here we draw the terrain of the planet
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawTerrain(C3DCamera *pCamera);
  
  /**
  * Here we draw the static objects that are placed on the terrain.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawStaticObjects(C3DCamera *pCamera);

  /**
  * Here we draw a water surface for planets with hasWater == true
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawWater(C3DCamera *pCamera);
  
  /**
  * Here we draw the sky effect of a planet. We draw a circle of quads around 
  * the camera position that clips with the horizon of the planet. This models 
  * an orthogonal cut through the atmosphere layer and gives the nice effect 
  * of air around the planet in space and nice color variations on ground.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawSky(C3DCamera* pCamera);
  
  /**
  * We compute if we are inside the atmosphere (if there is one) and how dense 
  * fogging should be. The visible distance can vary from from short view (1000m) 
  * up to no fog at all. We have to check the day/night cycle to adjust fog color.
  * @param camera   camera that defines our viewport in the virtual world
  */
  void updateFog(C3DCamera* camera);
  
  /**
  * We draw a fog sphere around the camera. In sky fogged geometry only gets fog 
  * color but underwater we can clip at far distance.
  * @param camera   camera that defines our viewport in the virtual world
  */
  void drawFog(C3DCamera* pCamera);

}; // end of JWPlanet

class JWStar : public JWPlanet
{
private:
  static int starcount;       // we count the number of stars created!
  static TEXTURE *texCorAll;  // a static corona texture for all stars that do not have their own!

public:
  TEXTURE *TexStar;           // texture for corona
  int lightnum;
  float lr,lg,lb;             // colors of light
  float rgbSunHot[3];         // 1st color of coronosphere
  float rgbSunCold[3];        // 2nd color of coronosphere

  JWStar();
  ~JWStar();

  void init(QDomElement node);
  void initLight(QDomElement node);
  void initCorona(QDomElement node);

  /**
  * This method creates a color value for each terrain vertex according to the planet type (see style def.) 
  * For now only height of terrain is regarded to create a color.
  * @param hf   the height value of a vertex for that we compute the terrain color
  * @param pos  vertex position
  * @return   quadrupel (r,g,b,a) as color
  */
  CFloatVector4 computeColor(const double hf, CIntVector pos = CIntVector(0,0,0));

  /** general draw method
  * @param camera   camera that defines our viewport in the virtual world ;-)
  */
  void draw(C3DCamera *camera);
  
  /**
  * Generate a point light source of color (r,g,b) and number lId at this star position
  * @param lId  the light number in OpenGL for that light
  * @param r    color attribute
  * @param g    color attribute
  * @param b    color attribute
  */
  void activateLight(int lId, float r, float g,float b);

  /** Deactivate the light of this star
  */
  void deactivateLight()  {   glDisable(lightnum);  };

  /**
  * Init the corona of a star with the colors (rh,gh,bh) and (rc,gc,bc) for hot and cold plasma zones
  * @param rh   color attribute
  * @param gh   color attribute
  * @param bh   color attribute
  * @param rc   color attribute
  * @param gc   color attribute
  * @param bc   color attribute
  * @param texfile  texture for corona, to do: compute this automatically, no bmp needed!
  */
  void initCorona(float rh, float gh, float bh,float rc,float gc,float bc, char *texfile);

  /** special draw method, shows a colored texture decal sprite
  * @param camera   camera that defines our viewport in the virtual world ;-)
  */
  void drawCorona(C3DCamera *pCamera);
};


#endif



