/************************************************************
 * Infinite Universe Engine:	        			                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2003 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWQuadtree.cpp                                    *
 *                                                          *
 * Content: Implementation of 3D quadtree structure         *
 *          to render planetary bodies                      *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#include <cmath>
#include "JWQuadtree.h"
#include "JWPlanet.h"
#include "JWRing.h"
#include "Matrix.h"
#include <qgl.h>
#include "glext.h"
#include "JWGameEngine.h"
#include "dem/ddgegrid.h"
#include "JWStaticObject.h"

int LOD_quality = 13; // LOD at which we switch to midpoint displacement instead of shpere interpolation + redefine Vertex positons!

JWQuadtree::JWQuadtree(const int pId, int pXmid, int pYmid, const int pLOD, LODObject *pJwp) 
{
	id = pId;		// give a dummy id value (seems not necessary)
	xmid = pXmid;
	ymid = pYmid;
	jwp = pJwp;		// reference pointer to planet
	this->SetParent(jwp); // also needed for 3D reference position!
  this->SetPosition(CDoubleVector(0,0,0));
  this->ResetRotation();
	// give a planet pointer if possible
	planet = dynamic_cast <JWPlanet *> (jwp);
	// give a ring pointer if possible
	ring = dynamic_cast <JWRing *> (jwp);
	LOD = pLOD;		// LOD of this patch
	isActive = false;	// arrays are not initialized yet, so not active
	no = false, nl = false, nu = false, nr = false;			// 4 neighbour patches have the same LOD (true/false)
	child[0] = 0, child[1] = 0, child[2] = 0, child[3] = 0;	// child patches	
	// create all arrays
	int gs = LODObject::gs;
	int gs_sqr = gs*gs;
	vertexData = new double[3*gs_sqr];
	if (planet && (planet->hasWater) && (LOD == LOD_quality))
		crackData	= new double[3*gs_sqr];
	else
		crackData = 0;

  normals = new GLfloat[3*gs_sqr];
	textureCoords = new GLfloat[2*gs_sqr];
	textureCoords2 = new GLfloat[2*gs_sqr];
	textureCoords3 = new GLfloat[2*gs_sqr];
	colorData = new GLfloat[4*gs_sqr];
	heightField = new double[gs_sqr];
	surfacePos = new CIntVector[gs_sqr];
	object = new JWStaticObject *[gs_sqr];
	objLOD = new int[gs_sqr];
	for (int i=0; i<gs_sqr; i++)
	{
		object[i] = 0;
		objLOD[i] = 0;
	}
	hmin = -jwp->H_RANGE;
	hmax = jwp->H_RANGE;
}

JWQuadtree::~JWQuadtree()
{
	if (vertexData != 0) { delete[] vertexData; vertexData = 0; }
	if (crackData != 0) { delete[] crackData; crackData = 0; }
	if (normals != 0) { delete[] normals; normals = 0; }
	if (textureCoords != 0) { delete[] textureCoords; textureCoords = 0; }
	if (textureCoords2 != 0) { delete[] textureCoords2; textureCoords2 = 0; }
	if (textureCoords3 != 0) { delete[] textureCoords3; textureCoords3 = 0; }
	if (colorData != 0) { delete[] colorData; colorData = 0; }
	if (surfacePos != 0) { delete[] surfacePos; surfacePos = 0; }
	if (heightField != 0) { delete[] heightField; heightField = 0; }
	if (object != 0) { delete[] object; object = 0; }
	if (objLOD != 0) { delete[] objLOD; objLOD = 0; }

	if (child[0] != 0) { delete child[0]; child[0] = 0; }
	if (child[1] != 0) { delete child[1]; child[1] = 0; }
	if (child[2] != 0) { delete child[2]; child[2] = 0; }
	if (child[3] != 0) { delete child[3]; child[3] = 0; }
}

void JWQuadtree::initAsCubeSide(const CIntVector base, const CIntVector dx, const CIntVector dy)
{
	int u,v; // etwas anders als sonst: u fr Zeilen, v fr Spalten
	CDoubleVector temp;
	CIntVector surfacePoint;
	double h;

	int gs = LODObject::gs;
	int grid_size = LODObject::grid_size;

	for(u=0; u<gs; u++)
	{
		for(v=0; v<gs; v++)
		{
			surfacePoint = base + (dx/grid_size) * (v-grid_size/2) + (dy/grid_size) * (u-grid_size/2);
			surfacePos[gs*u+v] = surfacePoint;

			h = jwp->get_HeightField_And_RandomOffset(surfacePoint, LOD ,0);
			
			heightField[gs*u+v] = h;

			temp = jwp->SurfaceVectorNormal(surfacePoint, id);
			normals[3*(gs*u+v) + 0] = (float)temp.x;
			normals[3*(gs*u+v) + 1] = (float)temp.y;
			normals[3*(gs*u+v) + 2] = (float)temp.z;

			colorData[4*(gs*u+v) + 0] = 1;
			colorData[4*(gs*u+v) + 1] = 1;
			colorData[4*(gs*u+v) + 2] = 1;
			colorData[4*(gs*u+v) + 3] = 1;

			if (planet && planet->hasWater)
				if (h < 0) h = 0;
      if (planet && (planet->style == JWPlanet::WATER)) h=0;
			if (ring) h=0;
			
			temp = jwp->SurfaceVectorToCoordinates(surfacePoint, id, h);
			vertexData[3*(gs*u+v) + 0] = temp.x;
			vertexData[3*(gs*u+v) + 1] = temp.y;
			vertexData[3*(gs*u+v) + 2] = temp.z;

			textureCoords[2*(gs*u+v) + 0] = (float)((double)v/grid_size);	// proper mapping for cube texture
			textureCoords[2*(gs*u+v) + 1] = (float)((double)u/grid_size);	// proper mapping for cube texture
			textureCoords2[2*(gs*u+v) + 0] = (float)((double)v/grid_size*jwp->HF_size);	// proper mapping for detail texture
			textureCoords2[2*(gs*u+v) + 1] = (float)((double)u/grid_size*jwp->HF_size);	// proper mapping for detail texture
			textureCoords3[2*(gs*u+v) + 0] = (float)((double)v/grid_size*jwp->HF_size*64);	// proper mapping for fine detail texture
			textureCoords3[2*(gs*u+v) + 1] = (float)((double)u/grid_size*jwp->HF_size*64);	// proper mapping for fine detail texture
		}
	}
	// lege das Zentrum des Quadtrees  in der Schwerpunkt aller Punkte (hier ntig wegen starker Krmmung)
	CDoubleVector center = CDoubleVector(&vertexData[0]);
	center += CDoubleVector( &vertexData[3*(0*gs + (gs-1))] );
	center += CDoubleVector( &vertexData[3*((gs-1)*gs + 0)] );
	center += CDoubleVector( &vertexData[3*((gs-1)*gs + (gs-1))] );
	center /= 4;
	center += CDoubleVector( &vertexData[3*((gs/2)*gs + (gs/2))] );
	center /= 2;
	SetPosition(CDoubleVector(center));
	// bestimme boundingRadius, here simplified because of symmetric geometry!
	boundingRadius = center.Distance(CDoubleVector( &vertexData[0] )) * 1.1 + jwp->H_RANGE; // security scale

	if (LOD == LOD_quality)	// For very small objects like phobos (7km radius) we go down to LOD_quality at start!
		offset = center;
	else
		offset = CDoubleVector(0,0,0);

	if (LOD <= LOD_quality)	// For very small objects like phobos (7km radius) we go down to LOD_quality at start!
	{
		// center is new center too
		// but transpose all vertices around center (so that they come near to the camera!)
		for (u=0; u<gs; u++)
		{
			for (v=0; v<gs; v++)
			{
				vertexData[3*(gs*u+v) + 0] -= offset.x;
				vertexData[3*(gs*u+v) + 1] -= offset.y;
				vertexData[3*(gs*u+v) + 2] -= offset.z;
			}
		}
	}

	isActive = true;

}

void JWQuadtree::createChild(const int k)
{
	int gs = LODObject::gs;
	int grid_size = LODObject::grid_size;
	int quadsize = (1 << LOD);

	int i = k/2;		// 0 oder 1
	int j = (k & 1);	// 0 oder 1
		
	int x0, y0;
	x0 = xmid - quadsize/4 + j*quadsize/2;
	y0 = ymid - quadsize/4 + i*quadsize/2;
	// create a new quadtree
	child[k] = new JWQuadtree(id, x0, y0, LOD-1, jwp);
	// copy & generate vertexData, normals, colorData etc.
	int u,v, uh, vh;
	register int gsuv, gsuv2, gsuv3, gsuhvh, gsuhvh2, gsuhvh3;	// fast shortcut indices!
	double h;
	CIntVector spos;
	CDoubleVector vpos;
	CFloatVector4 color;
	double hmin = 10*jwp->H_RANGE;
	double hmax = -10*jwp->H_RANGE;

	for (u=0; u<gs; u++)
	{
		uh = u/2 + (i*grid_size/2);
		
		for (v=0; v<gs;v++)
		{
			vh = v/2 + (j*grid_size/2);

			// compute shortcut indices only ONE time!
			gsuv    = gs*u+v;
			gsuv2   = 2*gsuv;
			gsuv3   = 3*gsuv;
			gsuhvh  = gs*uh+vh;
			gsuhvh2 = 2*gsuhvh;
			gsuhvh3 = 3*gsuhvh;
			
			if ((u & 1) == 0)
			{
				if ((v & 1) == 0)
				{
					child[k]->surfacePos[gsuv] = surfacePos[gsuhvh];
					spos = child[k]->surfacePos[gsuv];

					h = heightField[gs*uh+vh];

					child[k]->normals[gsuv3 + 0] = normals[gsuhvh3 + 0];
					child[k]->normals[gsuv3 + 1] = normals[gsuhvh3 + 1];
					child[k]->normals[gsuv3 + 2] = normals[gsuhvh3 + 2];
					child[k]->textureCoords[gsuv2 + 0] = textureCoords[gsuhvh2 + 0];
					child[k]->textureCoords[gsuv2 + 1] = textureCoords[gsuhvh2 + 1];					
					child[k]->textureCoords2[gsuv2 + 0] = textureCoords2[gsuhvh2 + 0];
					child[k]->textureCoords2[gsuv2 + 1] = textureCoords2[gsuhvh2 + 1];					
					child[k]->textureCoords3[gsuv2 + 0] = textureCoords3[gsuhvh2 + 0];
					child[k]->textureCoords3[gsuv2 + 1] = textureCoords3[gsuhvh2 + 1];					
				}
				else
				{	
					child[k]->surfacePos[gsuv] = (surfacePos[gsuhvh] + surfacePos[gsuhvh + 1])/2;
					spos = child[k]->surfacePos[gsuv];

					h = (heightField[gsuhvh]+heightField[gsuhvh+1])*0.5f;
					h = jwp->get_HeightField_And_RandomOffset(spos, LOD-1, h); // much better than before (REAL 3D NOISE)!!!

					CDoubleVector unit = CDoubleVector((normals[gsuhvh3 + 0] + normals[(gsuhvh3+3) + 0]),
													 (normals[gsuhvh3 + 1] + normals[(gsuhvh3+3) + 1]),
													 (normals[gsuhvh3 + 2] + normals[(gsuhvh3+3) + 2]));
					unit.Normalize();
					child[k]->normals[gsuv3 + 0] = (float)unit.x;
					child[k]->normals[gsuv3 + 1] = (float)unit.y;
					child[k]->normals[gsuv3 + 2] = (float)unit.z;

					child[k]->textureCoords[gsuv2 + 0] = (float)(textureCoords[gsuhvh2 + 0]+textureCoords[gsuhvh2 + 2])*0.5f;
					child[k]->textureCoords[gsuv2 + 1] = (float)(textureCoords[gsuhvh2 + 1]+textureCoords[gsuhvh2 + 3])*0.5f;
					child[k]->textureCoords2[gsuv2 + 0] = (float)(textureCoords2[gsuhvh2 + 0]+textureCoords2[gsuhvh2 + 2])*0.5f;
					child[k]->textureCoords2[gsuv2 + 1] = (float)(textureCoords2[gsuhvh2 + 1]+textureCoords2[gsuhvh2 + 3])*0.5f;
					child[k]->textureCoords3[gsuv2 + 0] = (float)(textureCoords3[gsuhvh2 + 0]+textureCoords3[gsuhvh2 + 2])*0.5f;
					child[k]->textureCoords3[gsuv2 + 1] = (float)(textureCoords3[gsuhvh2 + 1]+textureCoords3[gsuhvh2 + 3])*0.5f;
				}
				
			}
			else
			{
				if ((v & 1) == 0)
				{
					child[k]->surfacePos[gsuv] = (surfacePos[gsuhvh] + surfacePos[gsuhvh + gs])/2;
					spos = child[k]->surfacePos[gsuv];

					h = (heightField[gsuhvh]+heightField[gsuhvh + gs])*0.5f;
					h = jwp->get_HeightField_And_RandomOffset(spos, LOD-1, h); // much better than before (REAL 3D NOISE)!!!

					CDoubleVector unit = CDoubleVector((normals[gsuhvh3 + 0] + normals[gsuhvh3 + 3*gs + 0]),
														 (normals[gsuhvh3 + 1] + normals[gsuhvh3 + 3*gs + 1]),
														 (normals[gsuhvh3 + 2] + normals[gsuhvh3 + 3*gs + 2]));
					unit.Normalize();
					child[k]->normals[gsuv3 + 0] = (float)unit.x;
					child[k]->normals[gsuv3 + 1] = (float)unit.y;
					child[k]->normals[gsuv3 + 2] = (float)unit.z;
					child[k]->textureCoords[gsuv2 + 0] = (float)(textureCoords[gsuhvh2 + 0]+textureCoords[gsuhvh2 + 2*gs + 0])*0.5f;
					child[k]->textureCoords[gsuv2 + 1] = (float)(textureCoords[gsuhvh2 + 1]+textureCoords[gsuhvh2 + 2*gs + 1])*0.5f;										
					child[k]->textureCoords2[gsuv2 + 0] = (float)(textureCoords2[gsuhvh2 + 0]+textureCoords2[gsuhvh2 + 2*gs + 0])*0.5f;
					child[k]->textureCoords2[gsuv2 + 1] = (float)(textureCoords2[gsuhvh2 + 1]+textureCoords2[gsuhvh2 + 2*gs + 1])*0.5f;										
					child[k]->textureCoords3[gsuv2 + 0] = (float)(textureCoords3[gsuhvh2 + 0]+textureCoords3[gsuhvh2 + 2*gs + 0])*0.5f;
					child[k]->textureCoords3[gsuv2 + 1] = (float)(textureCoords3[gsuhvh2 + 1]+textureCoords3[gsuhvh2 + 2*gs + 1])*0.5f;										
				}
				else
				{
					child[k]->surfacePos[gs*u+v] = (surfacePos[gs*uh+vh] + surfacePos[gs*(uh+1)+vh + 1])/2;
					spos = child[k]->surfacePos[gs*u+v];

					h = (heightField[gs*uh+vh]+heightField[gs*(uh+1)+vh+1])*0.5f;
					h = jwp->get_HeightField_And_RandomOffset(spos, LOD-1, h); // much better than before (REAL 3D NOISE)!!!

					CDoubleVector unit = CDoubleVector((normals[gsuhvh3 + 0] + normals[gsuhvh3 + 3*gs + 3] + normals[gsuhvh3 + 3] + normals[gsuhvh3 + 3*gs + 0]),
													 (normals[gsuhvh3 + 1] + normals[gsuhvh3 + 3*gs + 4] + normals[gsuhvh3 + 4] + normals[gsuhvh3 + 3*gs + 1]),
													 (normals[gsuhvh3 + 2] + normals[gsuhvh3 + 3*gs + 5] + normals[gsuhvh3 + 5] + normals[gsuhvh3 + 3*gs + 2]));
					unit.Normalize();
					child[k]->normals[gsuv3 + 0] = (float)unit.x;
					child[k]->normals[gsuv3 + 1] = (float)unit.y;
					child[k]->normals[gsuv3 + 2] = (float)unit.z;
					child[k]->textureCoords[gsuv2 + 0] = (float)(textureCoords[gsuhvh2 + 0]+textureCoords[gsuhvh2 + 2*gs + 2])*0.5f;
					child[k]->textureCoords[gsuv2 + 1] = (float)(textureCoords[gsuhvh2 + 1]+textureCoords[gsuhvh2 + 2*gs + 3])*0.5f;										
					child[k]->textureCoords2[gsuv2 + 0] = (float)(textureCoords2[gsuhvh2 + 0]+textureCoords2[gsuhvh2 + 2*gs + 2])*0.5f;
					child[k]->textureCoords2[gsuv2 + 1] = (float)(textureCoords2[gsuhvh2 + 1]+textureCoords2[gsuhvh2 + 2*gs + 3])*0.5f;
					child[k]->textureCoords3[gsuv2 + 0] = (float)(textureCoords3[gsuhvh2 + 0]+textureCoords3[gsuhvh2 + 2*gs + 2])*0.5f;
					child[k]->textureCoords3[gsuv2 + 1] = (float)(textureCoords3[gsuhvh2 + 1]+textureCoords3[gsuhvh2 + 2*gs + 3])*0.5f;
				}				
			}

			hmin = Min(h,hmin);
			hmax = Max(h,hmax);
			child[k]->heightField[gsuv] = h;
			// now compute vertex color
			// fine LOD can be underwater (we have a special water surface), so use "normal" surface color!
			if (planet && (planet->hasWater) && (LOD <= LOD_quality))
			{
				planet->hasWater = false;
				color = planet->computeColor(h, spos);
				planet->hasWater = true;
			}
			else
				color = jwp->computeColor(h, spos);

			child[k]->colorData[4*gsuv + 0] = color.r;
			child[k]->colorData[4*gsuv + 1] = color.g;
			child[k]->colorData[4*gsuv + 2] = color.b;
			child[k]->colorData[4*gsuv + 3] = color.a;
			
			if (planet && (planet->hasWater) && (LOD > (LOD_quality+1)))
				if (h < 0) h = 0;
      if (planet && (planet->style == JWPlanet::WATER)) h=0;
  		if (ring) h=0;

			vpos = jwp->SurfaceVectorToCoordinates(spos, id, h);
			// for all LODs simply interpolate shpere position
			child[k]->vertexData[gsuv3 + 0] = vpos.x;
			child[k]->vertexData[gsuv3 + 1] = vpos.y;
			child[k]->vertexData[gsuv3 + 2] = vpos.z;
			
			// create water surface for LOD_quality
			if (planet && (planet->hasWater) && (LOD == (LOD_quality+1)))
			{
				// on patch borders we use uprounded crack vertices
				if ((u % (grid_size/2) == 0) || (v % (grid_size/2) == 0))
					vpos = jwp->SurfaceVectorToCoordinates(spos, id, Max(0,h));
				else
					vpos = jwp->SurfaceVectorToCoordinates(spos, id, Min(0,h)-100);

				child[k]->crackData[gsuv3 + 0] = vpos.x;
				child[k]->crackData[gsuv3 + 1] = vpos.y;
				child[k]->crackData[gsuv3 + 2] = vpos.z;
			}
			
			if (LOD == jwp->HF_LOD)
			{
				// 2002/11/08 i removed the wrong divisor "/grid_size"!
				child[k]->textureCoords[gsuv2 + 0] = (float)(0.5f*v); // this coords are used for detail tex
				child[k]->textureCoords[gsuv2 + 1] = (float)(0.5f*u);								
				child[k]->textureCoords2[gsuv2 + 0] = (float)(0.5f*v*64); // this one is for fine Detail
				child[k]->textureCoords2[gsuv2 + 1] = (float)(0.5f*u*64);								
				child[k]->textureCoords3[gsuv2 + 0] = (float)(0.5f*v*64*64); // this one is for superfine Detail and Water!
				child[k]->textureCoords3[gsuv2 + 1] = (float)(0.5f*u*64*64);								
			}
		}
	}

	// center is middle point of this patch
	CDoubleVector center = CDoubleVector(&child[k]->vertexData[3*((gs/2)*gs + (gs/2))]); //jwp->SurfaceVectorToCoordinates(spos, id, h);

	if (LOD == LOD_quality+1)
		child[k]->offset = center;
	else
		child[k]->offset = offset;

	if (LOD <= LOD_quality+1)
	{
		// center is new center too
		// but transpose all vertices around center (so that they come near to the camera!)
		for (u=0; u<gs; u++)
		{
			for (v=0; v<gs; v++)
			{
				gsuv    = gs*u+v;
				gsuv3   = 3*gsuv;

				child[k]->vertexData[gsuv3 + 0] -= child[k]->offset.x;
				child[k]->vertexData[gsuv3 + 1] -= child[k]->offset.y;
				child[k]->vertexData[gsuv3 + 2] -= child[k]->offset.z;
				if (planet && (planet->hasWater) && (LOD == LOD_quality+1))
				{
					child[k]->crackData[gsuv3 + 0] -= child[k]->offset.x;
					child[k]->crackData[gsuv3 + 1] -= child[k]->offset.y;
					child[k]->crackData[gsuv3 + 2] -= child[k]->offset.z;
				}
			}
		}
	}

	child[k]->SetPosition(center);	

	if (planet)
	// Now calculate accurate real normals for this patch (only for planets)!
	// I use a modified version of "Fast Heightfield Normal Calculation" in Game Programming Gems 3
	// since i must use the vectors instead of simple height differences which would be faster
	// I first tried to interpolate everything but now i recalc only the inner points!
	for (u=1; u<gs-1; u++)
	{
		for (v=1; v<gs-1; v++)
		{
			CDoubleVector currentVertex = CDoubleVector(&child[k]->vertexData[3*(gs*u+v)]); // short version because of double
			CDoubleVector currentNormal = CDoubleVector(child[k]->normals[3*(gs*u+v)+0], child[k]->normals[3*(gs*u+v)+1], child[k]->normals[3*(gs*u+v)+2]);
			CDoubleVector n1,n2,n3,n4;	// normal in "quadrant nx" relative to current vertex
			// compute normal of upper right quad patch relative to vertex (if possible)
			// create cross product of vectors to upper and right neighbour vertices
			n1 = (CDoubleVector(&child[k]->vertexData[3*(gs*u+(v+1))]) - currentVertex) ^ (CDoubleVector(&child[k]->vertexData[3*(gs*(u+1)+v)]) - currentVertex);
			// compute normal of upper left quad patch relative to vertex (if possible)
			// create cross product of vectors to upper and left neighbour vertices
			n2 = (CDoubleVector(&child[k]->vertexData[3*(gs*(u+1)+v)]) - currentVertex) ^ (CDoubleVector(&child[k]->vertexData[3*(gs*u+(v-1))]) - currentVertex);
			// compute normal of lower left quad patch relative to vertex (if possible)
			// create cross product of vectors to lower and left neighbour vertices
			n3 = (CDoubleVector(&child[k]->vertexData[3*(gs*u+(v-1))]) - currentVertex) ^ (CDoubleVector(&child[k]->vertexData[3*(gs*(u-1)+v)]) - currentVertex);
			// compute normal of lower right quad patch relative to vertex (if possible)

			// create cross product of vectors to lower and right neighbour vertices
			n4 = (CDoubleVector(&child[k]->vertexData[3*(gs*(u-1)+v)]) - currentVertex) ^ (CDoubleVector(&child[k]->vertexData[3*(gs*u+(v+1))]) - currentVertex);

			currentNormal = n1 + n2 + n3 + n4;
			currentNormal.Normalize();
			child[k]->normals[3*(gs*u+v)+0] = (float)currentNormal.x;
			child[k]->normals[3*(gs*u+v)+1] = (float)currentNormal.y;
			child[k]->normals[3*(gs*u+v)+2] = (float)currentNormal.z;
		}
	}
	child[k]->hmin = hmin;
	child[k]->hmax = hmax;

	// determine boundingRadius by evaluating center against all 4 corners of child patch
	center = CDoubleVector( &child[k]->vertexData[3*((gs/2)*gs + (gs/2))] );
	child[k]->boundingRadius = center.Distance(CDoubleVector( &child[k]->vertexData[0] )); 
	child[k]->boundingRadius = Max( center.Distance(CDoubleVector( &child[k]->vertexData[3*(gs-1)] )), child[k]->boundingRadius); // Sicherheitszuschlag!
	child[k]->boundingRadius = Max( center.Distance(CDoubleVector( &child[k]->vertexData[3*((gs-1)*gs)] )), child[k]->boundingRadius); // Sicherheitszuschlag!
	child[k]->boundingRadius = Max( center.Distance(CDoubleVector( &child[k]->vertexData[3*(gs*gs-1)] )), child[k]->boundingRadius); // Sicherheitszuschlag!
	// for LODs > 13 we need to surround the water surface and the terrain!
	if (planet && planet->hasWater && (LOD > LOD_quality+1))
		hmax = Max(hmax, 0);
  if (!planet || (planet && (planet->style != JWPlanet::WATER)))
	  child[k]->boundingRadius += (hmax-hmin); // essential addition!
	child[k]->boundingRadius *= (double)1.1; // numerical security scale!
	//if (LOD > jwp->HF_LOD)
	//	child[k]->boundingRadius *= (double)1.5; // high LOD need more tolerance due to precision problems!!!!

	// generate static objects for child patch
	generateStaticObjects(k);
}

void JWQuadtree::generateStaticObjects(const int k)
{
	int gs = LODObject::gs;
	int grid_size = LODObject::grid_size;
	// generate static objects in child patch

	int i = k/2;		// 0 oder 1
	int j = (k & 1);	// 0 oder 1
	int u,v, uh, vh;

	// we need to adjust the LOD measurement on cube face edges because there the
	// base size of a patch is smaller than in the middle of a cube side!
	// "compress" is the compression factor of the geometry
	CDoubleVector compvec = CDoubleVector(child[k]->surfacePos[(gs/2)*gs + (gs/2)].x,
										  child[k]->surfacePos[(gs/2)*gs + (gs/2)].y,
										  child[k]->surfacePos[(gs/2)*gs + (gs/2)].z);
	compvec /= (jwp->cube_size/2);
	float compress;	 // this reflects the compression of the geometry
	if (planet)
		compress = compvec.Magnitude();
	else if (ring)
	{
		compress = (child[k]->surfacePos[(gs/2)*gs + (gs/2)].y) / ((double)(jwp->cube_size)) + 0.5; // factor now from 0(r_min) to 1(r_max)
		compress = Max(1, ((ring->r_max + ring->r_min)/2) / ((1-compress)*ring->r_min + compress*ring->r_max));
	}

	// simple constant random number generator
	srand((unsigned int)((child[k]->hmax+127)*compress));
	
	bool cityZone = false;
	double alpha, beta, height;
	if (planet && (planet->cityDistance > 0))
	{
		CoordinatesToPolar(compvec.x, compvec.y, compvec.z, &alpha, &beta, &height);
		alpha = 1000*(alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, 1000]
		beta = 1000*(1-beta/PI);		  // convert from [0,PI] to [0,1000]
		cityZone = (((int)alpha % planet->cityDistance == 0) && ((int)beta % planet->cityDistance == 0));	// restrict city regions!
	}

	double delta_h = (child[k]->hmax - child[k]->hmin) / child[k]->boundingRadius; // simple roughness parameter!

	JWStaticObject *childobj = 0;	// obj that a child can get
	JWPlant *flower = 0;
	// setup parameters like LOD selection from objects in the vectors of jwp
	// 1st attempt: cycle through the objects list and add the first suitable candidate!
	std::vector<JWStaticObject *>::iterator theIterator;
	// lookup contents of the Vector objects
	for (theIterator = jwp->objects.begin(); (theIterator != jwp->objects.end()) && !childobj; theIterator++)
	{
		if ( ( (*theIterator)->baseLOD <= (child[k]->LOD - log2((unsigned int)grid_size)) ) 
			 && ( (*theIterator)->visibleLOD >= (child[k]->LOD - log2((unsigned int)grid_size)) ) )	// -value because of field size
		{
			height = child[k]->heightField[(gs/2)*gs + (gs/2)]; // take center of this patch
			if ((height > (*theIterator)->minHeight) && (height < (*theIterator)->maxHeight))
			{
				if (dynamic_cast <JWHouse *> (*theIterator) != 0)
				{
					if ((delta_h < 0.15) && (cityZone))// buildings must have a flat underground!
						childobj = (*theIterator);
				}
				else
				{
					childobj = (*theIterator);
				}
			}
		}
	}

	int baseLOD = 0;
	if (childobj)
	{
		flower = dynamic_cast <JWPlant *> (childobj);
		baseLOD = (int)(1.42 * childobj->width/2)+1;
	}
	int diff = 0;
	int vertexdist = (int)((1 << child[k]->LOD) / (compress * grid_size));

	float localdensity = 1.0f;
	int hx,hy, hw,hh;
	double rx, ry;
	int	byp;
	TEXTURE *structure = 0;
	if (ring)
	{
		structure = ring->structureTex;
		byp = structure->bpp / 8;
		hw = structure->width;
		hh = structure->height;
	}

	int rval;
	CFloatVector4 tcol;

	for (u=0; u<gs; u++)
	{
		uh = u/2 + (i*grid_size/2);
		
		for (v=0; v<gs;v++)
		{
			vh = v/2 + (j*grid_size/2);

			height = child[k]->heightField[u*gs + v]; // take height of current vertex
			
			if (childobj)
			{
				if (ring)
				{
					localdensity = child[k]->colorData[4*(u*gs + v)+3];
					// make localdensity also dependant on structureTex (* stuctureTex->imageData[alpha of pos])!!!
					// this means multiply with alpha of fine structure
					alpha = ((child[k]->surfacePos[gs*u+v].x / (double)jwp->cube_size) + 0.5) * (jwp->HF_size); // now in [0, HF_size]
					beta = ((child[k]->surfacePos[gs*u+v].y / (double)jwp->cube_size) + 0.5)* (jwp->HF_size);   // now in [0, HF_size]
					hx = (int)alpha;
					hy = (int)beta;
					rx = alpha-hx; // rest between 0 and 1
					ry = beta-hy; // rest between 0 and 1
					alpha = rx*(hw)-0.5;// now map into structuretex (-0.5 VERY IMPORTANT for repeated textures)
					beta = ry*(hh)-0.5;	// now map into structuretex (-0.5 VERY IMPORTANT for repeated textures)
					hx = (int)alpha;
					hy = (int)beta;
					rx = Cubic(alpha-hx); // rest between 0 and 1, bind stronger to nearest point
					ry = Cubic(beta-hy); // rest between 0 and 1, bind stronger to nearest point
					// reuse alpha as color alpha value!
					alpha = (1-rx)*(1-ry)*structure->imageData[byp*((hy % hh)*hw + (hx % hw)) + (byp-1)];
					alpha += rx*(1-ry)*structure->imageData[byp*((hy % hh)*hw + ((hx+1) % hw)) + (byp-1)];
					alpha += (1-rx)*ry*structure->imageData[byp*(((hy+1) % hh)*hw + (hx % hw)) + (byp-1)];
					alpha += rx*ry*structure->imageData[byp*(((hy+1) % hh)*hw + ((hx+1) % hw)) + (byp-1)];
					localdensity = localdensity * (alpha / 255.0f);
					if (localdensity < 0.6) // very good clamping value
						localdensity = 0;
					else
						localdensity = 1.0;
				}
				else if (planet)
				{
					if (flower)
					{
						tcol = CFloatVector4(&child[k]->colorData[4*(u*gs + v)]);
						// if grey is the dominant terrain color (stone, ice)
						if ((tcol.g > 0.9*tcol.r) && (tcol.g > 0.9*tcol.b) && (tcol.g < 1.1*tcol.r) && (tcol.g < 1.1*tcol.b))
							localdensity = 0;
						// if green is the dominant terrain color
						else if ((tcol.g > tcol.r) && (tcol.g > tcol.b))
							localdensity = 1.0f;
						// else if green is not dominant (blue or red values: sea, deserts ...)
						else
						{
							// redish, brown zones
							if ((tcol.g > 0.5*tcol.r) && (tcol.g > tcol.b))
								localdensity = 0.25f;	// lower plant density
							else
								localdensity = 0;	// be hard!
						}
					}
				}
			}
			
			if (((u & 1) == 0) && ((v & 1) == 0))
			{
				// this "even" points stay the same
				if ((object[gs*uh+vh] != 0) || (objLOD[gs*uh+vh] > 0))
				{
					child[k]->object[gs*u+v] = jwp->INVALID;		// this place is filled!
					child[k]->objLOD[gs*u+v] = objLOD[gs*uh+vh];	// decrease the LOD difference
				}
				else
				{
					// no higher LOD did some placement, so go on!
					if ((childobj != 0) && (height > childobj->minHeight) && (height < childobj->maxHeight))
					{
						// shuffle:
						rval = 255 - (int)(childobj->density * localdensity * 255);
						if ((rand() & 255) > rval)
						//if (localdensity > 0)
						{
							child[k]->object[gs*u+v] = childobj;
							if (childobj->isSolid)
								child[k]->objLOD[gs*u+v] = baseLOD; // store how small the childobj is
						}
					}
				}
			}
			else
			{
				if ((u & 1) == 0)
					diff = MAX( objLOD[gs*uh+vh], objLOD[gs*uh+vh+1]);
				else if ((v & 1) == 0)
					diff = MAX( objLOD[gs*uh+vh], objLOD[gs*(uh+1)+vh]);
				else
					diff = MAX( objLOD[gs*uh+vh], MAX(objLOD[gs*uh+vh+1], MAX(objLOD[gs*(uh+1)+vh], objLOD[gs*(uh+1)+vh+1])));

				diff =  diff + baseLOD - vertexdist;
				// check if we are allowed to generate an object
				if (diff > 0)
				{
					// we are not allowed to place an object!
					child[k]->objLOD[gs*u+v] = diff;
				}
				else
				{
					// the higher LOD placements were not full size, so go on!
					if ((childobj != 0) && (height > childobj->minHeight) && (height < childobj->maxHeight))
					{
						// shuffle:
						rval = 255 - (int)(childobj->density * localdensity * 255);
						if ((rand() & 255) > rval)
						//if (localdensity > 0)
						{
							child[k]->object[gs*u+v] = childobj;
							if (childobj->isSolid)
								child[k]->objLOD[gs*u+v] = baseLOD; // store how much too small the childobj is
						}
					}
				}
			}
		}
	}
}

void JWQuadtree::update(C3DCamera *pCamera, const CIntVector pos, int LODmin)
{
	// we have viewport clipping
	isVisible = pCamera->isVisible(this);

  int quadsize = (1 << LOD);
	int border = (2 + jwengine::jwgame->QualityScale)*quadsize;	// The border is the main influence factor for quality vs. performance variation!
	// sensible range for QualityScale is from 0 to 2, NOTE: we add a offset of 2 which is asolutetly necessary
	// because if it were 1 we would use a fine LOD only if we are inside this fine patch
	// the result was that there would exist borders of more than 1 LOD difference which is very BAD!
	int dist, dx,dy,dz;
	CIntVector center;
	
	int i,j,k;
	int x,y,gf;
	
	int gs = LODObject::gs;
	int grid_size = LODObject::grid_size;
	
	for (i=0; i<2; i++)
		for (j=0; j<2; j++)
		{
			k = 2*i+j; // k in [0,3]
			gf = grid_size/4;
			x = (1 + 2*j) * gf;
			y = (1 + 2*i) * gf;
			// center is middle point of this subpatch
			center = surfacePos[y*gs + x];
			dx = pos.x - center.x;
			if (dx < 0) dx = -dx;
			dy = pos.y - center.y;
			if (dy < 0) dy = -dy;
			dz = pos.z - center.z;
			if (dz < 0) dz = -dz;
			dist = dx + dy + dz;
			
			if ((LODmin < LOD) && (dist <= border) && isVisible)
			{
				if (child[k] == 0)
				{
					// create new quadpatch
					createChild(k);
				}
				child[k]->isActive = true;
				child[k]->update(pCamera, pos, LODmin);
			}
			else
			{
				if(child[k] != 0) 
				{
					delete child[k];
					child[k]=0;
				}
			}
		}
		// Nachbarschaft testen:
		CIntVector basePos  = surfacePos[(gs/2)*gs + (gs/2)];
		CIntVector dxPos = surfacePos[(gs-1)] - surfacePos[0];
		CIntVector dyPos = surfacePos[(gs-1)*gs] - surfacePos[0];
		// nl-case:
		center = basePos - dxPos;
		dx = pos.x - center.x;
		if (dx < 0) dx = -dx;
		dy = pos.y - center.y;
		if (dy < 0) dy = -dy;
		dz = pos.z - center.z;
		if (dz < 0) dz = -dz;
		dist = dx + dy + dz;
		if (dist <= 2*border) nl = true; else nl = false;
		// nu-case:
		center = basePos - dyPos;
		dx = pos.x - center.x;
		if (dx < 0) dx = -dx;
		dy = pos.y - center.y;
		if (dy < 0) dy = -dy;
		dz = pos.z - center.z;
		if (dz < 0) dz = -dz;
		dist = dx + dy + dz;
		if (dist <= 2*border) nu = true; else nu = false;
		// nr-case:
		center = basePos + dxPos;
		dx = pos.x - center.x;
		if (dx < 0) dx = -dx;
		dy = pos.y - center.y;
		if (dy < 0) dy = -dy;
		dz = pos.z - center.z;
		if (dz < 0) dz = -dz;
		dist = dx + dy + dz;
		if (dist <= 2*border) nr = true; else nr = false;
		// no-case:
		center = basePos + dyPos;
		dx = pos.x - center.x;
		if (dx < 0) dx = -dx;
		dy = pos.y - center.y;
		if (dy < 0) dy = -dy;
		dz = pos.z - center.z;
		if (dz < 0) dz = -dz;
		dist = dx + dy + dz;
		if (dist <= 2*border) no = true; else no = false;
		
		return;
}


// following function should be performed:
// user computes a virtual cube-position from current player position
// then you search the height of terrain at this position

// on the correct initial quad-patch of the planet
// NOTE: the search MUST start at the right parent quad-patch to avoid errors!
double JWQuadtree::getHeight_of_Terrain(CIntVector point)
{
  if (planet && (planet->style == JWPlanet::WATER))
    return 0; // global water height!

	int dist, dx,dy,dz, dmin;
	CIntVector center;
	
	int i,j,k, idx;
	int x,y,gf;
	
	int gs = LODObject::gs;
	int grid_size = LODObject::grid_size;
	
	dmin = 4*jwp->cube_size; // init bigger than every possible distance!
	for (i=0; i<2; i++)
		for (j=0; j<2; j++)
		{
			k = 2*i+j; // k in [0,3]
			gf = grid_size/4;
			x = (1 + 2*j) * gf;
			y = (1 + 2*i) * gf;
			// center is middle point of this subpatch
			center = surfacePos[y*gs + x];
			dx = point.x - center.x;
			if (dx < 0) dx = -dx;
			dy = point.y - center.y;
			if (dy < 0) dy = -dy;
			dz = point.z - center.z;
			if (dz < 0) dz = -dz;
			dist = dx + dy + dz;

			if (dist < dmin)
			{
				dmin = dist;
				idx = k;
			}
		}

	if (child[idx] != 0)
		return child[idx]->getHeight_of_Terrain(point);
	else
	{
		dmin = 4*jwp->cube_size; // init bigger than every possible distance!
		for (i=0; i<gs; i++)
			for (j=0; j<gs; j++)
			{
				center = surfacePos[i*gs + j];
				dx = point.x - center.x;
				if (dx < 0) dx = -dx;
				dy = point.y - center.y;
				if (dy < 0) dy = -dy;
				dz = point.z - center.z;
				if (dz < 0) dz = -dz;
				dist = dx + dy + dz;
				
				if (dist < dmin)
				{
					dmin = dist;
					idx = i*gs + j;
				}
			}
		return heightField[idx]; 	
		/*
		int cdx = 0;
		if (nu)	cdx += 1;
		if (nr)	cdx += 2;
		if (no)	cdx += 4;
		if (nl)	cdx += 8;
		// now look in this subpatches
		int side = 0;
		CDoubleVector o, d, v0,v1,v2;
		double u,v,t;
		for (int part=0; part<4; part++)
		{
			// use the correct indexList of current triangulation
			side = 8*idx + 2*part + (((1 << part) & (15-cdx)) >> part);
			// check all triangles in indexList
			for (i=0; i<LODObject::indexLenght[side]; i+=3)
			{
				o = CDoubleVector(point.x, point.y, point.z);
				d = o;
				d.Normalize();
				v0 = CDoubleVector(surfacePos[LODObject::indexList[side][i]].x, surfacePos[LODObject::indexList[side][i]].y, surfacePos[LODObject::indexList[side][i]].z);
				v1 = CDoubleVector(surfacePos[LODObject::indexList[side][i+1]].x, surfacePos[LODObject::indexList[side][i+1]].y, surfacePos[LODObject::indexList[side][i+1]].z);
				v2 = CDoubleVector(surfacePos[LODObject::indexList[side][i+2]].x, surfacePos[LODObject::indexList[side][i+2]].y, surfacePos[LODObject::indexList[side][i+2]].z);
				if (RayTriangleIntersection(o,d,v0,v1,v2,&u,&v,&t))
				{
					return ((1-u-v)*heightField[LODObject::indexList[side][i]] + u*heightField[LODObject::indexList[side][i+1]] + v*heightField[LODObject::indexList[side][i+2]]);
				}
			}
		}
		return 0;
		*/
	}
}

#if defined(_WIN32)
extern PFNGLACTIVETEXTUREARBPROC        glActiveTextureARB;
extern PFNGLCLIENTACTIVETEXTUREARBPROC  glClientActiveTextureARB;
#endif

void JWQuadtree::draw(C3DCamera *pCamera)
{
	int i;
	int gh = LODObject::grid_size/2;
	int n = 2*gh*gh;
	int gs = LODObject::gs;
	float hNN;
	if (planet)
		hNN = (float)((planet->GetFullPosition()-pCamera->GetFullPosition()).Magnitude() - planet->radius);
	else
		hNN = (float)jwp->getHeight_above_Surface(pCamera);
	float r,g,b;


	// if we currently have NO fineLOD mode we are not allowed to draw small LODs! 
	if ((!jwp->fineLOD) && (LOD <= LOD_quality))
		return;

  // we have viewport clipping
	//isVisible = pCamera->isVisible(this);

	if (!isVisible) //check Visibility of this node!
		return;

  // All LODs < HF_LOD are not covered by the large texture, so disable it!
	if(LOD == (jwp->HF_LOD -1))
	{
		if (planet)
		{
			// TEXTURE-UNIT #0		
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glBindTexture(GL_TEXTURE_2D, planet->TexDetail->texID);	// select detail Texture
			// TEXTURE-UNIT #1:
			glActiveTextureARB(GL_TEXTURE1_ARB);					// select fine detail
			glBindTexture(GL_TEXTURE_2D, planet->TexFineDetail->texID);		
			// TEXTURE-UNIT #2:
			glActiveTextureARB(GL_TEXTURE2_ARB);					// select fine detail
			glBindTexture(GL_TEXTURE_2D, planet->TexSuperFineDetail->texID);	
		}
		else if (ring)
		{
			// TEXTURE-UNIT #0		
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glBindTexture(GL_TEXTURE_2D, ring->structureTex->texID);	// select detail Texture
			// TEXTURE-UNIT #1:
			glActiveTextureARB(GL_TEXTURE1_ARB);			// select fine detail
			glBindTexture(GL_TEXTURE_2D, ring->structureTex->texID);		
			// TEXTURE-UNIT #2:
			glActiveTextureARB(GL_TEXTURE2_ARB);			// select fine detail
			glBindTexture(GL_TEXTURE_2D, ring->fadeOut->texID);	
		}
	}

	if (LOD == LOD_quality) //LOD_quality!!! see above
	{

		// Multiply the current view matrix by the model matrix and draw the planet
		CDoubleMatrix m = GetScaledModelMatrix(pCamera);
		C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
		glPushMatrix();
		// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
		glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);
	}

	if (LOD == jwengine::jwgame->lineLOD)
	{
		r = (rand() & 255)/256.0f;
		g = (rand() & 255)/256.0f;
		b = (rand() & 255)/256.0f;

    CDoubleVector p;

    // draw a boundingSphere!
		p = CDoubleVector(&vertexData[3*( (gs/2)*gs + (gs/2))]);
		glPushMatrix();
		glTranslated(p.x, p.y, p.z);

	  GLUquadricObj *qobj = gluNewQuadric();
    glColor3f(r,g,b);
	  gluQuadricDrawStyle(qobj, GLU_LINE); /* smooth shaded */
	  gluSphere(qobj, boundingRadius, 8, 8); // make big enough to avoid near clip plane, at cloud distance!!!
	  gluDeleteQuadric(qobj);

    glPopMatrix();
	}
	
	// 1 - check all 4 child regions and decide what to draw
	for (i=0; i<4; i++)
	{
		if ((child[i] != 0) && (child[i]->isVisible))
			child[i]->draw(pCamera);
		else
		{
			bool dopaint = true;

			if (planet)
			{
			// draw only fine LOD underwater because of fog that clips everything that is further away!
			if ((LOD > LOD_quality) && (planet->hasWater) && (hNN < 0))
				dopaint = false;
			
			// if this is a underwater terrain patch do not draw if we are far above water surface (no transparency)!
			if ((LOD <= LOD_quality) && (planet->hasWater) && (hNN > 1000) && (hmax < 0))
				dopaint = false;
			}

			// if we currently have a fineLOD mode we are not allowed to overdraw with LODs higher than LOD_quality! 
			if ((jwp->fineLOD) && (LOD > LOD_quality))
				dopaint = false;
				
			if (dopaint)
			{
				int cdx = 0;
				if (nu)	cdx += 1;
				if (nr)	cdx += 2;
				if (no)	cdx += 4;
				if (nl)	cdx += 8;

				// we disable this texture now and blend it over in a second pass!
				if(LOD == (jwp->HF_LOD))
				{
					// TEXTURE-UNIT #0		
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glBindTexture(GL_TEXTURE_2D, 0);	// deselect Texture for this cube side
				}

        bool useNormalArray = true; // most times we use normals
        bool useColorArray = false; // most times we do not use this array

        if ((colorData != 0) && (vertexData != 0) && (normals != 0))
				{

					// 2 - no child, draw current triangles
					glEnableClientState (GL_VERTEX_ARRAY);
					// TEXTURE-UNIT #0		
					glClientActiveTextureARB(GL_TEXTURE0_ARB);
					glTexCoordPointer(2, GL_FLOAT, 0, textureCoords);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					// TEXTURE-UNIT #1		
					glClientActiveTextureARB(GL_TEXTURE1_ARB);
					glTexCoordPointer(2, GL_FLOAT, 0, textureCoords2);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					// TEXTURE-UNIT #2		
					glClientActiveTextureARB(GL_TEXTURE2_ARB);
					glTexCoordPointer(2, GL_FLOAT, 0, textureCoords3);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					
					glVertexPointer (3, GL_DOUBLE, 0, vertexData);

					if (planet && (planet->style == JWPlanet::SUN))
            useNormalArray = false;

          if (useNormalArray)
          {
					  glEnableClientState (GL_NORMAL_ARRAY);
					  glNormalPointer (GL_FLOAT, 0, normals);
          }

          // we are at lineLOD
          if (LOD == jwengine::jwgame->lineLOD)
          {
            glColor3f(r,g,b);
          }
          // this is a sun and we do not draw fire colors (sunmode) but simple light
          else if (planet && planet->style == JWPlanet::SUN && !jwengine::jwgame->sunmode)
          {
            if (((JWStar *)jwp)->lightnum != -1)
              glColor3f(((JWStar *)jwp)->lr, ((JWStar *)jwp)->lg, ((JWStar *)jwp)->lb);
            else
              glColor3f(((JWStar *)jwp)->rgbSkyDayNight[0], ((JWStar *)jwp)->rgbSkyDayNight[1], ((JWStar *)jwp)->rgbSkyDayNight[2]);
          }
          else
          {
            if (LOD <= (jwp->HF_LOD))
            {
              glEnableClientState (GL_COLOR_ARRAY);
              glColorPointer (4, GL_FLOAT, 0, colorData);
              useColorArray = true;
              
              // the ring must use a modified color array because of double blending!!!
              if (ring && (LOD == (jwp->HF_LOD)))
              {							
                int x,y;
                int basex = (i & 1)*(gh);
                int basey = (i & 2)*(gh/2);
                
                for (y=basey; y < basey+gh+1; y++)
                {
                  for (x=basex; x < basex+gh+1; x++)
                  {
                    // now fill in the correct alpha value
                    ring->colorModify[4*(y*gs + x) + 0] = colorData[4*(y*gs + x) + 0];
                    ring->colorModify[4*(y*gs + x) + 1] = colorData[4*(y*gs + x) + 1];
                    ring->colorModify[4*(y*gs + x) + 2] = colorData[4*(y*gs + x) + 2];
                    ring->colorModify[4*(y*gs + x) + 3] = Clamp(0,1,colorData[4*(y*gs + x) + 3]*(1.7 - LODObject::colorBlend[cdx][4*(y*gs + x) + 3])); // do some inverse blending
                  }
                }
                // now use the modified array
                glColorPointer (4, GL_FLOAT, 0, ring->colorModify);
              }
            }
            else
              glColor3f(1,1,1);
          }
        }
        
        jwp->numTriangles += n;
        // draw this part of the patch with crack fixing!
        if (jwengine::jwgame->diamondStep)
        {
          int side = 0;
          for (int u=0; u<4; u++)
          {
            side = 8*i + 2*u + (((1 << u) & (15-cdx)) >> u);
            if (LODObject::tmode == LODObject::TRIANGLE_FAN)
            {
              glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, LODObject::fanindexList[side]);
              glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, (int *)(LODObject::fanindexList[side] + LODObject::fanindexLenght[side]/2));
            }
            else if (LODObject::tmode == LODObject::TRIANGLES)
              glDrawElements(GL_TRIANGLES, LODObject::indexLenght[side], GL_UNSIGNED_INT, LODObject::indexList[side]);
            else
              glDrawElements(GL_TRIANGLE_STRIP, LODObject::stripindexLenght[side], GL_UNSIGNED_INT, LODObject::stripindexList[side]);
          }
        }
        else
        {
          glDrawElements(GL_TRIANGLE_STRIP, /*3*n*/n + LODObject::grid_size, GL_UNSIGNED_INT, LODObject::RegularIndexList[i]);
        }
        
        // only at this LOD there can be some cracks because of water diffrences
        if (planet && (LOD == LOD_quality) && (planet->hasWater) && (hmin <= 0) && (hmax >= 0))
        {
          glVertexPointer (3, GL_DOUBLE, 0, crackData);
          int side = 0;
          for (int u=0; u<4; u++)
          {
            side = 8*i + 2*u + (((1 << u) & (15-cdx)) >> u);
            // only fix refinement parts!
            if (side & 1)
            {
              if (LODObject::tmode == LODObject::TRIANGLE_FAN)
              {
                glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, LODObject::fanindexList[side]);
                glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, (int *)(LODObject::fanindexList[side] + LODObject::fanindexLenght[side]/2));
              }
              else if (LODObject::tmode == LODObject::TRIANGLES)
                glDrawElements(GL_TRIANGLES, LODObject::indexLenght[side], GL_UNSIGNED_INT, LODObject::indexList[side]);
              else
                glDrawElements(GL_TRIANGLE_STRIP, LODObject::stripindexLenght[side], GL_UNSIGNED_INT, LODObject::stripindexList[side]);
            }
          }
        }
        
        // TEXTURE-UNIT #0		
        glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        // TEXTURE-UNIT #1		
        glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        // TEXTURE-UNIT #2		
        glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        
        glDisableClientState (GL_VERTEX_ARRAY);

        if (useColorArray)
          glDisableClientState (GL_COLOR_ARRAY);
          
        if (useNormalArray)
          glDisableClientState (GL_NORMAL_ARRAY);
        
        // now we need to overdraw in a second pass with the texture blend array (only for planets)!
        if(LOD == (jwp->HF_LOD))
        {
          // TEXTURE-UNIT #0		
          glActiveTextureARB(GL_TEXTURE0_ARB);
          if (planet)
            glBindTexture(GL_TEXTURE_2D, planet->Tex[id]->texID);	// select Texture for this cube side
          else
            glBindTexture(GL_TEXTURE_2D, ring->ringTex->texID);	// select Texture for this cube side
          
          // draw triangles a second time
          glDepthFunc(GL_EQUAL);
          glPushAttrib(GL_BLEND);
          glEnable(GL_BLEND);
          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
          
          glEnableClientState (GL_VERTEX_ARRAY);
          // TEXTURE-UNIT #0		
          glClientActiveTextureARB(GL_TEXTURE0_ARB);
          glTexCoordPointer(2, GL_FLOAT, 0, textureCoords);
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
          // TEXTURE-UNIT #1		
          glClientActiveTextureARB(GL_TEXTURE1_ARB);
          glTexCoordPointer(2, GL_FLOAT, 0, textureCoords2);
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
          // TEXTURE-UNIT #2		
          glClientActiveTextureARB(GL_TEXTURE2_ARB);
          glTexCoordPointer(2, GL_FLOAT, 0, textureCoords3);
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
          
          glVertexPointer (3, GL_DOUBLE, 0, vertexData);
          
          if (useNormalArray)
          {
            glEnableClientState (GL_NORMAL_ARRAY);
            glNormalPointer (GL_FLOAT, 0, normals);
          }
          
          glEnableClientState (GL_COLOR_ARRAY);
          
          // use 16 different blend arrays!
          glColorPointer (4, GL_FLOAT, 0, LODObject::colorBlend[cdx]);
          
          jwp->numTriangles += n;
          // draw this part of the patch with crack fixing!
          if (jwengine::jwgame->diamondStep)
          {
            int side = 0;
            for (int u=0; u<4; u++)
            {
              side = 8*i + 2*u + (((1 << u) & (15-cdx)) >> u);
              if (LODObject::tmode == LODObject::TRIANGLE_FAN)
              {
                glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, LODObject::fanindexList[side]);
                glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, (int *)(LODObject::fanindexList[side] + LODObject::fanindexLenght[side]/2));
              }
              else if (LODObject::tmode == LODObject::TRIANGLES)
                glDrawElements(GL_TRIANGLES, LODObject::indexLenght[side], GL_UNSIGNED_INT, LODObject::indexList[side]);
              else
                glDrawElements(GL_TRIANGLE_STRIP, LODObject::stripindexLenght[side], GL_UNSIGNED_INT, LODObject::stripindexList[side]);
            }
          }
          else
          {
            glDrawElements(GL_TRIANGLE_STRIP, /*3*n*/n + LODObject::grid_size, GL_UNSIGNED_INT, LODObject::RegularIndexList[i]);
          }
          
          // TEXTURE-UNIT #0		
          glClientActiveTextureARB(GL_TEXTURE0_ARB);
          glDisableClientState(GL_TEXTURE_COORD_ARRAY);
          // TEXTURE-UNIT #1		
          glClientActiveTextureARB(GL_TEXTURE1_ARB);
          glDisableClientState(GL_TEXTURE_COORD_ARRAY);
          // TEXTURE-UNIT #2		
          glClientActiveTextureARB(GL_TEXTURE2_ARB);
          glDisableClientState(GL_TEXTURE_COORD_ARRAY);
          
          glDisableClientState (GL_VERTEX_ARRAY);
          glDisableClientState (GL_COLOR_ARRAY);
          
          if (useNormalArray)
            glDisableClientState (GL_NORMAL_ARRAY);
          
          glPopAttrib(); // GL_BLEND!
          if (planet && planet->style == JWPlanet::CLOUDS)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

          glDepthFunc(GL_LESS);
        }
      }
    }
  }
  
  if(LOD == (jwp->HF_LOD-1))
  {
    if (planet)
    {
      // TEXTURE-UNIT #0		
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glBindTexture(GL_TEXTURE_2D, planet->Tex[id]->texID);	// select Texture for this cube side
      // TEXTURE-UNIT #1:
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glBindTexture(GL_TEXTURE_2D, planet->TexDetail->texID);// select detail Texture
      // TEXTURE-UNIT #2:
      glActiveTextureARB(GL_TEXTURE2_ARB);
      glBindTexture(GL_TEXTURE_2D, planet->TexFineDetail->texID);// select fine detail Texture
    }
    else if (ring)
    {
      // TEXTURE-UNIT #0		
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glBindTexture(GL_TEXTURE_2D, ring->ringTex->texID);	// select Texture for this cube side
      // TEXTURE-UNIT #1:
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glBindTexture(GL_TEXTURE_2D, ring->structureTex->texID);// select detail Texture
      // TEXTURE-UNIT #2:
      glActiveTextureARB(GL_TEXTURE2_ARB);
      glBindTexture(GL_TEXTURE_2D, ring->structureTex->texID);// select fine detail Texture
      
    }
  }
  
  if (LOD == LOD_quality) //LOD_quality!!! see above
  {
    glPopMatrix();
  }
}

void JWQuadtree::drawWater(C3DCamera *pCamera)
{
	int i;
	int n = 2*(LODObject::grid_size/2)*(LODObject::grid_size/2);
	int gs = LODObject::gs;
	float hNN = (float)((planet->GetFullPosition()-pCamera->GetFullPosition()).Magnitude() - planet->radius);

	CDoubleVector waterNormal = CDoubleVector(surfacePos[(gs/2)*gs].x, surfacePos[(gs/2)*gs].y, surfacePos[(gs/2)*gs].z);
	waterNormal.Normalize();

  //isVisible = pCamera->isVisible(this);

	if (!isVisible) //check Visibility of this node! (if LOD too fine ignore because of ocean depth error!)
		return;

	if (LOD == LOD_quality) //LOD_quality!!! see above
	{
		// Multiply the current view matrix by the model matrix and draw the planet
		CDoubleMatrix m = GetScaledModelMatrix(pCamera);
		C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
		glPushMatrix();
		// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
		glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);
		
	}
	// 1 - check all 4 child regions and decide what to draw
	for (i=0; i<4; i++)
	{
		if ((child[i] != 0) && (child[i]->isVisible))
			child[i]->drawWater(pCamera);
		// draw only if the terrain can get underwater!
    else if ((LOD <= LOD_quality) && (hmin < 0))
		{
		  // water should have higlights!
      /*
      float specfac = 0.02f*(LOD_quality-LOD);
		  GLfloat spec_mat[] = { specfac, specfac, specfac, specfac };
      GLfloat high_shininess[] = { specfac };
		  glMaterialfv(GL_FRONT, GL_SPECULAR, spec_mat);
		  glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);
      */

			// draw Water
			int cdx = 0;
			if (nu)	cdx += 1;
			if (nr)	cdx += 2;
			if (no)	cdx += 4;
			if (nl)	cdx += 8;

			glEnableClientState (GL_VERTEX_ARRAY);
			CFloatVector4 color = jwp->computeColor(0);

			if (hNN >= 0)
			{
				//glColor3f(r,g,b);
				// better: use modified color array!
				glEnableClientState (GL_COLOR_ARRAY);
				glColorPointer (4, GL_FLOAT, 0, colorData);
			}
			else
			{
				glColor4f(color.r, color.g, color.b,(float)Clamp(0,1,0.5-hNN/1000));
			}
			// TEXTURE-UNIT #0		
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, textureCoords);
			// TEXTURE-UNIT #1		
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, textureCoords2);
 			// TEXTURE-UNIT #2		
			glClientActiveTextureARB(GL_TEXTURE2_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, textureCoords3);
			
			glVertexPointer (3, GL_DOUBLE, 0, vertexData);
			
			glNormal3dv(waterNormal);
			
			jwp->numTriangles += n;
			// draw this part of the patch with crack fixing!
			if (jwengine::jwgame->diamondStep)
			{
				int side = 0;
				for (int u=0; u<4; u++)
				{
					side = 8*i + 2*u + (((1 << u) & (15-cdx)) >> u);
					if (LODObject::tmode == LODObject::TRIANGLE_FAN)
					{
						glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, LODObject::fanindexList[side]);
						glDrawElements(GL_TRIANGLE_FAN, LODObject::fanindexLenght[side]/2, GL_UNSIGNED_INT, (int *)(LODObject::fanindexList[side] + LODObject::fanindexLenght[side]/2));
					}
					else if (LODObject::tmode == LODObject::TRIANGLES)
						glDrawElements(GL_TRIANGLES, LODObject::indexLenght[side], GL_UNSIGNED_INT, LODObject::indexList[side]);
					else
						glDrawElements(GL_TRIANGLE_STRIP, LODObject::stripindexLenght[side], GL_UNSIGNED_INT, LODObject::stripindexList[side]);
				}
			}
			else
			{
				glDrawElements(GL_TRIANGLE_STRIP, /*3*n*/n + LODObject::grid_size, GL_UNSIGNED_INT, LODObject::RegularIndexList[i]);
			}
			
			glDisableClientState (GL_COLOR_ARRAY);
			// TEXTURE-UNIT #0		
			glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			// TEXTURE-UNIT #1		
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			// TEXTURE-UNIT #2		
			glClientActiveTextureARB(GL_TEXTURE2_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			
			glDisableClientState (GL_VERTEX_ARRAY);
			//glDisable(GL_BLEND);
		}					
	}

	if (LOD == LOD_quality) //LOD_quality!!! see above
	{
		glPopMatrix();
	}
}

void JWQuadtree::drawStaticObjects(C3DCamera *pCamera)
{
	int i, u, v;
	int n = 2*(LODObject::grid_size/2)*(LODObject::grid_size/2);
	int gs = LODObject::gs;

	if (!isVisible) //check Visibility of this node!
		return;

	if (LOD == LOD_quality) //LOD_quality!!! see above
	{
		// Multiply the current view matrix by the model matrix and draw the planet
		CDoubleMatrix m = GetScaledModelMatrix(pCamera);
		C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
		glPushMatrix();
		// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
		glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);
	}
	// 1 - check all 4 child regions and decide what to draw
	for (i=0; i<4; i++)
		if ((child[i] != 0) && (child[i]->isVisible))
			child[i]->drawStaticObjects(pCamera);

	// still here, now draw objects
	{
		CDoubleVector up = CDoubleVector(surfacePos[0].x, surfacePos[0].y, surfacePos[0].z);
		up.Normalize();
		// now compute 2 normal vectors to diff!
		CDoubleVector tmp;
		if (Abs(up | this->GetUpAxis()) < 0.7)
			tmp = this->GetUpAxis();
		else
			tmp = this->GetRightAxis();
		CDoubleVector u1 = tmp ^ up;
		CDoubleVector u2 = u1 ^ up;
		u1.Normalize();
		u2.Normalize();

		CDoubleMatrix mRot;
		mRot.ViewMatrix(CDoubleVector(0,0,0), u2, up, u1);
		mRot.Transpose();

    CDoubleVector p, objpos;
    CIntVector pdiff;

		for (u = 0; u < gs-1; u++)
			for (v = 0; v < gs-1; v++)
			{
				if ((object[u*gs+v] != 0) && (object[u*gs+v] != jwp->INVALID))
				{
					// draw this object
					p = CDoubleVector(&vertexData[3*( u*gs+ v)]);
					if (ring) // for rings the "terrain" (dust) stays at 0, so we must correct this!
						p += (CDoubleVector(0,1,0)*heightField[u*gs+v]);
          objpos = p + this->offset;

          pdiff = jwp->lastpos - this->surfacePos[u*gs+v]; // we compare camera position with nearest point

					glPushMatrix();
					glTranslated(p.x, p.y, p.z);
					glMultMatrixd(mRot);

          glColor3fv(&colorData[4*(u*gs+v)]); // use opaque color!
					//glColor3f(1,1,1);
          this->object[u*gs+v]->SetPosition(objpos);
          this->object[u*gs+v]->SetParent(jwp);
          if (pCamera->isVisible(object[u*gs+v]))
					  jwp->numTriangles += this->object[u*gs+v]->draw( ( ABS(pdiff.x)+ABS(pdiff.y)+ABS(pdiff.z) ) ); // fast LOD comparison!
					glPopMatrix();
				}
			}
	}
	
	if (LOD == LOD_quality) //LOD_quality!!! see above
	{
		glPopMatrix();
	}
}



