/************************************************************
 * Infinite Universe Engine:                                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2003 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:  JWQuadtree.h                                      *
 *                                                          *
 * Content: Header file of 3D quadtree structure            *
 *          to render planetary bodies                      *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#ifndef JWQUADTREE
#define JWQUADTREE

#include "JWObject.h"

class JWPlanet;
class JWRing;
class JWStaticObject;

class JWQuadtree : public C3DBase {
private:
  bool isVisible;       // save the visibiliy computation here
public:
  // 4 child nodes: top left[2], bottom left[0], top right[3], bottom right[1]
  JWQuadtree *child[4];

  LODObject *jwp;       // pointer to parent object (planet or ring)
  JWPlanet *planet;     // dynamic_cast of jwp to planet class
  JWRing *ring;         // dynamic_cast of jwp to ring class

  // ************************** patch data *********************************

  // a pointer field to the static objects that should be rendered
  JWStaticObject **object;
  // Here we store how much space each object consumes at its position.
  // For new objects this is sqrt(2)*width/2, otherwise the remaining 
  // overlap of an object at a neighbour vertex position.
  int *objLOD;          

  int id;                 // surface id (cube side index [0,5])
  int LOD;                // LOD of this patch, size is (2^LOD)
  int xmid, ymid;         // 2D center position of this patch (deprecated)
  double hmin, hmax;      // minimal and maximal height in this patch

  CDoubleVector offset;   // special vertex offset for geometry with LOD <= 13
  bool isActive;          // Is patch initialization completed?
  // These flags store the LOD state of the 4 neighbour patches. 
  // This is important to prevent cracks in the triangluation.
  // true = neighbour patch has the same LOD; false = neighbour's LOD is bigger
  bool no,nl,nu,nr;

  // All following fields contain data for 9*9 vertices.
  double *vertexData;   // vertex positions (x,y,z) in double space ([m]^3)
  double *crackData;    // modified vertex positions to prevent cracks
  GLfloat *normals;     // one normal vector for every vertex
  GLfloat *textureCoords;   // 2D-coordinates in [0,1]x[0,1]
  GLfloat *textureCoords2;  // 2D-coordinates in [0,1]x[0,1]
  GLfloat *textureCoords3;  // 2D-coordinates in [0,1]x[0,1]
  GLfloat *colorData;       // RGBA color value
  double *heightField;      // height value for every vertex
  CIntVector *surfacePos;   // cube position in range [-CUBE_HALF, CUBE_HALF]^3

  // ******************** function declarations ****************************

  /** 
  * Construction of a quadtree node with detail level pLOD.
  * @param pId            id of this quadtree node
  * @param pXmid, pYmid   2D position on this cube side (deprecated)
  * @param pLOD           level of detail for this patch, defines size
  * @param pJwp           pointer to parent object
  */
  JWQuadtree(const int pId, int pXmid, int pYmid,
             const int pLOD, LODObject *pJwp);

  virtual ~JWQuadtree(); // destructor

  /**
  * This method returns the height value for the given position.
  * @param point   integer position of vertex
  * @return        height at this position
  */
  double getHeight_of_Terrain(const CIntVector point);

  /**
  * Initialize geometry for a top-level quadtree node. This function is 
  * only called for the quadtree nodes that cover a whole cube side. 
  * All other patches are created by subdivision of these nodes.
  * @param base   center position of the quadtree for this cube side
  * @param dx,dy  vectors for 2-dimensional surface orientation
  */
  void initAsCubeSide(const CIntVector base,
                      const CIntVector dx, const CIntVector dy);
  /**
  * Core routine of the IU-LOD-algorithm. Here we subdivide the 
  * quadtree structure downto LODmin. The subdivision criteria is 
  * the L1-distance of each patch center to the camera position.
  * This method is recursive and calls createChild and 
  * generateStaticObjects to create and initialize new patches.
  * @param pCamera      Camera position in double space. This is needed 
                        because we also compute if a patch is visible or not.
  * @param pos          Camera position in integer space
  * @param LODmin       lower bound for subdivision of geometry
  */
  void update(C3DCamera *pCamera, const CIntVector pos, int LODmin);

  /**
  * Create all geometry data for the child[k]-patch.
  * @param k     create data for child[k]
  */
  void createChild(const int k);

  /**
  * Here we place static objects in the child[k]-patch.
  * @param k     create data for child[k]
  */
  void generateStaticObjects(const int k);

  /**
  * This method draws the whole terrain. Calls all child nodes recursivly.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void draw(C3DCamera *pCamera);

  /**
  * This method draws the water surface. Calls all child nodes recursivly.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawWater(C3DCamera *pCamera);

  /**
  * This methos draws the static objects. Calls all child nodes recursivly.
  * @param pCamera  camera that defines our viewport in the virtual world
  */
  void drawStaticObjects(C3DCamera *pCamera);

}; // end of JWQuadtree

#endif




