/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWRing.cpp                                      *
 *                                                          *
 * Content: Implementation of a fractal planet ring with all*
 *          necessary attributes                            *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#include <qgl.h>
#include "glext.h"
#include <cmath>
#include "JWRing.h"
#include "JWQuadtree.h"
#include "JWGameEngine.h"
#include "limits.h"
#include "JWStaticObject.h"
#include "JWNoise.h"

// begin JWRing


/**
 * This method converts an arbitrary (x,y,z)-vector into a virtual ring position (CIntVector)
 * We need the int^3-vector for LOD management and random functions!
 * @param x		x position
 * @param y		y position
 * @param z		z position
 * @param c_s	index of the responsible segment (0-127) for the position (x,y,z)
 * @param dist   distance from base geometry surface in meters
 * @return		Virtual ring position in range [-cube_size/2, cube_size/2]^3 and c_s
 */
CIntVector JWRing::CoordinatesToSurfaceVector(const double x, const double y, const double z, int *c_s, double *dist)
{
	double alpha, beta, htmp;
	CoordinatesToPolar(x, y, z, &alpha, &beta, &htmp);

	int xt,yt,zt, segm;
	int SEG_HALF = cube_size/2;
	alpha = num_segments*(alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, num_segments]
	segm = (int)alpha;
	*c_s = segm % num_segments; // to avoid overflow
	xt = (int)((alpha - segm) * cube_size - SEG_HALF);	// xt is now in range [-cube_size/2, cube_size/2] of segment "segm"
	yt = (int)(Clamp(0,1,(sin(beta)*htmp - r_min) / (r_max - r_min)) * cube_size - SEG_HALF); // yt is now in range [-cube_size/2, cube_size/2]
	// within a 2D ring we need no additional coordinate!
	zt = 0;

	double dx, dy;
	// horizontal distance
	dx = sin(beta)*htmp;
	if (dx < r_min)
		dx = r_min - dx;
	else if (dx > r_max)
		dx = dx - r_max;
	else
		dx = 0;
	// vertical distance
	dy = cos(beta)*htmp;

	*dist = sqrt(dx*dx+dy*dy);

	return CIntVector(xt,yt,zt);
}

CDoubleVector JWRing::SurfaceVectorToCoordinates(CIntVector pos, int c_s, double h)
{
	double alpha, beta, htmp;
	int SEG_HALF = cube_size/2;
	alpha = (pos.x + SEG_HALF) / (double)cube_size + c_s; // bring back to [0, num_segments]
	alpha = 2*PI*(alpha / num_segments) - PI;	  // bring back to [-PI, PI]
	beta = PI/2;
	htmp = ((pos.y + SEG_HALF) / (double)cube_size)*(r_max - r_min) + r_min;

	double x,y,z;
	PolarToCoordinates(&x,&y,&z,alpha, beta, htmp);
	return CDoubleVector(x, y+h, z);
}

CDoubleVector JWRing::SurfaceVectorNormal(CIntVector pos, int c_s)
{
	double alpha;
	int SEG_HALF = cube_size/2;
	alpha = (pos.x + SEG_HALF) / (double)cube_size + c_s; // bring back to [0, num_segments]
	alpha = 2*PI*(alpha / num_segments) - PI;	  // bring back to [-PI, PI]

	return CDoubleVector(sin(alpha),0,cos(alpha));
}


void JWRing::init(QDomElement node)
{
	std::string value;	// temporary container for attribute values

	// set the name of planet
	if (node.hasAttribute("NAME"))	// really needed attribute!
	{
		value = ( node.attribute("NAME") );
		this->setName(value.c_str());
	}
	
	// set the day duration of ring (direct write)
	if (node.hasAttribute("DAY_DURATION"))
		day_length = ( node.attribute("DAY_DURATION") ).toDouble();
	// get the radius of planet
	double prmin, prmax, pangle, phrange;
	if (node.hasAttribute("RMIN"))
		prmin = ( node.attribute("RMIN") ).toDouble();
	if (node.hasAttribute("RMAX"))
		prmax = ( node.attribute("RMAX") ).toDouble();
	if (node.hasAttribute("HEIGHT"))
		phrange = ( node.attribute("HEIGHT") ).toDouble();
	if (node.hasAttribute("AXIS_ANGLE"))
		pangle = ( node.attribute("AXIS_ANGLE") ).toDouble();
	else
		pangle = 0;

	// now we can init the geometry!
	this->initRing(prmin, prmax, phrange, pangle);
	this->SetPosition(CDoubleVector(0, 0, 0));	// default position to prevent chaos

	// get the detail texture parameters
	char *pringtex = 0;
	char *pfinetex = 0;
	if (node.hasAttribute("RINGTEX"))	// really needed attribute!
	{
		pringtex = new char[80];
		sprintf(pringtex, "%s", (const char *)( node.attribute("RINGTEX") ));
	}
	if (node.hasAttribute("FINETEX"))	// really needed attribute!
	{
		pfinetex = new char[80];
		sprintf(pfinetex, "%s", (const char *) ( node.attribute("FINETEX") ));
	}
	// now init the detail textures
	this->initTextures(pringtex, pfinetex);
	// create the segments!
	this->initGeometry();

	// before we complete generation we loop through the subnodes
	QDomNode n;
	JWStaticObject *obj;

	for (n = node.firstChild(); !n.isNull(); n = n.nextSibling())
	{
		// if we have a element node
		if (n.isElement())
		{
			// we have an asteroid object, for now reuse the plant class (not so famous)
			if (n.nodeName() == "ASTEROID")
			{
				obj = new JWAsteroid();
				obj->init(n.toElement());
				this->objects.push_back(obj);
			}
		}
	}

	// set the center of planet if it has one i.e. the center of moon is Earth! (call by "name")
	if (node.hasAttribute("CENTER"))	// really needed attribute!
	{
		value = node.attribute("CENTER");
		this->setCenter(value.c_str());
	}
}

	/**
	* Constructor of our ring. We initialize basic geometry and generate index lists for triangulation.
	* @param rmin			inner radius of the ring in meters!
	* @param rmax			outer radius of the ring in meters!
	* @param phrange        maximum height variance of ring
	* @param axis_angle		angle agaist planet axis
	*/
void JWRing::initRing(double rmin, double rmax, double phrange, double paxis_angle) {
	int LODmax;

	r_min = rmin;
	r_max = rmax;
	axis_angle = PI*paxis_angle/180;
	boundingRadius = 1.05*r_max;

	double rmed = (rmin + rmax)/2;
	num_segments = (int)Clamp(4,128,2.0*PI*rmed/(rmax-rmin));
	LODmax = log2((unsigned int)Clamp(1,0x04000000,2.0*PI*rmed/num_segments));
	cube_size = (1 << LODmax);	// resolution of each cube side
	element_size = (double)((2.0*PI*rmed)/(num_segments*cube_size)); // smallest voxel-size of ring (if possible ca. 1m)

	H_RANGE = phrange;		// asteroids field can be H_RANGE meters high
	H_LOG = log2((unsigned int)H_RANGE)+1+log2((unsigned int)LODObject::grid_size);	// every H_RANGE meters can have a height difference of H_RANGE (mountanious terrain!)
}

// constructor
JWRing::JWRing()
{
	numTriangles=0;	// total amount of triangles/frame
	fineLOD = false;	// the draw routines automatically set this property if needed!

	colorModify = new float[4*LODObject::gs*LODObject::gs];

	// init the random generator!
	jwfrac.Init(3, 4321);	// 3 dimensional noise with fixed seed 4321 (to do: make it more variable)
}

// destructor
JWRing::~JWRing()
{
	if (ringTex != 0) { delete ringTex; ringTex = 0; }
	if (structureTex != 0) { delete structureTex; structureTex = 0; }
	if (fadeOut != 0) { delete fadeOut; fadeOut = 0; }
	if (colorModify != 0) { delete colorModify; colorModify = 0; }
}	

/**
* This method inits our planetary textures which can also be fractal generated.
* We use 3 multiplicative layers of multitexturing for the planet's surface that vary the vertex colors
* @param rtex		1st texture layer (high LODs > 15)
* @param stex        2nd texture layer (medium LODs > 7)
*/
void JWRing::initTextures(char *rtex, char *stex)
{
	ringTex = new TEXTURE();
	// if texture loading fails create fractal!
	if (!ringTex->LoadImage(rtex, true))
		ringTex->createFractal(256,32, GL_RGBA);
	ringTex->keepRawData = true;
	ringTex->activate(TEXTURE::T_CUBEMAP);

	// loading texture file success
	// NOTE: HF_size is changing into user defined size from loaded file! (ignore pHF_size here)
	HF_width = ringTex->width;
	HF_height = ringTex->height;
	HF_size = 1 << log2((unsigned int)HF_height);
	// all LODs from LODmax downto HF_LOD have valid heightfield values
	HF_LOD = log2((unsigned int)cube_size)-log2((unsigned int)HF_size)+log2((unsigned int)LODObject::grid_size);

	structureTex = new TEXTURE();
	// if texture loading fails create fractal!
	if(!structureTex->LoadImage(stex, true))
		structureTex->createFractal(256,256, GL_LUMINANCE);
	structureTex->keepRawData = true;
	structureTex->activate(TEXTURE::T_DETAIL);

	fadeOut = new TEXTURE();
	fadeOut->LoadImage("Data/black.bmp", true);
	fadeOut->activate(TEXTURE::T_DETAIL);
}


void JWRing::initGeometry()
{
	int	LODmax = log2((unsigned int)cube_size);
	int i;
	// create num_segments starting quadtrees
	CIntVector start = CIntVector(0, 0, 0);
	dx_s = CIntVector(cube_size, 0, 0);
	dy_s = CIntVector(0, cube_size, 0);

	// init quadtrees
	for (i=0; i < this->num_segments; i++)
	{
		jwq[i] = new JWQuadtree(i, cube_size/2, cube_size/2, LODmax, this);
		jwq[i]->initAsCubeSide(start, dx_s, dy_s);
	}
	
	dx_s /= cube_size;
	dy_s /= cube_size;
}

/**
* Computes the height of the player object over terrain of this planet.
* We search the affected quadtree and delegate the query to it.
* @param player		the player object
* @return			height above terrain in meters
*/
double JWRing::getHeight_above_Surface(C3DBase *obj)
{
	// get Players Position in relation to Planet's coordinates
	C3DBase *cp = C3DBase::GetCommonParent(this, obj);
	CDoubleVector diff = this->GetCameraMatrix(cp) * (obj->GetFullPosition(cp) - this->GetFullPosition(cp));
	int c_s;
	double dist;
	CIntVector pos = CoordinatesToSurfaceVector(diff.x, diff.y, diff.z, &c_s, &dist);
	// returns real distance of player above the ring
	return dist;
}

/**
 * This is one of the core routines. Here we get a height value for every position on our planet
 * We first check the loaded DEMs (1), then use the global HF for high LODs (2) or 3D noise for fine LODs (3)
 * @param point		position of vertex on virtual cube [0,cube_size]^3
 * @param LOD		the level of detail we need for this vertex
 * @param h_old		interpolated height value from higher LODs (good for multifractal functions and coast line sharpness)
 * @return			the current height value at this position and this LOD
 */
double JWRing::get_HeightField_And_RandomOffset(CIntVector point, int LOD, double h_old)
{
	// small LODs return random values
	if (LOD < H_LOG)
	{
		// real 3D noise for small LODs
    double f[4] = {point.x, point.y, point.z, 0};
		// i think i must rework this but for now i have nice coast lines!!
		return (Min(H_RANGE, (1 << LOD))*jwfrac.Noise(f));

	}
	else
		return 0;
}


/**
* The "heart" of my algorithm. We need to recompute the adaptive LOD structure every frame. (TO DO: can we dynamically adjust this?)
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWRing::updateGeometry(C3DCamera *pCamera)
{	
	// get Players Position in relation to Planet's coordinates
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	CDoubleVector t = this->GetCameraMatrix(cp) * (pCamera->GetFullPosition(cp) - this->GetFullPosition(cp));

	CIntVector newpos;
	int c_s;
	double htmp;
	// We compute the new cube position that the camera has on the planets surface.
	newpos = CoordinatesToSurfaceVector(t.x, t.y, t.z, &c_s, &htmp);
	// Get the difference between these 2 vectors to speed up LOD updating!
	// Reason: when moving very fast we have to create too much new quadtrees!
	int dist = ABS(newpos.x - lastpos.x) + ABS(newpos.y - lastpos.y) + ABS(newpos.z - lastpos.z);
	// we just compared the movement for LOD purposes, now use the new position
	lastpos = newpos;
	LODalt = (int)computeLOD_min(pCamera);

	// now compute how much we'll have to update and adapt the quality!
	LOD_diff = log2((unsigned int)dist) - LODalt;	// how many scales of LOD does the distance vary?
	if (LOD_diff > 0)
		LODalt = CLAMP(LODalt, log2((unsigned int)cube_size), LODalt + LOD_diff);

	// update quadtrees
	int i;
	CIntVector segmentpos;
	for (i=0; i<num_segments; i++)
	{
		if (i == c_s)
			jwq[i]->update(pCamera, lastpos, LODalt);
		else if (i == ((c_s+1)%num_segments))
		{
			segmentpos = lastpos - CIntVector(cube_size,0,0);
			jwq[i]->update(pCamera, segmentpos, LODalt);
		}
		else if (i == ((c_s-1)%num_segments))
		{
			segmentpos = lastpos + CIntVector(cube_size,0,0);
			jwq[i]->update(pCamera, segmentpos, LODalt);
		}
		else
		{
			jwq[i]->update(pCamera, lastpos, log2((unsigned int)cube_size));
		}
	}
}

//extern int LOD_quality; // LOD at which we switch to midpoint displacement instead of shpere interpolation + redefine Vertex positons!

/**
* Implementation of the virtual draw method of all C3DBase objects. We draw a whole planet by calling special submethods!
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWRing::draw(C3DCamera *camera)
{
	numTriangles = 0;

	// high LODs (incl. water) should have higlights!
	GLfloat no_mat[] = { 0, 0, 0, 1.0 };
	GLfloat diff_mat[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat no_shininess[] = { 0.0 };
	glMaterialfv(GL_FRONT, GL_SPECULAR, no_mat);
	glMaterialfv(GL_FRONT, GL_EMISSION, no_mat);
	glMaterialfv(GL_FRONT, GL_SHININESS, no_shininess);
	// reset the ambient and diffuse properties (they could be corrupted!)
	glMaterialfv(GL_FRONT, GL_AMBIENT, diff_mat);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_mat);
	

	drawRing(camera);

	if (fineLOD) // draw objects through ring
	{
		glClear(GL_DEPTH_BUFFER_BIT);
		if (!objects.empty())
			this->drawStaticObjects(camera);
	}
}

#if defined(_WIN32)
extern PFNGLACTIVETEXTUREARBPROC       glActiveTextureARB;
extern PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;
#endif
	
/**
* Here we draw the ring of the planet
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWRing::drawRing(C3DCamera *pCamera)
{
	// Multiply the current view matrix by the model matrix and draw the planet
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glPushMatrix();
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	glDisable( GL_CULL_FACE );
	glDisable(GL_ALPHA_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,1.0f);

	//check visibility
	if (pCamera->isVisible(this))
	{
		int i;
		for (i=0; i<num_segments; i++)
		{
			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, ringTex->texID);	// select Texture for this cube side			
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, structureTex->texID);	// select detail Texture
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, structureTex->texID);	// select detail Texture
			}

			if ((jwq[i] != 0) && (jwq[i]->isActive)) jwq[i]->draw(pCamera);

			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			}
		}
	}

	if (jwengine::jwgame->BFculling)
		glEnable( GL_CULL_FACE );

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);

	glPopMatrix();
}


/**
* Here we draw the static objects that are placed on the terrain.
* @param pCamera	camera that defines our viewport in the virtual world ;-)
*/
void JWRing::drawStaticObjects(C3DCamera *pCamera)
{
	// Multiply the current view matrix by the model matrix and draw the planet
	CDoubleMatrix m = GetScaledModelMatrix(pCamera);
	C3DBase *cp = C3DBase::GetCommonParent(this, pCamera);
	glPushMatrix();
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(pCamera->GetCameraMatrix(cp)*m);

	glDisable( GL_CULL_FACE );	// Most objects don't use backface culling to have more flexibility (caves etc!)
	//checkVisibility;
	if (pCamera->isVisible(this))
	{
		int i;
		for (i=0; i<num_segments; i++)
		{
			if (jwengine::jwgame->tex_mode)
			{
				// TEXTURE-UNIT #2:
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
				// TEXTURE-UNIT #1:
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
				// TEXTURE-UNIT #0		
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, 0);		
			}

			if ((jwq[i] != 0) && (jwq[i]->isActive)) jwq[i]->drawStaticObjects(pCamera);

			if (jwengine::jwgame->tex_mode)
			{
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			}
		}
	}

	if (jwengine::jwgame->BFculling)
		glEnable( GL_CULL_FACE );

	glPopMatrix();
}


/**
* This method creates a color value for each terrain vertex according to the planet type (see style def.) 
* For now only height of terrain is regarded to create a color.
* @param hf		the height value of a vertex for that we compute the terrain color
* @param pos	vertex position
* @return		quadrupel (r,g,b,a) as color
*/
CFloatVector4 JWRing::computeColor(const double hf, CIntVector pos)
{
	float h = (float)(hf/(2*H_RANGE)); // height in [-0.5,0.5]
	double alpha, beta;
	int hw, hh;
	TEXTURE *coltex = 0; // this points to the color data (global map or DEM texture!)

	// no suitable DEM texture so use colorMap if present
	if (this->ringTex)
	{
		coltex = this->ringTex;
		hw = coltex->width;
		hh = coltex->height;
		alpha = hw*((pos.x + cube_size/2) / (double)cube_size); // convert from [-PI,PI] to [0, hw]
		beta = hh*((pos.y + cube_size/2) / (double)cube_size); // convert to [0,hh]
	}

	if (coltex)
	{
		// bilinear interpolation between the colorMap points
		float col[4];
		
		int bypp = coltex->bpp/8;

		int hx,hy;
		double rx, ry;
		hx = (int)alpha;
		hy = (int)beta;
		rx = Cubic(alpha-hx);
		ry = Cubic(beta-hy);
		for (int i=0; i<4; i++)
		{
			h = (1-rx)*(1-ry)*coltex->imageData[bypp*((hy % hh)*hw + (hx % hw)) + i];
			h += rx*(1-ry)*coltex->imageData[bypp*((hy % hh)*hw + ((hx+1) % hw)) + i];
			h += (1-rx)*ry*coltex->imageData[bypp*(((hy+1) % hh)*hw + (hx % hw)) + i];
			h += rx*ry*coltex->imageData[bypp*(((hy+1) % hh)*hw + ((hx+1) % hw)) + i];
			col[i] = h / 255.0f;
		}
		return CFloatVector4(col);
	}
	return CFloatVector4(1,1,1,1);
}

// end JWRing



