/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWRing.h                                        *
 *                                                          *
 * Content: Header file of a fractal planet ring with all   *
 *          necessary attributes                            *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */

#ifndef JWRING_H
#define JWRING_H

#include "Matrix.h"
#include "JWObject.h"
#include "Texture.h"
#include "qtxml.h"

class JWQuadtree;

class JWRing : public LODObject {
public:
	int num_segments;		// the real number of used segments to build the ring
	JWQuadtree *jwq[128];	// maximal number of segments
	CIntVector dx_s;		// Vectors needed for orientation on surface!
	CIntVector dy_s;
	
	TEXTURE *ringTex;		// Texture for each ring segment
	TEXTURE *structureTex;	// detail texture for asteroid field
	TEXTURE *fadeOut;		// when superfine we use fadeout!

	double r_min, r_max;
	float *colorModify;		// array used to smooth texture to vertexcolor border

	/**
	* This method converts an arbitrary (x,y,z)-vector into a virtual ring position (CIntVector)
	* We need the int^3-vector for LOD management and random functions!
	* @param x		x position
	* @param y		y position
	* @param z		z position
	* @param c_s	index of the responsible segment (0-127) for the position (x,y,z)
	* @return		Virtual ring position in range [-cube_size/2, cube_size/2]^3 and c_s
	*/
	virtual CIntVector CoordinatesToSurfaceVector(const double x, const double y, const double z, int *c_s, double *dist);
	virtual CDoubleVector SurfaceVectorToCoordinates(CIntVector pos, int c_s, double h);
	virtual CDoubleVector SurfaceVectorNormal(CIntVector pos, int c_s);
	
	virtual void init(QDomElement node);

	/**
	* Constructor of our ring. We initialize basic geometry and generate index lists for triangulation.
	* @param rmin			inner radius of the ring in meters!
	* @param rmax			outer radius of the ring in meters!
	* @param phrange        maximum height variance of ring
	* @param axis_angle		angle agaist planet axis
	*/
	void initRing(double rmin, double rmax, double phrange, double paxis_angle = 0);
	// constructor
	JWRing();
	// destructor
	virtual ~JWRing();

	/**
	* This method inits our planetary textures which can also be fractal generated.
	* We use 3 multiplicative layers of multitexturing for the planet's surface that vary the vertex colors
	* @param rtex		1st texture layer (high LODs > 15)
	* @param stex        2nd texture layer (medium LODs > 7)
	*/
	void initTextures(char *rtex=0, char *stex=0);
	
	void initGeometry();
	
	/**
	* Computes the height of the player object over terrain of this planet.
	* We search the affected quadtree and delegate the query to it.
	* @param player		the player object
	* @return			height above terrain in meters
	*/
	virtual double getHeight_above_Surface(C3DBase *obj);

	double get_HeightField_And_RandomOffset(CIntVector point, int LOD, double h_old);

	/**
	* The "heart" of my algorithm. We need to recompute the adaptive LOD structure every frame. (TO DO: can we dynamically adjust this?)
	* @param pCamera	camera that defines our viewport in the virtual world ;-)
	*/
	void updateGeometry(C3DCamera *pCamera);

	/**
	* Implementation of the virtual draw method of all C3DBase objects. We draw a whole planet by calling special submethods!
	* @param pCamera	camera that defines our viewport in the virtual world ;-)
	*/
	void draw(C3DCamera *camera);

	/**
	* Here we draw the rings of the planet
	* @param pCamera	camera that defines our viewport in the virtual world ;-)
	*/
	void drawRing(C3DCamera *pCamera);

	/**
	* Here we draw the static objects that are placed on the terrain.
	* @param pCamera	camera that defines our viewport in the virtual world ;-)
	*/
	void drawStaticObjects(C3DCamera *pCamera);
	
	/**
	* This method creates a color value for each terrain vertex according to the planet type (see style def.) 
	* For now only height of terrain is regarded to create a color.
	* @param hf		the height value of a vertex for that we compute the terrain color
	* @param pos	vertex position
	* @return		quadrupel (r,g,b,a) as color
	*/
	CFloatVector4 computeColor(const double hf, CIntVector pos);	
};

#endif



