/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2003 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWSolarSystem.cpp                               *
 *                                                          *
 * Content: Implementation of a solar system class          *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2003/02/03                                      *
 ************************************************************
 */

//#if defined(_CC_MSVC_)
//#pragma warning(disable:4786)
//#endif

#include <vector>
#include <algorithm>
#include "JWSolarSystem.h"
#include "JWPlanet.h"
#include "JWRing.h"
#include "JWGameEngine.h"

void JWSolarSystem::init(QDomElement start)
{
	if (start.nodeName() == "SOLARSYSTEM")
		qWarning("Initializing solar system");
	else
		return;

	this->setName( start.attribute("NAME", "unknown system") );
	
	if(!start.isNull())
	{
		// now loop through the planets list
		QDomNode n;
		LODObject *jwp;
		for (n = start.firstChild(); !n.isNull(); n = n.nextSibling())
		{
			// if we have a planet element node
			if (n.isElement())
			{
				if (n.nodeName() == "PLANET")
				{
					// initialize JWPlanet object
					jwp = new JWPlanet();
					jwp->init(n.toElement());
					// add object to solar system
					components.push_back(jwp);
				}
				else if (n.nodeName() == "STAR")
				{
					// initialize JWStar object
					jwp = new JWStar();
					jwp->init(n.toElement());
					// add object to solar system
					components.push_back(jwp);
				}
				else if (n.nodeName() == "RING")
				{
					// we have a planetary asteroids ring
					jwp = new JWRing();
					jwp->init(n.toElement());
					// add object to solar system
					components.push_back(jwp);
				}				
			}
		}
	}
	else
	{
		// if we have no planet definitions create one dummy system!
		JWStar *sun;
		sun = new JWStar();
		sun->init(start);
		components.push_back(sun);
		
		JWPlanet *moon;
		moon = new JWPlanet();
		moon->init(start);
		components.push_back(moon);
	}

	// now loop through our solar system objects and determine the boundingRadius of this system
	boundingRadius = 1.0;

	std::vector<C3DBase *>::iterator theIterator;
	LODObject *otmp;
	
	for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
	{
		// first check if this is a object we know how to handle
		otmp = dynamic_cast<LODObject *> (*theIterator);
		if (otmp)
		{
			boundingRadius = Max( (otmp->orbit + otmp->boundingRadius), boundingRadius);
		}
	}
	// enlarge a bit to really contain all objects
	boundingRadius *= 1.1;
};

/*
* A solar system contains a bunch of planets and other big space objects.
* These are all expected to be LODObjects, so that we adjust them in a specific manner
*/
void JWSolarSystem::updateGeometry(C3DCamera *pCamera)
{ 
	double year_cycle, day_cycle;

	std::vector<C3DBase *>::iterator theIterator;
	LODObject *otmp;
	std::string value;
	
	// Position all planets according to their rotation times (year/day)
	// now reclac the elliptical curve positions of planets ...
	for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
	{
		// first check if this is a object we know how to handle
		otmp = dynamic_cast<LODObject *> (*theIterator);
		if (otmp)
		{
			otmp->ResetRotation();
			
			otmp->SetPosition(CDoubleVector(0,0,0));
			
      // this object (planet + ring) has an angle against x-axis!
			if (otmp->axis_angle != 0)
				otmp->Rotate(otmp->GetRightAxis(), otmp->axis_angle);

      // this planet has a year cycle
      if (otmp->year_length != 0)
			{
				// compute year cycle in [0,2*PI] depending on year_length
				year_cycle = 2*PI*(Mod(jwengine::jwgame->virtualTime, Abs(otmp->year_length))) / otmp->year_length;
				otmp->SetPosition(CDoubleVector(otmp->orbit*sin(year_cycle), 0.0, otmp->orbit*cos(year_cycle)));
				// we must rotate to keep tangential orientation correct
        if (otmp->day_length == 0)
				  otmp->Rotate(otmp->GetUpAxis(),year_cycle); // regular rotation when moving on an elliptical curve
			}

      // this planet has a day cycle
			if (otmp->day_length != 0)
			{
				// compute day cycle in [0,2*PI] depending on day_length
				day_cycle = 2*PI*(Mod(Abs(jwengine::jwgame->virtualTime*24), Abs(otmp->day_length))) / otmp->day_length;
				otmp->Rotate(otmp->GetUpAxis(),day_cycle); // regular self-rotation
			}
		}
	}
	
	// now correct the rotation of moons about their centers!
	// we only can do this in a second step since all "normal" planets must be at right position!
	for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
	{
		otmp = dynamic_cast<LODObject *> (*theIterator);
		if (otmp)
		{
			value = otmp->getCenter();
			
			// this planet has a center (is a moon)
			if (value != "none")
			{
				// add position of center to this moon
				otmp->SetPosition(this->getObjectByName(value.c_str())->GetFullPosition() + otmp->GetPosition());
			}
		}
	}
	
 	// VERY IMPORTANT: due to camera movement we have to recalculate the clipping frustum
  jwengine::jwgame->initCamera();

	// most important step: update geometry (LOD structure) of the planets!
	if (jwengine::jwgame->update_mode)
	{
		for (theIterator = components.begin(); theIterator != components.end(); theIterator++)
		{
			(*theIterator)->updateGeometry(pCamera);
		}
	}
};

extern int LOD_quality;

/**
 * This method draws our whole solar sytem (for now some planets)
 * @param camera	The camera we use
 */
void JWSolarSystem::draw(C3DCamera *camera)
{
	numTriangles = 0;

	double dist = (GetFullPosition()-camera->GetFullPosition()).Magnitude();

	//check visibility
	if (!camera->isVisible(this))
		return;

	// draw a simple star dot for the whole system if we are far away!
	if (dist > this->boundingRadius)
	{
		// Multiply the current view matrix by the model matrix and draw the planet
		CDoubleMatrix m = GetScaledModelMatrix(camera);
		C3DBase *cp = C3DBase::GetCommonParent(this, camera);
		glPushMatrix();
		// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
		glLoadMatrixd(camera->GetCameraMatrix(cp)*m);

		glPointSize(2.0);
		glColor3f(1,1,1);
		glBegin(GL_POINTS);
		glVertex3f(0,0,0);
		glEnd();
		numTriangles = 1;

		glPopMatrix();		
		return;
	}	

	// get the nearest planet to the camera in this solar system
	// TO DO: change this so that we can handle all types of objects!!!
	LODObject *nearestObject = dynamic_cast <LODObject *> (getNearestObject(camera));
	JWPlanet *nearestPlanet = dynamic_cast <JWPlanet *> (nearestObject);

	// compute height above terrain and altitude (they are not the same!)
	float alt, hNN;

	alt = (float)nearestObject->getHeight_above_Surface(camera);
	
	if (nearestPlanet)
		hNN = (float)((nearestPlanet->GetFullPosition()-camera->GetFullPosition()).Magnitude() - nearestPlanet->radius);
	else
		hNN = alt;
	
	// First draw all other planets that are far away from camera
	// Sort the vector with increasing distance
	std::sort(components.begin(), components.end(), C3DBase::distanceSort);

	std::vector<C3DBase *>::iterator theIterator;
	// lookup contents of the Vector planets
	for (theIterator = components.end()-1; theIterator != components.begin(); theIterator--)
	{
		(*theIterator)->draw(camera);
		
		numTriangles += (*theIterator)->numTriangles;
	}
	
	// clear depth buffer to prevent overlay errors of terrain with planets far away!
	if (nearestPlanet)
		glClear(GL_DEPTH_BUFFER_BIT);
	
	nearestObject->draw(camera);

	numTriangles += nearestObject->numTriangles;

	glFlush(); // everything before must be completed!!!

	// only switch to near clipping plane here if we are near to the ground!	
	if ((alt < 2000) || (nearestPlanet && nearestPlanet->hasWater && (hNN < 2000)) || (nearestPlanet && nearestPlanet->hasClouds && (hNN < 12000)))
	{
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
    jwengine::jwgame->activeRange = JWGameEngine::V_NEAR;
    jwengine::jwgame->initViewFrustum();
		glMatrixMode(GL_MODELVIEW);
		glClear(GL_DEPTH_BUFFER_BIT); // clear to show fine LOD objects
		glEnable(GL_DEPTH_TEST);
		// we are "within" the landcape so render with near clipping planes
		nearestObject->fineLOD = true;
		// test if there is any fine LOD
		if (nearestObject->LODalt <= LOD_quality)
			nearestObject->draw(camera);
		nearestObject->fineLOD = false;
	}
	else
	{
		glClear(GL_DEPTH_BUFFER_BIT); // clear to show fine LOD objects
		// we are far away, there is no need for high precision!
		nearestObject->fineLOD = true;
		// test if there is any fine LOD
		if (nearestObject->LODalt <= LOD_quality)
			nearestObject->draw(camera);
		nearestObject->fineLOD = false;
		// now activate for other objects like ships etc.
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
    jwengine::jwgame->activeRange = JWGameEngine::V_NEAR;
    jwengine::jwgame->initViewFrustum();
		glMatrixMode(GL_MODELVIEW);
		glClear(GL_DEPTH_BUFFER_BIT); // clear to show fine LOD objects
		glEnable(GL_DEPTH_TEST);
	}
	
	numTriangles += nearestObject->numTriangles;
}




