
#include "JWStaticObject.h"
#include "qtxml.h"

/**
 * This creates a static object. They are randomly placed on terrain controled by specific parameters.
 * @param pwidth		width of the object
 * @param pheight		height of the object
 * @param tex			basis texture for that object
 * @param pdensity		how dense will this object be placed (0.0 - 1.0)
 * @param pminh			minimum height to place the object (means sth. like climatic zone e.g at coast 0-200m, mountains 6000-10000m)
 * @param pmaxh			maximum height to place the object
 */
void JWStaticObject::init(float pwidth, float pheight, char *tex, float pdensity, float pminh, float pmaxh)
{
	width = pwidth;
	height = pheight;
  boundingRadius = sqrt(2*width*width + height*height); // |3D Diagonale| because of base at bottom!
	// We consider only width because it is of interest how
	// the ground we place the object on is covered (tiling technique)!
	baseLOD = log2((unsigned int)Max(width,height)) + 1;
	visibleLOD = (int)Min(10, log2((unsigned int)Max(width,height)) + 3); // keep LOD under LOD_quality limit to have good coordinates!

	baseTex.LoadImage(tex, true);
	baseTex.activate(TEXTURE::T_TERRAIN);

	density = pdensity;
	minHeight = pminh;
	maxHeight = pmaxh;

  qWarning("boundingRadius of %s is: %f", getName(), boundingRadius);
}

void JWStaticObject::init(QDomElement node)
{
	char* stopstring;	// temporary variable for type conversion
	std::string value;	// temporary container for attribute values

	// set the name of object
	if (node.hasAttribute("NAME"))	// really needed attribute!
	{
		value = ( node.attribute("NAME") );
		setName(value.c_str());
	}

	// get the width of object
	double pwidth;
	if (node.hasAttribute("WIDTH"))
	{
		value = ( node.attribute("WIDTH") );
		pwidth = strtod( value.c_str(), &stopstring );
	}
	// get the height
	double pheight;
	if (node.hasAttribute("HEIGHT"))
	{
		value = ( node.attribute("HEIGHT") );
		pheight = strtod( value.c_str(), &stopstring );
	}
	// get the texture
	char *tex = 0;
	if (node.hasAttribute("TEXTURE"))	// really needed attribute!
	{
		tex = new char[80]; 
		sprintf(tex, "%s", (const char *)( node.attribute("TEXTURE") ));
	}
	double pdensity;
	if (node.hasAttribute("DENSITY"))
	{
		value = ( node.attribute("DENSITY") );
		pdensity = strtod( value.c_str(), &stopstring );
	}
	double pminh;
	if (node.hasAttribute("MIN_H"))
	{
		value = ( node.attribute("MIN_H") );
		pminh = strtod( value.c_str(), &stopstring );
	}
	double pmaxh;
	if (node.hasAttribute("MAX_H"))
	{
		value = ( node.attribute("MAX_H") );
		pmaxh = strtod( value.c_str(), &stopstring );
	}

	// now we can init the object!
	this->init(pwidth, pheight, tex, pdensity, pminh, pmaxh);
}


/**
 * draws a plant object which is a 3D decal sprite
 * @param dlod		LOD for that object
 */
int JWPlant::draw(int distance)
{
	int dlod = log2((unsigned int)distance);

	if (strstr(this->getName(), "nocolor") == NULL)
		glColor4f(1,1,1,1);

	if (List != 0)
		// take the precomputed display list
		glCallList(List);
	else
	{
		List = glGenLists(1); 
		glNewList(List,GL_COMPILE);		// Start With The List
		
		CDoubleVector up = CDoubleVector(0,1,0);
		CDoubleVector unit1 = CDoubleVector(1,0,0);
		CDoubleVector unit2 = CDoubleVector(0,0,1);
		// scaling is variable for different trees!
		up *= height;
		unit1 *= width/2;
		unit2 *= width/2;
		
		glBindTexture(GL_TEXTURE_2D, baseTex.texID);    
		
		glAlphaFunc(GL_GREATER,0.8f); // Set Alpha Testing  
		glEnable(GL_ALPHA_TEST);      // Enable Alpha Testing
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		CDoubleVector base;
		glNormal3d(0,1,0);

		glBegin(GL_QUADS);
		glTexCoord2f(0.0, 1.0);
		base =  - unit1 + up;
		glVertex3dv(base);
		glTexCoord2f(1.0, 1.0);
		base =   unit1 + up;
		glVertex3dv(base);
		glTexCoord2f(1.0, 0.0);
		base =   unit1;
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1;
		glVertex3dv(base);
		
		glTexCoord2f(0.0, 1.0);
		base =  - unit2 + up;
		glVertex3dv(base);
		glTexCoord2f(1.0, 1.0);
		base =   unit2 + up;
		glVertex3dv(base);
		glTexCoord2f(1.0, 0.0);
		base =   unit2;
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit2;
		glVertex3dv(base);
		glEnd(); 
		glEndList();
	}
	return 4;
}

/**
 * draws a house (skyscraper) as a textured 3D box
 * @param dlod		LOD for that object
 */
int JWHouse::draw(int distance)
{
	int triag = 0;
	int dlod = log2((unsigned int)distance);

	float fadeval = 1.0f / Max(1, (log(distance+1)/log(2)) - baseLOD-3); // parameter to fade in details!
	//glColor3f(fadeval, fadeval, fadeval);
	//glColor4f(1, 1, 1, fadeval);
	glColor3f(1, 1, 1);

  // we are inside model of building!
  if (dlod <= this->baseLOD)
    glEnable(GL_ALPHA_TEST);   // Enable Alpha Testing!
  else
    glDisable(GL_ALPHA_TEST);   // Disable Alpha Testing to draw blended geometry!
	glAlphaFunc(GL_GREATER,0.48f); // Set Alpha Testing  

	if (List != 0)
  {
		// take the precomputed display list
		glCallList(List);
  }
	else
	{
		List = glGenLists(1); 
		glNewList(List,GL_COMPILE);		// Start With The List
		
		CDoubleVector up = CDoubleVector(0,1,0);
		CDoubleVector unit1 = CDoubleVector(1,0,0);
		CDoubleVector unit2 = CDoubleVector(0,0,1);
		CDoubleVector nv = up;
		CDoubleVector base;
		
		// scaling is variable for different skyscapers!
		up *= height;
		unit1 *= width/2;
		unit2 *= width/2;
		nv *= width/2;	// lift side normals so that they are light on day and dark in the night!
		
		float sX, sY;
		sX = width/5;
		sY = height/5;

		glBindTexture(GL_TEXTURE_2D, baseTex.texID);    
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glDisable(GL_LIGHTING);

		glBegin(GL_QUADS);
		// solid roof (draw first to avoid blending errors)
		glTexCoord2f(0.0f, 0.1f);
		base =  - unit2 + up + unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.1f);
		base =  unit2 + up + unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0f);
		base =  unit2 + up - unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.0f, 0.0f);
		base =  - unit2 + up - unit1;
		//glNormal3dv(up);
		glVertex3dv(base);

		// solid floor (draw first to avoid blending errors)
		glTexCoord2f(0.0f, 0.1f);
		base =  - unit2 + unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.1f);
		base =  unit2 + unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0f);
		base =  unit2 - unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.0f, 0.0f);
		base =  - unit2 - unit1;
		//glNormal3dv(up);
		glVertex3dv(base);
		glEnd();

		// now draw semisolid backface geometry!
		glFrontFace(GL_CW);			// first draw backfaces! NOTE: Thus we must invert the normals!
		glEnable( GL_CULL_FACE );	// Most objects don't use backface culling but we need it!

		glBegin(GL_QUADS);
		// side 1
		glTexCoord2f(sX, sY);
		base =  unit1 + up - unit2;
		//glNormal3dv(unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, sY);
		base =  - unit1 + up - unit2;
		//glNormal3dv(unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 - unit2;
		//glNormal3dv(unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit1 - unit2;
		//glNormal3dv(unit2-nv);
		glVertex3dv(base);
		// side 3
		glTexCoord2f(0.0, sY);
		base =  - unit1 + up + unit2;
		//glNormal3dv(-unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, sY);
		base =  unit1 + up + unit2;
		//glNormal3dv(-unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit1 + unit2;
		//glNormal3dv(-unit2-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 + unit2;
		//glNormal3dv(-unit2-nv);
		glVertex3dv(base);
		// side 2
		glTexCoord2f(0.0, sY);
		base =  - unit2 + up - unit1;
		//glNormal3dv(unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, sY);
		base =   unit2 + up - unit1;
		//glNormal3dv(unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit2 - unit1;
		//glNormal3dv(unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base = - unit2 - unit1;
		//glNormal3dv(unit1-nv);
		glVertex3dv(base);
		// side 4
		glTexCoord2f(sX, sY);
		base =   unit2 + up + unit1;
		//glNormal3dv(-unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, sY);
		base =  - unit2 + up + unit1;
		//glNormal3dv(-unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit2 + unit1;
		//glNormal3dv(-unit1-nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =   unit2 + unit1;
		//glNormal3dv(-unit1-nv);
		glVertex3dv(base);
		glEnd(); 
		glFrontFace(GL_CCW);	// reset to standard		
		glEndList();
	}

  // we see a fine detail model of building!
	if (dlod <= this->baseLOD + 2)
	{
		// draw all etage ceilings/floors
		float fadeEtage = 1.0f / Max(1, (log(distance+1)/log(2)) - this->baseLOD + 1);
		glColor4f(1,1,1, fadeEtage);

		if (ListEtages != 0)
			// take the precomputed display list
			glCallList(ListEtages);
		else
		{
			ListEtages = glGenLists(1); 
			glNewList(ListEtages,GL_COMPILE);		// Start With The List
			
			CDoubleVector etup = CDoubleVector(0,1,0);
			CDoubleVector etunit1 = CDoubleVector(1,0,0);
			CDoubleVector etunit2 = CDoubleVector(0,0,1);
			CDoubleVector etbase;
			// scaling is variable for different skyscapers!
			etunit1 *= width/2;
			etunit2 *= width/2;
			
			glBindTexture(GL_TEXTURE_2D, baseTex.texID);    
			glDisable(GL_LIGHTING);
			glDisable( GL_CULL_FACE );	// Most objects don't use backface culling but we need it!
			
			// blended floors every 5m
			glBegin(GL_QUADS);
			for (int hscale = 5; hscale < height; hscale += 5)
			{
				glTexCoord2f(0.0f, 0.1f);
				etbase =  - etunit2 + etup*hscale + etunit1;
				//glNormal3dv(etup);
				glVertex3dv(etbase);
				glTexCoord2f(0.1f, 0.1f);
				etbase =  etunit2 + etup*hscale + etunit1;
				//glNormal3dv(etup);
				glVertex3dv(etbase);
				glTexCoord2f(0.1f, 0.0f);
				etbase =  etunit2 + etup*hscale - etunit1;
				//glNormal3dv(etup);
				glVertex3dv(etbase);
				glTexCoord2f(0.0f, 0.0f);
				etbase =  - etunit2 + etup*hscale - etunit1;
				//glNormal3dv(etup);
				glVertex3dv(etbase);
			}
			glEnd();
			glEndList();			
		}
		triag += (int)(2*height/5);
	}

	//glColor3f(fadeval, fadeval, fadeval);
	//glColor4f(1, 1, 1, fadeval);
	glColor3f(1, 1, 1);

	if (ListBlending != 0)
		// take the precomputed display list
		glCallList(ListBlending);
	else
	{
		ListBlending = glGenLists(1); 
		glNewList(ListBlending,GL_COMPILE);		// Start With The List
		
		CDoubleVector up = CDoubleVector(0,1,0);
		CDoubleVector unit1 = CDoubleVector(1,0,0);
		CDoubleVector unit2 = CDoubleVector(0,0,1);
		CDoubleVector nv = up;
		CDoubleVector base;
		
		// scaling is variable for different skyscapers!
		up *= height;
		unit1 *= width/2;
		unit2 *= width/2;
		nv *= width/2;	// lift side normals so that they are light on day and dark in the night!
		
		float sX, sY;
		sX = width/5;
		sY = height/5;

		// 2nd blending pass
		// now overdraw with frontfaces!
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_LIGHTING);
		glEnable( GL_CULL_FACE );	// Most objects don't use backface culling but we need it!
		glDepthFunc(GL_LEQUAL);

		glBegin(GL_QUADS);
		// solid roof (draw first to avoid blending errors)
		glTexCoord2f(0.0f, 0.1f);
		base =  - unit2 + up + unit1;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.1f);
		base =  unit2 + up + unit1;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0f);
		base =  unit2 + up - unit1;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.0f, 0.0f);
		base =  - unit2 + up - unit1;
		glNormal3dv(up);
		glVertex3dv(base);

		// solid basement floor + walls
		glTexCoord2f(0.1f, 0.1f);
		base =  unit2 + unit1 - up;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.0f, 0.1f);
		base =  - unit2 + unit1 - up;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.0f, 0.0f);
		base =  - unit2 - unit1 - up;
		glNormal3dv(up);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0f);
		base =  unit2 - unit1 - up;
		glNormal3dv(up);
		glVertex3dv(base);
		// side 1
		glTexCoord2f(0.1f, 0.1f);
		base =  unit1 + up - unit2 - up;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.1f);
		base =  - unit1 + up - unit2 - up;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 - unit2 - up;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0);
		base =  unit1 - unit2 - up;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		// side 3
		glTexCoord2f(0.0, 0.1f);
		base =  - unit1 + up + unit2 - up;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.1f);
		base =  unit1 + up + unit2 - up;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0);

		base =  unit1 + unit2 - up;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 + unit2 - up;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		// side 2
		glTexCoord2f(0.0, 0.1f);
		base =  - unit2 + up - unit1 - up;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.1f);
		base =   unit2 + up - unit1 - up;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0);
		base =  unit2 - unit1 - up;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base = - unit2 - unit1 - up;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		// side 4
		glTexCoord2f(0.1f, 0.1f);
		base =   unit2 + up + unit1 - up;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.1f);
		base =  - unit2 + up + unit1 - up;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit2 + unit1 - up;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.1f, 0.0);
		base =   unit2 + unit1 - up;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);

		// "normal" house walls
		// side 1
		glTexCoord2f(sX, sY);
		base =  unit1 + up - unit2;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, sY);
		base =  - unit1 + up - unit2;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 - unit2;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit1 - unit2;
		glNormal3dv(-unit2+nv);
		glVertex3dv(base);
		// side 3
		glTexCoord2f(0.0, sY);
		base =  - unit1 + up + unit2;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, sY);
		base =  unit1 + up + unit2;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit1 + unit2;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit1 + unit2;
		glNormal3dv(unit2+nv);
		glVertex3dv(base);
		// side 2
		glTexCoord2f(0.0, sY);
		base =  - unit2 + up - unit1;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, sY);
		base =   unit2 + up - unit1;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =  unit2 - unit1;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base = - unit2 - unit1;
		glNormal3dv(-unit1+nv);
		glVertex3dv(base);
		// side 4
		glTexCoord2f(sX, sY);
		base =   unit2 + up + unit1;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, sY);
		base =  - unit2 + up + unit1;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(0.0, 0.0);
		base =  - unit2 + unit1;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glTexCoord2f(sX, 0.0);
		base =   unit2 + unit1;
		glNormal3dv(unit1+nv);
		glVertex3dv(base);
		glEnd(); 

		glDisable( GL_CULL_FACE );	// Most objects don't use backface culling so deactivate it!
		glDepthFunc(GL_LESS);
		glEndList();
	}
	triag += 24;
	return triag;
}

/**
 * draws a house (skyscraper) as a textured 3D box
 * @param dlod		LOD for that object
 */
int JWCube::draw(int distance)
{
	glColor3f(1,1,1);

	if (List != 0)
		// take the precomputed display list
		glCallList(List);
	else
	{
		List = glGenLists(1); 
		glNewList(List,GL_COMPILE);		// Start With The List
		
		
		glBindTexture(GL_TEXTURE_2D, baseTex.texID);    
		
		glAlphaFunc(GL_GREATER,0.8f); // Set Alpha Testing  
		glEnable(GL_ALPHA_TEST);      // Enable Alpha Testing
		glDisable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_LIGHTING);
		// draw nested cubes
		for (int scale = 50; scale <= width/2; scale += 50)
		{
			if (scale == width/2)
			{
				glEnable(GL_LIGHTING);
				glDisable(GL_ALPHA_TEST);      // Enable Alpha Testing
			}
			
			CDoubleVector up = CDoubleVector(0,scale,0);
			CDoubleVector unit1 = CDoubleVector(scale,0,0);
			CDoubleVector unit2 = CDoubleVector(0,0,scale);
			CDoubleVector base;
			
			glBegin(GL_QUADS);
			// roof

			glTexCoord2f(0.0f, 1.0f);
			base =  - unit2 + up + unit1;
			glNormal3dv(up);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 1.0f);
			base =  unit2 + up + unit1;
			glNormal3dv(up);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0f);
			base =  unit2 + up - unit1;
			glNormal3dv(up);
			glVertex3dv(base);
			glTexCoord2f(0.0f, 0.0f);
			base =  - unit2 + up - unit1;
			glNormal3dv(up);
			glVertex3dv(base);
			// floor
			glTexCoord2f(1.0f, 1.0f);
			base =  unit2 - up + unit1;
			glNormal3dv(-up);
			glVertex3dv(base);
			glTexCoord2f(0.0f, 1.0f);
			base =  - unit2 - up + unit1;
			glNormal3dv(-up);
			glVertex3dv(base);
			glTexCoord2f(0.0f, 0.0f);
			base =  - unit2 - up - unit1;
			glNormal3dv(-up);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0f);
			base =  unit2 - up - unit1;
			glNormal3dv(-up);
			glVertex3dv(base);
			// side 1
			glTexCoord2f(1.0f, 1.0f);
			base =  unit1 + up - unit2;
			glNormal3dv(-unit2);
			glVertex3dv(base);
			glTexCoord2f(0.0, 1.0f);
			base =  - unit1 + up - unit2;
			glNormal3dv(-unit2);
			glVertex3dv(base);
			glTexCoord2f(0.0, 0.0);
			base =  - unit1 - up - unit2;
			glNormal3dv(-unit2);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0);
			base =  unit1 - up - unit2;
			glNormal3dv(-unit2);
			glVertex3dv(base);
			// side 3
			glTexCoord2f(0.0, 1.0f);
			base =  - unit1 + up + unit2;
			glNormal3dv(unit2);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 1.0f);
			base =  unit1 + up + unit2;
			glNormal3dv(unit2);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0);
			base =  unit1 - up + unit2;
			glNormal3dv(unit2);
			glVertex3dv(base);
			glTexCoord2f(0.0, 0.0);
			base =  - unit1 - up + unit2;
			glNormal3dv(unit2);
			glVertex3dv(base);
			// side 2
			glTexCoord2f(0.0, 1.0f);
			base =  - unit2 + up - unit1;
			glNormal3dv(-unit1);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 1.0f);
			base =   unit2 + up - unit1;
			glNormal3dv(-unit1);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0);
			base =  unit2 - up - unit1;
			glNormal3dv(-unit1);
			glVertex3dv(base);
			glTexCoord2f(0.0, 0.0);
			base = - unit2 - up - unit1;
			glNormal3dv(-unit1);
			glVertex3dv(base);
			// side 4
			glTexCoord2f(1.0f, 1.0f);
			base =   unit2 + up + unit1;
			glNormal3dv(unit1);
			glVertex3dv(base);
			glTexCoord2f(0.0, 1.0f);
			base =  - unit2 + up + unit1;
			glNormal3dv(unit1);
			glVertex3dv(base);
			glTexCoord2f(0.0, 0.0);
			base =  - unit2 - up + unit1;
			glNormal3dv(unit1);
			glVertex3dv(base);
			glTexCoord2f(1.0f, 0.0);
			base =   unit2 - up + unit1;
			glNormal3dv(unit1);
			glVertex3dv(base);
			glEnd(); 
		}		

		glEndList();
	}

	return 12; // cube has 6*2 triangles
}


#define X .525731112119133606f
#define Z .850650808352039932f

	GLfloat JWAsteroid::vdata[12][3] = {
	{-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
	{0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
	{Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
	};
	GLuint JWAsteroid::tindices[20][3] = {
	{0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},
	{8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
	{7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
	{6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11} };

JWAsteroid::JWAsteroid() : JWStaticObject()
{
	isSolid = true;
	ListAsteroid[0] = 0;
	ListAsteroid[1] = 0;
	ListAsteroid[2] = 0;
	ListAsteroid[3] = 0;
};

void JWAsteroid::init(float pwidth, float pheight, char *tex, float pdensity, float pminh, float pmaxh)
{
  JWStaticObject::init(pwidth, pheight, tex, pdensity, pminh, pmaxh);
  boundingRadius = pwidth/2;
  qWarning("boundingRadius of %s changed to: %f", getName(), boundingRadius);
};


/**
* draws a plant object which is a 3D decal sprite
* @param dlod		LOD for that object
*/
int JWAsteroid::draw(int distance)
{
	int dlod = log2((unsigned int)distance);
	
	int depth = CLAMP(0,3,this->visibleLOD - dlod + 3);

	if (ListAsteroid[depth] != 0)
		// take the precomputed display list
		glCallList(ListAsteroid[depth]);
	else
	{
		ListAsteroid[depth] = glGenLists(1); 
		glNewList(ListAsteroid[depth],GL_COMPILE);		// Start With The List
		
		glDisable(GL_BLEND);
		glBindTexture(GL_TEXTURE_2D, baseTex.texID);    
		
		for (int i = 0; i < 20; i++) {
			subdivide(&vdata[tindices[i][0]][0],
				&vdata[tindices[i][1]][0],
				&vdata[tindices[i][2]][0],
				depth);
		}
		glEndList();
	}
	
	return 20*pow(4,depth);
}

void JWAsteroid::drawtriangle(CFloatVector v1, CFloatVector v2, CFloatVector v3)
{
	double alpha, beta, htmp;
	CFloatVector cv;
	glBegin(GL_TRIANGLES);
	CoordinatesToPolar(v1.x, v1.y, v1.z, &alpha, &beta, &htmp);
	alpha = (alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, 1]
	beta = (1-beta/PI);		  // convert from [0,PI] to [0,1]
	glTexCoord2d(alpha, beta); 
	glNormal3fv(v1);
	cv = CFloatVector(v1.x*(width/2), v1.y*(height/2), v1.z*(width/2));
	glVertex3fv(cv);
	CoordinatesToPolar(v2.x, v2.y, v2.z, &alpha, &beta, &htmp);
	alpha = (alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, 1]
	beta = (1-beta/PI);		  // convert from [0,PI] to [0,1]
	glTexCoord2d(alpha, beta); 
	glNormal3fv(v2);
	cv = CFloatVector(v2.x*(width/2), v2.y*(height/2), v2.z*(width/2));
	glVertex3fv(cv);
	CoordinatesToPolar(v3.x, v3.y, v3.z, &alpha, &beta, &htmp);
	alpha = (alpha+PI)/(2*PI); // convert from [-PI,PI] to [0, 1]
	beta = (1-beta/PI);		  // convert from [0,PI] to [0,1]
	glTexCoord2d(alpha, beta); 
	glNormal3fv(v3);
	cv = CFloatVector(v3.x*(width/2), v3.y*(height/2), v3.z*(width/2));
	glVertex3fv(cv);
	glEnd();
}

void JWAsteroid::subdivide(CFloatVector v1, CFloatVector v2, CFloatVector v3, long depth)
{
	CFloatVector v12, v23, v31;

	if (depth == 0) {
		drawtriangle(v1, v2, v3);
		return;
	}
	v12 = v1+v2;
	v23 = v2+v3;
	v31 = v3+v1;
	v12.Normalize();
	v23.Normalize();
	v31.Normalize();
	subdivide(v1, v12, v31, depth-1);
	subdivide(v2, v23, v12, depth-1);
	subdivide(v3, v31, v23, depth-1);
	subdivide(v12, v23, v31, depth-1);
}



