/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	JWObject.h                                      *
 *                                                          *
 * Content: Header file for some basic 3D object classes    *
 *          e.g movable objects and cameras                 *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#ifndef JWSTATOBJECT_H
#define JWSTATOBJECT_H


#include "JWObject.h"		// Base 3D obejct classes
#include "Texture.h"

class QDomElement;
class CFLoatVector;


/****************************************************************************
* Class: JWStaticObject
*****************************************************************************
* This primitive class contains proxy objects that are dislayed on the terrain
* at different LODs
*****************************************************************************/
class JWStaticObject : public C3DBase
{
public:
	GLuint List;		// number for the display list entry if there is one
	int baseLOD;		// LOD that is the finest sensible resolution for that object
	int visibleLOD;		// LOD at which we can first see the object
	float width, height;// size of object

	// natural selection factors
	float density;  // value between 0 and 1, determines how dense a quad would be if we only use this object
	float minHeight, maxHeight;	// some objects are limited on specific areas e.g underwater, climatic zones...

	bool isSolid;	// this property allows or forbids objects of smaller LOD at the same position
					// e.g. a house is solid, we do not want grass inside it, but grass may surround a tree!

	TEXTURE baseTex;	// all objects have in common that their high LOD appearance is a decal sprite using baseTex!

	JWStaticObject() { List = 0; isSolid = true; };
	virtual ~JWStaticObject() {};

	/**
	* This creates a static object. They are randomly placed on terrain controled by specific parameters.
	* @param pwidth		width of the object
	* @param pheight		height of the object
	* @param tex			basis texture for that object
	* @param pdensity		how dense will this object be placed (0.0 - 1.0)
	* @param pminh			minimum height to place the object (means sth. like climatic zone e.g at coast 0-200m, mountains 6000-10000m)
	* @param pmaxh			maximum height to place the object
	*/
	virtual void init(float pwidth, float pheight, char *tex, float pdensity, float pminh, float pmaxh);

	void init(QDomElement node);

	virtual int draw(int distance) { return 0; };	// here it is important to know at what LOD we are, returns number of triang drawn

};

class JWPlant : public JWStaticObject
{
public:
	JWPlant() : JWStaticObject() {	isSolid = false; };
	/**
	* draws a plant object which is a 3D decal sprite
	* @param dlod		LOD for that object
	*/
	virtual int draw(int distance);	// here it is important to know at what LOD we are
};

class JWHouse : public JWStaticObject
{
public:
	GLuint ListEtages;
	GLuint ListBlending;

	JWHouse() : JWStaticObject() {	isSolid = true; ListEtages = 0; ListBlending = 0; };
	/**
	* draws a house (skyscraper) as a textured 3D box
	* @param dlod		LOD for that object
	*/
	virtual int draw(int distance);	// here it is important to know at what LOD we are
};

class JWCube : public JWStaticObject
{
public:
	JWCube() : JWStaticObject() {	isSolid = true; };
	/**
	* draws a borg cube (skyscraper) as a textured 3D box
	* @param dlod		LOD for that object
	*/
	virtual int draw(int distance);	// here it is important to know at what LOD we are
};

class JWAsteroid : public JWStaticObject
{
private:
	static GLfloat vdata[12][3] ;
	static GLuint tindices[20][3];

	GLuint ListAsteroid[4];
public:
	JWAsteroid();

  virtual void init(float pwidth, float pheight, char *tex, float pdensity, float pminh, float pmaxh);

	int draw(int distance);	// here it is important to know at what LOD we are
	void drawtriangle(CFloatVector v1, CFloatVector v2, CFloatVector v3);
	void subdivide(CFloatVector v1, CFloatVector v2, CFloatVector v3, long depth);
};

#endif




