/*
	The Universe Development Kit
	Copyright (C) 2000  Sean O'Neil
	s_p_oneil@hotmail.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __Matrix_h__
#define __Matrix_h__

#include <vector>
#include "Noise.h"	// Has some useful defines and inline functions

class CDoubleMatrix;
class CDoubleQuaternion;
class Frustum;
class C3DBase;
class C3DObject;

/*******************************************************************************
* Template Class: TVector
********************************************************************************
* This template class implements a simple 3D vector with x, y, and z coordinates.
* Several functions and operators are defined to make working with vectors easier,
* and because it's templatized, any numeric type can be used with it. Macros are
* defined for the most common types.

*******************************************************************************/
#define CFloatVector	TVector<float>
#define CDoubleVector	TVector<double>
#define CIntVector		TVector<int>
#define CByteVector		TVector<unsigned char>
template <class T> class TVector
{
public:
	T x, y, z;

	// Constructors
	TVector()									{}
	TVector(const T a, const T b, const T c)	{ x = a; y = b; z = c; }
	TVector(const T t)							{ x = y = z = t; }
	TVector(const T *pt)						{ x = pt[0]; y = pt[1]; z = pt[2]; }
	TVector(const TVector<T> &v)				{ x = v.x; y = v.y; z = v.z; }

	// Casting and unary operators
	operator TVector<double>()					{ return TVector<double>((double)x, (double)y, (double)z); }
	operator T*()								{ return &x; }
	T &operator[](const int n)					{ return (&x)[n]; }
	//operator const T*() const					{ return &x; }
	T operator[](const int n) const				{ return (&x)[n]; }
	TVector<T> operator-() const				{ return TVector<T>(-x, -y, -z); }

	// Equal and comparison operators
	void operator=(const T t)					{ x = y = z = t; }
	void operator=(const T *pt)					{ x = pt[0]; y = pt[1]; z = pt[2]; }
	void operator=(const TVector<T> &v)			{ x = v.x; y = v.y; z = v.z; }
	bool operator==(TVector<T> &v) const		{ return (Abs(x - v.x) <= (T)0.001f && Abs(y - v.y) <= 0.001f && Abs(z - v.z) <= 0.001f); }
	bool operator!=(TVector<T> &v) const		{ return !(*this == v); }
	// pareto-optimal comparison
	bool operator<=(TVector<T> &v) const		{ return ((x <= v.x) && (y <= v.y) && (z <= v.z)); }
	bool operator>=(TVector<T> &v) const		{ return ((x >= v.x) && (y >= v.y) && (z >= v.z)); }


	// Arithmetic operators (vector with scalar)
	TVector<T> operator+(const T t) const		{ return TVector<T>(x+t, y+t, z+t); }
	TVector<T> operator-(const T t) const		{ return TVector<T>(x-t, y-t, z-t); }
	TVector<T> operator*(const T t) const		{ return TVector<T>(x*t, y*t, z*t); }
	TVector<T> operator/(const T t) const		{ return TVector<T>(x/t, y/t, z/t); }
	void operator+=(const T t)					{ x += t; y += t; z += t; }
	void operator-=(const T t)					{ x -= t; y -= t; z -= t; }
	void operator*=(const T t)					{ x *= t; y *= t; z *= t; }
	void operator/=(const T t)					{ x /= t; y /= t; z /= t; }

	// Arithmetic operators (vector with vector)
	TVector<T> operator+(const TVector<T> &v) const	{ return TVector<T>(x+v.x, y+v.y, z+v.z); }
	TVector<T> operator-(const TVector<T> &v) const	{ return TVector<T>(x-v.x, y-v.y, z-v.z); }
	TVector<T> operator*(const TVector<T> &v) const	{ return TVector<T>(x*v.x, y*v.y, z*v.z); }
	TVector<T> operator/(const TVector<T> &v) const	{ return TVector<T>(x/v.x, y/v.y, z/v.z); }
	void operator+=(const TVector<T> &v)		{ x += v.x; y += v.y; z += v.z; }
	void operator-=(const TVector<T> &v)		{ x -= v.x; y -= v.y; z -= v.z; }
	void operator*=(const TVector<T> &v)		{ x *= v.x; y *= v.y; z *= v.z; }
	void operator/=(const TVector<T> &v)		{ x /= v.x; y /= v.y; z /= v.z; }

	// Dot and cross product operators
	T operator|(const TVector<T> &v) const		{ return x*v.x + y*v.y + z*v.z; }
	TVector<T> operator^(const TVector<T> &v) const	{ return TVector<T>(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x); }
	void operator^=(const TVector<T> &v)		{ *this = *this ^ v; }

	// Magnitude/distance methods
	double MagnitudeSquared() const					{ return (double)x*(double)x + (double)y*(double)y + (double)z*(double)z; }
	double Magnitude() const						{ return (double)sqrt(MagnitudeSquared()); }
	double DistanceSquared(const TVector<T> &v) const{ return (*this - v).MagnitudeSquared(); }
	double Distance(const TVector<T> &v) const		{ return (*this - v).Magnitude(); }
	TVector<T> Midpoint(const TVector<T> &v) const	{ return TVector<T>((*this - v) / 2 + v); }
	TVector<T> Average(const TVector<T> &v) const	{ return TVector<T>((*this + v) / 2); }

	// Advanced methods (should only be used with float or double types)
	void Normalize()							{ *this /= Magnitude(); }
	double Angle(const TVector<T> &v) const		{ return acos(*this | v); }
	TVector<T> Reflect(const TVector<T> &n) const
	{
		T t = (T)Magnitude();
		TVector<T> v = *this / t;
		return (v - n * (2 * (v | n))) * t;
	}
	TVector<T> Rotate(const T tAngle, const CDoubleVector &n) const
	{
		T tCos = (T)cos(tAngle);
		T tSin = (T)sin(tAngle);
		return TVector<T>(*this * tCos + ((n * *this) * (1 - tCos)) * n + (*this ^ n) * tSin);
	}
};

// Returns the normal vector of two vectors (the normalized cross product)
template <class T> inline TVector<T> NormalVector(const TVector<T> &v1, const TVector<T> &v2)
{
	TVector<T> v = v1 ^ v2;
	v.Normalize();
	return v;
}

// Returns the normal vector of a triangle or 3 points on a plane (assumes counter-clockwise orientation)
template <class T> inline TVector<T> NormalVector(const TVector<T> &p1, const TVector<T> &p2, const TVector<T> &p3)
{
	return NormalVector(p2-p1, p3-p1);
}

// Returns the direction vector between two points
template <class T> inline TVector<T> DirectionVector(const TVector<T> &p1, const TVector<T> &p2)
{
	TVector<T> v = p2 - p1;
	v.Normalize();
	return v;
}



/*******************************************************************************
* Template Class: TVector4
********************************************************************************
* This template class implements a simple 4D vector with x, y, z, and w
* coordinates. Like TVector, it is templatized and macros are defined for the
* most common types.
*******************************************************************************/
#define CFloatVector4	TVector4<float>
#define CDoubleVector4	TVector4<double>
#define CIntVector4		TVector4<int>
#define CByteVector4	TVector4<unsigned char>
template <class T> class TVector4
{
public:
	union {	T x; T r; };
	union {	T y; T g; };
	union {	T z; T b; };
	union {	T w; T a; };

	// Constructors
	TVector4()									{}
	TVector4(const T a, const T b, const T c, const T d)	{ x = a; y = b; z = c; w = d; }
	TVector4(const T t)							{ x = y = z = w = t; }
	TVector4(const T *pt)						{ x = pt[0]; y = pt[1]; z = pt[2]; w = pt[3]; }
	TVector4(const TVector<T> &v)				{ x = v.x; y = v.y; z = v.z; w = 0; }
	TVector4(const TVector4<T> &v)				{ x = v.x; y = v.y; z = v.z; w = v.w; }

	// Casting and unary operators
	operator T*()								{ return &x; }
	T &operator[](const int n)					{ return (&x)[n]; }
	operator const T*() const					{ return &x; }
	T operator[](const int n) const				{ return (&x)[n]; }
	TVector4<T> operator-() const				{ return TVector4<T>(-x, -y, -z, -w); }

	// Equal and comparison operators
	void operator=(const T t)					{ x = y = z = w = t; }
	void operator=(const T *pt)					{ x = pt[0]; y = pt[1]; z = pt[2]; w = pt[3]; }
	void operator=(const TVector<T> &v)			{ x = v.x; y = v.y; z = v.z; w = 0; }
	void operator=(const TVector4<T> &v)		{ x = v.x; y = v.y; z = v.z; w = v.w; }
	bool operator==(TVector4<T> &v) const		{ return (Abs(x - v.x) <= (T)DELTA && Abs(y - v.y) <= (T)DELTA && Abs(z - v.z) <= (T)DELTA && Abs(w - v.w) <= (T)DELTA); }
	bool operator!=(TVector4<T> &v) const		{ return !(*this == v); }

	// Arithmetic operators (vector with scalar)
	TVector4<T> operator+(const T t) const		{ return TVector4<T>(x+t, y+t, z+t, w+t); }
	TVector4<T> operator-(const T t) const		{ return TVector4<T>(x-t, y-t, z-t, w-t); }
	TVector4<T> operator*(const T t) const		{ return TVector4<T>(x*t, y*t, z*t, w*t); }
	TVector4<T> operator/(const T t) const		{ return TVector4<T>(x/t, y/t, z/t, w/t); }
	void operator+=(const T t)					{ x += t; y += t; z += t; w += t; }
	void operator-=(const T t)					{ x -= t; y -= t; z -= t; w -= t; }
	void operator*=(const T t)					{ x *= t; y *= t; z *= t; w *= t; }
	void operator/=(const T t)					{ x /= t; y /= t; z /= t; w /= t; }

	// Arithmetic operators (vector with vector)
	TVector4<T> operator+(const TVector4<T> &v) const	{ return TVector4<T>(x+v.x, y+v.y, z+v.z, w+v.w); }
	TVector4<T> operator-(const TVector4<T> &v) const	{ return TVector4<T>(x-v.x, y-v.y, z-v.z, w-v.w); }
	TVector4<T> operator*(const TVector4<T> &v) const	{ return TVector4<T>(x*v.x, y*v.y, z*v.z, w*v.w); }
	TVector4<T> operator/(const TVector4<T> &v) const	{ return TVector4<T>(x/v.x, y/v.y, z/v.z, w/v.w); }
	void operator+=(const TVector4<T> &v)		{ x += v.x; y += v.y; z += v.z; w += v.w; }
	void operator-=(const TVector4<T> &v)		{ x -= v.x; y -= v.y; z -= v.z; w -= v.w; }
	void operator*=(const TVector4<T> &v)		{ x *= v.x; y *= v.y; z *= v.z; w *= v.w; }
	void operator/=(const TVector4<T> &v)		{ x /= v.x; y /= v.y; z /= v.z; w /= v.w; }

	// Magnitude/normalize methods
	T MagnitudeSquared() const					{ return x*x + y*y + z*z + w*w; }
	T Magnitude() const							{ return (T)sqrt(MagnitudeSquared()); }
	void Normalize()							{ *this /= Magnitude(); }
};


/*******************************************************************************
* Class: CDoubleQuaternion
********************************************************************************
* This class implements a 4D quaternion. Several functions and operators are
* defined to make working with quaternions easier. Quaternions are often used to
* represent rotations, and have a number of advantages over other constructs.
* Their main disadvantage is that they are unintuitive.
*
* Note: This class is not templatized because integral data types don't make sense
*       and there's no need for double-precision.
*******************************************************************************/
class CDoubleQuaternion
{
public:
	double x, y, z, w;

	// Constructors
	CDoubleQuaternion()								{}
	CDoubleQuaternion(const double a, const double b, const double c, const double d)	{ x = a; y = b; z = c; w = d; }
	CDoubleQuaternion(const CDoubleVector &v, const double f){ SetAxisAngle(v, f); }
	CDoubleQuaternion(const CDoubleVector &v)				{ *this = v; }
	CDoubleQuaternion(const CDoubleQuaternion &q)			{ *this = q; }
	CDoubleQuaternion(const CDoubleMatrix &m)				{ *this = m; }
	CDoubleQuaternion(const double *p)					{ *this = p; }

	// Casting and unary operators
	operator double*()							{ return &x; }
	double &operator[](const int n)				{ return (&x)[n]; }
	//operator const double*() const				{ return &x; }
	double operator[](const int n) const			{ return (&x)[n]; }
	CDoubleQuaternion operator-() const				{ return CDoubleQuaternion(-x, -y, -z, -w); }

	// Equal and comparison operators
	void operator=(const CDoubleVector &v)			{ x = v.x; y = v.y; z = v.z; w = 0; }
	void operator=(const CDoubleQuaternion &q)		{ x = q.x; y = q.y; z = q.z; w = q.w; }
	void operator=(const CDoubleMatrix &m);
	void operator=(const double *p)				{ x = p[0]; y = p[1]; z = p[2]; w = p[3]; }

	// Arithmetic operators (quaternion and scalar)
	CDoubleQuaternion operator+(const double f) const	{ return CDoubleQuaternion(x+f, y+f, z+f, w+f); }
	CDoubleQuaternion operator-(const double f) const	{ return CDoubleQuaternion(x-f, y-f, z-f, w-f); }
	CDoubleQuaternion operator*(const double f) const	{ return CDoubleQuaternion(x*f, y*f, z*f, w*f); }
	CDoubleQuaternion operator/(const double f) const	{ return CDoubleQuaternion(x/f, y/f, z/f, w/f); }
	void operator+=(const double f)				{ x+=f; y+=f; z+=f; w+=f; }
	void operator-=(const double f)				{ x-=f; y-=f; z-=f; w-=f; }
	void operator*=(const double f)				{ x*=f; y*=f; z*=f; w*=f; }
	void operator/=(const double f)				{ x/=f; y/=f; z/=f; w/=f; }

	// Arithmetic operators (quaternion and quaternion)
	CDoubleQuaternion operator+(const CDoubleQuaternion &q) const	{ return CDoubleQuaternion(x+q.x, y+q.y, z+q.z, w+q.w); }
	CDoubleQuaternion operator-(const CDoubleQuaternion &q) const	{ return CDoubleQuaternion(x-q.x, y-q.y, z-q.z, w-q.w); }
	CDoubleQuaternion operator*(const CDoubleQuaternion &q) const;	// Multiplying quaternions is a special operation
	void operator+=(const CDoubleQuaternion &q)	{ x+=q.x; y+=q.y; z+=q.z; w+=q.w; }
	void operator-=(const CDoubleQuaternion &q)	{ x-=q.x; y-=q.y; z-=q.z; w-=q.w; }
	void operator*=(const CDoubleQuaternion &q)	{ *this = *this * q; }

	// Magnitude/normalize methods
	double MagnitudeSquared() const			{ return x*x + y*y + z*z + w*w; }
	double Magnitude() const					{ return (double)sqrt(MagnitudeSquared()); }
	void Normalize()						{ *this /= Magnitude(); }

	// Advanced quaternion methods
	CDoubleQuaternion Conjugate() const			{ return CDoubleQuaternion(-x, -y, -z, w); }
	CDoubleQuaternion Inverse() const				{ return Conjugate() / MagnitudeSquared(); }
	CDoubleQuaternion UnitInverse() const			{ return Conjugate(); }
	CDoubleVector RotateVector(const CDoubleVector &v) const	{ return (CDoubleVector)(*this * CDoubleQuaternion(v) * UnitInverse()); }

	void SetAxisAngle(const CDoubleVector &vAxis, const double fAngle)
	{
		// 4 muls, 2 trig function calls
		double f = (double)(fAngle * 0.5);
		*this = vAxis * (double)sin(f);
		w = (double)cos(f);
	}
	void GetAxisAngle(CDoubleVector &vAxis, double &fAngle) const
	{
		// 4 muls, 1 div, 2 trig function calls
		fAngle = (double)acos(w);
		vAxis = *this / (double)sin(fAngle);
		fAngle *= (double)2.0;
	}

	void ResetRotation() { x = 0.0; y = 0.0; z = 0.0; w = 1.0; };
	void Rotate(const CDoubleQuaternion &q)			{ *this = q * *this; }
	void Rotate(const CDoubleVector &vAxis, const double fAngle)
	{
		CDoubleQuaternion q;
		q.SetAxisAngle(vAxis, fAngle);
		Rotate(q);
	}

	CDoubleVector GetViewAxis() const
	{
		// 6 muls, 7 adds
		double x2 = x + x, y2 = y + y, z2 = z + z;
		double xx = x * x2, xz = x * z2;
		double yy = y * y2, yz = y * z2;
		double wx = w * x2, wy = w * y2;
		return -CDoubleVector(xz+wy, yz-wx, 1-(xx+yy));
	}
	CDoubleVector GetUpAxis() const
	{
		// 6 muls, 7 adds
		double x2 = x + x, y2 = y + y, z2 = z + z;
		double xx = x * x2, xy = x * y2;
		double yz = y * z2, zz = z * z2;
		double wx = w * x2, wz = w * z2;
		return CDoubleVector(xy-wz, 1-(xx+zz), yz+wx);
	}
	CDoubleVector GetRightAxis() const
	{
		// 6 muls, 7 adds
		double y2 = y + y, z2 = z + z;
		double xy = x * y2, xz = x * z2;
		double yy = y * y2, zz = z * z2;
		double wy = w * y2, wz = w * z2;
		return CDoubleVector(1-(yy+zz), xy+wz, xz-wy);
	}
};

extern CDoubleQuaternion Slerp(const CDoubleQuaternion &q1, const CDoubleQuaternion &q2, const double t);

/*******************************************************************************
* Class: CDoubleMatrix
********************************************************************************
* This class implements a 4x4 matrix. Several functions and operators are
* defined to make working with matrices easier. The values are kept in column-
* major order to make it easier to use with OpenGL. For performance reasons,
* most of the functions assume that all matrices are orthogonal, which means the
* bottom row is [ 0 0 0 1 ]. Since I plan to use the GL_PROJECTION matrix to
* handle the projection matrix, I should never need to use any other kind of
* matrix, and I get a decent performance boost by ignoring the bottom row.
*
* Note: This class is not templatized because integral data types don't make sense
*       and there's no need for double-precision.
*******************************************************************************/
class CDoubleMatrix
{
public:
	// This class uses column-major order, as used by OpenGL
	// Here are the ways in which the matrix values can be accessed:
	// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | f1[0] f1[4] f1[8]  f1[12] |   | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |
	// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | f1[1] f1[5] f1[9]  f1[13] |   | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |
	// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | = | f1[2] f1[6] f1[10] f1[14] | = | f2[0][2] f2[1][2] f2[2][2] f2[3][2] |
	// | f2[0][3] f2[1][3] f2[2][3] f2[3][3] |   | f1[3] f1[7] f1[11] f1[15] |   | f2[0][3] f2[1][3] f2[2][3] f2[3][3] |
	union
	{
		//struct { double f11, f12, f13, f14, f21, f22, f23, f24, f31, f32, f33, f34, f41, f42, f43, f44; };
		double f1[16];
		double f2[4][4];
	};

	CDoubleMatrix()							{}
	CDoubleMatrix(const double f)				{ *this = f; }
	CDoubleMatrix(const double *pf)			{ *this = pf; }
	CDoubleMatrix(const CDoubleQuaternion &q)		{ *this = q; }

	// Init functions
	void ZeroMatrix()
	{
		f2[0][0] = f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][1] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[2][2] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = f2[3][3] = 0;
	}
	void IdentityMatrix()
	{
		f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[0][0] = f2[1][1] = f2[2][2] = f2[3][3] = 1;
	}

	operator double*()								{ return f1; }
	double &operator[](const int n)					{ return f1[n]; }
	double &operator()(const int i, const int j)		{ return f2[i][j]; }
	//operator const double*() const					{ return f1; }
	double operator[](const int n) const				{ return f1[n]; }
	double operator()(const int i, const int j) const{ return f2[i][j]; }

	void operator=(const double f)					{ for(register int i=0; i<16; i++) f1[i] = f; }
	void operator=(const double *pf)					{ for(register int i=0; i<16; i++) f1[i] = pf[i]; }
	void operator=(const CDoubleQuaternion &q);

	CDoubleMatrix operator*(const CDoubleMatrix &m) const;
	void operator*=(const CDoubleMatrix &m)				{ *this = *this * m; }
	CDoubleVector operator*(const CDoubleVector &v) const		{ return TransformVector(v); }

	CDoubleVector TransformVector(const CDoubleVector &v) const
	{
		// 9 muls, 9 adds
		// | f11 f2[1][0] f2[2][0] f2[3][0] |   | v.x |   | f11*v.x+f2[1][0]*v.y+f2[2][0]*v.z+f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | v.y |   | f2[0][1]*v.x+f2[1][1]*v.y+f2[2][1]*v.z+f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | v.z | = | f2[0][2]*v.x+f2[1][2]*v.y+f2[2][2]*v.z+f2[3][2] |
		// | 0   0   0   1   |   | 1   |   | 1                           |
		return CDoubleVector((f2[0][0]*v.x+f2[1][0]*v.y+f2[2][0]*v.z+f2[3][0]),
					   (f2[0][1]*v.x+f2[1][1]*v.y+f2[2][1]*v.z+f2[3][1]),
					   (f2[0][2]*v.x+f2[1][2]*v.y+f2[2][2]*v.z+f2[3][2]));
	}
	CDoubleVector TransformNormal(const CDoubleVector &v) const
	{
		// 9 muls, 6 adds
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | v.x |   | f2[0][0]*v.x+f2[1][0]*v.y+f2[2][0]*v.z |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | v.y |   | f2[0][1]*v.x+f2[1][1]*v.y+f2[2][1]*v.z |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | v.z | = | f2[0][2]*v.x+f2[1][2]*v.y+f2[2][2]*v.z |
		// | 0   0   0   1   |   | 1   |   | 1                       |
		return CDoubleVector((f2[0][0]*v.x+f2[1][0]*v.y+f2[2][0]*v.z),
					   (f2[0][1]*v.x+f2[1][1]*v.y+f2[2][1]*v.z),
					   (f2[0][2]*v.x+f2[1][2]*v.y+f2[2][2]*v.z));
	}

	// Translate functions
	void TranslateMatrix(const double x, const double y, const double z)
	{
		// | 1  0  0  x |
		// | 0  1  0  y |
		// | 0  0  1  z |
		// | 0  0  0  1 |
		f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[2][3] = 0;
		f2[0][0] = f2[1][1] = f2[2][2] = f2[3][3] = 1;
		f2[3][0] = x; f2[3][1] = y; f2[3][2] = z;
	}
	void TranslateMatrix(const double *pf)		{ TranslateMatrix(pf[0], pf[1], pf[2]); }
	void Translate(const double x, const double y, const double z)
	{
		// 9 muls, 9 adds
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | 1  0  0  x |   | f2[0][0] f2[1][0] f2[2][0] f2[0][0]*x+f2[1][0]*y+f2[2][0]*z+f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | 0  1  0  y |   | f2[0][1] f2[1][1] f2[2][1] f2[0][1]*x+f2[1][1]*y+f2[2][1]*z+f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | 0  0  1  z | = | f2[0][2] f2[1][2] f2[2][2] f2[0][2]*x+f2[1][2]*y+f2[2][2]*z+f2[3][2] |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0   0   0   1                     |
		f2[3][0] = f2[0][0]*x+f2[1][0]*y+f2[2][0]*z+f2[3][0];
		f2[3][1] = f2[0][1]*x+f2[1][1]*y+f2[2][1]*z+f2[3][1];
		f2[3][2] = f2[0][2]*x+f2[1][2]*y+f2[2][2]*z+f2[3][2];
	}
	void Translate(const double *pf)				{ Translate(pf[0], pf[1], pf[2]); }

	// Scale functions
	void ScaleMatrix(const double x, const double y, const double z)
	{
		// | x  0  0  0 |
		// | 0  y  0  0 |
		// | 0  0  z  0 |
		// | 0  0  0  1 |
		f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[0][0] = x; f2[1][1] = y; f2[2][2] = z; f2[3][3] = 1;
	}
	void ScaleMatrix(const double *pf)			{ ScaleMatrix(pf[0], pf[1], pf[2]); }
	void Scale(const double x, const double y, const double z)
	{
		// 9 muls
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | x  0  0  0 |   | f2[0][0]*x f2[1][0]*y f2[2][0]*z f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | 0  y  0  0 |   | f2[0][1]*x f2[1][1]*y f2[2][1]*z f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | 0  0  z  0 | = | f2[0][2]*x f2[1][2]*y f2[2][2]*z f2[3][2] |
		// | 0   0   0   1   |   | 0  0  0  1 |   | 0     0     0     1   |
		f2[0][0] *= x; f2[1][0] *= y; f2[2][0] *= z;
		f2[0][1] *= x; f2[1][1] *= y; f2[2][1] *= z;
		f2[0][2] *= x; f2[1][2] *= y; f2[2][2] *= z;
	}
	void Scale(const double *pf)					{ Scale(pf[0], pf[1], pf[2]); }

	// Rotate functions
	void RotateXMatrix(const double fRadians)
	{
		// | 1 0    0     0 |
		// | 0 fCos -fSin 0 |
		// | 0 fSin fCos  0 |
		// | 0 0    0     1 |
		f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][3] = f2[2][0] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[0][0] = f2[3][3] = 1;

		double fCos = (double)cos(fRadians);
		double fSin = (double)sin(fRadians);
		f2[1][1] = f2[2][2] = fCos;
		f2[1][2] = fSin;
		f2[2][1] = -fSin;
	}
	void RotateX(const double fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | 1 0    0     0 |   | f2[0][0] f2[1][0]*fCos+f2[2][0]*fSin f2[2][0]*fCos-f2[1][0]*fSin f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | 0 fCos -fSin 0 |   | f2[0][1] f2[1][1]*fCos+f2[2][1]*fSin f2[2][1]*fCos-f2[1][1]*fSin f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | 0 fSin fCos  0 | = | f2[0][2] f2[1][2]*fCos+f2[2][2]*fSin f2[2][2]*fCos-f2[1][2]*fSin f2[3][2] |
		// | 0   0   0   1   |   | 0 0    0     1 |   | 0   0                 0                 1   |
		double fTemp, fCos, fSin;
		fCos = (double)cos(fRadians);
		fSin = (double)sin(fRadians);
		fTemp = f2[1][0]*fCos+f2[2][0]*fSin;
		f2[2][0] = f2[2][0]*fCos-f2[1][0]*fSin;
		f2[1][0] = fTemp;
		fTemp = f2[1][1]*fCos+f2[2][1]*fSin;
		f2[2][1] = f2[2][1]*fCos-f2[1][1]*fSin;
		f2[1][1] = fTemp;
		fTemp = f2[1][2]*fCos+f2[2][2]*fSin;
		f2[2][2] = f2[2][2]*fCos-f2[1][2]*fSin;
		f2[1][2] = fTemp;
	}
	void RotateYMatrix(const double fRadians)
	{
		// | fCos  0 fSin  0 |
		// | 0     1 0     0 |
		// | -fSin 0 fCos  0 |
		// | 0     0 0     1 |
		f2[0][1] = f2[0][3] = f2[1][0] = f2[1][2] = f2[1][3] = f2[2][1] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[1][1] = f2[3][3] = 1;

		double fCos = (double)cos(fRadians);
		double fSin = (double)sin(fRadians);
		f2[0][0] = f2[2][2] = fCos;
		f2[0][2] = -fSin;
		f2[2][0] = fSin;
	}
	void RotateY(const double fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | fCos  0 fSin  0 |   | f2[0][0]*fCos-f2[2][0]*fSin f2[1][0] f2[0][0]*fSin+f2[2][0]*fCos f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | 0     1 0     0 |   | f2[0][1]*fCos-f2[2][1]*fSin f2[1][1] f2[0][1]*fSin+f2[2][1]*fCos f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | -fSin 0 fCos  0 | = | f2[0][2]*fCos-f2[2][2]*fSin f2[1][2] f2[0][2]*fSin+f2[2][2]*fCos f2[3][2] |
		// | 0   0   0   1   |   | 0     0 0     1 |   | 0                 0   0                 1   |
		double fTemp, fCos, fSin;
		fCos = (double)cos(fRadians);
		fSin = (double)sin(fRadians);
		fTemp = f2[0][0]*fCos-f2[2][0]*fSin;
		f2[2][0] = f2[0][0]*fSin+f2[2][0]*fCos;
		f2[0][0] = fTemp;
		fTemp = f2[0][1]*fCos-f2[2][1]*fSin;
		f2[2][1] = f2[0][1]*fSin+f2[2][1]*fCos;
		f2[0][1] = fTemp;
		fTemp = f2[0][2]*fCos-f2[2][2]*fSin;
		f2[2][2] = f2[0][2]*fSin+f2[2][2]*fCos;
		f2[0][2] = fTemp;
	}
	void RotateZMatrix(const double fRadians)
	{
		// | fCos -fSin 0 0 |
		// | fSin fCos  0 0 |
		// | 0    0     1 0 |
		// | 0    0     0 1 |
		f2[0][2] = f2[0][3] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[2][2] = f2[3][3] = 1;

		double fCos = (double)cos(fRadians);
		double fSin = (double)sin(fRadians);
		f2[0][0] = f2[1][1] = fCos;
		f2[0][1] = fSin;
		f2[1][0] = -fSin;
	}
	void RotateZ(const double fRadians)
	{
		// 12 muls, 6 adds, 2 trig function calls
		// | f2[0][0] f2[1][0] f2[2][0] f2[3][0] |   | fCos -fSin 0 0 |   | f2[0][0]*fCos+f2[1][0]*fSin f2[1][0]*fCos-f2[0][0]*fSin f2[2][0] f2[3][0] |
		// | f2[0][1] f2[1][1] f2[2][1] f2[3][1] |   | fSin fCos  0 0 |   | f2[0][1]*fCos+f2[1][1]*fSin f2[1][1]*fCos-f2[0][1]*fSin f2[2][1] f2[3][1] |
		// | f2[0][2] f2[1][2] f2[2][2] f2[3][2] | * | 0    0     1 0 | = | f2[0][2]*fCos+f2[1][2]*fSin f2[1][2]*fCos-f2[0][2]*fSin f2[2][2] f2[3][2] |
		// | 0   0   0   1   |   | 0    0     0 1 |   | 0                 0                 0   1   |
		double fTemp, fCos, fSin;
		fCos = (double)cos(fRadians);
		fSin = (double)sin(fRadians);
		fTemp = f2[0][0]*fCos+f2[1][0]*fSin;
		f2[1][0] = f2[1][0]*fCos-f2[0][0]*fSin;
		f2[0][0] = fTemp;
		fTemp = f2[0][1]*fCos+f2[1][1]*fSin;
		f2[1][1] = f2[1][1]*fCos-f2[0][1]*fSin;
		f2[0][1] = fTemp;
		fTemp = f2[0][2]*fCos+f2[1][2]*fSin;
		f2[1][2] = f2[1][2]*fCos-f2[0][2]*fSin;
		f2[0][2] = fTemp;
	}
	void RotateMatrix(const CDoubleVector &v, const double f)
	{
		// 15 muls, 10 adds, 2 trig function calls
		double fCos = (double)cos(f);
		CDoubleVector vCos = v * (1 - fCos);
		CDoubleVector vSin = v * (double)sin(f);

		f2[0][3] = f2[1][3] = f2[2][3] = f2[3][0] = f2[3][1] = f2[3][2] = 0;
		f2[3][3] = 1;

		f2[0][0] = (v.x * vCos.x) + fCos;
		f2[1][0] = (v.x * vCos.y) - (vSin.z);
		f2[2][0] = (v.x * vCos.z) + (vSin.y);
		f2[0][1] = (v.y * vCos.x) + (vSin.z);
		f2[1][1] = (v.y * vCos.y) + fCos;
		f2[2][1] = (v.y * vCos.z) - (vSin.x);
		f2[0][2] = (v.z * vCos.x) - (vSin.y);
		f2[2][1] = (v.z * vCos.y) + (vSin.x);
		f2[2][2] = (v.z * vCos.z) + fCos;
	}
	void Rotate(const CDoubleVector &v, const double f)
	{
		// 51 muls, 37 adds, 2 trig function calls
		CDoubleMatrix mat;
		mat.RotateMatrix(v, f);
		*this *= mat;
	}

	void ModelMatrix(const CDoubleQuaternion &q, const CDoubleVector &vFrom)
	{
		*this = q;
		f2[3][0] = vFrom.x;
		f2[3][1] = vFrom.y;
		f2[3][2] = vFrom.z;
	}
	void ViewMatrix(const CDoubleQuaternion &q, const CDoubleVector &vFrom)
	{
		*this = q;
		f2[3][0] = -(vFrom.x*f2[0][0] + vFrom.y*f2[1][0] + vFrom.z*f2[2][0]);
		f2[3][1] = -(vFrom.x*f2[0][1] + vFrom.y*f2[1][1] + vFrom.z*f2[2][1]);
		f2[3][2] = -(vFrom.x*f2[0][2] + vFrom.y*f2[1][2] + vFrom.z*f2[2][2]);
	}
	void ViewMatrix(const CDoubleVector &vFrom, const CDoubleVector &vView, const CDoubleVector &vUp, const CDoubleVector &vRight)
	{
		// 9 muls, 9 adds
		f2[0][0] = vRight.x;	f2[1][0] = vRight.y;	f2[2][0] = vRight.z;	f2[3][0] = -(vFrom | vRight);
		f2[0][1] = vUp.x;	f2[1][1] = vUp.y;	f2[2][1] = vUp.z;	f2[3][1] = -(vFrom | vUp);
		f2[0][2] = -vView.x;	f2[1][2] = -vView.y;	f2[2][2] = -vView.z;	f2[3][2] = -(vFrom | -vView);
		f2[0][3] = 0;		f2[1][3] = 0;		f2[2][3] = 0;		f2[3][3] = 1;
	}
	void ViewMatrix(const CDoubleVector &vFrom, const CDoubleVector &vAt, const CDoubleVector &vUp)
	{
		CDoubleVector vView = vAt - vFrom;
		vView.Normalize();
		CDoubleVector vRight = vView ^ vUp;
		vRight.Normalize();
		CDoubleVector vTrueUp = vRight ^ vView;
		vTrueUp.Normalize();
		ViewMatrix(vFrom, vView, vTrueUp, vRight);
	}
	void ProjectionMatrix(const double fNear, const double fFar, const double fFOV, const double fAspect)
	{
		// 2 muls, 3 divs, 2 adds, 1 trig function call
		double h = (double)(1.0 / tan(DEGTORAD(fFOV * 0.5)));
		double Q = (double)(fFar / (fFar - fNear));
		f2[0][1] = f2[0][2] = f2[0][3] = f2[1][0] = f2[1][2] = f2[1][3] = f2[2][0] = f2[2][1] = f2[3][0] = f2[3][1] = f2[3][3] = 0;
		f2[0][0] = h / fAspect;
		f2[1][1] = h;
		f2[2][2] = Q;
		f2[2][3] = 1;
		f2[3][2] = -Q*fNear;
	}

	// For orthogonal matrices, I belive this also gives you the inverse.
	void Transpose()
	{

		double f;
		SWAP(f2[0][1], f2[1][0], f);
		SWAP(f2[0][2], f2[2][0], f);
		SWAP(f2[0][3], f2[3][0], f);
		SWAP(f2[1][2], f2[2][1], f);
		SWAP(f2[1][3], f2[3][1], f);
		SWAP(f2[2][3], f2[3][2], f);
	}
};

#endif //__Matrix_h__


