/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	Model.cpp                                       *
 *                                                          *
 * Content: Implementation for some basic 3D model classes  *
 *          (Milkshape, 3ds, obj, ...)                      *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#include <fstream.h>
#include "Model.h"
#include "JWGameEngine.h"
#include "JWAstro.h"

Model::Model()
{
	List[0] = 0; List[1] = 0; List[2] = 0;
	//baserotation = 0.0f;
	m_numMeshes = 0;
	m_pMeshes = NULL;
	m_numMaterials = 0;
	m_pMaterials = NULL;
	m_numTriangles = 0;
	m_pTriangles = NULL;
	m_numVertices = 0;
	m_pVertices = NULL;
}

Model::~Model()
{
	int i;
	for ( i = 0; i < m_numMeshes; i++ )
		delete[] m_pMeshes[i].m_pTriangleIndices;
	for ( i = 0; i < m_numMaterials; i++ )
		delete[] m_pMaterials[i].m_pTextureFilename;

	m_numMeshes = 0;
	if ( m_pMeshes != NULL )
	{
		delete[] m_pMeshes;
		m_pMeshes = NULL;
	}

	m_numMaterials = 0;
	if ( m_pMaterials != NULL )
	{
		delete[] m_pMaterials;
		m_pMaterials = NULL;
	}

	m_numTriangles = 0;
	if ( m_pTriangles != NULL )
	{
		delete[] m_pTriangles;
		m_pTriangles = NULL;
	}

	m_numVertices = 0;
	if ( m_pVertices != NULL )
	{
		delete[] m_pVertices;
		m_pVertices = NULL;
	}
}

float Model::GetSize()
{
	// we expect the model to be axis aligned and search the axis extends
	double minx, miny, minz, maxx, maxy, maxz;
	minx = miny = minz = 1.0e30;
	maxx = maxy = maxz = -1.0e30;
	CDoubleVector p;

	if (m_numMeshes == 0) return 2.0f;	// default if no model exists

	for ( int index = 0; index < m_numVertices; index++ )
	{	
		p = CDoubleVector( m_pVertices[index].m_location[0], m_pVertices[index].m_location[1], m_pVertices[index].m_location[2] );
		minx = Min(p.x, minx);
		miny = Min(p.y, miny);
		minz = Min(p.z, minz);
		maxx = Max(p.x, maxx);
		maxy = Max(p.y, maxy);
		maxz = Max(p.z, maxz);
	}

	return (float)( Max( (maxx-minx), Max( (maxy-miny), (maxz-minz) ) ) );
}

void Model::SetSize(float len)
{
	boundingRadius = 0.5 * len;
	float scale = len / GetSize();
	scaleMat.ScaleMatrix(scale, scale, scale);

	qWarning("Model has size %s", distStr(GetSize()));
}

/**
*	Draw the model with display list if possible else call drawModel
* @param camera			The camera that looks at model
*/
void Model::draw(C3DCamera *camera)
{
	glDisable(GL_COLOR_MATERIAL);		// Forbid color as material because models have complex material definitions!

	glPushAttrib(GL_NORMALIZE);
	glEnable(GL_NORMALIZE);

	glPushAttrib(GL_BLEND);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glAlphaFunc(GL_GREATER,0.1f); // Set Alpha Testing  
	glEnable(GL_ALPHA_TEST);	// if material is no glass we use alpha testing!	

	glDisable(GL_CULL_FACE);

	glEnable(GL_TEXTURE_2D);

	glColor3f(1,1,1);
	// Multiply the current view matrix by the model matrix and draw the model
	CDoubleMatrix m = GetScaledModelMatrix(camera);
	C3DBase *cp = C3DBase::GetCommonParent(this, camera);
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(camera->GetCameraMatrix(cp)*m);
	//glRotatef(baserotation,0.0,1.0,0.0); // base rotation for flipped models
	glMultMatrixd(this->scaleMat);

	drawModel(M_NORMAL);

	glMatrixMode(GL_TEXTURE);
	CDoubleMatrix rotMat = *this;
	glLoadMatrixd(rotMat);
	glMatrixMode(GL_MODELVIEW);

	drawModel(M_ENVMAP);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);


	drawModel(M_GLASS);

	glDisable(GL_ALPHA_TEST);	// glass is "really" blending
	glPopAttrib(); // GL_BLEND
	glPopAttrib(); // GL_NORMALIZE
	glEnable(GL_COLOR_MATERIAL);		// Allow color as material
}

void Model::drawModel(MAT_MODE m_mode) 
{
	int lindex;
	
	if (m_mode == M_NORMAL)
		lindex = 0;
	else if (m_mode == M_ENVMAP)
		lindex = 1;
	else if (m_mode == M_GLASS)
		lindex = 2;
	
	if (List[lindex] != 0)
		// take the precomputed display list
		glCallList(List[lindex]);
	else
	{
		List[lindex] = glGenLists(1); 
		glNewList(List[lindex],GL_COMPILE);							// Start With The Box List
		
		GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );
		
		bool drawMesh = false;
		
		// Draw by group and material mode
		for ( int i = 0; i < m_numMeshes; i++ )
		{
			drawMesh = false;
			
			int materialIndex = m_pMeshes[i].m_materialIndex;
			if ( materialIndex >= 0 )
			{
				glMaterialfv( GL_FRONT, GL_AMBIENT, m_pMaterials[materialIndex].m_ambient );
				glMaterialfv( GL_FRONT, GL_DIFFUSE, m_pMaterials[materialIndex].m_diffuse );
				glMaterialfv( GL_FRONT, GL_SPECULAR, m_pMaterials[materialIndex].m_specular );
				glMaterialfv( GL_FRONT, GL_EMISSION, m_pMaterials[materialIndex].m_emissive );
				glMaterialf( GL_FRONT, GL_SHININESS, m_pMaterials[materialIndex].m_shininess );
				
				if ( m_pMaterials[materialIndex].m_texture > 0 )
				{
					// look if material has a "envmap"-texture!
					if (strstr(m_pMaterials[materialIndex].m_pTextureFilename, "envmap"))
					{
						if (m_mode == M_ENVMAP)
							drawMesh = true;
					}
					// look if material has a "glass"-texture that needs blending!
					else if (strstr(m_pMaterials[materialIndex].m_pTextureFilename, "glass"))
					{
						if (m_mode == M_GLASS)
							drawMesh = true;
					}
					else
					{
						if (m_mode == M_NORMAL)
							drawMesh = true;
					}
					
					if (drawMesh)
					{
						glBindTexture( GL_TEXTURE_2D, m_pMaterials[materialIndex].m_texture );
						glEnable( GL_TEXTURE_2D );
						
						if (m_mode == M_ENVMAP)
						{
							// here we force an environment mapping
							glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
							glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
							glEnable(GL_TEXTURE_GEN_S);
							glEnable(GL_TEXTURE_GEN_T);
							glEnable(GL_AUTO_NORMAL);
						}
					}
				}
				else
				{
					glDisable( GL_TEXTURE_2D );
					if (m_mode == M_NORMAL)
						drawMesh = true;
				}
			}
			else
			{
				// Material properties?
				glDisable( GL_TEXTURE_2D );
				if (m_mode == M_NORMAL)
					drawMesh = true;
			}
			
			if (drawMesh)
			{
				glBegin( GL_TRIANGLES );
				{
					for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
					{
						int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
						const Triangle* pTri = &m_pTriangles[triangleIndex];
						
						for ( int k = 0; k < 3; k++ )
						{
							int index = pTri->m_vertexIndices[k];
							
							glNormal3fv( pTri->m_vertexNormals[k] );
							if (m_mode != M_ENVMAP)
								glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
							glVertex3fv( m_pVertices[index].m_location );
						}
					}
				}
				glEnd();
			}
		}
		
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		
		if ( texEnabled )
			glEnable( GL_TEXTURE_2D );
		else
			glDisable( GL_TEXTURE_2D );
		
		glEndList();
	}
}

// 2002/11/01: fixed bug under Linux depending on backslash insertions
void Model::reloadTextures()
{
	for ( int i = 0; i < m_numMaterials; i++ )
	{
		if ( strlen( m_pMaterials[i].m_pTextureFilename ) > 0 )
		{
			char *comment = strrchr(m_pMaterials[i].m_pTextureFilename, '\\');
			if (comment)
				comment++;
			else
				comment = m_pMaterials[i].m_pTextureFilename;
			char texfile[80];
			sprintf(texfile, "%s%s", modeldir, comment);
			m_pMaterials[i].m_texture = LoadGLTexture(texfile);
		}
		else
			m_pMaterials[i].m_texture = 0;
	}
}

BorgCube::BorgCube()
{
	bcube.init(1000.0f, 1000.0f, "Data/houses/borgcube.tga",0,0,0);
	this->SetSize(1000.0f);
}

BorgCube::~BorgCube()
{
}

bool BorgCube::loadModelData( const char *dir, const char *filename )
{
	return true;
}

/**
*	Draw the model with display list if possible else call drawModel
* @param camera			The camera that looks at model
*/
void BorgCube::draw(C3DCamera *camera)
{
	glEnable(GL_NORMALIZE);
	glDisable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
	// Multiply the current view matrix by the model matrix and draw the model
	CDoubleMatrix m = GetScaledModelMatrix(camera);
	C3DBase *cp = C3DBase::GetCommonParent(this, camera);
	// VERY vERY verY important!!!!!! do matrix multiplication by yourself!!!!
	glLoadMatrixd(camera->GetCameraMatrix(cp)*m);

	int distance = (int)Clamp(0,INT_MAX,(this->GetFullPosition(cp) - camera->GetFullPosition(cp)).Magnitude() );
	bcube.draw(distance);
	glDisable(GL_TEXTURE_2D);
}


// Here comes a concrete version of a model class that handles Milkshape models
MilkshapeModel::MilkshapeModel()
{
}

MilkshapeModel::~MilkshapeModel()
{
}

bool MilkshapeModel::loadModelData( const char *dir, const char *filename )
{
	char thefile[80];
	sprintf(modeldir, "%s", dir);
	sprintf(thefile, "%s%s", modeldir, filename);

	class ifstream inputFile( thefile, ios::in | ios::binary | ios::nocreate );
	if ( inputFile.fail())
		return false;	// "Couldn't open the model file."

	inputFile.seekg( 0, ios::end );
	long fileSize = inputFile.tellg();
	inputFile.seekg( 0, ios::beg );

	byte *pBuffer = new byte[fileSize];
	inputFile.read( pBuffer, fileSize );
	inputFile.close();

	const byte *pPtr = pBuffer;
	MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
	pPtr += sizeof( MS3DHeader );

	if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
		return false; // "Not a valid Milkshape3D model file."

	if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
		return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr; 
	m_numVertices = nVertices;
	m_pVertices = new Vertex[nVertices];
	pPtr += sizeof( word );

	int i;
	for ( i = 0; i < nVertices; i++ )
	{
		MS3DVertex *pVertex = ( MS3DVertex* )pPtr;
		m_pVertices[i].m_boneID = pVertex->m_boneID;
		memcpy( m_pVertices[i].m_location, pVertex->m_vertex, sizeof( GLfloat )*3 );
		pPtr += sizeof( MS3DVertex );
	}

	int nTriangles = *( word* )pPtr;
	m_numTriangles = nTriangles;
	m_pTriangles = new Triangle[nTriangles];
	pPtr += sizeof( word );

	for ( i = 0; i < nTriangles; i++ )
	{
		MS3DTriangle *pTriangle = ( MS3DTriangle* )pPtr;
		int vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
		GLfloat t[3] = { 1.0f-pTriangle->m_t[0], 1.0f-pTriangle->m_t[1], 1.0f-pTriangle->m_t[2] };
		memcpy( m_pTriangles[i].m_vertexNormals, pTriangle->m_vertexNormals, sizeof( GLfloat )*3*3 );
		memcpy( m_pTriangles[i].m_s, pTriangle->m_s, sizeof( GLfloat )*3 );
		memcpy( m_pTriangles[i].m_t, t, sizeof( GLfloat )*3 );
		memcpy( m_pTriangles[i].m_vertexIndices, vertexIndices, sizeof( int )*3 );
		pPtr += sizeof( MS3DTriangle );
	}

	int nGroups = *( word* )pPtr;
	m_numMeshes = nGroups;
	m_pMeshes = new Mesh[nGroups];
	pPtr += sizeof( word );
	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		word nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		int *pTriangleIndices = new int[nTriangles];
		for ( int j = 0; j < nTriangles; j++ )
		{
			pTriangleIndices[j] = *( word* )pPtr;
			pPtr += sizeof( word );
		}

		char materialIndex = *( char* )pPtr;
		pPtr += sizeof( char );
	
		m_pMeshes[i].m_materialIndex = materialIndex;
		m_pMeshes[i].m_numTriangles = nTriangles;
		m_pMeshes[i].m_pTriangleIndices = pTriangleIndices;
	}

	int nMaterials = *( word* )pPtr;
	m_numMaterials = nMaterials;
	m_pMaterials = new Material[nMaterials];
	pPtr += sizeof( word );
	for ( i = 0; i < nMaterials; i++ )
	{
		MS3DMaterial *pMaterial = ( MS3DMaterial* )pPtr;
		memcpy( m_pMaterials[i].m_ambient, pMaterial->m_ambient, sizeof( GLfloat )*4 );
		memcpy( m_pMaterials[i].m_diffuse, pMaterial->m_diffuse, sizeof( GLfloat )*4 );
		memcpy( m_pMaterials[i].m_specular, pMaterial->m_specular, sizeof( GLfloat )*4 );
		memcpy( m_pMaterials[i].m_emissive, pMaterial->m_emissive, sizeof( GLfloat )*4 );
		m_pMaterials[i].m_shininess = pMaterial->m_shininess;
		m_pMaterials[i].m_pTextureFilename = new char[strlen( pMaterial->m_texture )+1];
		strcpy( m_pMaterials[i].m_pTextureFilename, pMaterial->m_texture );
		pPtr += sizeof( MS3DMaterial );
	}

	reloadTextures();

	delete[] pBuffer;

	return true;
}



