/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	Model.h                                         *
 *                                                          *
 * Content: Header file for some basic 3D model classes     *
 *          (Milkshape, 3ds, obj, ...)                      *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#ifndef JWMODEL_H
#define JWMODEL_H

#include <qgl.h>
#include <cmath>
#include "JWObject.h"
#include "JWStaticObject.h"
#include "Texture.h"

inline GLuint LoadGLTexture(char *filename )						// Load Bitmaps And Convert To Textures
{
	GLuint texturem = 0;

  TEXTURE textemp;
  if (strstr(filename, "glass"))
	textemp.LoadImage(filename, true); // load a bmp, tga, jpeg texture
  else
	textemp.LoadImage(filename, false); // load a bmp, tga, jpeg texture

  textemp.activate(TEXTURE::T_NORMAL);
  texturem = textemp.texID;
  textemp.texID = 0;  // to avoid that destructor deletes the texture!

	return texturem;
};

class Model : public C3DObject
{
	typedef enum MAT_MODE { M_NORMAL, M_ENVMAP, M_GLASS };

	public:
		GLuint List[3];	// the number for the display list entry of this model
		// relative directory of this model (needed for textures)
		char modeldir[80];
		CDoubleMatrix scaleMat;
		// deprecated: float	baserotation; // angle (in degrees we have to flip the model)
		//	Mesh
		struct Mesh
		{
			int m_materialIndex;
			int m_numTriangles;
			int *m_pTriangleIndices;
		};

		//	Material properties
		struct Material
		{
			GLfloat m_ambient[4], m_diffuse[4], m_specular[4], m_emissive[4];
			GLfloat m_shininess;
			GLuint m_texture;
			char *m_pTextureFilename;
		};

		//	Triangle structure
		struct Triangle
		{
			GLfloat m_vertexNormals[3][3];
			GLfloat m_s[3], m_t[3];
			int m_vertexIndices[3];
		};

		//	Vertex structure
		struct Vertex
		{
			char m_boneID;	// for skeletal animation
			GLfloat m_location[3];
		};

	public:
		/*	Constructor. */
		Model();

		/*	Destructor. */
		virtual ~Model();

		/*	
			Load the model data into the private variables. 
				filename			Model filename
		*/
		virtual bool loadModelData( const char *dir, const char *filename ) = 0;

		float GetSize();
		void SetSize(float len);

		/**
		*	Draw the model with display list if possible else call drawModel
		* @param camera			The camera that looks at model
		*/
		void draw(C3DCamera *camera);

		/*
			Draw the model, internally uses display lists when possible!
		*/
		void drawModel(MAT_MODE m_mode);

		/*
			Called if OpenGL context was lost and we need to reload textures, display lists, etc.
		*/
		void reloadTextures();

	protected:
		//	Meshes used
		int m_numMeshes;
		Mesh *m_pMeshes;

		//	Materials used
		int m_numMaterials;
		Material *m_pMaterials;

		//	Triangles used
		int m_numTriangles;
		Triangle *m_pTriangles;

		//	Vertices Used
		int m_numVertices;
		Vertex *m_pVertices;
};

class BorgCube : public Model
{
public:
	JWCube bcube;

	BorgCube();
	virtual ~BorgCube();

	virtual void draw(C3DCamera *);
	virtual bool loadModelData( const char *dir, const char *filename );
};



class MilkshapeModel : public Model
{
	public:
		/*	Constructor. */
		MilkshapeModel();

		/*	Destructor. */
		virtual ~MilkshapeModel();

		/*	
			Load the model data into the private variables. 
				filename			Model filename
		*/
		virtual bool loadModelData( const char *dir, const char *filename );
};

/* 
	MS3D STRUCTURES 
*/

// byte-align structures
#ifdef _MSC_VER
#	pragma pack( push, packing )
#	pragma pack( 1 )
#	define PACK_STRUCT
#elif defined( __GNUC__ )
#	define PACK_STRUCT	__attribute__((packed))
#else
#	error you must byte-align these structures with the appropriate compiler directives
#endif

typedef unsigned char byte;
typedef unsigned short word;

// File header
struct MS3DHeader
{
	char m_ID[10];
	int m_version;
} PACK_STRUCT;

// Vertex information
struct MS3DVertex
{
	byte m_flags;
	GLfloat m_vertex[3];
	char m_boneID;
	byte m_refCount;
} PACK_STRUCT;

// Triangle information
struct MS3DTriangle
{
	word m_flags;
	word m_vertexIndices[3];
	GLfloat m_vertexNormals[3][3];
	GLfloat m_s[3], m_t[3];
	byte m_smoothingGroup;
	byte m_groupIndex;
} PACK_STRUCT;

// Material information
struct MS3DMaterial
{
    char m_name[32];
    GLfloat m_ambient[4];
    GLfloat m_diffuse[4];
    GLfloat m_specular[4];
    GLfloat m_emissive[4];
    GLfloat m_shininess;	// 0.0f - 128.0f
    GLfloat m_transparency;	// 0.0f - 1.0f
    byte m_mode;	// 0, 1, 2 is unused now
    char m_texture[128];
    char m_alphamap[128];
} PACK_STRUCT;

//	Joint information
struct MS3DJoint
{
	byte m_flags;
	char m_name[32];
	char m_parentName[32];
	GLfloat m_rotation[3];
	GLfloat m_translation[3];
	word m_numRotationKeyframes;
	word m_numTranslationKeyframes;
} PACK_STRUCT;

// Keyframe data
struct MS3DKeyframe
{
	GLfloat m_time;
	GLfloat m_parameter[3];
} PACK_STRUCT;

// Default alignment
#ifdef _MSC_VER
#	pragma pack( pop, packing )
#endif

#undef PACK_STRUCT

#endif



