/*
	The Universe Development Kit
	Copyright (C) 2000  Sean O'Neil
	s_p_oneil@hotmail.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __Noise_h__
#define __Noise_h__

#include <stdlib.h>
#include <cmath>
#include <cfloat>

// Defines
#define PI					3.1415926535898			// PI
#define HALF_PI				1.5707963267949		// PI / 2
#define TWO_PI				6.28318530718			// PI * 2
#define LOGHALF				-0.6931471805599	// log(0.5)
#define LOGHALFI			-1.442695040889		// Inverse of log(0.5)
#define DELTA				1e-6								// Small number for comparing floating point numbers
#define MAX_DIMENSIONS		4							// Maximum number of dimensions in a noise object
#define MAX_OCTAVES			128							// Maximum # of octaves in an fBm object

// Macros
#define SQUARE(a)			((a) * (a))
#define FLOOR(a)			((int)(a) - ((a) < 0 && (a) != (int)(a)))
#define CEILING(a)			((int)(a) + ((a) > 0 && (a) != (int)(a)))
#define MIN(a, b)			((a) < (b) ? (a) : (b))
#define MAX(a, b)			((a) > (b) ? (a) : (b))
#define ABS(a)				((a) < 0 ? -(a) : (a))
#define CLAMP(a, b, x)		((x) < (a) ? (a) : ((x) > (b) ? (b) : (x)))
#define LERP(a, b, x)		((a) + (x) * ((b) - (a)))
#define CUBIC(a)			((a) * (a) * (3 - 2*(a)))
#define STEP(a, x)			((x) >= (a))
#define BOXSTEP(a, b, x)	Clamp(0, 1, ((x)-(a))/((b)-(a)))
#define PULSE(a, b, x)		(((x) >= (a)) - ((x) >= (b)))
#define GAMMA(a, g)			pow(a, 1/g)
#define BIAS(a, b)			pow(a, log(b) * LOGHALFI)
#define EXPOSE(l, k)		(1 - exp(l * k))
#define DEGTORAD(x)			((x) * 0.01745329251994)
#define RADTODEG(x)			((x) * 57.29577951308)
#define SWAP(a, b, t)		t = a; a = b; b = t

// Inline functions (use instead of macros to avoid performing slow operations twice)
inline double Square(double a)						{ return a * a; }
inline int Floor(double a)							{ return ((int)a - (a < 0 && a != (int)a)); }
inline int Ceiling(double a)						{ return ((int)a + (a > 0 && a != (int)a)); }
inline double Min(double a, double b)				{ return (a < b ? a : b); }
inline double Max(double a, double b)				{ return (a > b ? a : b); }
inline double Abs(double a)							{ return (a < 0 ? -a : a); }
inline double Clamp(double a, double b, double x)	{ return (x < a ? a : (x > b ? b : x)); }
inline double Lerp(double a, double b, double x)	{ return a + x * (b - a); }
inline double Cubic(double a)						{ return a * a * (3 - 2*a); }
inline double Step(double a, double x)				{ return (double)(x >= a); }
inline double Boxstep(double a, double b, double x)	{ return Clamp(0, 1, (x-a)/(b-a)); }
inline double Pulse(double a, double b, double x)	{ return (double)((x >= a) - (x >= b)); }
inline double Gamma(double a, double g)				{ return (double)pow(a, 1/g); }
inline double Bias(double a, double b)				{ return (double)pow(a, log(b) * LOGHALFI); }
inline double Expose(double l, double k)			{ return (double)(1 - exp(-l * k)); }

inline unsigned int exp2(unsigned int x) { return (1 << x); };

inline int log2(unsigned int x)
{
	int log = -1;	// special case for log2(0)
	while (x != 0) { x >>= 1; log++; }
	return log;
};

// simple mathematical help routines

inline void rescaleVector(double *x, double *y, double *z, double len)
{
	// rescale vector (xyz) to lenght len.
	len /= (double)sqrt((*x)*(*x) + (*y)*(*y) + (*z)*(*z));
	*x = len*(*x);
	*y = len*(*y);
	*z = len*(*z);
};	

inline void PolarToCoordinates(double *x, double *y, double *z, double alpha, double beta, double h)
{
	*x=(double)(h*sin(beta)*sin(alpha));
	*y=(double)(h*cos(beta));
	*z=(double)(h*sin(beta)*cos(alpha));
};

inline void CoordinatesToPolar(double x, double y, double z, double *alpha, double *beta, double *h)
{
	// compute height above planet ball (in m)
	*h = (double)sqrt(x*x + y*y + z*z);
	// compute angle alpha in xz-plane [-Pi,Pi]
	*alpha = (double)atan2(x,z);
	// compute angle against vertical y-axis [0,Pi]
	*beta = (double)atan2(sqrt(x*x+z*z),y);
};	

inline int f4(int a, int b, int c, int d) {
	return (a + b + c + d)/4;
}

inline int f2(int a, int b) {
	return (a + b)/2;
}

inline int index(int x, int y, int w, int h)
{
	// (x,y) must lie in [0,w-1]*[0,h-1]
	// we have h rows (y) of w bytes (x)
	return ( (x & (w-1)) + (y & (h-1))*w );
}

inline double Gain(double a, double b)
{
	if(a <= DELTA)
		return 0;
	if(a >= 1-DELTA)
		return 1;

	register double p = (double)(log(1 - b) * LOGHALFI);
	if(a < 0.5)
		return (double)(pow(2 * a, p) * 0.5);
	else
		return (double)(1 - pow(2 * (1 - a), p) * 0.5);
}

inline double Smoothstep(double a, double b, double x)
{
	if(x <= a)
		return 0;
	if(x >= b)
		return 1;
	return Cubic((x - a) / (b - a));
}

inline double Mod(double a, double b)
{
	a -= ((int)(a / b)) * b;
	if(a < 0)
		a += b;
	return a;
}

inline int Mod(int a, int b)
{
	a -= ((int)(a / b)) * b;
	if(a < 0)
		a += b;
	return a;
}

inline void Normalize(double *f, int n)
{
	double fMagnitude = 0;
  int i;
	for(i=0; i<n; i++)
		fMagnitude += f[i]*f[i];
	fMagnitude = (double)(1 / sqrt(fMagnitude));
	for(i=0; i<n; i++)
		f[i] *= fMagnitude;
}

/*******************************************************************************
* Class: CRandom
********************************************************************************
* This class wraps a random number generator. I plan to implement my own random
* number generator so I can keep the seeds as member variables (which is more
* flexible than using statics or globals). I was using one I found on the
* Internet implemented in assembler, but I was having problems with it so I
* removed it for this demo.
*******************************************************************************/
class CRandom
{
public:
	CRandom()						{}
	CRandom(unsigned int nSeed)		{ Init(nSeed); }
	void Init(unsigned int nSeed)	{ srand(nSeed); }
	inline double Random()			{ return (double)rand()/(double)RAND_MAX; }

	inline double RandomD(double dMin, double dMax)
	{
		double dInterval = dMax - dMin;
		double d = dInterval * Random();
		return dMin + MIN(d, dInterval);
	}

	inline unsigned int RandomI(unsigned int nMin, unsigned int nMax)
	{
		unsigned int nInterval = nMax - nMin;
		unsigned int i = (unsigned int)((nInterval+1.0) * Random());
		return nMin + MIN(i, nInterval);
	}
};

/*******************************************************************************
* Class: CNoise
********************************************************************************
* This class implements the Perlin noise function. Initialize it with the number
* of dimensions (1 to 4) and a random seed. I got the source for the first 3
* dimensions from "Texturing & Modeling: A Procedural Approach". I added the
* extra dimension because it may be desirable to use 3 spatial dimensions and
* one time dimension. The noise buffers are set up as member variables so that
* there may be several instances of this class in use at the same time, each
* initialized with different parameters.
*******************************************************************************/
class CNoise
{
protected:
	int m_nDimensions;						// Number of dimensions used by this object
	unsigned char m_nMap[256];				// Randomized map of indexes into buffer
	double m_nBuffer[256][MAX_DIMENSIONS];	// Random n-dimensional buffer

	double Lattice(int ix, double fx, int iy=0, double fy=0, int iz=0, double fz=0, int iw=0, double fw=0)
	{
		int n[4] = {ix, iy, iz, iw};
		double f[4] = {fx, fy, fz, fw};
		int nIndex = 0;
    int i;
		for(i=0; i<m_nDimensions; i++)
			nIndex = m_nMap[(nIndex + n[i]) & 0xFF];
		double fValue = 0;
		for(i=0; i<m_nDimensions; i++)
			fValue += m_nBuffer[nIndex][i] * f[i];
		return fValue;
	}

public:
	CNoise()	{}
	CNoise(int nDimensions, unsigned int nSeed)	{ Init(nDimensions, nSeed); }
	void Init(int nDimensions, unsigned int nSeed);
	double Noise(double *f);
};

/*******************************************************************************
* Class: CFractal
********************************************************************************
* This class implements fBm, or fractal Brownian motion. Since fBm uses Perlin
* noise, this class is derived from CNoise. Initialize it with the number of
* dimensions (1 to 4), a random seed, H (roughness ranging from 0 to 1), and
* the lacunarity (2.0 is often used). Many of the fractal routines came from
* "Texturing & Modeling: A Procedural Approach". fBmTest() is my own creation,

* and I created it to generate my first planet.
*******************************************************************************/
class CFractal : public CNoise
{
protected:
	double m_fH;
	double m_fLacunarity;
	double m_fExponent[MAX_OCTAVES];

public:
	CFractal()	{}
	CFractal(int nDimensions, unsigned int nSeed, double fH, double fLacunarity)
	{
		Init(nDimensions, nSeed, fH, fLacunarity);
	}
	void Init(int nDimensions, unsigned int nSeed, double fH, double fLacunarity)
	{
		CNoise::Init(nDimensions, nSeed);
		m_fH = fH;
		m_fLacunarity = fLacunarity;
		double f = 1;
		for(int i=0; i<MAX_OCTAVES; i++) 
		{
			m_fExponent[i] = (double)pow(f, -m_fH);
			f *= m_fLacunarity;
		}
	}
	double fBm(double *f, double fOctaves);
	double Turbulence(double *f, double fOctaves);
	double Multifractal(double *f, double fOctaves, double fOffset);
	double Heterofractal(double *f, double fOctaves, double fOffset);
	double HybridMultifractal(double *f, double fOctaves, double fOffset, double fGain);
	double RidgedMultifractal(double *f, double fOctaves, double fOffset, double fThreshold);
	double fBmTest(double *f, int nStart, int nEnd, double fInitial=0.0);
	double SinusFun(double, double, double);
};

#endif // __Noise_h__



