/************************************************************
 * Infinite Universe Engine:				                *
 *                                                          *
 * Interactive simulation of a solar system and             *
 * real-time rendering of planetary bodies with OpenGL      *
 * copyright  2002 Jochen Winzen                           *
 * winzen@online.de                                         *
 * -------------------------------------------------------- *
 * File:	Texture.cpp                                     *
 *                                                          *
 * Content: Implementation of my own texture class with     *
 *          special functions like mipmap-blending,         *
 *          transparency... supports bmp, tga, jpeg         *
 *                                                          *
 * Author:  Jochen Winzen                                   *
 * Version: 2002/12/08                                      *
 ************************************************************
 */
#include <iostream>
#include <fstream>
#include <string>
#include "Texture.h"
#include "glext.h"
#include "Noise.h"
#include <qapplication.h>
#include <qimage.h>
#include <qcolor.h>

//Global variables
TGAHeader tgaheader;						// TGA header
TGA tga;									// TGA image data

GLubyte uTGAcompare[12] = {0,0,2, 0,0,0,0,0,0,0,0,0};	// Uncompressed TGA Header
GLubyte cTGAcompare[12] = {0,0,10,0,0,0,0,0,0,0,0,0};	// Compressed TGA Header

// TEXTURE CLASS DEFINITION
TEXTURE::TEXTURE() {
	width = 0;
	height = 0;
	imageData = 0;
	texID = 0;
	keepRawData = false;
}

// w,h MUST be powers of 2
void TEXTURE::init(int w, int h, GLuint	ptype) {
	width = w;
	height = h;
	type = ptype;
	switch (type)
	{
	case GL_ALPHA:
	case GL_LUMINANCE:
	case GL_INTENSITY:
		bpp = 8; break;
			
	case GL_LUMINANCE_ALPHA:	bpp = 16; break;
	case GL_RGB:		bpp = 24; break;
	case GL_RGBA:		bpp = 32; break;
	}
	int byte_count = bpp/8;

	if(imageData) 
	{
		delete[] imageData;
		imageData = 0;
	}
	imageData = new unsigned char[(width * height)*byte_count];
}

// w,h MUST be powers of 2
void TEXTURE::createFractal(int w, int h, GLuint ptype) {
	int i;
	int x, z;
	int bsize, csize;
	int r;
	int c;
	int byte_count;
	int *hf2 = new int[w*h];

	width = w;
	height = h;
	type = ptype;

	switch (type)
	{
	case GL_ALPHA:
	case GL_LUMINANCE:
	case GL_INTENSITY:
		bpp = 8; break;
			
	case GL_LUMINANCE_ALPHA:	bpp = 16; break;
	case GL_RGB:		bpp = 24; break;
	case GL_RGBA:		bpp = 32; break;
	}
	byte_count = bpp/8;

	if(!imageData) 
		imageData = new unsigned char[(width * height)*byte_count];

	for(i=0; i<(int)(width * height * byte_count); i++) {
		imageData[i] = (unsigned char) (255); 
	}

	for(i=0; i<(int)(width*height); i++) {
		hf2[i] = 0;
	}

	r = 512; // height values range

	bsize = w;

	csize = bsize/2;

	while(csize > 0) {
	
		for(x=0; x<(int)width; x+=bsize) {
			for(z=0; z<(int)height; z+=bsize) {			
				hf2[index(x+csize, z+csize, w, h)] = f4(hf2[index(x, z,w,h)],
												 hf2[index(x+bsize, z,w,h)],
												 hf2[index(x+bsize, z+bsize,w,h)],
												 hf2[index(x, z+bsize,w,h)]) + rand()%(r+1) - r/2;
			}
		} 

		for(x=0; x<(int)width; x+=bsize) {
			for(z=0; z<(int)height; z+=bsize) {			
				hf2[index(x+csize, z,w,h)] += f4(hf2[index(x, z,w,h)],
											hf2[index(x+bsize, z,w,h)],
											hf2[index(x+csize, z-csize,w,h)],
											hf2[index(x+csize, z+csize,w,h)]) + rand()%(r+1) - r/2;
				hf2[index(x, z+csize,w,h)] += f4(hf2[index(x, z,w,h)],
											hf2[index(x+csize, z+csize,w,h)],
											hf2[index(x, z+bsize,w,h)],
											hf2[index(x-csize, z+csize,w,h)]) + rand()%(r+1) - r/2;
			}
		} 

		if(csize<16) {
			r = r/2;
		}
			
		bsize /= 2;
		csize = bsize/2;
	}; 

	for(x=0; x<(int)(width*height); x++) {
		if(hf2[x] < 0) 
			hf2[x] /= 2;
	} 

	x = width*height;

	for(i=0; i<x; i++) {
		c = (220 + (hf2[i] - hf2[i+1]));

		if(c<0) c = 0;
		else if(c>255) c = 255;

		imageData[i*byte_count] = c;
		if (byte_count > 1)
		{
			imageData[i*byte_count+1] = c;
			imageData[i*byte_count+2] = c;
			if (byte_count == 4)
				imageData[i*byte_count+3] = 255;
		}
	}

	imageData[(x-1)*byte_count] = 255;
	if (byte_count > 1)
	{
		imageData[(x-1)*byte_count+1] = 255;
		imageData[(x-1)*byte_count+2] = 255;
		if (byte_count == 4)
			imageData[(x-1)*byte_count+3] = 255;
	}
}

TEXTURE::~TEXTURE()
{
	if (imageData)
	{
		delete[] imageData;
		imageData = 0;
	}
	glDeleteTextures(1, &texID);
}

void TEXTURE::build2DMipmaps(GLint type, int width, int height, GLubyte *imageData, TEX_MIPMAP mode)
{
  
		int levels = log2((unsigned int)MIN(width, height));
    int i,w,h,bypp;
    
    switch (type)
    {
    case GL_ALPHA:
    case GL_LUMINANCE:
    case GL_INTENSITY:
      bypp = 1; break;
      
    case GL_LUMINANCE_ALPHA:	bypp = 2; break;
    case GL_RGB:		bypp = 3; break;
    case GL_RGBA:		bypp = 4; break;
    }
    
    switch (mode)
    {
      
    case MIPMAP_NORMAL:
      
      gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
      break;
      
    case MIPMAP_WHITE:
    case MIPMAP_CHAOS:
      
    	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	    glPixelStorei(GL_PACK_ALIGNMENT, 1);
      glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
      // we make our own mipmaps that fade to white!
      GLubyte *masterData = imageData;
      w = width;
      h = height;
      GLubyte *mipmapData;
      float value;
      float a = 0.0625f;
      float b = 0.125f;
      
      for (i=1; i <= levels; i++)
      {
        w /= 2;
        h /= 2;
        mipmapData = new GLubyte[w*h*bypp];
        
        for (int y = 0; y < h; y++)
        {
          for (int x = 0; x < w; x++)
          {
            for (int n = 0; n < bypp; n++)
            {
              if (mode == MIPMAP_WHITE)
              {
              // "normal" summation of 4 pixels to replace with medium value
              value = (
                masterData[(2*y+0)*(2*w*bypp) + (2*x+0)*bypp + n]*b +
                masterData[(2*y+1)*(2*w*bypp) + (2*x+0)*bypp + n]*b +
                masterData[(2*y+0)*(2*w*bypp) + (2*x+1)*bypp + n]*b +
                masterData[(2*y+1)*(2*w*bypp) + (2*x+1)*bypp + n]*b
                );
              // add surround region values (simple gaussian filter)
              // 0aa0
              // abba   a=1/16, b=1/8 -> total weight = 1
              // abba
              // 0aa0
              value += (
                masterData[(2*y+0)*(2*w*bypp) + Mod(2*x-1, 2*w)*bypp + n]*a +
                masterData[(2*y+1)*(2*w*bypp) + Mod(2*x-1, 2*w)*bypp + n]*a +
                masterData[(2*y+0)*(2*w*bypp) + Mod(2*x+2, 2*w)*bypp + n]*a +
                masterData[(2*y+1)*(2*w*bypp) + Mod(2*x+2, 2*w)*bypp + n]*a
                );
              value += (
                masterData[Mod(2*y-1,2*h)*(2*w*bypp) + (2*x+0)*bypp + n]*a +
                masterData[Mod(2*y+2,2*h)*(2*w*bypp) + (2*x+0)*bypp + n]*a +
                masterData[Mod(2*y-1,2*h)*(2*w*bypp) + (2*x+1)*bypp + n]*a +
                masterData[Mod(2*y+2,2*h)*(2*w*bypp) + (2*x+1)*bypp + n]*a
                );
              }
              else
              {
                // primitive pixel zoom
                value = masterData[(2*y+0)*(w*bypp) + (2*x+0)*bypp + n];
              }
              // here we linearly fade to white!
              value = Cubic(i/(float)levels)*255 + Cubic(1 - i/(float)levels)*value;
              mipmapData[y*(w*bypp) + x*bypp + n] = (GLubyte) value;
            }
          }
        }
        
      	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	      glPixelStorei(GL_PACK_ALIGNMENT, 1);
        glTexImage2D(GL_TEXTURE_2D, i, type, w, h, 0, type, GL_UNSIGNED_BYTE, mipmapData);
        
        if (masterData != imageData)
          delete[] masterData;
        masterData = mipmapData;
      }
      delete[] mipmapData;
      
      break;
    }
};

void TEXTURE::activate(TEX_MODE mode) {
	// very important line for my user defined mipmaps!
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);

	// NEW Do Texture Initialisation Stuff here!
	switch(mode)
	{
	case T_NORMAL:
		{
			glGenTextures(1, &texID);				// Create The Texture
			glBindTexture(GL_TEXTURE_2D, texID);	// assign texID
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
			//glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
			gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
		}
		break;
		
	case T_CUBEMAP:
		{
			glGenTextures(1, &texID);				// Create The Texture
			glBindTexture(GL_TEXTURE_2D, texID);	// assign texID
			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
			//glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
			gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
		}
		break;

  case T_TERRAIN:
		{
			glGenTextures(1, &texID);
			glBindTexture(GL_TEXTURE_2D, texID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			//glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT); 
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			//glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_ADD_SIGNED_EXT); // seems to look a little bit nicer 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
			//glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
			//gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
			TEXTURE::build2DMipmaps(type, width, height, imageData, TEXTURE::MIPMAP_NORMAL);
		}
		break;
		
	case T_DETAIL:
		{
			glGenTextures(1, &texID);
			glBindTexture(GL_TEXTURE_2D, texID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			//glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT); 
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			//glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_ADD_SIGNED_EXT); // seems to look a little bit nicer 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
			//glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
			//gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
			TEXTURE::build2DMipmaps(type, width, height, imageData, TEXTURE::MIPMAP_WHITE);
		}
		break;

	case T_CHAOS:
		{
			glGenTextures(1, &texID);
			glBindTexture(GL_TEXTURE_2D, texID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			//glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT); 
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			//glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_ADD_SIGNED_EXT); // seems to look a little bit nicer 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
			//glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
			//gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData);
			TEXTURE::build2DMipmaps(type, width, height, imageData, TEXTURE::MIPMAP_CHAOS);
		}
		break;

  case T_NOMIPMAP:
		{
			glGenTextures(1, &texID);				// Create The Texture
			glBindTexture(GL_TEXTURE_2D, texID);	// assign texID
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData);
		}
		break;
		
	}
	// only delete imageData if we do not need to "keepRawData" of the image!
	if (!keepRawData && imageData) {
		delete[] imageData;
		imageData = NULL;
	}
}

// simple adapter that chosses the routines according to file format
bool TEXTURE::LoadImage(char *Filename, bool transparent)
{
	if (!Filename)											// Make Sure A Filename Was Given
		return false;											// If Not Return NULL
	std::string fname = Filename;
	std::string format = fname.substr(fname.length()-3, fname.length());
	if (strcmp(format.c_str(),"bmp") == 0)
		return this->LoadBMP(Filename, transparent);
	else if (strcmp(format.c_str(), "tga") == 0)
		return this->LoadTGA(Filename);
	else if (strcmp(format.c_str(), "jpg") == 0)
		return this->LoadJPEG(Filename);
	return false;
}


// 2001/11/28 added transpanrency feature to overcome tga-limitation! color of upper left corner is used as transparent value!
// 2001/11/11 new method to load bmp-textures
// Load Bitmap And Convert To Texture
bool TEXTURE::LoadBMP(char *Filename, bool transparent)
{
  bool isGlass = (strstr(Filename, "glass") != NULL);

  QImage tex1, buf;
  if ( buf.load( Filename ) ) // Load image from file
  {
    tex1 = QGLWidget::convertToGLFormat( buf );  // flipped 32bit RGBA

    bpp = 32;             // set parameters of TEXTURE-object
    type = GL_RGBA;
    width = tex1.width();
    height = tex1.height();

    if(!imageData)
      imageData = new unsigned char[(width * height)*4];
    unsigned char *data = tex1.bits();

    unsigned char lr, lg, lb;	// upper-left-corner color value
    unsigned char tr, tg, tb;	// temporal comparison value
    lr = data[0];
    lg = data[1];
    lb = data[2];
    for(int i=0; i<(int)(width * height); i++)
    {
       tr = data[4*i];
       tg = data[4*i+1];
       tb = data[4*i+2];
       imageData[i*4] = tr;
       imageData[i*4+1] = tg;
       imageData[i*4+2] = tb;
	   if (isGlass)
         imageData[i*4+3] = Max(Abs(tr-lr), Max(Abs(tg-lg),Abs(tb-lb)) );
       else if (transparent)
        // hard cutoff tolerance
        imageData[i*4+3] = (Max(Abs(tr-lr), Max(Abs(tg-lg),Abs(tb-lb)) ) < 10 ? 0 : 255);
       else
         imageData[i*4+3] = 255;
    }
  }
  else
  {
    qWarning("Could not open texture file %s", Filename);
    return false;
  }

  return true;
}

/*
// WriteBMP()
// desc: takes image data and saves it into a 24-bit RGB .BMP file
//       with width X height dimensions
bool TEXTURE::WriteBMP(char *filename)
{

	FILE			 *filePtr;			// file pointer
	BITMAPFILEHEADER bitmapFileHeader;	// bitmap file header
	BITMAPINFOHEADER bitmapInfoHeader;	// bitmap info header
	int				 imageIdx;			// used for swapping RGB->BGR
	unsigned char	 tempRGB;			// used for swapping

	// open file for writing binary mode
	filePtr = fopen(filename, "wb");
	if (!filePtr)
		return false;

	// define the bitmap file header
	bitmapFileHeader.bfSize = sizeof(BITMAPFILEHEADER);
	bitmapFileHeader.bfType = 0x4D42;
	bitmapFileHeader.bfReserved1 = 0;
	bitmapFileHeader.bfReserved2 = 0;
	bitmapFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	
	// define the bitmap information header
	bitmapInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
	bitmapInfoHeader.biPlanes = 1;
	bitmapInfoHeader.biBitCount = 24;						// 24-bit
	bitmapInfoHeader.biCompression = BI_RGB;				// no compression
	bitmapInfoHeader.biSizeImage = width * abs(height) * 3;	// width * height * (RGB bytes)
	bitmapInfoHeader.biXPelsPerMeter = 0;
	bitmapInfoHeader.biYPelsPerMeter = 0;
	bitmapInfoHeader.biClrUsed = 0;
	bitmapInfoHeader.biClrImportant = 0;
	bitmapInfoHeader.biWidth = width;						// bitmap width
	bitmapInfoHeader.biHeight = height;						// bitmap height

	// switch the image data from RGB to BGR
	for (imageIdx = 0; imageIdx < bitmapInfoHeader.biSizeImage; imageIdx+=3)
	{
		tempRGB = imageData[imageIdx];
		imageData[imageIdx] = imageData[imageIdx + 2];
		imageData[imageIdx + 2] = tempRGB;
	}

	// write the bitmap file header
	fwrite(&bitmapFileHeader, 1, sizeof(BITMAPFILEHEADER), filePtr);

	// write the bitmap info header
	fwrite(&bitmapInfoHeader, 1, sizeof(BITMAPINFOHEADER), filePtr);

	// write the image data
	fwrite(imageData, 1, bitmapInfoHeader.biSizeImage, filePtr);

	// close our file
	fclose(filePtr);

	return true;
}

*/

/********************************************************************************
 * name :		LoadTGA(char * filename)							*
 * function:  Open and test the file to make sure it is a valid TGA file			*	
 * parems:	texture, pointer to a TEXTURE structure								*
 *    			filename, string pointing to file to open							*
 ********************************************************************************/

bool TEXTURE::LoadTGA(char * filename)				// Load a TGA file
{
	FILE * fTGA;												// File pointer to texture file
	fTGA = fopen(filename, "rb");				// Open file for reading

	if(fTGA == NULL)											// If it didn't open....
	{
		qWarning("Could not open texture file %s", filename);	// Display an error message
		return false;														// Exit function
	}

	if(fread(&tgaheader, sizeof(TGAHeader), 1, fTGA) == 0)					// Attempt to read 12 byte header from file
	{
		qWarning("Could not read file header");		// If it fails, display an error message
		if(fTGA != NULL)													// Check to seeiffile is still open
		{
			fclose(fTGA);													// If it is, close it
		}
		return false;														// Exit function
	}

	if(memcmp(uTGAcompare, &tgaheader, sizeof(tgaheader)) == 0)				// See if header matches the predefined header of 
	{																		// an Uncompressed TGA image
		LoadUncompressedTGA(filename, fTGA);						// If so, jump to Uncompressed TGA loading code
	}
	else if(memcmp(cTGAcompare, &tgaheader, sizeof(tgaheader)) == 0)		// See if header matches the predefined header of
	{																		// an RLE compressed TGA image
		LoadCompressedTGA(filename, fTGA);							// If so, jump to Compressed TGA loading code
	}
	else																	// If header matches neither type
	{
		qWarning("Invalid image: TGA file must be type 2 or type 10 ");	// Display an error
		fclose(fTGA);
		return false;																// Exit function
	}
	return true;															// All went well, continue on
}

bool TEXTURE::LoadUncompressedTGA(char * filename, FILE * fTGA)	// Load an uncompressed TGA (note, much of this code is based on NeHe's 
{																			// TGA Loading code nehe.gamedev.net)
	if(fread(tga.header, sizeof(tga.header), 1, fTGA) == 0)					// Read TGA header
	{										
		qWarning("Could not read info header");		// Display error
		if(fTGA != NULL)													// if file is still open
		{
			fclose(fTGA);													// Close it
		}
		return false;														// Return failular
	}	

	width  = tga.header[1] * 256 + tga.header[0];					// Determine The TGA Width	(highbyte*256+lowbyte)
	height = tga.header[3] * 256 + tga.header[2];					// Determine The TGA Height	(highbyte*256+lowbyte)
	bpp	= tga.header[4];										// Determine the bits per pixel
	tga.Width		= width;										// Copy width into local structure						
	tga.Height	= height;										// Copy height into local structure
	tga.Bpp			= bpp;											// Copy BPP into local structure

	if((width <= 0) || (height <= 0) || ((bpp != 24) && (bpp !=32)))	// Make sure all information is valid
	{
		qWarning("Invalid texture information");	// Display Error
		if(fTGA != NULL)													// Check if file is still open
		{
			fclose(fTGA);													// If so, close it
		}
		return false;														// Return failed
	}

	if(bpp == 24)													//If the BPP of the image is 24...
	{
		type	= GL_RGB;											// Set Image type to GL_RGB
	}
	else																	// Else if its 32 BPP
	{
		type	= GL_RGBA;											// Set image type to GL_RGBA
	}

	tga.bytesPerPixel	= (tga.Bpp / 8);									// Compute the number of BYTES per pixel
	tga.imageSize		= (tga.bytesPerPixel * tga.Width * tga.Height);		// Compute the total amout ofmemory needed to store data
	imageData	= new unsigned char[tga.imageSize];
	if(imageData == NULL)											// If no space was allocated
	{
		qWarning("Could not allocate memory for image");	// Display Error
		fclose(fTGA);														// Close the file
		return false;														// Return failed
	}

	if(fread(imageData, 1, tga.imageSize, fTGA) != tga.imageSize)	// Attempt to read image data
	{
		qWarning("Could not read image data");		// Display Error
		if (imageData) {
			delete[] imageData;
			imageData = 0;
		}
		fclose(fTGA);														// Close file
		return false;														// Return failed
	}

	// Byte Swapping Optimized By Steve Thomas
	unsigned char swapelem;
	for(GLuint cswap = 0; cswap < tga.imageSize; cswap += tga.bytesPerPixel)
	{
		swapelem = imageData[cswap+2];
		imageData[cswap+2] = imageData[cswap];
		imageData[cswap] = swapelem;		
	}

	fclose(fTGA);															// Close file
	return true;															// Return success

}

bool TEXTURE::LoadCompressedTGA(char * filename, FILE * fTGA)		// Load COMPRESSED TGAs
{ 
	if(fread(tga.header, sizeof(tga.header), 1, fTGA) == 0)					// Attempt to read header
	{
		qWarning("Could not read info header");		// Display Error
		if(fTGA != NULL)													// If file is open
		{
			fclose(fTGA);													// Close it
		}
		return false;														// Return failed
	}

	width  = tga.header[1] * 256 + tga.header[0];					// Determine The TGA Width	(highbyte*256+lowbyte)
	height = tga.header[3] * 256 + tga.header[2];					// Determine The TGA Height	(highbyte*256+lowbyte)
	bpp	= tga.header[4];										// Determine Bits Per Pixel
	tga.Width		= width;										// Copy width to local structure
	tga.Height	= height;										// Copy width to local structure
	tga.Bpp			= bpp;											// Copy width to local structure

	if((width <= 0) || (height <= 0) || ((bpp != 24) && (bpp !=32)))	//Make sure all texture info is ok
	{
		qWarning("Invalid texture information");	// If it isnt...Display error
		if(fTGA != NULL)													// Check if file is open
		{
			fclose(fTGA);													// Ifit is, close it
		}
		return false;														// Return failed
	}

	tga.bytesPerPixel	= (tga.Bpp / 8);									// Compute BYTES per pixel
	tga.imageSize		= (tga.bytesPerPixel * tga.Width * tga.Height);		// Compute amout of memory needed to store image
	imageData	= new unsigned char[tga.imageSize];

	if(imageData == NULL)											// If it wasnt allocated correctly..
	{
		qWarning("Could not allocate memory for image");	// Display Error
		fclose(fTGA);														// Close file
		return false;														// Return failed
	}

	GLuint pixelcount	= tga.Height * tga.Width;							// Nuber of pixels in the image
	GLuint currentpixel	= 0;												// Current pixel being read
	GLuint currentbyte	= 0;												// Current byte 
	GLubyte * colorbuffer = (GLubyte *)malloc(tga.bytesPerPixel);			// Storage for 1 pixel


	do
	{
		GLubyte chunkheader = 0;											// Storage for "chunk" header

		if(fread(&chunkheader, sizeof(GLubyte), 1, fTGA) == 0)				// Read in the 1 byte header
		{
			qWarning("Could not read RLE header");	// Display Error
			if(fTGA != NULL)												// If file is open
			{
				fclose(fTGA);												// Close file
			}
			if (imageData) {
				delete[] imageData;
				imageData = 0;
			}
			return false;													// Return failed
		}

		if(chunkheader < 128)												// If the ehader is < 128, it means the that is the number of RAW color packets minus 1
		{																	// that follow the header
			chunkheader++;													// add 1 to get number of following color values
			for(short counter = 0; counter < chunkheader; counter++)		// Read RAW color values
			{
				if(fread(colorbuffer, 1, tga.bytesPerPixel, fTGA) != tga.bytesPerPixel) // Try to read 1 pixel
				{
					qWarning("Could not read image data");		// IF we cant, display an error

					if(fTGA != NULL)													// See if file is open
					{
						fclose(fTGA);													// If so, close file
					}

					if(colorbuffer != NULL)												// See if colorbuffer has data in it
					{
						free(colorbuffer);												// If so, delete it
					}

					if (imageData) {
						delete[] imageData;
						imageData = 0;
					}
					
					return false;														// Return failed
				}
																						// write to memory
				imageData[currentbyte		] = colorbuffer[2];				    // Flip R and B vcolor values around in the process 
				imageData[currentbyte + 1	] = colorbuffer[1];
				imageData[currentbyte + 2	] = colorbuffer[0];

				if(tga.bytesPerPixel == 4)												// if its a 32 bpp image
				{
					imageData[currentbyte + 3] = colorbuffer[3];				// copy the 4th byte
				}

				currentbyte += tga.bytesPerPixel;										// Increase thecurrent byte by the number of bytes per pixel
				currentpixel++;															// Increase current pixel by 1

				if(currentpixel > pixelcount)											// Make sure we havent read too many pixels
				{
					qWarning("Too many pixels read");			// if there is too many... Display an error!

					if(fTGA != NULL)													// If there is a file open
					{
						fclose(fTGA);													// Close file
					}	

					if(colorbuffer != NULL)												// If there is data in colorbuffer
					{
						free(colorbuffer);												// Delete it

					}

					if (imageData) {
						delete[] imageData;
						imageData = 0;
					}
					
					return false;														// Return failed
				}
			}
		}
		else																			// chunkheader > 128 RLE data, next color reapeated chunkheader - 127 times
		{
			chunkheader -= 127;															// Subteact 127 to get rid of the ID bit
			if(fread(colorbuffer, 1, tga.bytesPerPixel, fTGA) != tga.bytesPerPixel)		// Attempt to read following color values
			{	
				qWarning("Could not read from file");			// If attempt fails.. Display error (again)

				if(fTGA != NULL)														// If thereis a file open
				{
					fclose(fTGA);														// Close it
				}

				if(colorbuffer != NULL)													// If there is data in the colorbuffer
				{
					free(colorbuffer);													// delete it
				}

				if (imageData) {
					delete[] imageData;
					imageData = 0;
				}
				return false;															// return failed
			}

			for(short counter = 0; counter < chunkheader; counter++)					// copy the color into the image data as many times as dictated 
			{																			// by the header
				imageData[currentbyte		] = colorbuffer[2];					// switch R and B bytes areound while copying
				imageData[currentbyte + 1	] = colorbuffer[1];
				imageData[currentbyte + 2	] = colorbuffer[0];

				if(tga.bytesPerPixel == 4)												// If TGA images is 32 bpp
				{
					imageData[currentbyte + 3] = colorbuffer[3];				// Copy 4th byte
				}

				currentbyte += tga.bytesPerPixel;										// Increase current byte by the number of bytes per pixel
				currentpixel++;															// Increase pixel count by 1

				if(currentpixel > pixelcount)											// Make sure we havent written too many pixels
				{
					qWarning("Too many pixels read");			// if there is too many... Display an error!

					if(fTGA != NULL)													// If there is a file open
					{
						fclose(fTGA);													// Close file
					}	

					if(colorbuffer != NULL)												// If there is data in colorbuffer
					{
						free(colorbuffer);												// Delete it
					}

					if (imageData) {
						delete[] imageData;
						imageData = 0;
					}
					
					return false;														// Return failed
				}
			}
		}
	}

	while(currentpixel < pixelcount);													// Loop while there are still pixels left
	fclose(fTGA);																		// Close the file
	return true;																		// return success
}

/*
 * Include file for users of JPEG library.
 * You will need to have included system headers that define at least
 * the typedefs FILE and size_t before you can include jpeglib.h.
 * (stdio.h is sufficient on ANSI-conforming systems.)
 * You may also wish to include "jerror.h".
 */
// NOTE: the "libjpeg-6b" directory must be in your include directory
extern "C"
{
#if defined(_WIN32)
#include "jpeglibwin.h"
#else
#include "jpeglib.h"
#endif
}
/*
 * <setjmp.h> is used for the optional error recovery mechanism shown in
 * the second part of the example.
 */
#include <setjmp.h>

/*
 * ERROR HANDLING:
 *
 * The JPEG library's standard error handler (jerror.c) is divided into
 * several "methods" which you can override individually.  This lets you
 * adjust the behavior without duplicating a lot of code, which you might
 * have to update with each future release.
 *
 * Our example here shows how to override the "error_exit" method so that
 * control is returned to the library's caller when a fatal error occurs,
 * rather than calling exit() as the standard error_exit method does.
 *
 * We use C's setjmp/longjmp facility to return control.  This means that the
 * routine which calls the JPEG library must first execute a setjmp() call to
 * establish the return point.  We want the replacement error_exit to do a
 * longjmp().  But we need to make the setjmp buffer accessible to the
 * error_exit routine.  To do this, we make a private extension of the
 * standard JPEG error handler object.  (If we were using C++, we'd say we
 * were making a subclass of the regular error handler.)
 *
 * Here's the extended error handler struct:
 */

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
  // cinfo->err really points to a my_error_mgr struct, so coerce pointer
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  // Always display the message.
  // We could postpone this until after returning, if we chose.
  (*cinfo->err->output_message) (cinfo);

  // Return control to the setjmp point
  longjmp(myerr->setjmp_buffer, 1);
}


/*
 * Sample routine for JPEG compression.  We assume that the target file name
 * and a compression quality factor are passed in.
 */
bool TEXTURE::WriteJPEG (char * filename, int quality)
{
  /* This struct contains the JPEG compression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   * It is possible to have several such structures, representing multiple
   * compression/decompression processes, in existence at once.  We refer
   * to any one struct (and its associated working data) as a "JPEG object".
   */
  struct jpeg_compress_struct cinfo;
  /* This struct represents a JPEG error handler.  It is declared separately
   * because applications often want to supply a specialized error handler
   * (see the second half of this file for an example).  But here we just
   * take the easy way out and use the standard error handler, which will
   * print a message on stderr and call exit() if compression fails.
   * Note that this struct must live as long as the main JPEG parameter
   * struct, to avoid dangling-pointer problems.
   */
  struct jpeg_error_mgr jerr;
  /* More stuff */
  FILE * outfile;		/* target file */
  JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
  int row_stride;		/* physical row width in image buffer */

  /* Step 1: allocate and initialize JPEG compression object */

  /* We have to set up the error handler first, in case the initialization
   * step fails.  (Unlikely, but it could happen if you are out of memory.)
   * This routine fills in the contents of struct jerr, and returns jerr's
   * address which we place into the link field in cinfo.
   */
  cinfo.err = jpeg_std_error(&jerr);
  /* Now we can initialize the JPEG compression object. */
  jpeg_create_compress(&cinfo);

  /* Step 2: specify data destination (eg, a file) */
  /* Note: steps 2 and 3 can be done in either order. */

  /* Here we use the library-supplied code to send compressed data to a
   * stdio stream.  You can also write your own code to do something else.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to write binary files.
   */
  if ((outfile = fopen(filename, "wb")) == NULL) {
    qWarning("Can't write image %s\n", filename);
	  //MessageBox(NULL, buf, "ERROR", MB_OK);	// Display an error message
    return false;
  }
  jpeg_stdio_dest(&cinfo, outfile);

  /* Step 3: set parameters for compression */

  /* First we supply a description of the input image.
   * Four fields of the cinfo struct must be filled in:
   */
  cinfo.image_width = width; 	/* image width and height, in pixels */
  cinfo.image_height = height;
  cinfo.input_components = 3;		/* # of color components per pixel */
  cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
  /* Now use the library's routine to set default compression parameters.
   * (You must set at least cinfo.in_color_space before calling this,
   * since the defaults depend on the source color space.)
   */
  jpeg_set_defaults(&cinfo);
  /* Now you can set any non-default parameters you wish to.
   * Here we just illustrate the use of quality (quantization table) scaling:
   */
  jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);

  /* Step 4: Start compressor */

  /* TRUE ensures that we will write a complete interchange-JPEG file.

   * Pass TRUE unless you are very sure of what you're doing.
   */
  jpeg_start_compress(&cinfo, TRUE);

  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */

  /* Here we use the library's state variable cinfo.next_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   * To keep things simple, we pass one scanline per call; you can pass
   * more if you wish, though.
   */
  row_stride = width * 3;	/* JSAMPLEs per row in imageData */

  while (cinfo.next_scanline < cinfo.image_height) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass
     * more than one scanline at a time if that's more convenient.
     */
    row_pointer[0] = & imageData[(cinfo.image_height-1 - cinfo.next_scanline) * row_stride];
    (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }

  /* Step 6: Finish compression */

  jpeg_finish_compress(&cinfo);
  /* After finish_compress, we can close the output file. */
  fclose(outfile);

  /* Step 7: release JPEG compression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_compress(&cinfo);

  /* And we're done! */
  return true;
}

/*
 * Sample routine for JPEG decompression.  We assume that the source file name
 * is passed in.  We want to return true on success, false on error.
 */
 bool TEXTURE::LoadJPEG(char * filename)
{
  /* This struct contains the JPEG decompression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   */
  struct jpeg_decompress_struct cinfo;
  /* We use our private extension JPEG error handler.
   * Note that this struct must live as long as the main JPEG parameter
   * struct, to avoid dangling-pointer problems.
   */
  struct my_error_mgr jerr;
  /* More stuff */
  FILE * infile;		/* source file */
  JSAMPARRAY buffer;		/* Output row buffer */
  int row_stride;		/* physical row width in output buffer */

  /* In this example we want to open the input file before doing anything else,
   * so that the setjmp() error recovery below can assume the file is open.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to read binary files.
   */

  if ((infile = fopen(filename, "rb")) == NULL) {
    qWarning("Can't open image %s\n", filename);
	  //MessageBox(NULL, buf, "ERROR", MB_OK);	// Display an error message
    return false;
  }

  /* Step 1: allocate and initialize JPEG decompression object */

  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object, close the input file, and return.
     */
    jpeg_destroy_decompress(&cinfo);
    fclose(infile);
    return 0;
  }
  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(&cinfo);

  /* Step 2: specify data source (eg, a file) */

  jpeg_stdio_src(&cinfo, infile);

  /* Step 3: read file parameters with jpeg_read_header() */

  (void) jpeg_read_header(&cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */

  /* Step 4: set parameters for decompression */

  /* In this example, we don't need to change any of the defaults set by
   * jpeg_read_header(), so we do nothing here.
   */

  /* Step 5: Start decompressor */

  (void) jpeg_start_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  width = cinfo.image_width; 	/* image width and height, in pixels */
  height = cinfo.image_height;
  bpp = cinfo.num_components*8;		/* # of color components per pixel */
  type = GL_RGB;
  imageData = new unsigned char[(width * height)*cinfo.num_components];


  /* We may need to do some setup of our own at this point before reading
   * the data.  After jpeg_start_decompress() we have the correct scaled
   * output image dimensions available, as well as the output colormap
   * if we asked for color quantization.
   * In this example, we need to make an output work buffer of the right size.
   */ 
  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo.output_width * cinfo.output_components;
  /* Make a one-row-high sample array that will go away when done with image */
  buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  while (cinfo.output_scanline < cinfo.output_height) {
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    /* Assume put_scanline_someplace wants a pointer and sample count. */
	memcpy(&imageData[(cinfo.output_height-cinfo.output_scanline) * row_stride], buffer[0], row_stride);
  }


  /* Step 7: Finish decompression */

  (void) jpeg_finish_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* Step 8: Release JPEG decompression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_decompress(&cinfo);

  /* After finish_decompress, we can close the input file.
   * Here we postpone it until after no more JPEG errors are possible,
   * so as to simplify the setjmp error logic above.  (Actually, I don't
   * think that jpeg_destroy can do an error exit, but why assume anything...)
   */
  fclose(infile);

  /* At this point you may want to check to see whether any corrupt-data
   * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
   */

  /* And we're done! */
  return true;
}



